/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include <iostream>
#include <string>
#include <cstdio>
#include <Matrix_LA.h>
#include <Vektor_LA.h>
#include <Vektor_mpf_t.h>
#include <head.h>
#include <GF256.h>

#define Problem_76 1
//#define G_Norm 1
#define Euklid 1

unsigned char antilog_table[256];
unsigned char log_table[256];
mpz_t alpha_division;
mpz_t C_0_wurzel;
mpz_t C_1_wurzel;
mpz_t E_0_wurzel;
mpz_t E_1_wurzel;
mpz_t zwei_hoch_k_hoch_minus_1;
mpz_t zwei_hoch_f;
mpz_t halb;
mpz_t eins_Wurzel;
mpz_t Wurzel_zwei_hoch_f;
mpf_t Wurzel_2;
mpz_t Wurzel_2_z;
mpz_t Drei;
mpz_t rem;
mpz_t rem_q_1;
mpz_t grosser_Modulus;
mpz_t kleiner_Modulus;
gmp_randstate_t rState1;
mpz_t * lambda;
mpz_t * lambda_Rekonstrukt;
mpz_t * RISS;
mpz_t ** RISS_Koeffizienten_q1;
mpz_t ** RISS_Koeffizienten_q;
GF256 ** RISS_Koeffizienten_F2M;
GF256 * lambda_256;
GF256 * lambda_256_Rekonstrukt;
int **Teilmengen;
int Theta_wurzel;
int Theta_Division;
int n_ueber_t;

using namespace std;

int Berechne_J(Matrix_LA& J, Matrix_LA& R, Matrix_LA& A, Matrix_LA& B, Matrix_LA& G, Vektor_LA& Aktive_Menge, int n, int t, int k, mpz_t& q, mpz_t& q_1, int Iteration);//Neuberechnung der Matrizen
int Aktualisiere_J(Matrix_LA& J, Matrix_LA& R, Matrix_LA& A, Matrix_LA& G, Matrix_LA& B, Matrix_LA& P, Vektor_LA& n_plus_minus, Share& Nummer_NB, Vektor_LA& Nummern_AM, Vektor_LA& Indikatorvektor, Vektor_LA& AM_kurz, int n, int t, int k, mpz_t& q, mpz_t& q_1, int Iteration);//Neuberechnung der Matrizen
int Berechne_z_r(Matrix_LA& J, Matrix_LA& R, Vektor_LA& n_p, Vektor_LA& Aktive_Menge_kurz, Vektor_LA& z, Vektor_LA& r, int n, int t, mpz_t& q, mpz_t& q_1, int k, int Iteration);//Berechnung von z und r
void Max_G_Norm(Matrix_LA& A, Matrix_LA& G, Vektor_LA& b, Vektor_LA& x, Vektor_LA& I, Vektor_LA& AM, Vektor_LA& s, Vektor_LA& NN, int n, int t, int k, mpz_t& q, mpz_t& q_1,int Iterationszaehler);
void Max_Residuum(Matrix_LA& A, Vektor_LA& b, Vektor_LA& x, Vektor_LA& I, int n, int t, int k, mpz_t& q, mpz_t& q_1);
void Max_Eukl_Norm(Matrix_LA& A, Vektor_LA& b, Vektor_LA& x, Vektor_LA& I, Vektor_LA& AM, int n, int t, int k, mpz_t& q, mpz_t& q_1);

void Programmende(clock_t start, clock_t stop, Vektor_LA& x, Vektor_mpf_t& x_korrekt, Vektor_LA& u, Share& ff, mpf_t& f_korrekt, int Iterationszaehler, int t, mpz_t& q);

int main()
{
int i,j,n=5, t=2, k=128, Iterationszaehler=1;
mpf_set_default_prec(300);

mpz_t q;
mpz_t q_1;
mpz_t basis;
mpz_t modulo;
mpz_t modulo_q1;

mpz_init(q);
mpz_init(q_1);
mpz_init_set_ui(basis,2);
mpz_init(modulo);
mpz_init(modulo_q1);

mpz_pow_ui(q,basis,Modullaenge);
mpz_pow_ui(q_1,basis,Modullaenge_kurz);

do{
mpz_nextprime(q,q);
mpz_nextprime(q_1,q_1);
mpz_mod_ui(modulo,q, 4);
mpz_mod_ui(modulo_q1,q_1, 4);
}
while(mpz_cmp_ui(modulo,3)!=0 || mpz_cmp_ui(modulo_q1,3)!=0); //Erzeuge geeignete Primzahl fr Berechnung des modulus (mu kongruent 3 mod 4 sein)

haupt::Berechne_Koeffizienten(n,t,k,f,q,q_1);

mpz_init_set(grosser_Modulus,q);
mpz_init_set(kleiner_Modulus,q_1);

mpz_t unendlich;
mpz_init(unendlich);

clock_t start, stop;

mpz_t seed_mpz;
mpz_init(seed_mpz);
time_t seed;
double seed_d;

seed=clock();
seed_d=(double) seed;
mpz_set_d(seed_mpz,seed_d);

gmp_randstate_t state;
gmp_randinit_default(state);
gmp_randseed(state,seed_mpz);

mpz_set_ui(unendlich,1);
mpz_mul_2exp(unendlich,unendlich,k-1);
mpz_sub_ui(unendlich,unendlich,1);

double ** Werte, ** Q_Werte;

#ifdef Problem_21
Vektor_LA x(2,n,t,q,q_1,k,f);
Vektor_LA d(2,n,t,q,q_1,k,f);
Vektor_LA c(2,n,t,q,q_1,k,f);//ist hier =0
Matrix_LA G(2,2,n,t,q,q_1,k,f);
Vektor_LA b(5,n,t,q,q_1,k,f);
Matrix_LA A(5,2,n,t,q,q_1,k,f);


Werte=new double *[5];
for(i=0; i<5; i++){
Werte[i]=new double[2];
}

Q_Werte=new double*[2];
Q_Werte[0]=new double[2];
Q_Werte[1]=new double[2];

Q_Werte[0][0]=0.02;
Q_Werte[0][1]=0.0;
Q_Werte[1][0]=0.0;
Q_Werte[1][1]=2.0;

Werte[0][0]=10.0;
Werte[0][1]=-1.0;
Werte[1][0]=-1.0;
Werte[1][1]=0.0;
Werte[2][0]=1.0;
Werte[2][1]=0.0;
Werte[3][0]=0.0;
Werte[3][1]=-1.0;
Werte[4][0]=0.0;
Werte[4][1]=1.0;

b.Vector_LA.at(0).Erzeugung_double(10.0,t,q);
b.Vector_LA.at(1).Erzeugung_double(-50.0,t,q);
b.Vector_LA.at(2).Erzeugung_double(2.0,t,q);
b[3].Erzeugung_double(-50.0,t,q);
b[4].Erzeugung_double(-50.0,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],2.0);
mpf_set_d(x_korrekt.V[1],0.0);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,0.04);
#endif

#ifdef Problem_35
Vektor_LA x(3,n,t,q,q_1,k,f);
Vektor_LA d(3,n,t,q,q_1,k,f);
Vektor_LA c(3,n,t,q,q_1,k,f);
Vektor_LA b(4,n,t,q,q_1,k,f);
Matrix_LA A(4,3,n,t,q,q_1,k,f);
Matrix_LA G(3,3,n,t,q,q_1,k,f);

Werte=new double *[4];
for(i=0; i<4; i++){
Werte[i]=new double[3];
}


Q_Werte=new double*[3];
for(i=0; i<5; i++){
    Q_Werte[i]=new double[3];
    }

Q_Werte[0][0]=4.0;
Q_Werte[0][1]=2.0;
Q_Werte[0][2]=2.0;

Q_Werte[1][0]=2.0;
Q_Werte[1][1]=4.0;
Q_Werte[1][2]=0.0;

Q_Werte[2][0]=2.0;
Q_Werte[2][1]=0.0;
Q_Werte[2][2]=2.0;


Werte[0][0]=-1.0;
Werte[0][1]=-1.0;
Werte[0][2]=-2.0;

Werte[1][0]=1.0;
Werte[1][1]=0.0;
Werte[1][2]=0.0;

Werte[2][0]=0.0;
Werte[2][1]=1.0;
Werte[2][2]=0.0;

Werte[3][0]=0.0;
Werte[3][1]=0.0;
Werte[3][2]=1.0;


c[0].Erzeugung_double(-8.0,t,q);
c[1].Erzeugung_double(-6.0,t,q);
c[2].Erzeugung_double(-4.0,t,q);

b[0].Erzeugung_double(-3.0,t,q);
b[1].Erzeugung_double(0.0,t,q);
b[2].Erzeugung_double(0.0,t,q);
b[3].Erzeugung_double(0.0,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],4.0);
mpf_div_ui(x_korrekt.V[0],x_korrekt.V[0],3);
mpf_set_d(x_korrekt.V[1],7.0);
mpf_div_ui(x_korrekt.V[1],x_korrekt.V[1],9);
mpf_set_d(x_korrekt.V[2],4.0);
mpf_div_ui(x_korrekt.V[2],x_korrekt.V[2],9);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,-8.0);
mpf_div_ui(f_korrekt,f_korrekt,9);
mpf_sub_ui(f_korrekt,f_korrekt,8);//f=-8 8/9
#endif

#ifdef Problem_76
Vektor_LA x(4,n,t,q,q_1,k,f);
Vektor_LA d(4,n,t,q,q_1,k,f);
Vektor_LA c(4,n,t,q,q_1,k,f);
Vektor_LA b(7,n,t,q,q_1,k,f);
Matrix_LA A(7,4,n,t,q,q_1,k,f);
Matrix_LA G(4,4,n,t,q,q_1,k,f);

Werte=new double *[7];
for(i=0; i<7; i++){
Werte[i]=new double[4];
}


Q_Werte=new double*[4];
for(i=0; i<5; i++){
    Q_Werte[i]=new double[4];
    }

Q_Werte[0][0]=2.0;
Q_Werte[0][1]=0.0;
Q_Werte[0][2]=-1.0;
Q_Werte[0][3]=0.0;

Q_Werte[1][0]=0.0;
Q_Werte[1][1]=1.0;
Q_Werte[1][2]=0.0;
Q_Werte[1][3]=0.0;

Q_Werte[2][0]=-1.0;
Q_Werte[2][1]=0.0;
Q_Werte[2][2]=2.0;
Q_Werte[2][3]=1.0;

Q_Werte[3][0]=0.0;
Q_Werte[3][1]=0.0;
Q_Werte[3][2]=1.0;
Q_Werte[3][3]=1.0;


Werte[0][0]=-1.0;
Werte[0][1]=-2.0;
Werte[0][2]=-1.0;
Werte[0][3]=-1.0;

Werte[1][0]=-3.0;
Werte[1][1]=-1.0;
Werte[1][2]=-2.0;
Werte[1][3]=1.0;

Werte[2][0]=0.0;
Werte[2][1]=1.0;
Werte[2][2]=4.0;
Werte[2][3]=0.0;

Werte[3][0]=1.0;
Werte[3][1]=0.0;
Werte[3][2]=0.0;
Werte[3][3]=0.0;

Werte[4][0]=0.0;
Werte[4][1]=1.0;
Werte[4][2]=0.0;
Werte[4][3]=0.0;

Werte[5][0]=0.0;
Werte[5][1]=0.0;
Werte[5][2]=1.0;
Werte[5][3]=0.0;

Werte[6][0]=0.0;
Werte[6][1]=0.0;
Werte[6][2]=0.0;
Werte[6][3]=1.0;

c[0].Erzeugung_double(-1.0,t,q);
c[1].Erzeugung_double(-3.0,t,q);
c[2].Erzeugung_double(1.0,t,q);
c[3].Erzeugung_double(-1.0,t,q);

b[0].Erzeugung_double(-5.0,t,q);
b[1].Erzeugung_double(-4.0,t,q);
b[2].Erzeugung_double(1.5,t,q);
b[3].Erzeugung_double(0.0,t,q);
b[4].Erzeugung_double(0.0,t,q);
b[5].Erzeugung_double(0.0,t,q);
b[6].Erzeugung_double(0.0,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],3.0);
mpf_div_ui(x_korrekt.V[0],x_korrekt.V[0],11);//0,272727272727... =3/11
mpf_set_d(x_korrekt.V[1],23.0);
mpf_div_ui(x_korrekt.V[1],x_korrekt.V[1],11);//2,09090909090909... =23/11
mpf_set_d(x_korrekt.V[2],0.26);
mpf_div_ui(x_korrekt.V[2],x_korrekt.V[2],10000000000);
mpf_set_d(x_korrekt.V[3],6.0);
mpf_div_ui(x_korrekt.V[3],x_korrekt.V[3],11);//0,545454545454... =6/11

mpf_t f_korrekt;
mpf_init(f_korrekt);

mpf_t h;
mpf_init(h);

mpf_mul(f_korrekt,x_korrekt.V[0],x_korrekt.V[0]);
mpf_mul(h,x_korrekt.V[1],x_korrekt.V[1]);
mpf_div_ui(h,h,2);
mpf_add(f_korrekt,f_korrekt,h);
mpf_mul(h,x_korrekt.V[2],x_korrekt.V[2]);
mpf_add(f_korrekt,f_korrekt,h);
mpf_mul(h,x_korrekt.V[3],x_korrekt.V[3]);
mpf_div_ui(h,h,2);
mpf_add(f_korrekt,f_korrekt,h);
mpf_mul(h,x_korrekt.V[0],x_korrekt.V[2]);
mpf_sub(f_korrekt,f_korrekt,h);
mpf_mul(h,x_korrekt.V[2],x_korrekt.V[3]);
mpf_add(f_korrekt,f_korrekt,h);
mpf_sub(f_korrekt,f_korrekt,x_korrekt.V[0]);
mpf_mul_ui(h,x_korrekt.V[1],3);
mpf_sub(f_korrekt,f_korrekt,h);
mpf_add(f_korrekt,f_korrekt,x_korrekt.V[2]);
mpf_sub(f_korrekt,f_korrekt,x_korrekt.V[3]);
#endif

#ifdef Problem_224
Vektor_LA x(2,n,t,q,q_1,k,f);
Vektor_LA d(2,n,t,q,q_1,k,f);
Vektor_LA c(2,n,t,q,q_1,k,f);//ist hier =0
Matrix_LA G(2,2,n,t,q,q_1,k,f);
Vektor_LA b(4,n,t,q,q_1,k,f);
Matrix_LA A(4,2,n,t,q,q_1,k,f);

Werte=new double *[5];
for(i=0; i<5; i++){
Werte[i]=new double[2];
}

Q_Werte=new double*[2];
Q_Werte[0]=new double[2];
Q_Werte[1]=new double[2];

Q_Werte[0][0]=4.0;
Q_Werte[0][1]=0.0;
Q_Werte[1][0]=0.0;
Q_Werte[1][1]=2.0;

Werte[0][0]=1.0;
Werte[0][1]=3.0;
Werte[1][0]=-1.0;
Werte[1][1]=-3.0;
Werte[2][0]=1.0;
Werte[2][1]=1.0;
Werte[3][0]=-1.0;
Werte[3][1]=-1.0;

b.Vector_LA.at(0).Erzeugung_double(0.0,t,q);
b.Vector_LA.at(1).Erzeugung_double(-18.0,t,q);
b.Vector_LA.at(2).Erzeugung_double(0.0,t,q);
b[3].Erzeugung_double(-8.0,t,q);

c[0].Erzeugung_double(-48.0,t,q);
c[1].Erzeugung_double(-40.0,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],4.0);
mpf_set_d(x_korrekt.V[1],4.0);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,-304.0);
#endif

#ifdef Prioblem_268
Vektor_LA x(5,n,t,q,q_1,k,f);
Vektor_LA d(5,n,t,q,q_1,k,f);
Vektor_LA c(5,n,t,q,q_1,k,f);
Vektor_LA b(5,n,t,q,q_1,k,f);
Matrix_LA A(5,5,n,t,q,q_1,k,f);
Matrix_LA G(5,5,n,t,q,q_1,k,f);

Werte=new double *[5];
for(i=0; i<5; i++){
Werte[i]=new double[5];
}


Q_Werte=new double*[5];
for(i=0; i<5; i++){
    Q_Werte[i]=new double[5];
    }

Q_Werte[0][0]=20394.0;
Q_Werte[0][1]=-24908.0;
Q_Werte[0][2]=-2026.0;
Q_Werte[0][3]=3896.0;
Q_Werte[0][4]=658.0;

Q_Werte[1][0]=-24908.0;
Q_Werte[1][1]=41818.0;
Q_Werte[1][2]=-3466.0;
Q_Werte[1][3]=-9828.0;
Q_Werte[1][4]=-372.0;

Q_Werte[2][0]=-2026.0;
Q_Werte[2][1]=-3466.0;
Q_Werte[2][2]=3510.0;
Q_Werte[2][3]=2178.0;
Q_Werte[2][4]=-348.0;

Q_Werte[3][0]=3896.0;
Q_Werte[3][1]=-9828.0;
Q_Werte[3][2]=2178.0;
Q_Werte[3][3]=3030.0;
Q_Werte[3][4]=-44.0;

Q_Werte[4][0]=658.0;
Q_Werte[4][1]=-372.0;
Q_Werte[4][2]=-348.0;
Q_Werte[4][3]=-44.0;
Q_Werte[4][4]=54.0;


Werte[0][0]=-1.0;
Werte[0][1]=-1.0;
Werte[0][2]=-1.0;
Werte[0][3]=-1.0;
Werte[0][4]=-1.0;

Werte[1][0]=10.0;
Werte[1][1]=10.0;
Werte[1][2]=-3.0;
Werte[1][3]=5.0;
Werte[1][4]=4.0;

Werte[2][0]=-8.0;
Werte[2][1]=1.0;
Werte[2][2]=-2.0;
Werte[2][3]=-5.0;
Werte[2][4]=-.0;

Werte[3][0]=8.0;
Werte[3][1]=-1.0;
Werte[3][2]=2.0;
Werte[3][3]=5.0;
Werte[3][4]=-3.0;

Werte[4][0]=-4.0;
Werte[4][1]=-2.0;
Werte[4][2]=3.0;
Werte[4][3]=-5.0;
Werte[4][4]=1.0;


b[0].Erzeugung_double(-5.0,t,q);
b[1].Erzeugung_double(20.0,t,q);
b[2].Erzeugung_double(-40.0,t,q);
b[3].Erzeugung_double(11.0,t,q);
b[4].Erzeugung_double(-30.0,t,q);

c[0].Erzeugung_double(18340.0,t,q);
c[1].Erzeugung_double(-34198.0,t,q);
c[2].Erzeugung_double(4542.0,t,q);
c[3].Erzeugung_double(8672.0,t,q);
c[4].Erzeugung_double(86.0,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],1.0);
mpf_set_d(x_korrekt.V[1],2.0);
mpf_set_d(x_korrekt.V[2],-1.0);
mpf_set_d(x_korrekt.V[3],3.0);
mpf_set_d(x_korrekt.V[4],-4.0);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,-14463.0);
#endif

#ifdef Problem_A
Vektor_LA x(2,n,t,q,q_1,k,f);
Vektor_LA d(2,n,t,q,q_1,k,f);
Vektor_LA c(2,n,t,q,q_1,k,f);
Matrix_LA G(2,2,n,t,q,q_1,k,f);
Vektor_LA b(5,n,t,q,q_1,k,f);
Matrix_LA A(5,2,n,t,q,q_1,k,f);


Werte=new double *[5];
for(i=0; i<5; i++){
Werte[i]=new double[2];
}

Q_Werte=new double*[2];
Q_Werte[0]=new double[2];
Q_Werte[1]=new double[2];

Q_Werte[0][0]=2.0;
Q_Werte[0][1]=0.0;
Q_Werte[1][0]=0.0;
Q_Werte[1][1]=2.0;

Werte[0][0]=1.0;
Werte[0][1]=-2.0;
Werte[1][0]=-1.0;
Werte[1][1]=-2.0;
Werte[2][0]=-1.0;
Werte[2][1]=2.0;
Werte[3][0]=1.0;
Werte[3][1]=0.0;
Werte[4][0]=0.0;
Werte[4][1]=1.0;

b[0].Erzeugung_double(-2.0,t,q);
b[1].Erzeugung_double(-6.0,t,q);
b[2].Erzeugung_double(-2.0,t,q);
b[3].Erzeugung_double(0.0,t,q);
b[4].Erzeugung_double(0.0,t,q);

c.Vector_LA.at(0).Erzeugung_double(-2.0,t,q);
c.Vector_LA.at(1).Erzeugung_double(-5.0,t,q);

Vektor_mpf_t x_korrekt(2);
mpf_set_d(x_korrekt.V[0],1.4);
mpf_set_d(x_korrekt.V[1],1.7);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,-6.45);
#endif


G.Erzeugung_double(Q_Werte);
A.Erzeugung_double(Werte);

printf("\nDualer Algorithmus");
printf("\nParameter des Problems:");
G.Ausgabe_skaliert("G");
c.Ausgabe_skaliert("c");
A.Ausgabe_skaliert("A");
b.Ausgabe_skaliert("b");

mpz_t eins;
mpz_init_set_ui(eins,1);

Vektor_LA AM_Zq(b.d,n,t,q,q_1,k,f);
Vektor_LA Index_AM(b.d,n,t,q,q_1,k,f);
Vektor_LA NN(b.d,n,t,q,q_1,k,f);
Vektor_LA AM_kurz(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA AM_kurz_alt(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA AM_kurz_Arbeit(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA AM_kurz_Arbeit_C(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA AM_Arbeit(b.d,n,t,q,q_1,k,f);///Aktive Menge bei der ausgewhle, veletzte NB provisorisch inzugegfgt wird
Vektor_LA AM_Arbeit_C(b.d,n,t,q,q_1,k,f);///Komplement der temporren aktiven Menge

Share ff(n,t,q);

Vektor_LA s(b.d,n,t,q,q_1,k,f);
Vektor_LA S=s;///s wird noch bei der Berechnung von t_2 gebraucht
Share s_p(n,t,q);

Vektor_LA u(A.Zeilenzahl,n,t,q,q_1,k,f);//Vektor der Lagrange-Multiplikatoren
Vektor_LA u_plus(A.Zeilenzahl,n,t,q,q_1,k,f);//Vektor der nderungen der Lagrange-Multiplikatoren

Matrix_LA Kopie_G=G;

Vektor_LA z(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA tz(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA r(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA r_Arbeit(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA r_GTZ(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA r_GTZ_C(G.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA r_gross(A.Zeilenzahl,n,t,q,q_1,k,f);//wird fr die Berechnung der Lagrange-Multiplikatoren und ihre Zuordnung zu den einzelnen NBen bentigt.

Vektor_LA t_1_Vektor(A.Zeilenzahl,n,t,q,q_1,k,f);//Zur Bestimmung von t_1 bentigt
Vektor_LA t_1_Index(A.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA t_1_Index_kurz(A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA u_plus_kurz(A.Spaltenzahl,n,t,q,q_1,k,f);

Matrix_LA B(A.Spaltenzahl,A.Spaltenzahl,n,t,q,q_1,k,f);

Vektor_LA n_plus_minus(A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA n_p(A.Spaltenzahl,n,t,q,q_1,k,f);

Vektor_LA y(A.Spaltenzahl,n,t,q,q_1,k,f);

Vektor_LA Hilfsvektor(A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA Hilfsvektor2(A.Spaltenzahl,n,t,q,q_1,k,f);

Share V(n,t,q);
Share Hilfe(n,t,q);
Share qq(n,t,q);///Anzahl der aktiven Gleichungen
Share t_1(n,t,q);
Share t_1_2(n,t,q);
Share Norm_z(n,t,q);
Share t_2(n,t,q);
Share zn(n,t,q);
Share T(n,t,q);
Share T_Hilfe(n,t,q);
Share T_Vergleich_t_2(n,t,q);
Share T_Vergleich_t_1(n,t,q);
Share eins_minus_T_Vergleich_t1(n,t,q);
Share eins_minus_T_Vergleich_t2(n,t,q);
Share t_2_Vergleich_unendlich(n,t,q);//Wird in 1. Iteration noch nicht gebraucht
Share eins_minus_t_2_Vergleich_unendlich(n,t,q);

Share q_Vergleich_Null(n,t,q);
Share r_kleiner_Null(n,t,q);
Share Schalter_t_1(n,t,q);

Vektor_LA nn(A.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA Additionshilfe(A.Zeilenzahl,n,t,q,q_1,k,f);

Vektor_LA beta(x.d-1,n,t,q,q_1,k,f);

Matrix_LA P(A.Zeilenzahl,A.Zeilenzahl,n,t,q,q_1,k,f);
Matrix_LA P2(A.Zeilenzahl,A.Zeilenzahl,n,t,q,q_1,k,f);
Matrix_LA P_t(A.Zeilenzahl,A.Zeilenzahl,n,t,q,q_1,k,f);
Matrix_LA P_alt(A.Zeilenzahl,A.Zeilenzahl,n,t,q,q_1,k,f);

Matrix_LA N(A.Spaltenzahl, A.Zeilenzahl,n,t,q,q_1,k,f);///Matrizen, die fr die Berechnung von z und r bentigt werden
Matrix_LA J(G.Zeilenzahl,G.Zeilenzahl,n,t,q,q_1,k,f);
Matrix_LA R(G.Zeilenzahl,G.Zeilenzahl,n,t,q,q_1,k,f);
Matrix_LA RR(1,1,n,t,q,q_1,k,f);

Vektor_LA Nummerierung(A.Zeilenzahl,n,t,q,q_1,k,f);
mpz_t Zahl_mpz;
mpz_init(Zahl_mpz);
for(i=0; i<A.Zeilenzahl; i++){
    mpz_set_ui(Zahl_mpz,i+1);//Nummerierung startet bei 1; so wird bei nicht-besetzten Eintrgen in Nummern_AM erkannt, dass keine nderung vorzunehmen ist
    Nummerierung[i].Erzeugung(Zahl_mpz,t,q);//Fr jede NB eine Nummer erzeugen
    }
Share Nummer(n,t,q);
Vektor_LA Nummern_AM(J.Spaltenzahl,n,t,q,q_1,k,f);

start=clock();

/**Schritt 0: Bestimme als Startwert das Minimum ohne Nebenbedingungen x=G^{-1}c */

x=c;
Kopie_G.Loese_GLS_Cholesky(x.Vector_LA,x.d);
x*=-1;

c.SkalarMult(x,ff);
ff.TruncPr(k,f,t,q,q_1);
ff.Div_Konst_int(2,q,q_1,k); ///ff=1/2*c^tx

/*H=G^{-1} wird nicht bestimmt; ist in Kopie_Q implizit gespeichert. Aktive Menge ist leer*/

/** Schritt 1: Aktive Menge bestimmen*/

///Bestimme verletzte Nebenbedingungen
x.MatrixMult(A,s);
s.TruncPr();
s=s-b;///s=Ax-b

S=s;
for(i=0; i<s.d; i++)
    {
    S[i].LTZ_approximativ(t,q,q_1,k);
    V=V+S[i];//V zhlt Anzahl der verletzten NBen
    }

i=V.EQZPub();
if(i==1){///Ist keine Nebenbedingung verletzt, Ende.
    stop=clock();
    Programmende(start,stop,x,x_korrekt,u,ff,f_korrekt,Iterationszaehler,t,q);
    return 0;
    }

Iterationszaehler++;

#ifdef G_Norm
Max_G_Norm(A,Kopie_G,b,x,nn,S,s,NN,n,t,k,q,q_1,Iterationszaehler);//S: Die Stellen an denen eine auswhlbare NB sitzt
#endif

#ifdef Euklid
Max_Eukl_Norm(A,b,x,nn,S,n,t,k,q,q_1);
#endif

for(i=0; i<A.Zeilenzahl; i++){
    for(j=0; j<A.Spaltenzahl; j++){
        n_p[j]=A.M[i][j]*nn[i]+n_p[j];
        }
    }

///*Berechne Schrittrichtung (Schritt 2a)*/

y=n_p;

y.VWSubs_mit_Div(Kopie_G);//In Tests war noch die alte Version enthalten
z.RWSubs(Kopie_G,y);

///*Berechne Schrittweite; im 1. Schritt ist t_1 immer gleich unendlich!*/
z.Gleich_Null_approximativ(Norm_z);
z.SkalarMult(n_p,zn);
zn.TruncPr(k,f,t,q,q_1);

s_p=s[nn];
s_p.FPDiv(zn,t,q,q_1,k);
s_p*=-1;
t_2=Norm_z*unendlich;
Hilfe=-Norm_z;
Hilfe+=1;

t_2+=Hilfe*s_p;
T=t_2;

///*Schritt 2c*/
T_Hilfe=T-unendlich;

T_Vergleich_t_2=T;
T_Vergleich_t_1=T;
eins_minus_T_Vergleich_t1=T_Vergleich_t_1;
eins_minus_T_Vergleich_t2=T_Vergleich_t_2;

i=T_Hilfe.EQZPub();
if(i==1){
    printf("\nNebenbedingungen widerspruechlich. Abbruch");
    stop=clock();
    Programmende(start,stop,x,x_korrekt,u,ff,f_korrekt,Iterationszaehler,t,q);
    exit(1);
    }

t_2_Vergleich_unendlich=t_2;//Wird in 1. Iteration noch nicht gebraucht
eins_minus_t_2_Vergleich_unendlich=-t_2_Vergleich_unendlich+1;

tz=z;
tz*=T;
tz.TruncPr(); //tz=T*z
x=x+tz;

///Bestimme verletzte Nebenbedingungen
x.MatrixMult(A,s);
s.TruncPr();
s=s-b;//s=Ax-b

///Bestimme verletzte NB gleich nach Neuberechnung von x

S=s;//s wird noch bei der Berechnung von t_2 gebraucht
V=0;
for(i=0; i<s.d; i++)
    {
    S[i].LTZ_approximativ(t,q,q_1,k);
    V=V+S[i];//V zhlt Anzahl der verletzten NBen
    }

i=V.EQZPub();

zn=zn*T*T;
zn.Div_Konst_int(2,q,q_1,2*k);//Division vor TruncPr ist teurer, aber genauer
zn.TruncPr(2*k,2*f,t,q,q_1);


ff+=zn;//f=f+tz^tn_p(0.5t+u); (u=0) in der ersten Iteration!!!

u_plus=nn*T;//Lagrange-Multiplikator der aktivierten Gleichung wird auf T gesetzt (u_plus ist am Anfang Null)

u=u_plus;

/** Aktive Menge aktualisieren*/
Additionshilfe=nn;
Additionshilfe+=AM_Zq;
AM_Zq=nn;

if(i==1){//Ist keine Nebenbedingung verletzt, Ende.
    stop=clock();
    Programmende(start,stop,x,x_korrekt,u,ff,f_korrekt,Iterationszaehler,t,q);
    return 0;
    }

qq=1;

//Aktive Menge kompaktifizieren
/** N^{\ast},H aktualisieren!!!*/
P.Ersetze_Zeile(AM_Zq.Vector_LA,0,AM_Zq.d);//Das so definierte P schiebt als Pemutaionsmatrix die 1. aktivierte NB an die Stelle 1
AM_kurz[0].Erzeugung(eins,t,q);//Nach dem 1. Schritt is genau eine NB aktiv; im verkrten Vekto AM_kurz steht diese an Stelle 0

int check;
check=Berechne_J(J,RR,A,B,Kopie_G,nn,n,t,k,q,q_1,Iterationszaehler);//nn statt AM_Zq, da im ersten Schritt nur eine NB aktiv ist

AM_Arbeit=AM_Zq;///Aktive Menge bei der ausgewhle, veletzte NB provisorisch inzugegfgt wird
AM_Arbeit_C=AM_Arbeit;///Komplement der temporren aktiven Menge

for(Iterationszaehler=3; Iterationszaehler<50; Iterationszaehler++){

///Bestimme verletzte Nebenbedingungen - Verschoben ans Ende; nach der Neuberechnung

#ifdef G_Norm
Max_G_Norm(A,Kopie_G,b,x,nn,S,s,NN,n,t,k,q,q_1,Iterationszaehler);
#endif

#ifdef Euklid
Max_Eukl_Norm(A,b,x,nn,S,n,t,k,q,q_1);
#endif

#ifdef FindFirst
S.FindFirst(nn);
#endif

AM_Arbeit=AM_Zq+nn;
for(i=0; i<AM_Arbeit.d; i++){
    AM_Arbeit_C[i]=AM_Arbeit[i];
    AM_Arbeit_C[i]*=-1;
    AM_Arbeit_C[i]+=1;
    }//Bestimmt Komplement der temporren aktiven Menge

Index_AM=AM_Arbeit;
Index_AM.Vektor_Zusammenschieben_binaer(AM_Arbeit.Vector_LA,P2);//Aktualisiere die Matrix P, mit deren Hilfe die aktiven Ungleichungen kompaktifziert werden

AM_kurz_Arbeit.Oben_einpassen(Index_AM);

for(i=0; i<AM_kurz_Arbeit.d; i++){
    AM_kurz_Arbeit_C[i]=AM_kurz_Arbeit[i];
    AM_kurz_Arbeit_C[i]*=-1;
    AM_kurz_Arbeit_C[i]+=1;
    }

n_p=0;
for(i=0; i<A.Zeilenzahl; i++){
    for(j=0; j<A.Spaltenzahl; j++){
        n_p[j]=A.M[i][j]*nn[i]+n_p[j];
        }
    }
Nummer=Nummerierung[nn];//Nummer der NB auslesen

/** Berechne Schrittrichtung (Schritt 2a)*/

int v=haupt::min(Iterationszaehler,G.Zeilenzahl);
Matrix_LA R_klein(v-1,v-1,n,t,q,q_1,k,f);//Gre hngt von Iterationszahl ab; mu deshalb in jeder Runde neu erzeugt werden
R_klein.Links_oben_einpassen(R);
if(Iterationszaehler==3){
    check=Berechne_z_r(J,RR,n_p,AM_kurz,z,r,n,t,q,q_1,k,Iterationszaehler);
    }
else
    check=Berechne_z_r(J,R_klein,n_p,AM_kurz,z,r,n,t,q,q_1,k,Iterationszaehler);
/** Berechne Schrittweite; im 1. Schritt ist t_1 immer gleich unendlich!*/

z.Gleich_Null_approximativ(Norm_z);

z.SkalarMult(n_p,zn);
zn.TruncPr(k,f,t,q,q_1);
Hilfe=zn;

/** Berechne t_1*/
q_Vergleich_Null=qq;
q_Vergleich_Null.EQZ(t,q,q_1,k);
r_kleiner_Null=r[0];
r_kleiner_Null.GTZ_approximativ(t,q,q_1,k);
r_GTZ[0]=r_kleiner_Null;
for(i=1; i<r.d; i++){
    r_GTZ[i]=r[i];
    r_GTZ[i].GTZ_approximativ(t,q,q_1,k);
    r_kleiner_Null=r_kleiner_Null+r_GTZ[i]-r_kleiner_Null*r_GTZ[i];//r_kleiner_Null ist 1, wenn mindestens eine Komponente grer als 0 ist
    }

Schalter_t_1=r_kleiner_Null*(-q_Vergleich_Null+1);

r_gross=0;
r_gross.Oben_einpassen(r);

P_t=P;
P_t.Transponiere_quadratisch();

r_gross.MatrixMult(P_t,r_gross);//Ordne die r's ihren Indizes zu

t_1_Vektor=u_plus;
t_1_Vektor.MatrixMult(P,t_1_Vektor);//permutiere die Eintrge von u^+ an die oberen Stellen

u_plus_kurz.Oben_einpassen(t_1_Vektor);

for(i=0; i<r.d; i++){
    r_GTZ_C[i]=r_GTZ[i];
    r_GTZ_C[i]*=-1;
    r_GTZ_C[i]+=1;
    }

for(i=0; i<A.Spaltenzahl; i++){// TODO (LocalAdmin#1#): Kriteriumsvektor verwenden!
    u_plus_kurz[i]=u_plus_kurz[i]*AM_kurz_Arbeit[i]+AM_kurz_Arbeit_C[i]*unendlich;
    u_plus_kurz[i]=r_GTZ[i]*u_plus_kurz[i]+r_GTZ_C[i]*unendlich;
    r_Arbeit[i]=r[i]*AM_kurz_Arbeit[i]+AM_kurz_Arbeit_C[i]*zwei_hoch_f;//r_gross wird spter noch fr die Aktualisierung von u bentigt
    r_Arbeit[i]=r_Arbeit[i]*r_GTZ[i]+r_GTZ_C[i]*zwei_hoch_f;
    }

u_plus_kurz.Min_Bruch(r_Arbeit,t_1_Index_kurz);
t_1_2=u_plus_kurz[t_1_Index_kurz];

t_1_2.FPDiv(r_Arbeit[t_1_Index_kurz],t,q,q_1,k);
t_1=t_1_2;

t_1_Index=0;
t_1_Index.Oben_einpassen(t_1_Index_kurz);
t_1_Index.MatrixMult(P_t,t_1_Index);//Index von t_1 an die richtige Stelle permutieren

/** Berechne t_2*/
s_p=s[nn];
s_p.FPDiv(zn,t,q,q_1,k);
s_p*=-1;
t_2=Norm_z*unendlich;
Hilfe=-Norm_z;
Hilfe+=1;

t_2+=Hilfe*s_p;

T.min(t_1,t_2,q,q_1,t,k,Hilfe);

/**Schritt 2c*/
T_Hilfe=T;
T_Hilfe-=unendlich;
i=T_Hilfe.EQZPub();

if(i==1){
    stop=clock();
    Programmende(start,stop,x,x_korrekt,u,ff,f_korrekt,Iterationszaehler,t,q);
    printf("\nNebenbedingungen widerspruechlich. Abbruch");
    exit(1);
    }

t_2_Vergleich_unendlich=t_2;
t_2_Vergleich_unendlich.EQ_mpz(unendlich,t,q,q_1,k);//t_2=?unendlich
eins_minus_t_2_Vergleich_unendlich=-t_2_Vergleich_unendlich;
eins_minus_t_2_Vergleich_unendlich+=1;

T_Vergleich_t_1=Hilfe;//Hilfe==1 <=> T=t_1, sonst 0
T_Vergleich_t_2=-Hilfe+1;

eins_minus_T_Vergleich_t1=T_Vergleich_t_1;
eins_minus_T_Vergleich_t2=T_Vergleich_t_2;
eins_minus_T_Vergleich_t1*=-1;
eins_minus_T_Vergleich_t1+=1;
eins_minus_T_Vergleich_t2*=-1;
eins_minus_T_Vergleich_t2+=1;

///xNeu
tz=z;
tz*=T;
tz.TruncPr(); //tz=T*z

x+=tz;
x*=eins_minus_t_2_Vergleich_unendlich;
Hilfsvektor=x*t_2_Vergleich_unendlich;
x+=Hilfsvektor;

///fNeu
zn*=T;//wird bentigt fr die Aktualisierung von f
zn.TruncPr(k,f,t,q,q_1);
Hilfe=u_plus[nn];//Hilfe=u^+_{q+1}
Hilfe*=2;
Hilfe+=T;
Hilfe.Div_Konst_int(2,q,q_1,k);
zn*=Hilfe;
zn.TruncPr(k,f,t,q,q_1);

ff=eins_minus_t_2_Vergleich_unendlich*(ff+zn)+t_2_Vergleich_unendlich*ff;;//f=f+tz^tn_p(0.5t+u); (u=0) in der ersten Iteration!!!

///uNeu
u_plus=u_plus+nn*T;

r_gross*=T;
r_gross.TruncPr(k,f);

u_plus-=r_gross;//u^+=u^+t(-r 1)^t

u=u_plus*T_Vergleich_t_2+u*eins_minus_T_Vergleich_t2;//If T=t_2 ->u=u_plus

///Aktualisierung der aktiven Menge
Additionshilfe=nn+AM_Zq;

AM_Zq=AM_Zq*eins_minus_T_Vergleich_t2+Additionshilfe*T_Vergleich_t_2;//neuen Index zur aktiven Menge hinzufgen

Additionshilfe=AM_Zq-t_1_Index;
AM_Zq=AM_Zq*eins_minus_T_Vergleich_t1+Additionshilfe*T_Vergleich_t_1;

qq=(qq+1)*T_Vergleich_t_2+(qq-1)*eins_minus_T_Vergleich_t2; //Aktualisierung des Zhlers der aktiven Menge

///berprfen der Zulssigkeit des neuen Punkts

x.MatrixMult(A,s);
s.TruncPr();
s=s-b;//s=Ax-b

S=s;//s wird noch bei der Berechnung von t_2 gebraucht
V=0;
for(i=0; i<s.d; i++)
    {
    S[i].LTZ_approximativ(t,q,q_1,k);
    V=V+S[i];//V zhlt Anzahl der verletzten NBen
    }

i=V.EQZPub();
if(i==1){//Ist keine Nebenbedingung verletzt, Ende.
    stop=clock();
    Programmende(start,stop,x,x_korrekt,u,ff,f_korrekt,Iterationszaehler,t,q);
    break;
    }

/** N^{\ast},H aktualisieren!!!*/
int check;

Index_AM=AM_Zq;
P_alt=P;
Index_AM.Vektor_Zusammenschieben_binaer(AM_Zq.Vector_LA,P);
AM_kurz_alt=AM_kurz;
AM_kurz.Oben_einpassen(Index_AM);

Matrix_LA R_gross(v,v,n,t,q,q_1,k,f);
if(A.Spaltenzahl<=3){//Bei Gre 3 msste bei Aktualisierung von Matrix der Gre 2 UND 3 berechnet werden. Da ist Neuberechnung billiger
    check=Berechne_J(J,R_gross,A,B,Kopie_G,AM_Zq,n,t,k,q,q_1,Iterationszaehler);
    }
else{
    if(Iterationszaehler==3)
        J.Ausgabe_skaliert("J");
    Hilfe=qq-1;//qq ist um eins grer als die Spalte, die es indiziert
    Hilfsvektor.BitMask(Hilfe);//HV ist Bitvektorkodierung der Gre der aktiven Menge

    t_1_Index.MatrixMult(P_alt,Additionshilfe);//Additionshilfe=P*t_1_Index; multipliziere t_1_Index an eine vordere Position; verwende P_alt, da neues P auf t_1_I nicht anspricht

    Hilfsvektor2.Oben_einpassen(Additionshilfe);//AH ist einer der ersten Eintrge
    Hilfsvektor=Hilfsvektor*T_Vergleich_t_2;//HV ist Bitvektorkodierung von qq, wenn neue NB hinzugefgt wird (dann steht die Spalte - qq - an der sie in J eingetragen wir, fest)
    Hilfe=-T_Vergleich_t_2+1;
    Hilfsvektor+=(Hilfsvektor2*Hilfe);//If T=t_2 -> HV<-HV else HV<-HV2

    n_plus_minus=(n_p*T_Vergleich_t_2);//T!=t_2 ->n+- = 0 und somit wird in B die 0-Spalte substituiert

    AM_kurz_Arbeit.Oben_einpassen(Additionshilfe);
    AM_kurz_Arbeit=AM_kurz_alt-AM_kurz_Arbeit;//Aus alter, verkrzter AM gelschten Index entfernen
    AM_kurz_Arbeit=AM_kurz*T_Vergleich_t_2+AM_kurz_Arbeit*T_Vergleich_t_1;//neue verkrzte AM bergeben, wenn NB hinzugefgt wurde; sonst manipulierte alte, vekrzte AM
    Nummer=Nummer*T_Vergleich_t_2+t_1_Index[Nummerierung]*T_Vergleich_t_1;//Index der zu lschenden Spalte bergeben, wenn T=t_1
    T_Vergleich_t_1.Rekonstrukt_Ausgabe_unskaliert("T=?t_1");
    check=Aktualisiere_J(J,R_gross,A,Kopie_G,B,P,n_plus_minus,Nummer,Nummern_AM,Hilfsvektor,AM_kurz_Arbeit,n,t,k,q,q_1,Iterationszaehler);
    }
R.Links_oben_einpassen(R_gross);
}


return 0;
}

void Programmende(clock_t start, clock_t stop, Vektor_LA& x, Vektor_mpf_t& x_korrekt, Vektor_LA& u, Share& ff, mpf_t& f_korrekt, int Iterationszaehler, int t, mpz_t& q){

int i;

mpz_t C;
mpz_init(C);


printf("\n\n\nLaufzeit des Programms: %f s",((double)stop-(double)start)/CLOCKS_PER_SEC);
printf("\nAnzahl der Iterationen: %d",Iterationszaehler);

    FILE* datei1;
    datei1=fopen("Ausgabe.txt","a");
    cout<<"\nOptimaler Punkt x erreicht"<<endl;
    fprintf(datei1,"\n\n\n------------------------------------------\n");
    time_t Zeit;
    struct tm *ts;//Variable zur Zeitmessung

    time(&Zeit);
    ts=localtime(&Zeit);
    fprintf(datei1,"\n\n\nZeit: %s\n\n",asctime(ts));
    fprintf(datei1,"\nOptimaler Punkt x erreicht");
    fprintf(datei1,"\nAnzahl der Iterationen: %d",Iterationszaehler);
    fprintf(datei1,"\nLaufzeit des Programms: %f s\n\n",((double)stop-(double)start)/CLOCKS_PER_SEC);
    mpf_t Z;
    mpf_init(Z);
    for(i=0; i<x.d; i++){
        x[i].Rekonstrukt(C,t,q);
        mpf_set_z(Z,C);
        mpf_div_2exp(Z,Z,f);
        gmp_fprintf(datei1,"\nx(%d)=%.10Ff",i,Z);
        }
    fprintf(datei1,"\n\n");
    for(i=0; i<u.d; i++){
        u[i].Rekonstrukt(C,t,q);
        mpf_set_z(Z,C);
        mpf_div_2exp(Z,Z,f);
        gmp_fprintf(datei1,"\nu(%d)=%.10Ff",i,Z);
        }
    x.Ausgabe_skaliert("x");
    Vektor_mpf_t x_Werte(x.d);
    x.Ausgabe_skaliert_Vektor_mpf_t(x_Werte);
    x_Werte-=x_korrekt;
    x_Werte.Ausgabe_wissenschaftlich("x_Diff");
    x_Werte.Ausgabe_wissenschaftlich_Datei("x_Diff",datei1);
    double Norm_Differenz=x_Werte.Norm();
    printf("\n\nNorm der Differenz: %e",Norm_Differenz);
    fprintf(datei1,"\n\nNorm der Differenz: %e", Norm_Differenz);


    fprintf(datei1,"\n\n");
    ff.Rekonstrukt_Ausgabe("Zielfunktion");
    ff.Rekonstrukt(C,t,q);
    mpf_set_z(Z,C);
    mpf_div_2exp(Z,Z,f);
    gmp_fprintf(datei1,"\nZielfunktion: %.5Ff",Z);
    mpf_sub(Z,Z,f_korrekt);
    gmp_printf("\nDifferenz zu exaktem Wert der Zielfunktion: %.10Fe",Z);
    gmp_fprintf(datei1,"\nDifferenz zu exaktem Wert der Zielfunktion: %.10Fe",Z);

    u.Ausgabe_skaliert("u");

    fclose(datei1);
    getchar();

mpf_clear(Z);
mpz_clear(C);
}

int Berechne_J(Matrix_LA& J, Matrix_LA& R, Matrix_LA& A, Matrix_LA& B, Matrix_LA& G, Vektor_LA& Aktive_Menge, int n, int t, int k, mpz_t& q, mpz_t& q_1, int Iteration){//Neuberechnung der Matrizen
//in G ist eine Choleskay-Zerlegung gespeichert
//Ab dim(G)=4 ist es billiger B zu aktualisieren
int i,j;

int r;

if(Iteration<A.Spaltenzahl)
    r=Iteration;
else
    r=A.Spaltenzahl;

Matrix_LA B_Hilfe(B.Zeilenzahl,r,n,t,q,q_1,k,f);
Matrix_LA N_Kompakt(A.Spaltenzahl,A.Zeilenzahl,n,t,q,q_1,k,f);
Matrix_LA N_quadratisch(A.Spaltenzahl,r,n,t,q,q_1,k,f);

N_Kompakt=A.Transponiere();
N_Kompakt.Multipliziere_Spaltenweise(Aktive_Menge.Vector_LA);
N_Kompakt.Matrix_spaltenweise_zusammenschieben_binaer(Aktive_Menge.Vector_LA);
N_quadratisch.Kopiere_links(N_Kompakt);

Vektor_LA Hilfsvektor(G.Zeilenzahl,n,t,q,q_1,k,f);//Berechne B=L^{-1}N
Vektor_LA x(G.Zeilenzahl,n,t,q,q_1,k,f);
for(i=0; i<r; i++){
    Hilfsvektor.getSpaltenvektor(N_quadratisch,i);
    Hilfsvektor.VWSubs_mit_Div(G);
    B.Ersetze_Spalte(Hilfsvektor.Vector_LA,i,Hilfsvektor.d);
    }
Matrix_LA Merke_B=B;

if(r!=A.Spaltenzahl)
    B_Hilfe.Kopiere_links(B);

Vektor_LA beta(B.Zeilenzahl-1,n,t,q,q_1,k,f);
Vektor_LA b(B.Zeilenzahl-1,n,t,q,q_1,k,f);

if(r!=A.Spaltenzahl){
    B_Hilfe.QR_Householder_neu_Rechtecksmatrix(b.Vector_LA,beta.Vector_LA);
    B.Kopiere_links(B_Hilfe);
    }
else
    B.QR_Householder_neu(b.Vector_LA,beta.Vector_LA);

R=0;

R.Links_oben_einpassen(B);

Matrix_LA Q(B.Zeilenzahl,B.Zeilenzahl,n,t,q,q_1,k,f);
for(i=1; i<B.Zeilenzahl; i++){//Kopiere die Householder Vektoren aus B in die Matrix Q, da die Funktion Berechne_Q eine quadratische Matrix im Scope voraussetzt
    for(j=0; j<i; j++){
        if(j<r){
            Q.M[i][j]=B.M[i][j];
            }
        }
    }

if(r!=A.Spaltenzahl)
    Q.Berechne_Q_neu(b.Vector_LA,beta.Vector_LA,Iteration);
else
    Q.Berechne_Q_neu(b.Vector_LA,beta.Vector_LA);

for(i=0; i<G.Spaltenzahl; i++){
    Hilfsvektor.getSpaltenvektor(Q,i);
    x.RWSubs(G,Hilfsvektor);
    J.Ersetze_Spalte(x.Vector_LA,i,x.d);
    }

B=Merke_B;
return 0;//Methode korrekt durchlaufen
}

int Aktualisiere_J(Matrix_LA& J, Matrix_LA& R, Matrix_LA& A, Matrix_LA& G, Matrix_LA& B, Matrix_LA& P, Vektor_LA& n_plus_minus, Share& Nummer_NB, Vektor_LA& Nummern_AM, Vektor_LA& Indikatorvektor, Vektor_LA& AM_kurz, int n, int t, int k, mpz_t& q, mpz_t& q_1, int Iteration){//Neuberechnung der Matrizen
//in G ist eine Cholesky-Zerlegung gespeichert
//Ab dim(G)=4 ist es billiger B zu aktualisieren
int i,j,r;

if(Iteration<A.Spaltenzahl){
    r=Iteration;
    }
else{
    r=A.Spaltenzahl;
    }
Matrix_LA B_Hilfe(A.Spaltenzahl,r,n,t,q,q_1,k,f);

Vektor_LA Hilfsvektor=n_plus_minus;//Berechne B=L^{-1}N
Vektor_LA x(G.Zeilenzahl,n,t,q,q_1,k,f);

Hilfsvektor.VWSubs_mit_Div(G);

///Schaltervariablen
Share Hilfe(n,t,q);
Share Hilfe2(n,t,q);
Share Index_Gleich(n,t,q);

Share Minus_Hilfe(n,t,q);
Share Schalter(n,t,q);
Share Nummer_Gleich_Null(n,t,q);
mpz_t eins;
mpz_init_set_ui(eins,1);
Schalter.Erzeugung(eins,t,q);
Vektor_LA HV(B.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA HV2(B.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA Nullvektor(B.Spaltenzahl,n,t,q,q_1,k,f);
int Groesse_Spalten=ceil(log((double)B.Spaltenzahl)/log(2.0));//kleinste 2er-Potenz grer als B.Spaltenzahl

for(i=0; i<B.Spaltenzahl; i++){
    Hilfe=Nummer_NB;
    Nummer_Gleich_Null=Nummern_AM[i];
    Index_Gleich=Nummern_AM[i];
    Nummer_Gleich_Null.EQZ(t,q,q_1,k);//berprfe, ob aktuelle Nummer des Index gleich 0; in diesem Fall wird die erste solche Spalte mit der neuen Spalte besetzt; die anderen bleiben unverndert
    Index_Gleich.EQ(Nummer_NB,t,q,q_1,k);//bergebener Index gleich aktuellem Index der AM?
    Hilfe.LQ(Nummern_AM[i],t,q,q_1,k);//berprfe, ob Index der neuen NB <= einem vorhandenen ist ("=" kommt bei Entfernen der NB zum Tragen)
    Hilfe2=Schalter*Nummer_Gleich_Null;
    Hilfe.OR(Hilfe2);
    Minus_Hilfe=-Hilfe+1;

    ///Spalte und Nummerierung ersetzen
    HV.getSpaltenvektor(B,i);//i-ten Spaltenvekor auslesen
    HV2=Hilfsvektor*Hilfe+HV*Minus_Hilfe;
    B.Ersetze_Spalte(HV2.Vector_LA,i,HV2.d);//Neuen Vektor in Spalte schreiben, WENN Index kleiner ist, als der des vorerigen Eintrags
    Hilfe2=Nummern_AM[i];//Merke den aktuellen Index; er wird fr die Aktualisierung von Nummer_NB bentigt.
    Nummern_AM[i]=Nummer_NB*Hilfe+Nummern_AM[i]*(-Hilfe+1);//Nummerierung der Vektoren einsetzen

    ///bergebenen Vektor mit Index erstellen
    Nummer_NB=Hilfe2*Hilfe*(-Index_Gleich+1)+Nummer_NB*(-Hilfe+1)+Index_Gleich*(A.Zeilenzahl+1);//Merke die Nummer des gemerkten (d.h. in die nchste Iteration hinbergetragenen Vektors) Vektors; wenn eine spalte gelscht wird, wird Nummer_NB auf A.Zeilenzahl+1 gesetzt: auf diese Weise werden die folgenden Spalten nicht verndert und der Inhalt von Hilfsvektor spielt keine Rolle
    Hilfsvektor=HV*Hilfe+Hilfsvektor*Minus_Hilfe;//Merke den nicht geschriebenen Vektor fr die nchste Iteration; er kann dort eingesetzt werden
    Hilfe2=-Index_Gleich+1;
    Hilfsvektor=Hilfsvektor*Hilfe2+Nullvektor*Index_Gleich;//bei Streichen einer Spalte den an die nchste Iteration bergbenen Vektor auf 0 setzen
    Schalter=Schalter*(-Nummer_Gleich_Null+1);//Schalter umlegen, wenn Nummer gleich 0
    }

Matrix_LA PP(B.Spaltenzahl,B.Spaltenzahl,n,t,q,q_1,k,f);
B.Matrix_spaltenweise_zusammenschieben_binaer(AM_kurz.Vector_LA,PP);//falls Spalte gelscht wurde, mu Matrix kompaktifiziert werden
Nummern_AM.MatrixMult(PP,Nummern_AM);

Matrix_LA Merke_B=B;//B merken
if(r<A.Spaltenzahl)
    B_Hilfe.Kopiere_links(B);//QR_householder_neu_Rechtecksmatrix bentigt kleinere Matrix

Vektor_LA beta(B.Zeilenzahl-1,n,t,q,q_1,k,f);
Vektor_LA b(B.Zeilenzahl-1,n,t,q,q_1,k,f);

if(r!=A.Spaltenzahl){
    B_Hilfe.QR_Householder_neu_Rechtecksmatrix(b.Vector_LA,beta.Vector_LA);
    B.Kopiere_links(B_Hilfe);
    }
else
    B.QR_Householder_neu(b.Vector_LA,beta.Vector_LA);

R=0;
R.Links_oben_einpassen(B);

Matrix_LA Q(B.Zeilenzahl,B.Zeilenzahl,n,t,q,q_1,k,f);
for(i=1; i<B.Zeilenzahl; i++){//Kopiere die Householder Vektoren aus B in die Matrix Q, da die Funktion Berechne_Q eine quadratische Matrix im Scope voraussetzt
    for(j=0; j<i; j++){
        if(j<r){
            Q.M[i][j]=B.M[i][j];
            }
        }
    }

if(r!=A.Spaltenzahl)
    Q.Berechne_Q_neu(b.Vector_LA,beta.Vector_LA,Iteration);
else
    Q.Berechne_Q_neu(b.Vector_LA,beta.Vector_LA);

for(i=0; i<G.Spaltenzahl; i++){
    Hilfsvektor.getSpaltenvektor(Q,i);
    x.RWSubs(G,Hilfsvektor);
    J.Ersetze_Spalte(x.Vector_LA,i,x.d);
    }

B=Merke_B;
return 0;//Methode korrekt durchlaufen
}

int Berechne_z_r(Matrix_LA& J, Matrix_LA& R, Vektor_LA& n_p, Vektor_LA& Aktive_Menge_kurz, Vektor_LA& z, Vektor_LA& r, int n, int t, mpz_t& q, mpz_t& q_1, int k, int Iteration){

Matrix_LA J_1=J;
Matrix_LA J_2=J;
Matrix_LA J_t=J;
Vektor_LA d(J.Spaltenzahl, n, t, q, q_1, k, f);

J_t.Transponiere_quadratisch();
n_p.MatrixMult(J_t,d);
d.TruncPr();//D=J^t*n_p

Vektor_LA d_1=d;
d_1.Multipliziere_Zeilenweise(Aktive_Menge_kurz);

J_1.Multipliziere_Spaltenweise(Aktive_Menge_kurz.Vector_LA);
J_2=J-J_1;


/// Explizite Berechnung von H nicht ntig

d.MatrixMult(J_2,z);
z.TruncPr(k,f);

Vektor_LA D_1(R.Zeilenzahl,n,t,q,q_1,k,f);
D_1.Oben_einpassen(d_1);
if(Iteration>3){
    r=0;
    Vektor_LA r_klein(D_1.d,n,t,q,q_1,k,f);
    r_klein.Oben_einpassen(r);
    r_klein.RWSubs(R,d_1);
    r.Oben_einpassen(r_klein);
    }
else{//Fr die erste Berechnung von r wird nur eine Division bentigt
    d_1[0].FPDiv(R.M[0][0],t,q,q_1,k);
    r[0]=d_1[0];
    }

return 0;
}

void Max_G_Norm(Matrix_LA& A, Matrix_LA& G, Vektor_LA& b, Vektor_LA& x, Vektor_LA& I, Vektor_LA& S, Vektor_LA& s, Vektor_LA& N, int n, int t, int k, mpz_t& q, mpz_t& q_1, int Iterationszaehler){
//Bestimme den maximalen Abstand des aktuellen Punktes zu verletzten NBen bzgl. der  G-Norm
int i;

mpz_t minus_unendlich;
mpz_init_set_ui(minus_unendlich,1);
mpz_mul_2exp(minus_unendlich,minus_unendlich,k-1);
mpz_mul_si(minus_unendlich,minus_unendlich,-1);
mpz_add_ui(minus_unendlich,minus_unendlich,1);//minus_unendlich=-2^{k-1}+1

Vektor_LA Z(A.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA Kopie_N=N;
Vektor_LA Kopie_b=b;
Vektor_LA HV(A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA ZV(A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA Lsg(A.Spaltenzahl,n,t,q,q_1,k,f);

Share Hilfe_Z(n,t,q);
for(i=0; i<A.Zeilenzahl; i++){

    Z[i]=s[i];
    Z[i].Quadrat();
    Z[i].TruncPr(k,f,t,q,q_1);

    if(Iterationszaehler<=2){
        ZV.getZeilenvektor(A,i);
        HV=ZV;
        HV.VWSubs_mit_Div(G);
        Lsg.RWSubs(G,HV);
        ZV.SkalarMult(Lsg,N[i]); //Der jeweilige Nenner ist c^tG^{-1}c (G-Norm-Strategie; vgl. Korn)
        N[i].TruncPr(k,f,t,q,q_1);
        Kopie_N[i]=N[i];
        }

    Hilfe_Z=-S[i]+1;
    Z[i]=S[i]*Z[i]+Hilfe_Z*minus_unendlich;//Ersetze bereits aktive NBen durch -unendlich im Zhler.
    Kopie_N[i]=S[i]*N[i]-Hilfe_Z*zwei_hoch_f;//Ersetze Nenner durch 1, falls die NB bereits aktiv ist.// TODO (LocalAdmin#1#): Nenner richten
    }

Z.Max_Bruch(Kopie_N,I);

mpz_clear(minus_unendlich);
}

void Max_Eukl_Norm(Matrix_LA& A, Vektor_LA& b, Vektor_LA& x, Vektor_LA& I, Vektor_LA& AM, int n, int t, int k, mpz_t& q, mpz_t& q_1){
//Bestimme den maximalen Abstand des aktuellen Punktes zu verletzten NBen bzgl. der eukl. Norm
int i;

mpz_t minus_unendlich;
mpz_init_set_ui(minus_unendlich,1);
mpz_mul_2exp(minus_unendlich,minus_unendlich,k-1);
mpz_mul_si(minus_unendlich,minus_unendlich,-1);
mpz_add_ui(minus_unendlich,minus_unendlich,1);//minus_unendlich=-2^{k-1}+1

Vektor_LA Z(A.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA N(A.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA ZV(A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA Kopie_b=b;

Share Hilfe_Z(n,t,q);

for(i=0; i<A.Zeilenzahl; i++){

    ZV.getZeilenvektor(A,i);
    ZV.SkalarMult(x,Z[i]);
    Kopie_b[i].Mul_Konst(zwei_hoch_f,q);
    Z[i]-=Kopie_b[i];
    Z[i].Quadrat();
    Z[i].TruncPr(3*k,3*f,t,q,q_1);

    ZV.SkalarMult(ZV,N[i]); //Der jeweilige Nenner ist c^tG^{-1}c (G-Norm-Strategie; vgl. Korn)
    N[i].TruncPr(k,f,t,q,q_1);

    Hilfe_Z=-AM[i]+1;
    Z[i]=AM[i]*Z[i]+Hilfe_Z*minus_unendlich;//Ersetze bereits aktive NBen durch unendlich im Zhler.
    N[i]=AM[i]*N[i]+(Hilfe_Z*zwei_hoch_f)*zwei_hoch_f;//Ersetze Nenner durch 1, falls die NB bereits aktiv ist.
    }

Z.Max_Bruch(N,I);

mpz_clear(minus_unendlich);
}

void Max_Residuum(Matrix_LA& A, Vektor_LA& b, Vektor_LA& x, Vektor_LA& I, int n, int t, int k, mpz_t& q, mpz_t& q_1){

int i,m=A.Spaltenzahl;
Vektor_LA nn(I.d,n,t,q,q_1,k,f);
Vektor_LA HV(A.Spaltenzahl,n,t,q,q_1,k,f);

for(i=0; i<A.Zeilenzahl; i++){//
    HV.getZeilenvektor(A,i);
    HV.SkalarMult(x,nn[i]);
    nn[i].TruncPr(k,f,t,q,q_1);
    nn[i]-=b[i];
    }

nn.Min(I,A.Zeilenzahl);
}

