#ifndef __FIELD_H
#define __FIELD_H

/**Represents an element in the \f$GF(2^8)\f$ field.
 *
 * \remarks The class name is not quite correct, as the class instances
 * cannot represent the whole field, but rather one element thereof.
 * But then again, the class itself can be considered as a field, which
 * can "contain" all elements as instances.
 */
class GF256 {
	private:
		static const int modulus = 256;

		//static unsigned int inst_counter;
	//	static unsigned char log_table[256];
		//static unsigned char antilog_table[256];

		//void generate_tables();
	public:
        unsigned char value;

		GF256();
		inline ~GF256(){};
		GF256(unsigned char value);

		void operator = (const GF256 &gf);
		void operator = (const int value);
		void operator = (const unsigned char value);

		GF256 operator +(const GF256 &gf);
		GF256 operator -(const GF256 &gf);
		GF256 operator ^(const GF256 &gf);
		GF256 operator *(const GF256 &gf);
		GF256 operator /(const GF256 &gf);

		void operator +=(const GF256 &gf);
		void operator -=(const GF256 &gf);
		void operator ^=(const GF256 &gf);
		void operator *=(const GF256 &gf);
		void operator /=(const GF256 &gf);


		GF256 inverse();
		unsigned char getValue();
};

#endif
