/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#ifndef MATRIX_LA_H
#define MATRIX_LA_H

#include<Share.h>
#include<Matrix_mpf_t.h>


class Matrix_LA//Matrix-Klasse; Objekte sind Matrizen von Sharings
{
    public:
        int Zeilenzahl;
        int Spaltenzahl;

        Share ** M;

        ///Konstruktoren
        Matrix_LA(int a, int b, int n, int t, mpz_t& grosser_Modulus, mpz_t& kleiner_Modulus, int Bitlaenge, int Nachkommastellen);
        Matrix_LA(int a, int b, int n, int t, const mpz_t& grosser_Modulus, const mpz_t& kleiner_Modulus, int Bitlaenge, int Nachkommastellen);

        ///Destruktor
        virtual ~Matrix_LA();

        ///Copy-Konstruktor
        Matrix_LA(const Matrix_LA& other);

        ///Operatoren
        Matrix_LA& operator=(const Matrix_LA& other);
        Matrix_LA operator=(int n);//Initialisiere die Matrix mit einem konstanten Wert (typischerweise 0)
        Matrix_LA& operator=(const Matrix_mpf_t& Init);//Schnittstelle zu Matrix_mpf_t
        Matrix_LA operator+(const Matrix_LA& Matrix_2);//Addition zweier Matrizen; getestet am 12.4.2011
        Matrix_LA& operator+=(const Matrix_LA& Matrix_2);//Addition zweier Matrizen; Ergebnis wird in den Scope geschrieben
        Matrix_LA operator-(const Matrix_LA& Matrix_2);//Subtraktion zweier Matrizen; getestet am 12.4.2011
        Matrix_LA& operator-=(const Matrix_LA& Matrix_2);//Subtraktion zweier Matrizen; getestet am 12.4.2011
        Matrix_LA operator*(const Matrix_LA& A);//Matrizenmultiplikation
        Matrix_LA operator*(const Share& S);
        Matrix_LA& operator*=(const Matrix_LA& A);//Matrizenmultiplikation. Ergebnis wird in den Scope geschrieben
        Matrix_LA& operator*=(int c);

        ///Transponieren von Matrizen und andere Manipulationen
        Matrix_LA Transponiere();//Getestet am 10.6.2011
        void Transponiere_quadratisch();
        void Loesche_links_unten();//Lsche unterhalb der Diagonale
        void Ersetze_Spalte(std::vector<Share>& Spalte, int Spaltenindex, int l);//Ersetze die Spalte Spaltenindex durch Spalte
        void Ersetze_Zeile(std::vector<Share>& Zeile, int Zeilenindex, int l);
        void Normiere_Spaltenweise();//Normiere Spaltenweise auf [-1,1]; nicht in Freigabe enthalten

        ///Erstellung von Matrizen
        void Erzeugung(mpz_t ** Werte);//Belege die Matrix mit den Werten aus "Werte"
        void Erzeugung_int(int ** Werte);//Belege die Matrix mit den Werten aus "Werte"
        void Erzeugung_double(double ** Werte);//Belege die Matrix mit den Werten aus "Werte"

        ///Erstellung besonderer Matrizen
        void Einheitsmatrix(mpz_t& Q);//Erzeuge Einheitsmatrix mit Shares in Z/QZ
        void Einheitsmatrix_Fixpunkt(mpz_t& Q);//Einheitsmatrix skaliert mit 2^f
        void Zufallsmatrix(int k);

        ///Ausgabefunktionen
        void Ausgabe(std::string Name);//Ausgabe der Matrix
        void Ausgabe_skaliert(std::string Name);//skalierte Ausgabe der Matrix (in Fixpunktarithmetik)
        void Ausgabe_skaliert_Matrix(mpf_t** N, int m, int n);//Ausgabe der Werte in N
        int Ausgabe_skaliert_Datei(std::string Name, FILE* datei1);//skalierte Ausgabe in Datei Ausgabe.txt; Rckgabe zeigt an, ob ffnen der Datei erfolgreich

        ///Abschneiden von Nachkommastellen nach Multiplikationen
        void TruncPr(int k, int r);

        ///Zeilen-/Spaltenweise Multiplikationen mit Vektoren
        void Multipliziere_Zeilenweise(std::vector<Share> * Binaerer_Vektor);
        void Multipliziere_Spaltenweise(std::vector<Share>& Binaerer_Vektor);
        void Multipliziere_Zeilenweise_ab_Zeile_X(std::vector<Share> * Binaerer_Vektor, int X);//wie Multipliziere_Zeilenweise; nur erst ab Zeile X

        ///Matrizenmultiplikation
        Matrix_LA MatrixMultMatrix(const Matrix_LA& A);//Matrizenmultiplikation ohnr Trunc; die Rckgabematrix wird in der Methode erzeugt
        Matrix_LA MatrixMultMatrix_ab_Zeile_l(Matrix_LA& A, int l);//Die ersten l Zeilen des ergebnisses sind mit den Ersten l Zeilen von A identisch. Voraussetzung: Scope-Matrix quadratisch
        void Matrix_aus_Vektor(std::vector<Share>& v_Strich,std::vector<Share>& w_Strich, int Laenge_v, int Laenge_w);//Bilde Produkt vw^t

        ///Kopieren aus/in grere Matrizen
        void Rechts_unten_einpassen(Matrix_LA& N, int z, int s);//Matrix N rechts unten in Matrix einpassen
        void Links_oben_einpassen(Matrix_LA& N);//Analog
        void Rechts_unten_auspassen(Matrix_LA& N, int z, int s);//Rechte untere Ecke in kleinere Matrix N kopieren
        void Rechts_einpassen(Matrix_LA& A);//A von rechts in die Scope-Matrix einpassen
        void Kopiere_links(Matrix_LA& L);//Kopiere den linken Teil der Matrix L in die Scope-Matrix
        void Kopiere_rechts(Matrix_LA& L);//Kopiere den rechten Teil der Matrix L in die Scope-Matrix
        void Oben_einpassen(Matrix_LA& A);
        void Unten_einpassen(Matrix_LA& A);

        ///Eliminieren von Nullzeilen/-spalten
        void Matrix_Zusammenschieben(std::vector<Share>& Aktive_Menge_q, Matrix_LA& P);//Eliminiere Leerzeilen in der Scope-Matrix
        void Matrix_Zusammenschieben_LT(std::vector<Share>& Aktive_Menge_q);//Wie Matrix_Zusammenschieben, nur da es ausreicht, wenn in Aktive_Menge_q ein Index unterschritten wird
        void Matrix_Zusammenschieben_binaer(std::vector<Share>& Aktive_Menge_q, Matrix_LA& P);//Aktive_Menge_q wird als binr vorausgesetzt mit 1 an den Stellen, die zu Nullzeilen korrespondieren
        void Matrix_spaltenweise_zusammenschieben_binaer(std::vector<Share>& Aktive_Menge);//wie Matrix_Zusammenschieben_binaer nur spaltenweise
        void Matrix_spaltenweise_zusammenschieben_binaer(std::vector<Share>& Aktive_Menge, Matrix_LA& P);

        ///Normberechnungen
        void Eins_Norm(Share& Norm);//Berechne 1-Norm der Matrix

        ///Lsen von linearen Gleichungssystemen Ax=b
        void Loese_GLS_LR(std::vector<Share>& b, int d);//Loese das Gleichungssystem mit LR-Zerlegung mit Pivotisierung; die Lsung wird in b egschrieben
        void Loese_GLS_QR(std::vector<Share>&b, int d);//Loese das Gleichungssystem mit QR-Zerlegung; die Lsung wird in b egschrieben
        void Loese_GLS_Cholesky(std::vector<Share>& b_Strich, int d);//Loese das GLS mti Hilfe der Cholesky-Zerlegung (geht nur bei s.p.d- Matrizen)
        void CG(std::vector<Share>& b_Strich, std::vector<Share>& x_Strich, int D);//Lse das GlS Ax=b mit Hilfe des CG algorithmus; eine prfung auf s.p.d. der Matrix erfolgt natrlich nicht

        ///Hilfsroutinen zur LR-Zerlegung
        void Vorwaertselimination_SMPC_mit_Pivot(std::vector<Share>& b_Strich);///bei Freigabe entfernen; Vorwrtselimination mit Pivotisierung
        void Vorwaertselimination_SMPC_mit_Pivot(Matrix_LA& P);//Die verwendete Permutationsmatrix wird in P gespeichert
        void Erstelle_Permutationsmatrix(std::vector<Share>& Spalte, int l);//Sucht das betragsmig grte Element in Zeile k oder grer und vertauscht die entsprechende Zeile mit der k-ten

        ///Sicheres Schreiben von Zeilen/Spalten
        void SecWriteSpalte(std::vector<Share>& V, std::vector<Share>& B, int V_d, int B_d);//schreibe V in die in B kodierte Spalte
        void SecWriteZeile(std::vector<Share>& V, std::vector<Share>& B, int V_d, int B_d);//Schreibe V in die durch B kodierte Zeile

        ///Rechnen mit Householder-Vektoren und Matrizen
        void Householder_Matrix(std::vector<Share>& v, int d);///bei Freigabe entfernen; Erstelle Householder-Matrix mit Hilfe des Vektors v;
        void Multipliziere_Householder_Prae(std::vector<Share>& v_Strich, Share& beta, int d);//Multiplikation der Matrix mit dem Householder-Vektor v'
        void Multipliziere_Householder_Prae_beta_bekannt(std::vector<Share>& v_Strich, Share& beta, int d);//Pr-Householder-Multiplikation mit bekanntem beta
        void Multipliziere_Householder_Post_beta_bekannt(std::vector<Share>& v_Strich, Share& beta, int d);//Post-Householder-Multiplikation mit bekanntem beta, vgl. Golub
        void QR_Householder_neu(std::vector<Share>& b, std::vector<Share>& beta);
        void QR_Householder_neu_Rechtecksmatrix(std::vector<Share>& b, std::vector<Share>& beta);
        void QR_Householder_konditional(std::vector<Share>& b, std::vector<Share>& Index);//wie QR_Householder, es werden allerdings nur die in 'Index' gekennzeichneten Vektoren bercksichtigt
        void QR_Householder(Matrix_LA& Q);//QR-Zerlegung nach Householder; R wird im Scope gespeichert; die Householder-Matrix in Q
        void Berechne_Q_konditional(std::vector<Share>& beta, std::vector<Share>& Index);//Wie Berechne_Q; allerdings werden nur die HH-Vektoren bercksichtigt, die in 'Index' vermerkt sind
        void Householder_Matrix_neu(std::vector<Share>& v, int d);
        void Berechne_Q_neu(std::vector<Share>& beta, std::vector<Share>& b);//in b sind die 1. Komponenten der HH-Vektoren gespeichert
        void Berechne_Q_neu(std::vector<Share>& beta, std::vector<Share>& b, int Anzahl_HH_Vektoren);//in b sind die oberen Werte der HH-Vektoren gespeichert; fr den Fall, da weniger als dim Q-1 Vektoren vorgesehen sind

        ///Cholesky-Zerlegung
        void Cholesky();//Berechne Cholesky-Zerlegung; Die untere Dreiecksmatrix wird berschrieben.

    protected:
        int Spielerzahl;
        int Schwellenwert;
        int k;
        int N;
        mpz_t q;
        mpz_t q_1;
    private:

};

#endif // MATRIX_LA_H
