/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#ifndef MATRIX_MPF_T_H
#define MATRIX_MPF_T_H

#include<gmp.h>

//Tools, um die Qualitt der MPC Ergebnisse zu berprfen
class Matrix_mpf_t
{
    public:
        mpf_t ** M;
        int Hoehe, Breite;
        Matrix_mpf_t(int m, int n);
        Matrix_mpf_t(int m, int n, double** MM);
        /** Default destructor */
        virtual ~Matrix_mpf_t();
        /** Copy constructor
         *  \param other Object to copy from
         */
        Matrix_mpf_t(const Matrix_mpf_t& other);
        /** Assignment operator
         *  \param other Object to assign from
         *  \return A reference to this
         */
        Matrix_mpf_t& operator=(const Matrix_mpf_t& other);
        Matrix_mpf_t& operator+=(const Matrix_mpf_t& A);
        Matrix_mpf_t& operator-=(const Matrix_mpf_t& A);
        Matrix_mpf_t& operator=(int n);//n in ale Eintrge der Matrix schreiben
        Matrix_mpf_t operator*(const Matrix_mpf_t& B);//Matrizenmultiplikation
        Matrix_mpf_t& operator*=(double d);//Multiplikation mit Skalar

        void Einheitsmatrix();
        void Zufallsmatrix();
        void InverseMatrix();

        void SpalteEinfuegen(int Spalte, mpf_t* v, int Dimension);
        void Rechts_unten_auspassen(Matrix_mpf_t& N, int z, int s);
        void Rechts_unten_einpassen(Matrix_mpf_t& N, int z, int s);
        Matrix_mpf_t Transponiere();
        void Loesche_links_unten();
        Matrix_mpf_t Matrix_aus_Vektor(mpf_t * H, mpf_t * HG);
        void Kopiere_rechts(Matrix_mpf_t& R);//Rechten Teil der Matrix R in die Scope-Matrix kopieren
        void Kopiere_links(Matrix_mpf_t& L);//Linken Teil der Matrix R in die Scope-Matrix kopieren; neu gegenber Freigabe


        void Norm_Frobenius(mpf_t& Norm);//Berechne die Frobenius-Norm der Matrix
        double Kondition();//Berechne die Kondition anhand der Frobenius-Norm

        void VWElimination(mpf_t * b);
        void GaussAlgorithmus(mpf_t * b);
        void Pivot(mpf_t * b, int k);

        void Ausgabe(std::string Name);
        void Ausgabe_Datei(std::string Name, std::string DateiName); //Ausgabe in DateiName

        void Multipliziere_Householder_Prae(mpf_t * v, int d);
        void QR_Householder(Matrix_mpf_t& Q);
    protected:
    private:
};

#endif // MATRIX_MPF_T_H
