/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#ifndef SVM_H
#define SVM_H

#include <Matrix_LA.h>
#include <Vektor_LA.h>
#include <Matrix_mpf_t.h>


class SVM
{
    public:
        Matrix_LA Daten;

        int Hoehe_Daten;
        int Breite_Daten;

        int Spielerzahl;
        int Schwellenwert;

        mpz_t q;
        mpz_t q_1;

        int k;
        int N;

        Matrix_LA K;//Kernel-Matrix
        Matrix_LA G;//Matrix des Optimierungsproblems
        Vektor_LA Y;//Klassenindikatoren
        Vektor_LA Alpha;//Lsungsvektor fr die Lagrange-Multiplikatoren
        Vektor_LA SV;//Support-Vektoren binr kodiert
        Share b;//Konstanter Term der ges. Hyperflche
        Share Lambda;//Toleranzparameter fr SM-SVM
        Share Koeff;//Faktor bei quadratischem Kernel
        Share r;//Konstanter Term bei quadratischem Kernel

        /** Default constructor */
        SVM(int m, int n, int Anzahl_Spieler, int PolyGrad, mpz_t& Modulus, mpz_t& kleiner_Modulus, int kk, int ff);
        /** Default destructor */
        virtual ~SVM();
        /** Copy constructor
         *  \param other Object to copy from
         */
        SVM(const SVM& other);
        /** Assignment operator
         *  \param other Object to assign from
         *  \return A reference to this
         */

        ///Operatoren
        SVM& operator=(const Matrix_mpf_t& rhs);//ldt in Meber-Variable "Daten"
        SVM& operator=(const SVM& other);

        ///Methoden zur Kernel-Berechnung
        void Linearer_Kernel();
        void Quadratischer_Kernel(Share& Koeff, Share& Konst);
        void Quadratischer_Kernel(double Koeff, double Konst);//ruft QK auf

        ///Berechnung der Matrix des QPP
        void Berechne_G();//Berechne die Matrix des Optimierungsproblems \sum_i,j y_iy_jK_ij

        ///Berechnung der Translation der Hyperebene
        void Berechne_b(); //Berechnung des konstanten Terms in der Entscheidungsregel

        ///Berechnung der Support-Vektoren
        void Berechne_SV();//Berechne SVen

        ///Entscheidungsregeln
        void Entscheidungsregel_linear(Share& Ergebnis, Vektor_LA& x);
        void Entscheidungsregel_Test(std::string Name);//Teste die SVM mit den Trainingsdaten
        void Entscheidungsregel_quadratisch(Share& Ergebnis, Vektor_LA& x);
        void SVM_Test_linear(Matrix_LA& Daten_Test, Vektor_LA& Y_Test,std::string Name);//Klassifiziere die Testdaten und gib den korrekt klassifizierten Prozentsatz an Daten aus
        void SVM_Test_quadratisch(Matrix_LA& Daten_Test, Vektor_LA& Y_Test,std::string Name);//w.o. nur quadatischer Kernel

        ///Speicherung de relevanten Daten auf Festplatte
        void Save_SVM(std::string Dateiname);//Speichern der SVM in Dateien
    protected:
    private:
};

#endif // SVM_H
