/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#ifndef SHARE_H
#define SHARE_H

#include<cstring>
#include<vector>
#include<Vektor.h>
#include<Share_F2M.h>

class Share: public Vektor
{
    public:
        int Anzahl_Spieler;
        int Schwelle;
        mpz_t Modulus;

        Share(int n, int t, mpz_t& q):Vektor(n),Anzahl_Spieler(n),Schwelle(t){
            mpz_init_set(Modulus,q);
            }

        Share(){
            Anzahl_Spieler=0;
            Schwelle=0;
            }

        inline Share (const Share& Ursprungsshare):Vektor(Ursprungsshare.Anzahl_Spieler),Anzahl_Spieler(Ursprungsshare.Anzahl_Spieler),Schwelle(Ursprungsshare.Schwelle){

        mpz_init_set(Modulus,Ursprungsshare.Modulus);

        for(int i=0; i<=Anzahl_Spieler; i++){
            mpz_set(Vector[i],Ursprungsshare.Vector[i]);
            }
        }

        ~Share(){
            Anzahl_Spieler=0;
            Schwelle=0;
        mpz_clear(Modulus);
        }

        inline Share& operator=(const Share& Ursprungsshare){

        int i;

        for(i=0; i<=Anzahl_Spieler; i++){
            mpz_set(Vector[i],Ursprungsshare.Vector[i]);
            }

        mpz_set(Modulus,Ursprungsshare.Modulus);
        return *this;
        }

        Share operator+(const Share& A);
        Share operator+(int c);
        Share operator+(mpz_t& c);
        Share& operator+=(const Share& A);
        Share& operator+=(int c);
        Share& operator+=(mpz_t& c);
        Share operator-(const Share& A);
        Share operator-();
        Share& operator-=(const Share& A);
        Share& operator-=(mpz_t& Subtrahend);
        Share operator-(int c);
        Share operator-(mpz_t& c);
        Share& operator-=(int c);
        Share operator*(const Share& A);
        Share& operator*=(const Share& A);
        Share& operator*=(mpz_t& c);
        Share& operator*=(int c);
        Share operator*(mpz_t& a);
        Share operator*(int n);
        Share& operator=(int n);//Setze ein Sharing konstant auf n
        Share& operator=(mpz_t n);//Setze ein Sharing konstant auf n

        ///Arithmetische Operationen
        void Quadrat();
        void Betrag(mpz_t& q_1, int k);
        void Signum(int t, mpz_t& q, mpz_t&q_1, int k);
        void Mul(mpz_t *b, int t, mpz_t& q); //Multiplikation zweier verteilter Werte; Multiplikationsprotokoll von Peter mit beiden Verbesserungen; getestet am 19.1.2011; //Ergebnis der Mutliplikation wird in den Share-Vektor des Scopes geschrieben
        void Mul_GRR(mpz_t * b, int t, mpz_t& q);//GRR-Multiplikation zum Vergleich; Das Ergebnis mu nach der Berechnung rekonstruiert werden; getestet am 19.1.2011
        void MUL_Lory3(mpz_t *b, int t, mpz_t& q);//Multiplikation mit rekursiv berechneten Lagrange-Koeffizienten
        void MulPub(Share& b, mpz_t& Ausgabe);//Multiplikation mit anschlieender Ausgabe; getestet am 2.5.2011
        void FPDiv(const Share& Nenner, int t, mpz_t& q, mpz_t& q_1, int k);//Division; (k,f)-Fixpnktzahlen werden vorausgesetzt
        void DivNR(Share& Nenner, mpz_t& q_1, int k);//Division nach Newton-Raphson
        void Inverses(mpz_t& q_1, int k);//1/*this
        void Inverses_ohne_Trunc(mpz_t& q_1, int k);//Mu nach weiterer Multiplikation TruncPr(3k-f,2k-f) ausfhren!
        void Wurzel(int k, mpz_t& q, mpz_t& q_1);//Berechnung der Wurzel nach Goldschmidt
        void Div_Konst(mpz_t& d, int t, mpz_t& q, int k, mpz_t& q_1);//Division durch eine Konstante
        void Div_Konst_int(int d, mpz_t& q, mpz_t& q_1, int k);
        void Invertiere();//Invertiere den Share mod q; getestet am 6.5.2011

        inline void Mul_Konst(mpz_t c, mpz_t q){//Multiplikation mit Konstanten (mpz_t); Getestet am 19.1.2011
        register int i;
        for (i=1; i<=Anzahl_Spieler; i++){
            mpz_mul(Vector[i],Vector[i],c);
            mpz_mod(Vector[i],Vector[i],q);
            }
        }

        inline void Mul_Konst_int(int c, mpz_t q){//Multiplikation mit integer; getestet am 26.1.2010
            mpz_t Konstante;
            mpz_init_set_si(Konstante,c);
            Mul_Konst(Konstante, q);
            mpz_clear(Konstante);
        }

        inline void Add_Konst(mpz_t& Konstante, mpz_t& q){//Addieren einer Konstante zu einem Sharing mit n Teilnehmern; getestet am 20.1.2011
        register int i;
        for (i=0; i<=Anzahl_Spieler; i++){
            mpz_add(Vector[i],Vector[i],Konstante);
            mpz_mod(Vector[i],Vector[i],q);
            }
        }

        inline void Add_Konst_int(int c, mpz_t& q){
            mpz_t k;
            mpz_init_set_si(k,c);
            Add_Konst(k,q);
            mpz_clear(k);
            }

        ///Erzeugung, Rekonstruktion und Ausgabe von Shares
        inline void print(std::string shareName, int n){/* Funktion zum Ausgeben eines Shares */
            char * Name;
            Name=new char[shareName.size()+1];
            strcpy(Name,shareName.c_str());
            printf("\n%s = \n", Name);
            for (int i=0; i<=n; i++){
                gmp_printf("(%d,%Zd)\n",i,Vector[i]);
                }
        }

        void Erzeugung(mpz_t& s, int t, mpz_t& q);//Konstruktion eines Sharings
        void Erzeugung_double(double d, int t, mpz_t& q);//Erzeugung auf Grundlage einer double
        void Rekonstrukt(mpz_t& s, int t, mpz_t& q);//Rekonstruktion eines verteilten Werts
        void Rekonstrukt_Ausgabe(std::string ShareName);//Rekonstruktion mit Ausgabe; die Skalierung der Fixpunktarithmetik wird herausgerechnet
        void Rekonstrukt_Ausgabe(std::string ShareName, int n);
        void Rekonstrukt_GRR(mpz_t& s, mpz_t& q, int t);//Rekonstruktion mit Hilfe der Lagrange-Interpolation
        void Rekonstrukt_GRR_mit_VZ(mpz_t& s, mpz_t& q, int t);
        void Rekonstrukt_Ausgabe_skaliert_hexadezimal(std::string ShareName);//Rekonstruktion mit Ausgabe; die Skalierung der Fixpunktarithmetik wird herausgerechnet
        void Rekonstrukt_Ausgabe_unskaliert(std::string ShareName);//Rekonstruktion mit unskalierter Ausgabe - geeignet fr Bit-Ausgaben
        void Rekonstrukt_Ausgabe_unskaliert(std::string ShareName, int n);
        void Rekonstrukt_Ausgabe_unskaliert(std::string ShareName, mpz_t& p);//Primzahl mitgeben
        void Rekonstrukt_Spieler_beliebig(mpz_t& s, int * Spielerliste, int t, mpz_t& q);//Rekonstruktion an der Stelle 0 unter Verwendung beliebiger Spieler; getestet am 7.2.2011
        void Rekonstrukt_Spieler_beliebig_Stelle_beliebig(mpz_t& s, int * Spielerliste, int t, int r, mpz_t& q);//Rekonstruktion eines beliebigen Werts eines Spielers; getestet am 8.2.2011
        int Ausgabe_skaliert_Datei(std::string ShareName, FILE * datei);//Ausgabe in Datei; diese muss vorher geffnet werden

        ///Erzeugung von Zahlen und Zufallszahlen
        void Zufallszahl(int t, mpz_t& q);//Erzeugen einer verteilten Zufallszahl; getestet am 20.1.2011
        void PRandInt(int t, int Laenge, mpz_t& q);// vgl. "Secure Computation with Fixed-Point Numbers" PRSS simuliert.
        void PRandBitD(Share_F2M& F2M_Bit, int t, mpz_t& q, mpz_t& q_1);//vgl. SecureSCM-D.9.2
        void RandBit(int t, mpz_t& q);//Erzeugung eines zuflligen, verteilten Bits in Z_q; getestet am 20.1.2011

        inline void gemeinsame_1(mpz_t& q){//gemeinsame Erzeugung der 1
            mpz_t eins;
            mpz_init_set_ui(eins,1);
            Erzeugung(eins,Schwelle,q);
            mpz_clear(eins);
            }

        inline void gemeinsame_0(mpz_t& q){//gemeinsame Erzeugung der 0
            mpz_t null;
            mpz_init_set_ui(null,0);
            Erzeugung(null,Schwelle,q);
            mpz_clear(null);
            }

        ///Hilfsfunktionen fr Division, Wurzel und Multiplikation
        void AppRcr(int k, int t, mpz_t& q, mpz_t& q_1);//Ursprngliche Approximation von 1/b durch 2.9142-2*c; c ist auf (0.5,1) normalisiertes b
        void Norm(Share& v, int k, int t, mpz_t& q, mpz_t& q_1);//Normiere Werte auf Intervall (0.5,1)
        void Norm_Find_First(Share& v, int k, int t, mpz_t& q, mpz_t& q_1);
        void Norm_mit_Wurzel(Share& v, Share& SQ, Share& Wert, int k, int t, mpz_t& q, mpz_t& q_1);//Variante von Nom zur Berchnung der Wurzel
        void RecItNR(int k, mpz_t& q_1);//Hilfsprotokoll dazu;
        void RandNumbers(mpz_t **randNumbers, int t, mpz_t& q, mpz_t *a, mpz_t *b);//Hilfsfunktion zum Erzeugen von zuflligen Funktionswerten bei der Neuverteilung bei der Mutliplikation; getestet mit Multiplikationsprotokoll

        ///Umwandlung von Sharings
        void Umwandlung(int t, mpz_t& q, mpz_t& q_1);//Umwandlung von shares modulo q_1 in Shares modulo q mittels RISS
        void Umwandlung_Bit_Zq_F2M(Share_F2M& r_F2M, Share * RISS_q, mpz_t ** RKq, int t, mpz_t& q, mpz_t& z);//Umwandung eines Zufallsbits von Z_q nach F_2m; getestet am 30.3.2011
        void F2MtoZq(mpz_t& q, mpz_t& q_1, Share_F2M& Bit, int t);//Umwandlung des Shares im Scope in einen Share A in Z/qZ

        ///Vergleiche

        void BitLT_F28(int t, int k, mpz_t& Zahl_1, std::vector<Share>& Bits, mpz_t& q_1);//Wie BitLT, aber die Operationen werden in F2M druchgefhrt
        void LTZ(int t, mpz_t& q, mpz_t& q_1, int k);//<?0
        void LTZ_F28(int t, mpz_t& q, mpz_t& q_1, int k);
        void LT(Share& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k);//(a<b)? 1:0;getestet am 20.4.2011
        void LT(int c, int t, mpz_t& q, mpz_t& q_1, int k);
        void LT(mpz_t& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k);//*this<?Argument_1 1:0
        void LQ(Share& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k);//<=?; nach Freigabe hinzugefgt
        void LT_Konst_Betrag(mpz_t& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k);//Vergleich, ob Quadrat kleiner als "Argument"
        void GTZ(int t, mpz_t& q, mpz_t& q_1, int k);//wie LTZ
        void GT(Share& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k);//(a>b)? 1:0
        void GT(mpz_t& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k);//(a>b)? 1:0
        void GTZ_approximativ(int t, mpz_t& q, mpz_t& q_1, int k);//berprfe, ob Wert grer als 2^{-f+10} ist
        void LTZ_approximativ(int t, mpz_t& q, mpz_t& q_1, int k);//berprfe, ob Wert kleiner als 2^{-f+10} ist
        void GQ(int t, mpz_t& q, mpz_t& q_1, int k);//(a>=0)? 1:0
        void EQZ(int t, mpz_t& q, mpz_t& q_1, int k);//(a=0)? 1:0; getestet am
        int EQZPub();
        void EQ(Share& Argument, int t, mpz_t& q, mpz_t& q_1, int k);//Vergleich mit Argument
        void EQ_mpz(mpz_t& Argument, int t, mpz_t&q, mpz_t& q_1, int k);
        void EQ_int(int Wert, int t, mpz_t& q, mpz_t& q_1, int k);//Vergleich mit "Wert"

        ///Logische Operatoren
        void OR(Share& Argument);//OR von Bit-Shares
        void AND(Share& Argument);//AND von Bit-Shares

        void min(const Share& a, const Share& b, mpz_t& q, mpz_t& q_1, int t, int k, Share& Hilfe);//In "Hilfe" wird der Index des minimierenden Elements zurckgegeben: Hilfe=1<=>a<b; 0 sonst

        ///Abschneidefunktionen
        void TruncPr(int k, int m, int t, mpz_t& q, mpz_t& q_1);//Abschneiden berzhliger Stellen nach Multiplikation; getestet am 25.1.2011; probabilistisch
        void Trunc_F28(int t, mpz_t& q, mpz_t& q_1, int k, int m);//Trunc Function; bentigt fr LTZ
        void Mod2m_F28(int k, int m, mpz_t& q, mpz_t& q_1, int t);
        void Mod2(int k, mpz_t& q, mpz_t& q_1, int t);

        ///Hilfsfunktionen fr Rundung
        void CarryOutL_1offen(mpz_t& Argument_1, std::vector<Share>& Argument_2, int t, mpz_t& q, int k);
        void CarryOutL_cin_1offen(mpz_t& Argument_1, std::vector<Share>& Argument_2, int t, mpz_t& q, int k, int carry_in);//Wie CrryOutL_1offen nur mit zustzlichem Input-Bit
        void CarryOutAux(std::vector<Share>& d_1, std::vector<Share>& d_2, int k, int t, mpz_t& q);//Hilfprogramm
        void CarryOutL_cin_1offen_F2M(mpz_t& Argument_1, std::vector<Share_F2M>& Argument_2, int t, int k, int carry_in, mpz_t& q_1);//Wie CarryOutL_cin_1offen, aber mit Operationen in F2M
        void CarryOutAux_F2M(std::vector<Share_F2M>& d_1, std::vector<Share_F2M>& d_2, int k, int t);//Wie CarryOutAux nur mit Operationen in F28

        ///Bitzerlegung und Hilfsfunktionen
        void BitDec(std::vector<Share_F2M> * Bits, int k, int m, int t, mpz_t& q, mpz_t& q_1);//Erzeugung eines Bits sowohl in Z/qZ als auch in F_2^{8}; ersteres wird in den Scope-Vektor geschrieben; getestet am 5.4.2011
        void KOrCL(std::vector<Share> * Bits, int k, int t, mpz_t& q, mpz_t& q_1);//k-faches Oder der Sharings in Bits
        void KOrCS(std::vector<Share>& Bits, int k, int t, mpz_t& q);//Hilfsroutine fr KOrCL

    protected:
    private:
};

#endif // SHARE_H
