/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#ifndef SIMPLEX_STARTWERT_CATRINA_H_INCLUDED
#define SIMPLEX_STARTWERT_CATRINA_H_INCLUDED

#include <Matrix_LA.h>
#include <Vektor_LA.h>


class Simplex_Startwert_Catrina : public Matrix_LA
{
    public:
        /** Default constructor */
        Simplex_Startwert_Catrina(int m, int n, int Anzahl_Ungleichheitsbedingungen, int Anzahl_Spieler, int Schwelle, mpz_t& qmod, mpz_t& q_1mod, int kl, int ffl);
        /** Default destructor */
        virtual ~Simplex_Startwert_Catrina();
        /** Copy constructor
         *  \param other Object to copy from
         */
        Simplex_Startwert_Catrina(const Simplex_Startwert_Catrina& other);
        /** Assignment operator
         *  \param other Object to assign from
         *  \return A reference to this
         */
        Simplex_Startwert_Catrina& operator=(const Simplex_Startwert_Catrina& other);

        void Tableau_Erzeugen(Matrix_LA& A_alt, Vektor_LA& b, Vektor_LA& b_Signum, Vektor_LA&F, Vektor_LA& x0, int Anzahl_Gleichheitsbedingungen, int Anzahl_Ungleichheitsbedingungen);
        int Simplex_Algorithmus(Matrix_LA& A_alt, Vektor_LA&b, Vektor_LA& c, Vektor_LA& x0, Vektor_LA& X, int Anzahl_Gleichheitsbedingungen, int Anzahl_Ungleichheitsbedingungen, Vektor_LA& B, Vektor_LA& N);

        void UpdTab(Vektor_LA& C, Vektor_LA& R, Vektor_LA& V, Vektor_LA& W, Share& p);//Update des Tableaus
        int GetPivColumn(Vektor_LA& V);
        int GetPivRow(Vektor_LA& C, Vektor_LA& W);//D Eingabe-, W Rckgabevektor. Eingabevektor nicht als referenz, da im algorithmus manipuliert
        void UpdVar(Vektor_LA& B, Vektor_LA& N, Vektor_LA& V, Vektor_LA& W);//Aktualisieren von Basis- und Nichtbasisvariablen
        void GetSolution(Vektor_LA& B, Vektor_LA& X);
        void UpdCost(Vektor_LA& B, Matrix_LA& A, Vektor_LA b, int d);//Aktualisiere die Kostenfunktion
        void TransformCost(Vektor_LA& W, Matrix_LA& A, Vektor_LA& b_Signum);//Stelle die aktualisierte Kostenfunktion in Abhngigkeit von der Menge N dar
        void TransformCost(Matrix_LA& A, Vektor_LA& b_Signum);//Stelle die aktualisierte Kostenfunktion in Abhngigkeit von der Menge N dar; Vektor C_B wird nicht mit dem Signum von b multipliziert
    protected:
    private:
        int Tableau_Zeilen;
        int Tableau_Spalten;
        int Spielerzahl;
        int Schwellenwert;
        mpz_t q;
        mpz_t q_1;
        int k;
        int ff;
};

#endif // SIMPLEX_STARTWERT_CATRINA_H_INCLUDED
