/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#ifndef VEKTOR_LA_H
#define VEKTOR_LA_H

#include <string>
#include <Share.h>
#include <Matrix_LA.h>
#include <Vektor_mpf_t.h>

class Vektor_LA
{
    public:
        std::vector<Share> Vector_LA;
        int d;//Dimension

        ///Konstruktor
        Vektor_LA(int Laenge, int n, int t, mpz_t& grosser_Modulus, mpz_t& kleiner_Modulus, int Bitlaenge, int Nachkommastellen);

        ///Standard-Konstruktor
        Vektor_LA(){
        d=0;
        mpz_init(q);
        mpz_init(q_1);
        }

        ///Copy-Konstruktor
        Vektor_LA(const Vektor_LA& other);

        ///Destruktor
        virtual ~Vektor_LA();

        ///Operatoren
        Vektor_LA& operator=(const Vektor_LA& other);
        void operator=(int c);//Vektor von konstanten Sharings initialisieren
        Vektor_LA& operator=(const Vektor_mpf_t& Init);//Schnittstelle zu Vektor_mpf_t.h
        Vektor_LA operator+(const Vektor_LA& VLA);//Vektoraddition
        void operator+=(const Vektor_LA& VLA); //Vektoraddititon; Ergebnis wird in den Scope geschrieben
        Vektor_LA operator-(const Vektor_LA& VLA);//Vektorsubtraktion
        void operator-=(const Vektor_LA& VLA);//Vektorsubtraktion; Ergebnis wird in den Scope geschrieben
        Vektor_LA operator*(mpz_t& Skalar);//Multiplikation mit Skalar
        Vektor_LA operator*(int c);//Multiplikation mit int-Zahl
        Vektor_LA& operator*=(mpz_t& Skalar);//Multiplikation mit Skalar; Ergebnis wird in den Scope geschrieben
        Vektor_LA& operator*=(int c);//Multiplikation mit Skalar; Ergebnis wird in den Scope geschrieben;
        Vektor_LA operator*(Share& Faktor);//Multiplikation mit verteiltem Faktor
        Vektor_LA& operator*=(Share& Faktor);//Multiplikation mit verteilten Faktor; Ergebnis wird in den Scope geschrieben
        Share operator[](const Vektor_LA& Index);//Zugriff auf das in Index binr gespeicherte Element
        Share& operator[](int i);//Zugriffsoperator

        ///Sonstige elementare Operationen
        void Unskaliert_einlesen(const Vektor_mpf_t& rhs);//wie =const Vektor_mpf_t& rhs nur unskaliert
        void Summe_Vektor(Share& Summe, mpz_t& Modulus);//Summiere die Komponenten des Vektors auf und fhre die Modulo-Reduktion erst am Schluss durch

        ///Vergrern, Verkleinern und Umsortieren von Vektoren
        void Kopiere_unten(Vektor_LA& V);//den Vektor V von unten in den Scope schreiben;
        void Oben_einpassen(Vektor_LA& V);//Analog
        void Umdrehen();//Dreht die Reihenfolge des Vektors um
        void Concat_Vektor(Vektor_LA * a, Vektor_LA *b);//Vektoren konkatenieren

        ///Vergleich von Vektoren
        void Gleich_Null(Share& Bit);//Berechne Gleichheit mit Null anhand der euklidischen Norm
        void Gleich_Null_approximativ(Share& Bit);//Prfe ob, alle Komponenten nicht weiter als 2^-10 von Null entfernt sind

        ///Matrix- und Skalarmultiplikationen
        void Multipliziere_Zeilenweise(Vektor_LA& Binaerer_Vektor);
        void MatrixMult(const Matrix_LA& A, Vektor_LA& Ergebnis_Vektor);//Matrixmultiplikationmit anschlieendem TruncPr; Ergebnis wird in Ergebnis_Vektor geschrieben
        void MatrixMult_ab_Zeile(const Matrix_LA& A, Vektor_LA& Ergebnis_Vektor, int m, int n, int Zeilenzahl);
        void SkalarMult(const Vektor_LA& b, Share& Ergebnis);//Skalarprodukt; Ergebnis wird in "Ergebnis" geschrieben

        ///Auslesen von (Teil-)zeilen und (Teil-)spalten
        void getZeilenvektor(const Matrix_LA& Ma, int Zeilenindex);//Zeile Zeilenindex in Vector_LA kopieren
        void getZeilenvektor_ab_Spalte(const Matrix_LA& Ma, int Zeilenindex, int Spaltenindex);//Zeilenvektor ab Position Spaltenindex in den Scope kopieren
        void getSpaltenvektor(const Matrix_LA& Ma, int Spaltenindex);//Spalte Spaltenindex in Vector_LA kopieren
        void getSpaltenvektor_ab_Zeile(const Matrix_LA& Ma, int Spaltenindex, int Zeilenindex);//Eintrge ab Zeile 'Zeilenindex' in Spalte 'Spaltenindex' in Vektor kopieren
        void getSpaltenvektor_ab_Zeile(Vektor_LA& b, int Zeile);//Teilspalte aus Vektor auslesen
        void getSpaltenvektor_bis_Zeile(Vektor_LA& b, int Zeile);
        void getZeilenvektor_bis_Spalte(const Matrix_LA& Ma, int Zeilenindex, int Spaltenindex);//Teilzeile aus Matrix auslesen
        void getSpaltenvektor_bis_Zeile(const Matrix_LA& Ma, int Spaltenindex, int Zeilenindex);//Teilspalte aus Matrix auslesen

        ///Ausgabefunktionen
        void Ausgabe(std::string Name);//Ausgabe des Vektors
        void Ausgabe_skaliert(std::string Name);//wie "Ausgabe"; Skalierung mit 2^f herausgerechnet
        void Ausgabe_skaliert_Vektor_mpf_t(Vektor_mpf_t& V);//Ausgabe nach Vektor_mpf_t
        void Ausgabe_Shares(std::string Name);//Ausgabe der Shares
        int Ausgabe_skaliert_Datei(std::string Name, FILE* datei1);//Ausgabe in Datei
        int Ausgabe_Datei(std::string Name, FILE* datei1);
        void Ausgabe(std::string Name, mpz_t& p);

        ///Abschneiden berzhliger Stellen
        void TruncPr();//Komponentenweises Abschneiden des Vektors mit Hilfe des TruncPr-Protokolls
        void TruncPr(int l, int r);//Komponentenweises Abschneiden des Vektors mit Hilfe des TruncPr-Protokolls,; Lnge der Zahl: k Bits, davon f Nachkommastellen

        ///Vorwrts- und Rckwrtssubstitutionen; Subfunktionen
        void VWSubs(const Matrix_LA& L);//Innere Schleife zu Skalarprodukten zusammengefasst
        void VWSubs_mit_Div(const Matrix_LA& L);//Diagonalelement !=1
        void RWSubs(const Matrix_LA& A, Vektor_LA& b);//fasse die innere Schleife zu Skalarpoduk zusammen
        void getPivot(Vektor_LA& Index);//Pivotelement bestimmen

        ///Geometrische Funktionen
        void Norm(Share& N);//Normberechnung mit Hilfe des Goldschmidt-Algorithmus
        void Abstand_zu_Ebene(Vektor_LA& Hesseform, Share& b, Share& Abstand);//Berechnet den Abstand des Scope-Vektors zu der in HNF kodierten Ebene; b steht auf der rechten Seite der HNF

        ///Funktionen zur Berechnung von und mit Householder-Vektoren
        void house_neu();
        void Mul_house_Pre_beta_bekannt(Vektor_LA& v, Share& beta);//Pr-Multiplikation mit dem Householder-Vektor v und dem bekannten beta; analog zur selben Funktion in Matrix_LA
        void Qt_mult_b(Vektor_LA& beta, Matrix_LA& M, int d);//Multipliziere den Scope-Vektor mit den, aus den in Matrix M unterhalb der Diagonalen gespeicherten Householder-Vektoren abgeleiteten Householder-Matrizen
        void Qt_mult_b_neu(Vektor_LA& beta, Vektor_LA& beta_HH, Matrix_LA& M, int d);

        ///Manipulation von Index-Vektoren
        void binaerer_Index(Share * Index, int Obergrenze);//Wandelt den in "Index" gespeicherten Index in einen binren Vektor um, der im Scope gespeichert wird
        void erhoehe_Index();//Schiebt alle Elemente im Scope-Vektor um 1 nach oben; ist dieser ein binrer Indikator-Vektor wird so der Index um 1 erhht
        void erhoehe_Index_konditional(Share& Bedingung, mpz_t& q);//wie "erhhe_Index" nur unter der Bedingung "Bedingung"

        ///Bestimmung von maximalen und minimalen Elementen
        void Max_Neu(Vektor_LA& Index, int K);//Bestimme das Maximum des Scope-Vektors und speichere den Index binr in "Index"; der Scope-Vektor wird verndert; vgl. Protocol 13 von Tofts Dissertation
        void Max_Bruch(Vektor_LA& B, Vektor_LA& Index); //Wie Max nur da Brche betrachtet werden
        void Min_Bruch(Vektor_LA& C, Vektor_LA& W);//basiert auf Max_Bruch
        void Min(Vektor_LA& Index, int K);//Bestimme das Minimum durch: Min(x)=Max(-x)
        void Max_Betrag(Vektor_LA& Index);//Index des betragsmig grten Werts identifizieren; nich tin Freigabe-Version enthalten

        ///Logische Funktionen
        void AND(Vektor_LA& Argument_2);//Komponentenweises AND
        void OR(Vektor_LA& Argument_2);//OR von verteilten (Bit-)Vektoren;
        void Pre_Mul_C();//PreMulC nach Protocols 4.2 in: "Improved Primitives for Secure Multiparty Integer Computation"
        void PreOrC();
        void KOR(Share& Ausgabe, int r, mpz_t& p);//Standard-Algorithmus

        ///Eliminieren von Nullzeilen
        void Vektor_Zusammenschieben(std::vector<Share>& Aktive_Menge_q, Matrix_LA& P);//Entferne Nullelemente aus Vektor; P ist zugehrige Permutationsmatrix
        void Vektor_Zusammenschieben_binaer(std::vector<Share>& Aktive_Menge_q, Matrix_LA& P);//Wie Vektor_Zusammenschieben; die aktiven Elemente sind bereits binr kodiert in Aktive_Menge_q

        ///Verdecktes Lesen und Schreiben von Elementen
        void SecReadRow(Matrix_LA& A, Vektor_LA& I,  int l);//sicheres Auslesen einer versteckten Zeile
        void SecWrite(Vektor_LA& V, Share& s);
        void Spalte_auslesen(Matrix_LA& A, Vektor_LA& I, int l);///bei Freigabe entfernen

        ///Bestimmung des ersten Nicht-Null-Elements
        void SelectFirst();//vgl. Secure SCM D3.2
        void FindFirst(Vektor_LA& Index);

        ///Funktionen zum Umwandeln von Zahlen in binre Vektoren
        void BitMask(Share& x);//vgl. Secure SCM D3.2, Variante

    protected:
    private:
        int Spielerzahl;
        int Schwellenwert;
        int k;
        int N;
        mpz_t q;
        mpz_t q_1;
};

#endif // VEKTOR_LA_H
