/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#ifndef VEKTOR_MPF_T_H
#define VEKTOR_MPF_T_H

#include <gmp.h>
#include <Matrix_mpf_t.h>


class Vektor_mpf_t
{
    public:
        int Dimension;
        mpf_t * V; //!< Member variable "V"
        /** Default constructor */
        Vektor_mpf_t(int d);
        Vektor_mpf_t(int d, double*w);
        Vektor_mpf_t(int d, mpf_t*w);
        /** Default destructor */
        virtual ~Vektor_mpf_t();
        /** Copy constructor
         *  \param other Object to copy from
         */
        Vektor_mpf_t(const Vektor_mpf_t& other);
        /** Assignment operator
         *  \param other Object to assign from
         *  \return A reference to this
         */
        Vektor_mpf_t& operator=(const Vektor_mpf_t& other);
        /** Access V
         * \return The current value of V
         */
        Vektor_mpf_t operator+(const Vektor_mpf_t& W);
        Vektor_mpf_t& operator+=(const Vektor_mpf_t& W);
        Vektor_mpf_t operator-(const Vektor_mpf_t& W);
        Vektor_mpf_t& operator-=(const Vektor_mpf_t& W);
        Vektor_mpf_t& operator*=(mpf_t& Faktor);
        void operator=(int n);

        void SkalarProdukt(Vektor_mpf_t& v2, mpf_t& Ergebnis);
        void MatrixMultVektor(Matrix_mpf_t& M, Vektor_mpf_t& Ergebnis);

        void Zufallsvektor();

        double Norm();

        void SetV(mpf_t * val) { V = val; }
        void getSpaltenvektor(int Spalte, Matrix_mpf_t& A); //Spaltenvektor auslesen
        void getSpaltenvektor_ab_Zeile(int Spalte, Matrix_mpf_t& A, int Zeile);//Spaltenvektor ab Zeile "Zeile" auslesen

        void VWSubs(Matrix_mpf_t& L);
        void RWSubs(Matrix_mpf_t& L);

        void Ausgabe(std::string Name);
        void Ausgabe_wissenschaftlich(std::string Name);//in exponentieller Notation
        void Ausgabe_Datei(std::string Name, FILE* datei);
        void Ausgabe_wissenschaftlich_Datei(std::string Name, FILE* datei);//Ausgabe in Datei

        int Max();

        void Householder_Vektor();
    protected:
    private:

};

#endif // VEKTOR_MPF_T_H
