/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#ifndef HEAD_H_INCLUDED
#define HEAD_H_INCLUDED

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <unistd.h>
#include <GF256.h>
#include <Share_F2M.h>
#include <Vektor_LA.h>
#include <Matrix_LA.h>

#define Modullaenge 1024
#define Modullaenge_kurz 64

#define f 64
#define kappa 10
#define precision 500 //Przision fr Rechnen mit float

extern unsigned char antilog_table[256];
extern unsigned char log_table[256];
extern mpz_t alpha_division;
extern mpz_t C_0_wurzel;
extern mpz_t C_1_wurzel;
extern mpz_t zwei_hoch_k_hoch_minus_1;
extern mpz_t zwei_hoch_f;
extern mpz_t halb;
extern mpz_t eins_Wurzel;
extern mpz_t Wurzel_zwei_hoch_f;
extern mpf_t Wurzel_2;
extern mpz_t Wurzel_2_z;
extern mpz_t Drei;
extern mpz_t invers_zwei;
extern mpz_t rem;
extern mpz_t rem_q_1;
extern int **Teilmengen;
extern mpz_t * lambda;
extern mpz_t * lambda_Rekonstrukt;
extern GF256 * lambda_256;
extern GF256 * lambda_256_Rekonstrukt;
extern int Theta_wurzel;
extern int Theta_Division;
extern mpz_t ** RISS_Koeffizienten_q1;
extern mpz_t ** RISS_Koeffizienten_q;
extern mpz_t * RISS;
extern GF256 ** RISS_Koeffizienten_F2M;
extern gmp_randstate_t rState1;
extern mpz_t * r_RISS;
extern int n_ueber_t;
extern mpz_t grosser_Modulus;
extern mpz_t kleiner_Modulus;

namespace haupt{

inline void Erzeuge_ThetaWurzel(int k);
inline void Erzeuge_TruncPr_Koeff(int d, mpz_t& q);
inline void Initialisiere_Zufallszahlen(mpz_t& q);
inline void Initialisiere_Zufallszahlen_F2M();
inline void Erzeuge_Theta_Division(int k);
inline void generate_tables() ;
inline void Erzeuge_alpha_Division(int k);
inline void Erzeuge_Koeff_Wurzel(int l);
inline void Berechne_n_ueber_t(int n, int t);
inline void nt_Teilmengen(int n, int t);
inline void Berechne_RISS_Koeffizienten_q(int Anzahl_Spieler, mpz_t q);
inline void Berechne_RISS_Koeffizienten_q1(int Anzahl_Spieler, mpz_t q_1);
inline void Berechne_RISS_Koeffizienten_F2M(int Anzahl_Spieler);
inline void LagIntKoeff(int t);
inline void LagIntKoeff_Rekonstrukt(int t);
inline void LagIntKoeff_GF256(int n, int t);
inline void LagIntKoeff_GF256_Rekonstrukt(int t);
inline void Berechne_2_hoch_f();
inline void Berechne_Wurzel_Konstanten();
inline void Berechne_rem(mpz_t& q, mpz_t& q_1);
inline void Waehle_Modulus(mpz_t& Q, mpz_t& rem);


inline void Berechne_Koeffizienten(int n, int t, int k, int d, mpz_t& q, mpz_t& q_1){

Erzeuge_ThetaWurzel(k);
Erzeuge_TruncPr_Koeff(d,q);
Initialisiere_Zufallszahlen(q);
Initialisiere_Zufallszahlen_F2M();
Erzeuge_Theta_Division(k);
generate_tables();
Erzeuge_alpha_Division(k);
Erzeuge_Koeff_Wurzel(k);
Berechne_n_ueber_t(n,t);
nt_Teilmengen(n,t);
Berechne_RISS_Koeffizienten_q1(n,q_1);
Berechne_RISS_Koeffizienten_q(n,q);
Berechne_RISS_Koeffizienten_F2M(n);
LagIntKoeff(t);
LagIntKoeff_Rekonstrukt(t);
LagIntKoeff_GF256(n,t);
LagIntKoeff_GF256_Rekonstrukt(t);
Berechne_2_hoch_f();
Berechne_Wurzel_Konstanten();
Berechne_rem(q,q_1);
}

inline void Diskrete_Wurzel(mpz_t * w, mpz_t& r, mpz_t& q){//Berechnet die diskrete Wurzel modulo q. q kongruent 3 mod 4 ist Voraussetzung; vgl. Cohen: A course in computational algebraic number theory, p. 31
            mpz_t e;
            mpz_init(e);
            mpz_add_ui(e,q,1);
            mpz_div_ui(e,e,4); //e=(p+1)/4; ganze Zahl, da n.V. q\equiv 3mod 4
            mpz_powm(*w,r,e,q);
            mpz_clear(e);
        }

inline void Erzeuge_ThetaWurzel(int k){

double Zahl=log(((double)k)/5.5)/(log((double)2));
Theta_wurzel=ceil(Zahl);
}

inline void Erzeuge_TruncPr_Koeff(int d, mpz_t& q){

mpz_init_set_ui(zwei_hoch_k_hoch_minus_1,1);
mpz_mul_2exp(zwei_hoch_k_hoch_minus_1,zwei_hoch_k_hoch_minus_1,d);
mpz_invert(zwei_hoch_k_hoch_minus_1,zwei_hoch_k_hoch_minus_1,q);
}

inline void Initialisiere_Zufallszahlen(mpz_t& q){

mpz_t seed_mpz;
mpz_init(seed_mpz);
time_t seed;
double seed_d;

seed=clock();
seed_d=(double) seed;
mpz_set_d(seed_mpz,seed_d);
mpz_t prime;
mpz_init(prime);
mpz_nextprime(prime,q);
mpz_mul(seed_mpz,seed_mpz,prime);
mpz_add_ui(seed_mpz,seed_mpz,1);

gmp_randinit_default(rState1);
gmp_randseed(rState1,seed_mpz); //Erzeugen einer Zufallszahl
mpz_clear(seed_mpz);
mpz_clear(prime);
}

inline void Initialisiere_Zufallszahlen_F2M(){

srand(clock() + getpid());
}

inline void Erzeuge_Theta_Division(int k){

double theta_d=((double)k)/3.5;
theta_d=log(theta_d)/log((double)2);
Theta_Division=ceil(theta_d);
}

inline void PRandBitM_SMPC(int k, int m, int t, mpz_t& q, mpz_t& q_1, std::vector<Share> * r_Liste, int Anzahl_Spieler){//Erzeugen von r_0,...,r_{m-1} Zufallsbits aus Z_{q_1}, r'=\sum_{i=0}^{m-1}2^ir_i und r''<2^{k+kappa-m} (in dieser Reihenfolge)
//getestet am 10.2.2011;
int i;
mpz_t zwei_hoch;

mpz_init_set_ui(zwei_hoch,1);

Share r_2Strich(Anzahl_Spieler,t,q);
r_2Strich.PRandInt(t,k+kappa-m,q);

Share r_Strich(Anzahl_Spieler,t,q);
Share r(Anzahl_Spieler,t,q_1);

if(m<Modullaenge_kurz){//kann so unntige Umwandlungen vermeiden
for(i=0;i<m;i++){
    r.RandBit(t,q_1);
    r_Liste->push_back(r);
    r.Mul_Konst(zwei_hoch,q_1);
    r_Strich.Plus_Vektor(r.Vector,Anzahl_Spieler,q_1);
    mpz_mul_ui(zwei_hoch,zwei_hoch,2);
}

r_Strich.Umwandlung(t,q,q_1);
}
else{
for(i=0;i<m;i++){//Verwendung von q_1 spart ca. 6s\equiv 1/3 der Rechenzeit
    r.RandBit(t,q_1);
    r_Liste->push_back(r);
    r.Umwandlung(t,q,q_1);
    r.Mul_Konst(zwei_hoch,q);
    r_Strich.Plus_Vektor(r.Vector,Anzahl_Spieler,q);
    mpz_mul_ui(zwei_hoch,zwei_hoch,2);
    }
}

r_Liste->push_back(r_Strich);
r_Liste->push_back(r_2Strich);

mpz_clear(zwei_hoch);
}


inline void Bitzerlegung(mpz_t Argument_1, int k, int * Bits_1){//Zerlegt eine Zahl der Bitlnge k in ihre Bits; getestet am 25.2.2011
int i,check;//Bitzerlegung von c

for(i=0; i<k; i++){
    Bits_1[i]=0;
    }

mpz_t Abbruch;//bentige um folgende while-Schleife zu verlassen
mpz_t Arg_1_zerlegt;
mpz_mod_2exp(Argument_1,Argument_1,k);
mpz_t eins;
mpz_init_set_si(eins,1);
mpz_init(Abbruch);
mpz_init_set_si(Arg_1_zerlegt,0);
if(mpz_cmp_si(Argument_1,0)!=0){ //Zerlege Argument_1 in Bits, falls Argument_1!=0
    i=0;
    while(i<=k){
        check=mpz_scan1(Argument_1, i);
        Bits_1[check]=1;
        i=check+1;
        mpz_mul_2exp(Abbruch,eins,check);//Abbruch=2^check
        mpz_add(Arg_1_zerlegt,Arg_1_zerlegt,Abbruch);//Summiere die erhaltenen Bits auf
        check=mpz_cmp(Arg_1_zerlegt,Argument_1);//Vergleiche Arg_1_zerlegt mit Abbruch
        if(check==0){
            break;}
        };
}

mpz_clear(Arg_1_zerlegt);
mpz_clear(Abbruch);
mpz_clear(eins);
}

/** generate tables for fast finite field operations.
 *  code taken from http://www.samiam.org/galois.html
 */

inline void generate_tables() {
	unsigned char a = 1;
	unsigned char d;
	for(unsigned char c=0; c<255; c++) {
		antilog_table[c] = a;
		d = a & 0x80;
		a <<= 1;
		if(d == 0x80) {
			a ^= 0x1b;
		}
		a ^= antilog_table[c];
		log_table[antilog_table[c]] = c;
	}
	antilog_table[255] = antilog_table[0];
	log_table[0] = 0;
}

inline void Erzeuge_alpha_Division(int k){
mpf_t alpha;
mpf_init_set_d(alpha,2.9142);
mpz_init(alpha_division);

mpf_mul_2exp(alpha,alpha,k);
mpz_set_f(alpha_division,alpha);

mpf_clear(alpha);
}

inline void Erzeuge_Koeff_Wurzel(int l){
mpf_t C,D;

mpf_init_set_d(C,-0.8099868542);//Minimierung des rel. Fehlers//0.2863735989 0.8284271240 0.2928932190
mpf_mul_2exp(C,C,l);
mpz_init(C_1_wurzel);
mpz_set_f(C_1_wurzel,C);

mpf_init_set_d(D,1.787727479);//Minimerung des rel. Fehlers//1.264114224 1.801688366 1.273986062
mpf_mul_2exp(D,D,2*l);
mpz_init(C_0_wurzel);
mpz_set_f(C_0_wurzel,D);

mpf_clear(C);
mpf_clear(D);
}

inline void Berechne_n_ueber_t(int n, int t){

mpz_t binomial;
mpz_init(binomial);

mpz_bin_uiui(binomial,n,t);
n_ueber_t=mpz_get_ui(binomial);

mpz_clear(binomial);
}

inline void nt_Teilmengen(int n, int t){

int i,j,k=1, zwei_hoch_int;
mpz_t zwei_hoch;
mpz_init_set_ui(zwei_hoch,1);

Teilmengen=new int * [n_ueber_t+1];//Array fr unqualifizierte Gruppen

mpz_mul_2exp(zwei_hoch,zwei_hoch,n);
zwei_hoch_int=mpz_get_ui(zwei_hoch);

for(i=0;i<=n_ueber_t;i++){
    Teilmengen[i]=new int[n+1];
    }

for(i=0; i<zwei_hoch_int; i++){//fr alle Teilmengen mit t Elementen von n
        if (__builtin_popcount(i)==t){
            for(j=0; j<n; j++){//Wenn Teilmenge t Spieler enthlt
                if(1&i>>j) //bertrage die Teilmenge als binre Zeile ins Array Teilmengen
                    Teilmengen[k][j+1]=1;
                else
                    Teilmengen[k][j+1]=0;

                }
            k++;
            }
        }

mpz_clear(zwei_hoch);
}

inline void Berechne_RISS_Koeffizienten_q1(int Anzahl_Spieler, mpz_t q_1){

int i,j,k;
mpz_t ka;
mpz_init(ka);
mpz_t hilfe;
mpz_init(hilfe);
mpz_t p;
mpz_init(p);

RISS_Koeffizienten_q1=new mpz_t*[Anzahl_Spieler+1];
RISS_Koeffizienten_q1[0]=new mpz_t[n_ueber_t+1];//Anzahl
RISS=new mpz_t[n_ueber_t+1];//Anzahl
mpz_init_set_ui(RISS[0],0);
///RISS to Shamir q_1 nach Secure SCM-D.9.2; berechne fr jeden Spieler?
for(i=1; i<=Anzahl_Spieler; i++){
    RISS_Koeffizienten_q1[i]=new mpz_t[n_ueber_t+1];//Anzahl
    mpz_init(RISS_Koeffizienten_q1[i][0]);
    for(j=1; j<=n_ueber_t; j++){//Anzhal
        mpz_init(RISS_Koeffizienten_q1[i][j]);
        if(Teilmengen[j][i]==0){
            mpz_set_ui(p,1);
            for(k=1; k<=Anzahl_Spieler; k++){
                mpz_set_ui(ka,k);
                if(Teilmengen[j][k]==1){
                    mpz_sub_ui(hilfe,ka,i);
                    mpz_mul(p,p,hilfe);
                    mpz_invert(ka,ka,q_1);
                    mpz_mul(p,p,ka);
                    }
                }
            }
        mpz_set(RISS_Koeffizienten_q1[i][j],p);
        mpz_init_set_ui(RISS[j],0);
        }
    }
mpz_clear(ka);
mpz_clear(p);
mpz_clear(hilfe);
}

inline void reset_RISS(){

mpz_set_ui(RISS[0],0);
    for(int i=1; i<=n_ueber_t; i++){ //Erzeugung der RISS//Anzahl
        mpz_urandomb(RISS[i],rState1,kappa+10);
        }
}

inline void Berechne_RISS_Koeffizienten_q(int Anzahl_Spieler, mpz_t q){

int i,j,k;
mpz_t ka;
mpz_init(ka);
mpz_t hilfe;
mpz_init(hilfe);
mpz_t p;
mpz_init(p);

RISS_Koeffizienten_q=new mpz_t*[Anzahl_Spieler+1];
RISS_Koeffizienten_q[0]=new mpz_t[n_ueber_t+1];//Anzahl

for(i=1; i<=Anzahl_Spieler; i++){
    RISS_Koeffizienten_q[i]=new mpz_t[n_ueber_t+1];//Anzahl
    mpz_init(RISS_Koeffizienten_q[i][0]);
    for(j=1; j<=n_ueber_t; j++){//Anzahl
        mpz_init(RISS_Koeffizienten_q[i][j]);
        if(Teilmengen[j][i]==0){
            mpz_set_ui(p,1);
            for(k=1; k<=Anzahl_Spieler; k++){
                mpz_set_ui(ka,k);
                if(Teilmengen[j][k]==1){
                    mpz_sub_ui(hilfe,ka,i);
                    mpz_mul(p,p,hilfe);
                    mpz_invert(ka,ka,q);
                    mpz_mul(p,p,ka);
                    }
                }
            }
        mpz_set(RISS_Koeffizienten_q[i][j],p);
        }
    }

mpz_clear(p);
mpz_clear(ka);
mpz_clear(hilfe);
}

inline void Berechne_RISS_Koeffizienten_F2M(int Anzahl_Spieler){

int i,j,k;
GF256 i_256;
GF256 s;
GF256 p;
GF256 hilfe;
GF256 k_256;


RISS_Koeffizienten_F2M=new GF256*[Anzahl_Spieler+1];
RISS_Koeffizienten_F2M[0]=new GF256[n_ueber_t+1];//Anzahl

for(i=1; i<=Anzahl_Spieler; i++){
        RISS_Koeffizienten_F2M[i]=new GF256[n_ueber_t+1];//Anzahl
        i_256=i;
        s=0;
        for(j=1; j<=n_ueber_t; j++){//Anzahl
            p=0;
            if(Teilmengen[j][i]==0){
                p=1;
                for(k=1; k<=Anzahl_Spieler; k++){
                    k_256=k;
                    if(Teilmengen[j][k]==1){//Teilmengen wird in head.h. definiert und initialisiert
                        hilfe=k_256-i_256;
                        p=p*hilfe;
                        k_256=k_256.inverse();
                        p=p*k_256;
                        }
                    }
                }
            RISS_Koeffizienten_F2M[i][j]=p;
        }
    }

}

inline void LagIntKoeff(int t){
	/* Laufvariablen */
	int i, k;				/* 1: Debug-Ausgaben */

	mpz_t zaehler, nenner, rest;
	mpz_init(zaehler);
	mpz_init(nenner);
	mpz_init(rest);

	lambda= new mpz_t[2*t+2];
	for(i=0; i<=2*t+1; i++){
        mpz_init(lambda[i]);
        }



	/* Berechnung der Lagrange Interpolationskoeffizienten */
	for(i=1; i<=2*t+1; i++){
	   mpz_set_si(zaehler, 1);
	   mpz_set_si(nenner, 1);
	   for(k=1; k<=2*t+1; k++){
	      if(i!=k){
	         mpz_mul_si(zaehler, zaehler, k);
	         /* mpz_mod(zaehler, zaehler, q); */
	         mpz_mul_si(nenner, nenner, k-i);
	         /* mpz_mod(nenner, nenner, q); */
	      }
	   }
	   mpz_fdiv_qr(lambda[i], rest, zaehler, nenner);
	}
	/* Speicherfreigabe fr nicht mehr verwendete Variablen */
	mpz_clear(rest);
	mpz_clear(nenner);
	mpz_clear(zaehler);
}

inline void LagIntKoeff_Rekonstrukt(int t){
	/* Laufvariablen */
	int i, k;				/* 1: Debug-Ausgaben */

	mpz_t zaehler, nenner, rest;
	mpz_init(zaehler);
	mpz_init(nenner);
	mpz_init(rest);

	lambda_Rekonstrukt= new mpz_t[t+2];
	for(i=0; i<=t+1; i++){
        mpz_init(lambda_Rekonstrukt[i]);
        }



	/* Berechnung der Lagrange Interpolationskoeffizienten */
	for(i=1; i<=t+1; i++){
	   mpz_set_si(zaehler, 1);
	   mpz_set_si(nenner, 1);
	   for(k=1; k<=t+1; k++){
	      if(i!=k){
	         mpz_mul_si(zaehler, zaehler, k);
	         mpz_mul_si(nenner, nenner, k-i);
	      }
	   }

	   mpz_fdiv_qr(lambda_Rekonstrukt[i], rest, zaehler, nenner);
	}
	/* Speicherfreigabe fr nicht mehr verwendete Variablen */
	mpz_clear(rest);
	mpz_clear(nenner);
	mpz_clear(zaehler);
}


inline void LagIntKoeff_GF256(int n, int t){
	/* Laufvariablen */
	int i, k;
	/* Hilfsvariablen */
	GF256 zaehler, nenner, K, I;
	lambda_256=new GF256[n+1];

	for(i=0;i<=n; i++){
        lambda_256[i]=0;
        }

	/* Berechnung der Lagrange Interpolationskoeffizienten */
	for(i=1; i<=n; i++){
        zaehler=1;
        nenner=1;
        I=i;
        for(k=1; k<=n; k++){
        K=k;
	      if(i!=k){
            zaehler=zaehler*K;
            nenner=nenner*(K-I);
            }
	   }
	   nenner=nenner.inverse();
	   lambda_256[i]=zaehler*nenner;
	}
}

inline void LagIntKoeff_GF256_Rekonstrukt(int t){
	/* Laufvariablen */
	int i, k;
	/* Hilfsvariablen */
	GF256 zaehler, nenner, K, I;
	lambda_256_Rekonstrukt=new GF256[t+2];

	for(i=0;i<=t+1; i++){
        lambda_256_Rekonstrukt[i]=0;
        }

	/* Berechnung der Lagrange Interpolationskoeffizienten */
	for(i=1; i<=t+1; i++){
        zaehler=1;
        nenner=1;
        I=i;
        for(k=1; k<=t+1; k++){
        K=k;
	      if(i!=k){
            zaehler=zaehler*K;
            nenner=nenner*(K-I);
            }
	   }
	   nenner=nenner.inverse();
	   lambda_256_Rekonstrukt[i]=zaehler*nenner;
	}
}

inline void Berechne_2_hoch_f(){

mpz_init_set_ui(zwei_hoch_f,1);
mpz_mul_2exp(zwei_hoch_f,zwei_hoch_f,f);
}

inline void Berechne_Wurzel_Konstanten(){

mpz_init_set_ui(halb,1);
mpz_mul_2exp(halb,halb,f-1);

mpz_init_set_ui(eins_Wurzel,1);
mpz_mul_2exp(eins_Wurzel,eins_Wurzel,f);

mpz_init_set_ui(Wurzel_zwei_hoch_f,1);
mpz_mul_2exp(Wurzel_zwei_hoch_f,Wurzel_zwei_hoch_f,f/2);

mpf_init_set_d(Wurzel_2,2.0);
mpf_sqrt(Wurzel_2,Wurzel_2);
mpf_mul_2exp(Wurzel_2,Wurzel_2,f);
mpz_init(Wurzel_2_z);
mpz_set_f(Wurzel_2_z,Wurzel_2);

mpz_init_set_ui(Drei,3);
mpz_mul_2exp(Drei,Drei,3*f);
}

inline void Berechne_rem(mpz_t& q, mpz_t& q_1){

mpz_init(rem);
int k=mpz_cdiv_q_ui(rem,q,2); //rem=q/2

mpz_init(rem_q_1);
k=mpz_cdiv_q_ui(rem_q_1,q_1,2);
}

inline void Waehle_Modulus(mpz_t& Q, mpz_t& REM){

int k=mpz_cmp(Q,grosser_Modulus);

if(k==0){
    mpz_set(REM,rem);
    }
else
    mpz_set(REM,rem_q_1);
}

inline int min(int a, int b){

if(b<a){
    return b;
    }
else{
    return a;
    }
}

inline void Koeffizienten_Bit_Mask(int m, int k, mpz_t q, mpz_t * Koeff){

int i,r;

mpz_t h;
mpz_init(h);
for(i=0; i<m; i++){
    mpz_set_ui(Koeff[i],1);

    if(i<=k && i>0 && k!=0){
        mpz_set_ui(h,i-1);
        mpz_sub_ui(h,h,k);
        mpz_mul(Koeff[i],Koeff[i-1],h);
        }

    if(i>k && i>0 && k!=0){
        mpz_set_ui(h,i);
        mpz_sub_ui(h,h,k);
        mpz_mul(Koeff[i],Koeff[i-1],h);
        }

    if(k==0 && i>0){
        mpz_set_ui(h,i);
        mpz_mul(Koeff[i],Koeff[i-1],h);
        }
    }

for(i=1; i<m; i++){
    mpz_invert(Koeff[i],Koeff[i],q);
    r=i%2;
    if(r==1){
        mpz_mul_si(Koeff[i],Koeff[i],-1);
        }
    }
mpz_clear(h);
}
}

#endif // HEAD_H_INCLUDED
