/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/
#include <iostream>
#include <string>
#include <fstream>
#include <Matrix_LA.h>
#include <Simplex_Startwert_Catrina.h>
#include <Vektor_LA.h>
#include <Share.h>
#include <gmp.h>
#include <head.h>


#define Problem_76 1
#define LR 1

unsigned char antilog_table[256];
unsigned char log_table[256];
mpz_t alpha_division;
mpz_t C_0_wurzel;
mpz_t C_1_wurzel;
mpz_t E_0_wurzel;
mpz_t E_1_wurzel;
mpz_t zwei_hoch_k_hoch_minus_1;
mpz_t zwei_hoch_f;
mpz_t halb;
mpz_t eins_Wurzel;
mpz_t Wurzel_zwei_hoch_f;
mpf_t Wurzel_2;
mpz_t Wurzel_2_z;
mpz_t Drei;
mpz_t grosser_Modulus;
mpz_t kleiner_Modulus;
mpz_t rem;
mpz_t rem_q_1;
gmp_randstate_t rState1;
mpz_t * RISS;
mpz_t * lambda;
mpz_t * lambda_Rekonstrukt;
GF256 * lambda_256;
mpz_t ** RISS_Koeffizienten_q1;
mpz_t ** RISS_Koeffizienten_q;
GF256 ** RISS_Koeffizienten_F2M;
GF256 * lambda_256_Rekonstrukt;
int n_ueber_t;

int **Teilmengen;
int Anzahl;
int Theta_wurzel;
int Theta_Division;


using namespace std;

void Erstelle_Vektor(Vektor_LA& V, Vektor_LA& b, int m, int n);
void Erstelle_Gleichungssystem(Matrix_LA& M, Matrix_LA& Q, Matrix_LA& A);
void Berechne_alpha(const Matrix_LA& A,const Vektor_LA&d,const Vektor_LA&x,Vektor_LA&b, Share& alpha, Vektor_LA& Minimum, Vektor_LA& AM_Zq, Share& alpha_kleiner_eins, int Dimension, int n, int t, int k, mpz_t& q, mpz_t& q_1);
void Zielfunktion(Matrix_LA& Q, Vektor_LA& x, Vektor_LA& c, mpf_t& Wert, int n, int t, int k, mpz_t& q, mpz_t& q_1);
void Bestimme_Aktive_Menge(Matrix_LA& A, Vektor_LA& x, Vektor_LA& b, Vektor_LA& AM_Zq, int d, int k, int n, int t, mpz_t& q, mpz_t& q_1);

int main()
{
int i,j,n=5, t=2, k=128, Iterationszaehler=1;
mpf_set_default_prec(300);

mpz_t s, Ende;
clock_t start_programm,stop_programm;

mpz_init(Ende);
mpz_init_set_ui(s,1);

mpz_t q;
mpz_t q_1;
mpz_t basis;
mpz_t modulo;
mpz_t modulo_q1;

mpz_init(q);
mpz_init(q_1);
mpz_init_set_ui(basis,2);
mpz_init(modulo);
mpz_init(modulo_q1);

mpz_pow_ui(q,basis,Modullaenge);
mpz_pow_ui(q_1,basis,63);

do{
mpz_nextprime(q,q);
mpz_nextprime(q_1,q_1);
mpz_mod_ui(modulo,q, 4);
mpz_mod_ui(modulo_q1,q_1, 4);
}
while(mpz_cmp_ui(modulo,3)!=0 || mpz_cmp_ui(modulo_q1,3)!=0); //Erzeuge geeignete Primzahl fr Berechnung des modulus (mu kongruent 3 mod 4 sein)

haupt::Berechne_Koeffizienten(n,t,k,f,q,q_1);

mpz_init_set(grosser_Modulus,q);
mpz_init_set(kleiner_Modulus,q_1);


int Zaehler_voller_Schritt=0, Zaehler_halber_Schritt=0;

mpz_t unendlich;
mpz_init_set_ui(unendlich,1);
mpz_mul_2exp(unendlich,unendlich,k-2);
mpz_sub_ui(unendlich,unendlich,1);//unendlich=2^{k-1}-1

double ** Werte, **Q_Werte;

#ifdef Problem 21
Vektor_LA x(2,n,t,q,q_1,k,f);
Vektor_LA d(2,n,t,q,q_1,k,f);
Vektor_LA c(2,n,t,q,q_1,k,f);//ist hier =0
Matrix_LA Q(2,2,n,t,q,q_1,k,f);
Vektor_LA b(5,n,t,q,q_1,k,f);
Matrix_LA A(5,2,n,t,q,q_1,k,f);


Werte=new double *[5];
for(i=0; i<5; i++){
Werte[i]=new double[2];
}

Q_Werte=new double*[2];
Q_Werte[0]=new double[2];
Q_Werte[1]=new double[2];

Q_Werte[0][0]=0.02;
Q_Werte[0][1]=0.0;
Q_Werte[1][0]=0.0;
Q_Werte[1][1]=2.0;

Werte[0][0]=-10.0;
Werte[0][1]=1.0;
Werte[1][0]=1.0;
Werte[1][1]=0.0;
Werte[2][0]=-1.0;
Werte[2][1]=0.0;
Werte[3][0]=0.0;
Werte[3][1]=1.0;
Werte[4][0]=0.0;
Werte[4][1]=-1.0;

b.Vector_LA.at(0).Erzeugung_double(-10.0,t,q);
b.Vector_LA.at(1).Erzeugung_double(50.0,t,q);
b.Vector_LA.at(2).Erzeugung_double(-2.0,t,q);
b[3].Erzeugung_double(50.0,t,q);
b[4].Erzeugung_double(50.0,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],2.0);
mpf_set_d(x_korrekt.V[1],0.0);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,0.04);
#endif

#ifdef Problem_35
Vektor_LA x(3,n,t,q,q_1,k,f);
Vektor_LA d(3,n,t,q,q_1,k,f);
Vektor_LA c(3,n,t,q,q_1,k,f);
Vektor_LA b(4,n,t,q,q_1,k,f);
Matrix_LA A(4,3,n,t,q,q_1,k,f);
Matrix_LA Q(3,3,n,t,q,q_1,k,f);

int Gleichheitsbedingungen=0;
int Ungleichheitsbedingungen=b.d;

Werte=new double *[4];
for(i=0; i<4; i++){
Werte[i]=new double[3];
}


Q_Werte=new double*[3];
for(i=0; i<5; i++){
    Q_Werte[i]=new double[3];
    }

Q_Werte[0][0]=4.0;
Q_Werte[0][1]=2.0;
Q_Werte[0][2]=2.0;

Q_Werte[1][0]=2.0;
Q_Werte[1][1]=4.0;
Q_Werte[1][2]=0.0;

Q_Werte[2][0]=2.0;
Q_Werte[2][1]=0.0;
Q_Werte[2][2]=2.0;


Werte[0][0]=1.0;
Werte[0][1]=1.0;
Werte[0][2]=2.0;

Werte[1][0]=-1.0;
Werte[1][1]=0.0;
Werte[1][2]=0.0;

Werte[2][0]=0.0;
Werte[2][1]=-1.0;
Werte[2][2]=0.0;

Werte[3][0]=0.0;
Werte[3][1]=0.0;
Werte[3][2]=-1.0;


c[0].Erzeugung_double(-8.0,t,q);
c[1].Erzeugung_double(-6.0,t,q);
c[2].Erzeugung_double(-4.0,t,q);

b[0].Erzeugung_double(3.0,t,q);
b[1].Erzeugung_double(0.0,t,q);
b[2].Erzeugung_double(0.0,t,q);
b[3].Erzeugung_double(0.0,t,q);

x[0].Erzeugung_double(0.5,t,q);
x[1].Erzeugung_double(0.5,t,q);
x[2].Erzeugung_double(0.5,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],4.0);
mpf_div_ui(x_korrekt.V[0],x_korrekt.V[0],3);
mpf_set_d(x_korrekt.V[1],7.0);
mpf_div_ui(x_korrekt.V[1],x_korrekt.V[1],9);
mpf_set_d(x_korrekt.V[2],4.0);
mpf_div_ui(x_korrekt.V[2],x_korrekt.V[2],9);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,-8.0);
mpf_div_ui(f_korrekt,f_korrekt,9);
mpf_sub_ui(f_korrekt,f_korrekt,8);//f=-8 8/9
#endif

#ifdef Problem_76
Vektor_LA x(4,n,t,q,q_1,k,f);
Vektor_LA d(4,n,t,q,q_1,k,f);
Vektor_LA c(4,n,t,q,q_1,k,f);
Vektor_LA b(7,n,t,q,q_1,k,f);
Matrix_LA A(7,4,n,t,q,q_1,k,f);
Matrix_LA Q(4,4,n,t,q,q_1,k,f);

Werte=new double *[7];
for(i=0; i<7; i++){
Werte[i]=new double[4];
}


Q_Werte=new double*[4];
for(i=0; i<5; i++){
    Q_Werte[i]=new double[4];
    }

Q_Werte[0][0]=2.0;
Q_Werte[0][1]=0.0;
Q_Werte[0][2]=-1.0;
Q_Werte[0][3]=0.0;

Q_Werte[1][0]=0.0;
Q_Werte[1][1]=1.0;
Q_Werte[1][2]=0.0;
Q_Werte[1][3]=0.0;

Q_Werte[2][0]=-1.0;
Q_Werte[2][1]=0.0;
Q_Werte[2][2]=2.0;
Q_Werte[2][3]=1.0;

Q_Werte[3][0]=0.0;
Q_Werte[3][1]=0.0;
Q_Werte[3][2]=1.0;
Q_Werte[3][3]=1.0;


Werte[0][0]=1.0;
Werte[0][1]=2.0;
Werte[0][2]=1.0;
Werte[0][3]=1.0;

Werte[1][0]=3.0;
Werte[1][1]=1.0;
Werte[1][2]=2.0;
Werte[1][3]=-1.0;

Werte[2][0]=0.0;
Werte[2][1]=-1.0;
Werte[2][2]=-4.0;
Werte[2][3]=0.0;

Werte[3][0]=-1.0;
Werte[3][1]=0.0;
Werte[3][2]=0.0;
Werte[3][3]=0.0;

Werte[4][0]=0.0;
Werte[4][1]=-1.0;
Werte[4][2]=0.0;
Werte[4][3]=0.0;

Werte[5][0]=0.0;
Werte[5][1]=0.0;
Werte[5][2]=-1.0;
Werte[5][3]=0.0;

Werte[6][0]=0.0;
Werte[6][1]=0.0;
Werte[6][2]=0.0;
Werte[6][3]=-1.0;

c[0].Erzeugung_double(-1.0,t,q);
c[1].Erzeugung_double(-3.0,t,q);
c[2].Erzeugung_double(1.0,t,q);
c[3].Erzeugung_double(-1.0,t,q);

b[0].Erzeugung_double(5.0,t,q);
b[1].Erzeugung_double(4.0,t,q);
b[2].Erzeugung_double(-1.5,t,q);
b[3].Erzeugung_double(0.0,t,q);
b[4].Erzeugung_double(0.0,t,q);
b[5].Erzeugung_double(0.0,t,q);
b[6].Erzeugung_double(0.0,t,q);

x[0].Erzeugung_double(0.5,t,q);
x[1].Erzeugung_double(0.5,t,q);
x[2].Erzeugung_double(0.5,t,q);
x[3].Erzeugung_double(0.5,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],3.0);
mpf_div_ui(x_korrekt.V[0],x_korrekt.V[0],11);//0,272727272727... =3/11
mpf_set_d(x_korrekt.V[1],23.0);
mpf_div_ui(x_korrekt.V[1],x_korrekt.V[1],11);//2,09090909090909... =23/11
mpf_set_d(x_korrekt.V[2],0.26);
mpf_div_ui(x_korrekt.V[2],x_korrekt.V[2],10000000000);
mpf_set_d(x_korrekt.V[3],6.0);
mpf_div_ui(x_korrekt.V[3],x_korrekt.V[3],11);//0,545454545454... =6/11

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,-4.681818181);
#endif

#ifdef Problem_224
Vektor_LA x(2,n,t,q,q_1,k,f);
Vektor_LA d(2,n,t,q,q_1,k,f);
Vektor_LA c(2,n,t,q,q_1,k,f);//ist hier =0
Matrix_LA Q(2,2,n,t,q,q_1,k,f);
Vektor_LA b(4,n,t,q,q_1,k,f);
Matrix_LA A(4,2,n,t,q,q_1,k,f);

Werte=new double *[5];
for(i=0; i<5; i++){
Werte[i]=new double[2];
}

Q_Werte=new double*[2];
Q_Werte[0]=new double[2];
Q_Werte[1]=new double[2];

Q_Werte[0][0]=4.0;
Q_Werte[0][1]=0.0;
Q_Werte[1][0]=0.0;
Q_Werte[1][1]=2.0;

Werte[0][0]=-1.0;
Werte[0][1]=-3.0;
Werte[1][0]=1.0;
Werte[1][1]=3.0;
Werte[2][0]=-1.0;
Werte[2][1]=-1.0;
Werte[3][0]=1.0;
Werte[3][1]=1.0;

b.Vector_LA.at(0).Erzeugung_double(0.0,t,q);
b.Vector_LA.at(1).Erzeugung_double(18.0,t,q);
b.Vector_LA.at(2).Erzeugung_double(0.0,t,q);
b[3].Erzeugung_double(8.0,t,q);

c[0].Erzeugung_double(-48.0,t,q);
c[1].Erzeugung_double(-40.0,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],4.0);
mpf_set_d(x_korrekt.V[1],4.0);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,-304.0);
#endif

#ifdef Problem_268
Vektor_LA x(5,n,t,q,q_1,k,f);
Vektor_LA d(5,n,t,q,q_1,k,f);
Vektor_LA c(5,n,t,q,q_1,k,f);
Vektor_LA b(5,n,t,q,q_1,k,f);
Matrix_LA A(5,5,n,t,q,q_1,k,f);
Matrix_LA Q(5,5,n,t,q,q_1,k,f);

Werte=new double *[5];
for(i=0; i<5; i++){
Werte[i]=new double[5];
}


Q_Werte=new double*[5];
for(i=0; i<5; i++){
    Q_Werte[i]=new double[5];
    }

Q_Werte[0][0]=20394.0;
Q_Werte[0][1]=-24908.0;
Q_Werte[0][2]=-2026.0;
Q_Werte[0][3]=3896.0;
Q_Werte[0][4]=658.0;

Q_Werte[1][0]=-24908.0;
Q_Werte[1][1]=41818.0;
Q_Werte[1][2]=-3466.0;
Q_Werte[1][3]=-9828.0;
Q_Werte[1][4]=-372.0;

Q_Werte[2][0]=-2026.0;
Q_Werte[2][1]=-3466.0;
Q_Werte[2][2]=3510.0;
Q_Werte[2][3]=2178.0;
Q_Werte[2][4]=-348.0;

Q_Werte[3][0]=3896.0;
Q_Werte[3][1]=-9828.0;
Q_Werte[3][2]=2178.0;
Q_Werte[3][3]=3030.0;
Q_Werte[3][4]=-44.0;

Q_Werte[4][0]=658.0;
Q_Werte[4][1]=-372.0;
Q_Werte[4][2]=-348.0;
Q_Werte[4][3]=-44.0;
Q_Werte[4][4]=54.0;


Werte[0][0]=1.0;
Werte[0][1]=1.0;
Werte[0][2]=1.0;
Werte[0][3]=1.0;
Werte[0][4]=1.0;

Werte[1][0]=-10.0;
Werte[1][1]=-10.0;
Werte[1][2]=3.0;
Werte[1][3]=-5.0;
Werte[1][4]=-4.0;

Werte[2][0]=8.0;
Werte[2][1]=-1.0;
Werte[2][2]=2.0;
Werte[2][3]=5.0;
Werte[2][4]=-3.0;

Werte[3][0]=-8.0;
Werte[3][1]=1.0;
Werte[3][2]=-2.0;
Werte[3][3]=-5.0;
Werte[3][4]=3.0;

Werte[4][0]=4.0;
Werte[4][1]=2.0;
Werte[4][2]=-3.0;
Werte[4][3]=5.0;
Werte[4][4]=-1.0;


b[0].Erzeugung_double(5.0,t,q);
b[1].Erzeugung_double(-20.0,t,q);
b[2].Erzeugung_double(40.0,t,q);
b[3].Erzeugung_double(-11.0,t,q);
b[4].Erzeugung_double(30.0,t,q);

c[0].Erzeugung_double(18340.0,t,q);
c[1].Erzeugung_double(-34198.0,t,q);
c[2].Erzeugung_double(4542.0,t,q);
c[3].Erzeugung_double(8672.0,t,q);
c[4].Erzeugung_double(86.0,t,q);

x[0].Erzeugung_double(1.0,t,q);
x[1].Erzeugung_double(1.0,t,q);
x[2].Erzeugung_double(1.0,t,q);
x[3].Erzeugung_double(1.0,t,q);
x[4].Erzeugung_double(1.0,t,q);

Vektor_mpf_t x_korrekt(x.d);
mpf_set_d(x_korrekt.V[0],1.0);
mpf_set_d(x_korrekt.V[1],2.0);
mpf_set_d(x_korrekt.V[2],-1.0);
mpf_set_d(x_korrekt.V[3],3.0);
mpf_set_d(x_korrekt.V[4],-4.0);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,-14463.0);
#endif

#ifdef Problem_A
Vektor_LA x(2,n,t,q,q_1,k,f);
Vektor_LA d(2,n,t,q,q_1,k,f);
Vektor_LA c(2,n,t,q,q_1,k,f);
Matrix_LA Q(2,2,n,t,q,q_1,k,f);
Vektor_LA b(5,n,t,q,q_1,k,f);
Matrix_LA A(5,2,n,t,q,q_1,k,f);


Werte=new double *[5];
for(i=0; i<5; i++){
Werte[i]=new double[2];
}

Q_Werte=new double*[2];
Q_Werte[0]=new double[2];
Q_Werte[1]=new double[2];

Q_Werte[0][0]=2.0;
Q_Werte[0][1]=0.0;
Q_Werte[1][0]=0.0;
Q_Werte[1][1]=2.0;

Werte[0][0]=-1.0;
Werte[0][1]=2.0;
Werte[1][0]=1.0;
Werte[1][1]=2.0;
Werte[2][0]=1.0;
Werte[2][1]=-2.0;
Werte[3][0]=-1.0;
Werte[3][1]=0.0;
Werte[4][0]=0.0;
Werte[4][1]=-1.0;

b.Vector_LA.at(0).Erzeugung_double(2.0,t,q);
b.Vector_LA.at(1).Erzeugung_double(6.0,t,q);
b.Vector_LA.at(2).Erzeugung_double(2.0,t,q);
b[3].Erzeugung_double(0.0,t,q);
b[4].Erzeugung_double(0.0,t,q);

c.Vector_LA.at(0).Erzeugung_double(-2.0,t,q);
c.Vector_LA.at(1).Erzeugung_double(-5.0,t,q);

Vektor_mpf_t x_korrekt(2);
mpf_set_d(x_korrekt.V[0],1.4);
mpf_set_d(x_korrekt.V[1],1.7);

mpf_t f_korrekt;
mpf_init_set_d(f_korrekt,-6.45);
#endif

printf("\nPrimaler Algorithmus");
Q.Erzeugung_double(Q_Werte);
A.Erzeugung_double(Werte);

Vektor_LA S(b.d,n,t,q,q_1,k,f);
Share Summe(n,t,q);

for(i=0; i<b.d; i++){
    S[i]=b[i];
    S[i].GQ(t,q,q_1,k);//b_i>=0?
    Summe+=S[i];
    }

Summe-=b.d;
int Summe_int=Summe.EQZPub();

clock_t start_simplex, stop_simplex;
start_simplex=clock();
stop_simplex=start_simplex;
clock_t start_Phase2, stop_Phase_2;

int Anzahl_Simplex_Iterationen=0;

Vektor_LA x0(A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA X(A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA B(A.Zeilenzahl,n,t,q,q_1,k,f);

start_programm=clock();
if(Summe_int!=1){
Simplex_Startwert_Catrina S(A.Zeilenzahl,A.Spaltenzahl,A.Zeilenzahl,n,t,q,q_1,k,f);

Vektor_LA C(S.Spaltenzahl-1,n,t,q,q_1,k,f);
for(i=C.d-1; i>=A.Spaltenzahl;i--){
    C[i].Erzeugung_double(1.0,t,q);
    }

Vektor_LA N(S.Spaltenzahl-1,n,t,q,q_1,k,f);

start_programm=clock();
start_simplex=clock();
Anzahl_Simplex_Iterationen=S.Simplex_Algorithmus(A,b,C,x0,X,0,A.Zeilenzahl,B,N);
stop_simplex=clock();
}


A*=-1;//VZ umdrehen fr Phase 2
b*=-1;
x0.Oben_einpassen(X);

///Ende Phase 1

Matrix_LA G(Q.Zeilenzahl+A.Spaltenzahl,Q.Spaltenzahl+A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA l(Q.Zeilenzahl+A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA lambda(b.d,n,t,q,q_1,k,f);
Vektor_LA lambda_min=lambda;
Vektor_LA lambda_kurz(A.Spaltenzahl,n,t,q,q_1,k,f);
Share Bit(n,t,q);
Share Hilfe(n,t,q);
Share lambda_q (n,t,q);//Speichere das aktuelle lambda
Vektor_LA Hilfe_Vektor(A.Spaltenzahl,n,t,q,q_1,k,f);
Share alpha(n,t,q);
Vektor_LA Index(A.Zeilenzahl,n,t,q,q_1,k,f);//wird zur Bestimmung des Minimums von alpha bentigt
Vektor_LA Index_kurz(A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA Index_AM(A.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA ax(A.Zeilenzahl,n,t,q,q_1,k,f);
Vektor_LA AM_Zq(b.d,n,t,q,q_1,k,f);
Vektor_LA Kopie_AM(b.d,n,t,q,q_1,k,f);
Vektor_LA AM_kurz(A.Spaltenzahl,n,t,q,q_1,k,f);
Matrix_LA Kopie_A=A;
Matrix_LA A_Kompakt(A.Spaltenzahl,A.Spaltenzahl,n,t,q,q_1,k,f);
Vektor_LA Kopie_c=c;
Vektor_LA Kopie_x=x;
Vektor_LA Ergebnis=x;
Share Hilfe_2=Hilfe;
Matrix_LA P(A.Zeilenzahl,A.Zeilenzahl,n,t,q,q_1,k,f);
Matrix_LA P_t(A.Zeilenzahl,A.Zeilenzahl,n,t,q,q_1,k,f);//Die Transponierte von P
Share alpha_kleiner_eins(n,t,q);

///Beginn Phase 2
start_Phase2=clock();
x=x0;//x ist der in Phase 1 bestimmte Startwert

Bestimme_Aktive_Menge(A,x,b,AM_Zq,b.d,k,n,t,q,q_1);

for(i=0; i<A.Zeilenzahl; i++){//Bestimme die initialen Werte ax(i)
    Hilfe_Vektor.getZeilenvektor(A,i);
    Hilfe_Vektor.SkalarMult(x,ax[i]);
    }

for(Iterationszaehler=1; Iterationszaehler<51; Iterationszaehler++){

Kopie_c=c;
Kopie_x=x;

Kopie_x.MatrixMult(Q,Kopie_x);
Kopie_x.TruncPr();
Kopie_c+=Kopie_x;//Kopie_c=Q*x+c

Kopie_A.Multipliziere_Zeilenweise(&AM_Zq.Vector_LA);
Kopie_A.Matrix_Zusammenschieben_binaer(AM_Zq.Vector_LA,P);

A_Kompakt.Oben_einpassen(Kopie_A);

P_t=P;
P_t.Transponiere_quadratisch();

AM_Zq.MatrixMult(P,Kopie_AM);
AM_kurz.Oben_einpassen(Kopie_AM);

G=0;
l=0;
Erstelle_Gleichungssystem(G,Q,A_Kompakt);
Erstelle_Vektor(l,Kopie_c,c.d,l.d);

#ifdef QR
G.Loese_GLS_QR(l.Vector_LA,l.d);
#endif

#ifdef LR
G.Loese_GLS_LR(l.Vector_LA,l.d);
#endif

j=0;
for(i=c.d; i<l.d; i++){
    lambda.Vector_LA.at(j)=l.Vector_LA.at(i);
    j++;
    }

lambda.MatrixMult(P_t,lambda);//Setze lambda an die richtigen Stellen zurck; die ungltigen Stellen werden automatisch mit Null besetzt

for(i=0; i<lambda_kurz.d; i++){
    lambda_kurz[i]=l[i+c.d];//Kompaktifizierte Form des Vektors lambda erstellen
    lambda_kurz[i]=lambda_kurz[i]*AM_kurz[i]+(-AM_kurz[i]+1)*unendlich;//Ersetze nicht-aktive Komponenten mit unendlich; diese sind wg. Komplementrbedingung = 0
    }


for(i=0; i<x.d; i++){
    d.Vector_LA.at(i)=l.Vector_LA.at(i);
    }

d.Gleich_Null_approximativ(Bit);

#ifdef Test_voller_oder_halber_Schritt
Bit.Rekonstrukt(s,t,q);
Bit.Rekonstrukt_Ausgabe_unskaliert("d=?0");
hs=mpz_get_si(s);
if(hs==1)
    Zaehler_halber_Schritt+=1;
if(hs==0)
    Zaehler_voller_Schritt+=1;
#endif

///Schritt 3
Index_kurz=0;
lambda_kurz.Min(Index_kurz,k);

lambda_q=0;
lambda_q=lambda_kurz[Index_kurz];

///Verkrzten Vektor lambda wieder auf die richtige Lnge expandieren
lambda_min.Oben_einpassen(Index_kurz);
lambda_min.MatrixMult(P_t,lambda_min);

lambda_q.GTZ_approximativ(t,q,q_1,k);
lambda_q.MulPub(Bit,Ende);

if(mpz_cmp_si(Ende,1)==0){
    stop_programm=clock();
    stop_Phase_2=clock();
    FILE* datei1;
    datei1=fopen("Ausgabe.txt","a");

    fprintf(datei1,"\n\n\n---------------------------------------------------");
    time_t Zeit;
    struct tm *ts;//Variable zur Zeitmessung
    time(&Zeit);
    ts=localtime(&Zeit);
    fprintf(datei1,"\n\n\nZeit: %s\n\n",asctime(ts));

    cout<<"\n\nOptimaler Punkt x erreicht"<<endl;
    Q.Ausgabe_skaliert("G");
    c.Ausgabe_skaliert("c");
    A.Ausgabe_skaliert("A");
    b.Ausgabe_skaliert("b");

    Q.Ausgabe_skaliert_Datei("G",datei1);
    c.Ausgabe_skaliert_Datei("c",datei1);
    A.Ausgabe_skaliert_Datei("A",datei1);
    b.Ausgabe_skaliert_Datei("b",datei1);
    fprintf(datei1,"\nOptimaler Punkt x erreicht");

    double D=((double)stop_programm-(double)start_programm)/CLOCKS_PER_SEC;
    printf("\n\nLaufzeit des Programms: %f s",D);
    printf("\nZeit fuer Phase 1: %f s",((double)stop_simplex-(double)start_simplex)/CLOCKS_PER_SEC);
    fprintf(datei1,"\n\nZeit fuer Phase 1: %f s",((double)stop_simplex-(double)start_simplex)/CLOCKS_PER_SEC);
    fprintf(datei1,"\nZeit fuer Phase 2: %f s",((double)stop_Phase_2-(double)start_Phase2)/CLOCKS_PER_SEC);
    printf("\nZeit fuer Phase 2: %f s",((double)stop_Phase_2-(double)start_Phase2)/CLOCKS_PER_SEC);
    printf("\n\nAnzahl der Iterationen: %d", Iterationszaehler);
    printf("\nAnzahl volle Schritte: %d",Zaehler_voller_Schritt);
    printf("\nAnzahl halbe Schritte: %d",Zaehler_halber_Schritt);
    printf("\nAnzahl der Simplex-Iterationen: %d\n",Anzahl_Simplex_Iterationen);
    fprintf(datei1,"\n\nLaufzeit des Programms: %f s",D);
    fprintf(datei1,"\n\nAnzahl der Iterationen: %d", Iterationszaehler);
    fprintf(datei1,"\nAnzahl der Simplex-Iterationen: %d",Anzahl_Simplex_Iterationen);
    fprintf(datei1,"\nAnzahl volle Schritte: %d",Zaehler_voller_Schritt);
    fprintf(datei1,"\nAnzahl halbe Schritte: %d\n\n",Zaehler_halber_Schritt);
    mpf_t Z;
    mpf_init(Z);
    Zielfunktion(Q,x,c,Z,n,t,k,q,q_1);
    gmp_printf("\nZielfunktion: %.10Ff",Z);
    gmp_fprintf(datei1,"\nZielfunktion: %.10Ff",Z);
    mpf_sub(Z,Z,f_korrekt);
    gmp_printf("\nDifferenz zu exaktem Wert der Zielfunktion: %.10Fe",Z);
    gmp_fprintf(datei1,"\nDifferenz zu exaktem Wert der Zielfunktion: %.10Fe",Z);
    Vektor_mpf_t x_Werte(x.d);
    x.Ausgabe_skaliert_Vektor_mpf_t(x_Werte);
    x_Werte-=x_korrekt;
    x_Werte.Ausgabe_wissenschaftlich("x_Diff");
    x_Werte.Ausgabe_wissenschaftlich_Datei("x_Diff",datei1);
    double Norm_Differenz=x_Werte.Norm();
    printf("\n\nNorm der Differenz: %e",Norm_Differenz);
    fprintf(datei1,"\n\nNorm der Differenz: %e", Norm_Differenz);
    for(i=0; i<x.d; i++){
        x[i].Rekonstrukt(s,t,q);
        mpf_set_z(Z,s);
        mpf_div_2exp(Z,Z,f);
        gmp_fprintf(datei1,"\nx(%d)=%.10Ff",i,Z);
        }
    x.Ausgabe_skaliert("x");

    AM_Zq.Ausgabe("AM");


    fclose(datei1);
    getchar();
    return 0;
    }
///Schritt 2
//d auf Null setzen, falls der (appr.) Test dies ergibt; korrigiere so fr Vektoren, die nur auf Grund von Rundungsfehlern!=0 sind
Hilfe=Bit;
Hilfe*=-1;
Hilfe+=1;
for(i=0; i<x.d; i++){
    d[i]=d[i]*Hilfe;
    }

Berechne_alpha(A,d,x,b,alpha,Index,AM_Zq,alpha_kleiner_eins,A.Zeilenzahl,n,t,k,q,q_1);

Hilfe=Bit;

Hilfe*=-1;
Hilfe.Add_Konst_int(1,q);//Hilfe=1-(d=?0)
d*=alpha;
d*=Hilfe;//korrigiere d
d.TruncPr();
x+=d;//x_{k+1}=x_k+alpha_k*d_k

Hilfe=Bit;

Index*=alpha_kleiner_eins;//Index_AM=Index_AM*(alpha_k<?1)
Hilfe*=-1;
Hilfe.Add_Konst_int(1,q);//Hilfe=1-(d=?0)
Index_AM*=Hilfe;//Index=Index_AM*(alpha_k<?=1)*(1-(d=?0))
AM_Zq+=Index;//W_{k+1}=W_k+I*(alpha<?0)*(1-(d=?0))

lambda_min*=-1;
lambda_min*=Bit;
AM_Zq+=lambda_min;//W_{k+1}=W_k+I*(alpha<?0)*(1-(d=?0))-d*I_lambda_min

for(i=0; i<A.Zeilenzahl; i++){//Bestimme die neuen Werte ax(i)
    Hilfe_Vektor.getZeilenvektor(A,i);
    Hilfe_Vektor.SkalarMult(x,ax[i]);
    }

Kopie_A=A;
Kopie_x=x;
Kopie_c=c;
}

getchar();

mpz_clear(s);
mpz_clear(unendlich);
return 0;
}

void Berechne_alpha(const Matrix_LA& A, const Vektor_LA& d, const Vektor_LA&x, Vektor_LA& b, Share& alpha, Vektor_LA& Index, Vektor_LA& AM_Zq, Share& alpha_kleiner_eins, int Dimension, int n, int t, int k, mpz_t& q, mpz_t& q_1){

int i;
Vektor_LA ax(Dimension, n, t, q, q_1, k, f);
Vektor_LA ad(Dimension, n, t, q, q_1, k, f);
Vektor_LA ad_Vergleich_Null(Dimension, n, t, q, q_1, k, f);
Vektor_LA Hilfe_Vektor(A.Spaltenzahl, n, t, q, q_1, k, f);
Vektor_LA eins_minus_AM_Zq=AM_Zq;
Vektor_LA Kriterium(Dimension, n, t, q, q_1, k, f);
Vektor_LA Kriterium_Komplement(Dimension, n, t, q, q_1, k, f);
Vektor_LA alpha_ad(Dimension, n, t, q, q_1, k, f);

mpz_t unendlich;
mpz_init_set_ui(unendlich,1);
mpz_mul_2exp(unendlich,unendlich,k-1);
mpz_sub_ui(unendlich,unendlich,1);

mpz_t minus_eins;
mpz_init_set(minus_eins,zwei_hoch_f);
mpz_mul_si(minus_eins,minus_eins,-1);


for(i=0; i<Dimension; i++){
    Hilfe_Vektor.getZeilenvektor(A,i);
    Hilfe_Vektor.SkalarMult(x,ax[i]);
    ax[i].TruncPr(k,f,t,q,q_1);
    }

for(i=0; i<A.Zeilenzahl; i++){
    Hilfe_Vektor.getZeilenvektor(A,i);
    Hilfe_Vektor.SkalarMult(d,ad[i]);

    ad[i].TruncPr(k,f,t,q,q_1);
    ad_Vergleich_Null[i]=ad[i];//berprfe, ob ad[i]<0 ist; die betreffenden Komponenten werden bei der Berechnung von alpha nicht betrachtet

    ad_Vergleich_Null[i].LTZ_approximativ(t,q,q_1,k);

    eins_minus_AM_Zq[i].Mul_Konst_int(-1,q);
    eins_minus_AM_Zq[i].Add_Konst_int(1,q);//eins_minus_AM_Zq[i]=1-AM_Zq[i]

    Kriterium[i]=eins_minus_AM_Zq[i];
    Kriterium[i].AND(ad_Vergleich_Null[i]);
    Kriterium_Komplement[i]=Kriterium[i];
    Kriterium_Komplement[i]*=-1;
    Kriterium_Komplement[i]+=1;

    alpha_ad[i]=b[i]-ax[i];
    alpha_ad[i]=Kriterium[i]*alpha_ad[i]-Kriterium_Komplement[i]*unendlich;
    ad[i]=Kriterium[i]*ad[i]+Kriterium_Komplement[i]*minus_eins;//nicht anwendbare Nenner auf Eins setzen
    }

alpha_ad.Min_Bruch(ad,Index);
Share Zaehler(n,t,q);
Share Nenner(n,t,q);
Zaehler=alpha_ad[Index];

Nenner=ad[Index];

alpha=Zaehler;
alpha.FPDiv(Nenner,t,q,q_1,k);//alpha=min((alpha_ad[i]-ax[i])/ad[i])

Zaehler=alpha;
Zaehler.LT(zwei_hoch_f,t,q,q_1,k);//berprfe, ob das Minimum grer 1 ist

alpha_kleiner_eins=Zaehler;
alpha=Zaehler*alpha+(-Zaehler+1)*zwei_hoch_f;

mpz_clear(minus_eins);
mpz_clear(unendlich);
}

void Zielfunktion(Matrix_LA& Q, Vektor_LA& x, Vektor_LA& c, mpf_t& Wert, int n, int t, int k, mpz_t& q, mpz_t& q_1){

Share Ergebnis(n,t,q);
Share Ergebnis2(n,t,q);
mpz_t z;
mpz_init(z);

Vektor_LA Kopie_x=x;
Kopie_x.MatrixMult(Q,Kopie_x);
Kopie_x.SkalarMult(x,Ergebnis);
Ergebnis.TruncPr(2*k,2*f,t,q,q_1);
Ergebnis.Div_Konst_int(2,q,q_1,k);

x.SkalarMult(c,Ergebnis2);
Ergebnis2.TruncPr(k,f,t,q,q_1);

Ergebnis+=Ergebnis2;
Ergebnis.Rekonstrukt(z,t,q);

mpf_set_z(Wert,z);
mpf_div_2exp(Wert,Wert,f);

mpz_clear(z);
}

void Erstelle_Gleichungssystem(Matrix_LA& M, Matrix_LA& Q, Matrix_LA& A){
//Erstelle das Gleichungssystem, dessen Lsung Lsung des Optimierungsproblems unter den Gleichheitsbedingungen ist

int i,j,l;

for(i=0; i<Q.Zeilenzahl; i++){//Linke obere Ecke mit Q befllen
    for(j=0; j<Q.Spaltenzahl; j++){
        M.M[i][j]=Q.M[i][j];
        }
    l=0;
    for(j=Q.Spaltenzahl;j<Q.Spaltenzahl+A.Zeilenzahl; j++){//Rechte obere Ecke mit -A^t befllen
        M.M[i][j]=-A.M[l][i];
        l++;
    }
}

l=0;
for(i=Q.Zeilenzahl; i<Q.Zeilenzahl+A.Zeilenzahl; i++){
    for(j=0; j<A.Spaltenzahl; j++){
        M.M[i][j]=A.M[l][j];
        }
    l++;
    }
}

void Erstelle_Vektor(Vektor_LA& V, Vektor_LA& b, int m, int n){//m=dim b, n=dim *this; n>=m; bentigt fr Optimierung
//Wird bentigt im Iterationsschritt der generischen aktiven Mengen Strategie
if(n<m || n<0 || m<0){
    printf("\nWarnung: Erstelle_Vektor");
    printf("\nDimensionen falsch. Erstellen des Vektors nicht mglich. Abbruch");
    exit(1);
    }

int i;

for(i=0; i<m; i++){
    V.Vector_LA.at(i)=b[i];
    V.Vector_LA.at(i)*=-1;//*this(i)=-b_i
    }
}

void Bestimme_Aktive_Menge(Matrix_LA& A, Vektor_LA& x, Vektor_LA& b, Vektor_LA& AM_Zq, int d, int k, int n, int t, mpz_t& q, mpz_t& q_1){//Bestimmung der aktiven Menge

int i;
mpz_t z;
mpz_init(z);

Vektor_LA Hilfsvektor=x;
Hilfsvektor.MatrixMult(A,AM_Zq);
AM_Zq.TruncPr();
AM_Zq-=b;

mpz_set_ui(z,8);
for(i=0; i<d; i++){
AM_Zq[i].Quadrat();
AM_Zq[i].TruncPr(k,f,t,q,q_1);
AM_Zq[i].LT(z,t,q,q_1,k);
}

mpz_clear(z);
}
