/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include<cstdlib>
#include<cstring>
#include<cstdio>
#include<ctime>
#include<Matrix_LA.h>
#include<Vektor_LA.h>
#include<head.h>


Matrix_LA::Matrix_LA(int a, int b, int n, int t, mpz_t& grosser_Modulus, mpz_t& kleiner_Modulus, int Bitlaenge, int Nachkommastellen)
{
    int i,j,l;

    Zeilenzahl=a;
    Spaltenzahl=b;
    Spielerzahl=n;
    Schwellenwert=t;
    mpz_init_set(q,grosser_Modulus);
    mpz_init_set(q_1,kleiner_Modulus);
    k=Bitlaenge;
    N=Nachkommastellen;

    M=new Share * [a]; //Initialisierung mit Standardkonstruktor
    for(i=0; i<a; i++){
        M[i]=new Share [b];
        }

    for(i=0; i<a; i++){
        for(j=0; j<b; j++){
            M[i][j].Vector=new mpz_t[n+1]; //Erzeugung des konkreten Vektors
            for(l=0; l<=n; l++){//Einsetzen der korrekten Werte;
                mpz_init_set_ui(M[i][j].Vector[l],0);
                }
            M[i][j].Laenge=n;
            M[i][j].Anzahl_Spieler=n;
            M[i][j].Schwelle=t;
            mpz_init_set(M[i][j].Modulus,grosser_Modulus);
            }
        }
}

Matrix_LA::Matrix_LA(int a, int b, int n, int t, const mpz_t& grosser_Modulus, const mpz_t& kleiner_Modulus, int Bitlaenge, int Nachkommastellen)
{
    int i,j,l;

    Zeilenzahl=a;
    Spaltenzahl=b;
    Spielerzahl=n;
    Schwellenwert=t;
    mpz_init_set(q,grosser_Modulus);
    mpz_init_set(q_1,kleiner_Modulus);
    k=Bitlaenge;
    N=Nachkommastellen;

    M=new Share * [a]; //Initialisierung mit Standardkonstruktor
    for(i=0; i<a; i++){
        M[i]=new Share [b];
        }

    for(i=0; i<a; i++){
        for(j=0; j<b; j++){
            M[i][j].Vector=new mpz_t[n+1]; //Erzeugung des konkreten Vektors
            for(l=0; l<=n; l++){//Einsetzen der korrekten Werte;
                mpz_init_set_ui(M[i][j].Vector[l],0);
                }
            M[i][j].Laenge=n;
            M[i][j].Anzahl_Spieler=n;
            M[i][j].Schwelle=t;
            mpz_init_set(M[i][j].Modulus,grosser_Modulus);
            }
        }
}

Matrix_LA::~Matrix_LA()
{
    int i;

    for(i=Zeilenzahl-1; i>=0; i--){
        delete []M[i];
        M[i]=NULL;
        }
    delete [] M;
    M=NULL;

    mpz_clear(q);
    mpz_clear(q_1);
}

Matrix_LA::Matrix_LA(const Matrix_LA& other):Zeilenzahl(other.Zeilenzahl), Spaltenzahl(other.Spaltenzahl), Spielerzahl(other.Spielerzahl), Schwellenwert(other.Schwellenwert), k(other.k), N(other.N)
{
    int i,j,l;

    mpz_init_set(q,other.q);
    mpz_init_set(q_1,other.q_1);

    M=new Share * [Zeilenzahl]; //Initialisierung mit Standardkonstruktor
    for(i=0; i<Zeilenzahl; i++){
        M[i]=new Share [Spaltenzahl];
        }


    for(i=0; i<Zeilenzahl; i++){
        for(j=0; j<Spaltenzahl; j++){
            M[i][j].Vector=new mpz_t[Spielerzahl+1]; //Erzeugung des konkreten Vektors
            for(l=0; l<=Spielerzahl; l++){//Einsetzen der korrekten Werte;
                mpz_init_set(M[i][j].Vector[l],other.M[i][j].Vector[l]);
                }
            M[i][j].Laenge=other.M[i][j].Laenge;
            M[i][j].Anzahl_Spieler=other.M[i][j].Anzahl_Spieler;
            M[i][j].Schwelle=other.M[i][j].Schwelle;
            mpz_init_set(M[i][j].Modulus,q);
            }
        }
}

Matrix_LA& Matrix_LA::operator=(const Matrix_LA& rhs)
{
    if (this == &rhs) return *this; // handle self assignment

    if((Zeilenzahl!=rhs.Zeilenzahl) || (Spaltenzahl!=rhs.Spaltenzahl)){
        printf("\n\nMatrix_LA::operator=");
        printf("\nZuweisung nicht moeglich. Dimensionen verschieden.");
        printf("\nZeilenzahl=%d, Spaltenzahl=%d, rhs.Zeilenzahl=%d, rhs.Spaltenzahl=%d",Zeilenzahl,Spaltenzahl,rhs.Zeilenzahl,rhs.Spaltenzahl);
        exit(1);
        }

    int i,j,l;

    for(i=0; i<Zeilenzahl; i++){
        for(j=0; j<Spaltenzahl; j++){
            for(l=0; l<=Spielerzahl; l++){//Einsetzen der korrekten Werte;
                mpz_set(M[i][j].Vector[l],rhs.M[i][j].Vector[l]);
                }
            }
        }

    return *this;
}

Matrix_LA& Matrix_LA::operator=(const Matrix_mpf_t& Init){

if(Zeilenzahl!=Init.Hoehe){
    printf("\nWarnung: Matrix_LA::operator=(const Matrix_mpf_t& Init)");
    printf("\nZuweisung der Matrix Init nicht moeglich, da Zeilenzahl verschieden. Abbruch.");
    exit(1);
    }

if(Spaltenzahl!=Init.Breite){
    printf("\nWarnung: Matrix_LA::operator=(const Matrix_mpf_t& Init)");
    printf("\nZuweisung der Matrix Init nicht moeglich, da Spaltenzahl verschieden. Abbruch.");
    exit(1);
    }

int i,j;

double d;
for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        d=mpf_get_d(Init.M[i][j]);
        M[i][j].Erzeugung_double(d,Schwellenwert,q);
        }
    }

return *this;
}

void Matrix_LA::Zufallsmatrix(int k){//Erzeugt Matrix mit zuflligen Eintrgen. Werte werden nicht skaliert.

int i,j;

mpz_t seed_mpz;
mpz_init(seed_mpz);

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        mpz_urandomb(seed_mpz,rState1,k);
        M[i][j].Erzeugung(seed_mpz,Schwellenwert,q);
        }
    }
mpz_clear(seed_mpz);
}

Matrix_LA Matrix_LA::operator=(int n){//Initialisiere die Matrix mit einem konstanten Wert (typischerweise 0)

    int i,j;

    for(i=0; i<Zeilenzahl; i++){
        for(j=0; j<Spaltenzahl; j++){
            M[i][j]=n;
            }
    }
    return *this;
}

Matrix_LA Matrix_LA::operator+(const Matrix_LA& Matrix_2){

assert(Zeilenzahl==Matrix_2.Zeilenzahl && Spaltenzahl==Matrix_2.Spaltenzahl);

int i,j;

Matrix_LA Ergebnis=*this;

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        Ergebnis.M[i][j].Plus_Vektor(Matrix_2.M[i][j].Vector,Spielerzahl,q);
        }
    }

return Ergebnis;
}

Matrix_LA& Matrix_LA::operator+=(const Matrix_LA& Matrix_2){

assert(Zeilenzahl==Matrix_2.Zeilenzahl && Spaltenzahl==Matrix_2.Spaltenzahl);

int i,j;


for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        this->M[i][j].Plus_Vektor(Matrix_2.M[i][j].Vector,Spielerzahl,q);
        }
    }

return *this;
}

Matrix_LA Matrix_LA::operator-(const Matrix_LA& Matrix_2){


assert(Zeilenzahl==Matrix_2.Zeilenzahl && Spaltenzahl==Matrix_2.Spaltenzahl);

int i,j;

Matrix_LA Ergebnis=*this;

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        Ergebnis.M[i][j].minus(Matrix_2.M[i][j].Vector,Spielerzahl,q);
        }
    }


return Ergebnis;
}

Matrix_LA& Matrix_LA::operator-=(const Matrix_LA& Matrix_2){


assert(Zeilenzahl==Matrix_2.Zeilenzahl && Spaltenzahl==Matrix_2.Spaltenzahl);

int i,j;


for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        this->M[i][j].minus(Matrix_2.M[i][j].Vector,Spielerzahl,q);
        }
    }
return *this;
}

Matrix_LA Matrix_LA::operator*(const Matrix_LA& A){

if(Spaltenzahl!=A.Zeilenzahl){
    printf("\nMatrixmultiplikation nicht mglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }
Matrix_LA Ergebnis(Zeilenzahl, A.Spaltenzahl, Spielerzahl, Schwellenwert, q, q_1, k, f);//Matrix, in die das Ergebnis geschrieben wird

Ergebnis=MatrixMultMatrix(A);

return Ergebnis;
}

Matrix_LA Matrix_LA::operator*(const Share& S){

Matrix_LA N(Zeilenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);

for(int i=0; i<Zeilenzahl; i++){
    for(int j=0; j<Spaltenzahl; j++){
        N.M[i][j]=M[i][j]*S;
        }
    }

return N;
}

Matrix_LA& Matrix_LA::operator*=(const Matrix_LA& A){

if(Zeilenzahl!=Spaltenzahl || Zeilenzahl!=A.Zeilenzahl || Zeilenzahl!=A.Spaltenzahl){
    printf("\nMatrixmultiplikation ohne Rckgabe nicht mglich, da Matrizen nicht beide quadratisch. Abbruch");
    exit(1);
    }

int i,j,s=A.Spaltenzahl;

Matrix_LA Ergebnis(Zeilenzahl, s, Spielerzahl, Schwellenwert, q, q_1, k, f);//Matrix, in die das Ergebnis geschrieben wird
Vektor_LA Hilfsvektor1(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);//Hilfsvektoren, in denen die Zeilen7Spalten der Scope-/Matrix2 gespeichert werden
Vektor_LA Hilfsvektor2(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);

for(i=0; i<Zeilenzahl; i++){
    Hilfsvektor1.getZeilenvektor(*this,i);
    for(j=0; j<s; j++){
        Hilfsvektor2.getSpaltenvektor(A,j);
        Hilfsvektor2.SkalarMult(Hilfsvektor1,Ergebnis.M[i][j]);
        }
    }

*this=Ergebnis;
return *this;
}

Matrix_LA& Matrix_LA::operator*=(int c){

mpz_t Konstante;
mpz_init_set_si(Konstante,c);

for(int i=0; i<Zeilenzahl; i++){
    for(int j=0; j<Spaltenzahl; j++){
        M[i][j]*=Konstante;
        }
    }

mpz_clear(Konstante);

return *this;
}

Matrix_LA Matrix_LA::Transponiere(){

Matrix_LA A(Spaltenzahl,Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);

int i,j;

for(i=0; i<Spaltenzahl; i++){
    for(j=0; j<Zeilenzahl; j++){
        A.M[i][j]=M[j][i];
        }
    }

return A;
}

void Matrix_LA::Transponiere_quadratisch(){

if(Zeilenzahl!=Spaltenzahl){
    printf("\nTransponiere_Quadratisch nicht anwendbar, da Matrix nicht quadratisch. Abbruch");
    exit(1);
    }

Matrix_LA A(Zeilenzahl,Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);

int i,j;

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Zeilenzahl; j++){
        A.M[i][j]=M[j][i];
        }
    }

*this=A;
}

void Matrix_LA::Erzeugung(mpz_t ** Werte){

int i,j;

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j].Erzeugung(Werte[i][j],Schwellenwert,q);
        }
    }
}

void Matrix_LA::Erzeugung_int(int ** Werte){

int i,j;
mpz_t ** Werte_mpz;
Werte_mpz=new mpz_t * [Zeilenzahl+1];

for(i=0; i<Zeilenzahl; i++){
    Werte_mpz[i]=new mpz_t[Spaltenzahl];
    for(j=0; j<Spaltenzahl; j++){
        mpz_init_set_si(Werte_mpz[i][j],Werte[i][j]);
        mpz_mul_2exp(Werte_mpz[i][j],Werte_mpz[i][j],f);
        }
    }

Erzeugung(Werte_mpz);

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        mpz_clear(Werte_mpz[i][j]);
        }
    delete []Werte_mpz[i];
    Werte_mpz[i]=NULL;
    }
delete Werte_mpz;
Werte_mpz=NULL;
}

void Matrix_LA::Erzeugung_double(double ** Werte){

int i,j;
mpf_t ** Werte_mpf;
mpz_t ** Werte_mpz;
Werte_mpf=new mpf_t * [Zeilenzahl];
Werte_mpz=new mpz_t * [Zeilenzahl];

for(i=0; i<Zeilenzahl; i++){
    Werte_mpf[i]=new mpf_t[Spaltenzahl];
    Werte_mpz[i]=new mpz_t[Spaltenzahl];
    for(j=0; j<Spaltenzahl; j++){
        mpf_init_set_d(Werte_mpf[i][j],Werte[i][j]);
        mpf_mul_2exp(Werte_mpf[i][j],Werte_mpf[i][j],f);
        mpz_init(Werte_mpz[i][j]);
        mpz_set_f(Werte_mpz[i][j],Werte_mpf[i][j]);
        }
    }

Erzeugung(Werte_mpz);

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        mpz_clear(Werte_mpz[i][j]);
        mpf_clear(Werte_mpf[i][j]);
        }
    delete []Werte_mpz[i];
    delete []Werte_mpf[i];
    Werte_mpf[i]=NULL;
    Werte_mpz[i]=NULL;
    }
delete []Werte_mpz;
delete []Werte_mpf;
Werte_mpf=NULL;
Werte_mpz=NULL;
}

void Matrix_LA::Ausgabe(std::string Name){

int i, j;
mpz_t c;
mpz_init(c);

char * s;
s=new char[Name.size()+1];
strcpy(s,Name.c_str());

printf("\n\n");

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j].Rekonstrukt_GRR(c,q,Schwellenwert);
        gmp_printf("%s[%d][%d]=%Zd ",s,i,j,c);
        }
    printf("\n");
    }
mpz_clear(c);
delete []s;
s=NULL;
}

void Matrix_LA::Ausgabe_skaliert(std::string Name){

int i, j;
mpz_t c;
mpz_init(c);
mpf_t s;
mpf_init(s);

char * r;
r=new char[Name.size()+1];
strcpy(r,Name.c_str());

printf("\n\n");

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j].Rekonstrukt_GRR_mit_VZ(c,q,Schwellenwert);
        mpf_set_z(s,c);
        mpf_div_2exp(s,s,f);
        gmp_printf("%s[%d][%d]=%.2Ff ",r,i,j,s);
        }
    printf("\n");
    }

mpz_clear(c);
mpf_clear(s);
delete []r;
r=NULL;
}

int Matrix_LA::Ausgabe_skaliert_Datei(std::string Name, FILE* datei1){

int i, j;
mpz_t c;
mpz_init(c);
mpf_t s;
mpf_init(s);

char * r;
r=new char[Name.size()+1];
strcpy(r,Name.c_str());

if(datei1==NULL){
    printf("\nFehler beim ffnen der Datei Ausgabe.txt");//Fehler beim ffnen
    return 1;
    }

fprintf(datei1, "\n");
for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j].Rekonstrukt_GRR_mit_VZ(c,q,Schwellenwert);
        mpf_set_z(s,c);
        mpf_div_2exp(s,s,f);
        gmp_fprintf(datei1,"%s[%d][%d]=%.4Ff ",r,i,j,s);
        }
    fprintf(datei1, "\n");
    }


mpz_clear(c);
delete []r;
r=NULL;
return 0;
}

void Matrix_LA::Ausgabe_skaliert_Matrix(mpf_t** N, int m, int n){

if(m!=Zeilenzahl || n!=Spaltenzahl){
    printf("\nAusgabe in Matrix nicht mglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

int i, j;
mpz_t c;
mpz_init(c);

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j].Rekonstrukt_GRR_mit_VZ(c,q,Schwellenwert);
        mpf_set_z(N[i][j],c);
        mpf_div_2exp(N[i][j],N[i][j],f);
        }
    }

mpz_clear(c);
}

void Matrix_LA::TruncPr(int k, int r){

int i,j;

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j].TruncPr(k,r,Schwellenwert,q,q_1);
        }
    }
}

void Matrix_LA::Multipliziere_Zeilenweise(std::vector<Share> * Binaerer_Vektor){

if(Binaerer_Vektor->size()!=(unsigned int)Zeilenzahl){
    printf("\nDimensionen verschieden. Zeilenweise Multiplikation nicht moeglich. Abbruch.");
    printf("\nBinaerer_Vektor hat %d Bits, Matrix hat %d Zeilen",(int)Binaerer_Vektor->size(),Zeilenzahl);
    exit(1);
    }

int i,j;

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j]*=Binaerer_Vektor->at(i);
        }
    }
}

void Matrix_LA::Multipliziere_Spaltenweise(std::vector<Share>& Binaerer_Vektor){

if(Binaerer_Vektor.size()!=(unsigned int)Spaltenzahl){
    printf("\nDimensionen verschieden. Spaltenweise Multiplikation nicht moeglich. Abbruch.");
    printf("\nBinaerer Vektor hat %d Bits, Matrix hat %d Spalten",(int)Binaerer_Vektor.size(),Spaltenzahl);
    exit(1);
    }

int i,j;

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j]*=Binaerer_Vektor[j];
        }
    }
}

void Matrix_LA::Multipliziere_Zeilenweise_ab_Zeile_X(std::vector<Share> * Binaerer_Vektor, int X){

if(Binaerer_Vektor->size()!=(unsigned int)Zeilenzahl){
    printf("\nDimensionen verschieden. Zeilenweise Multiplikation nicht moeglich. Abbruch.");
    printf("\nBinaerer_Vektor hat %d Bits, Matrix hat %d Zeilen",(int)Binaerer_Vektor->size(),Zeilenzahl);
    exit(1);
    }

int i,j;

for(i=X; i<Zeilenzahl; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j]*=Binaerer_Vektor->at(i);
        }
    }
}

void Matrix_LA::Rechts_unten_einpassen(Matrix_LA& N, int z, int s){

if(z>Zeilenzahl || s>Spaltenzahl){
    printf("\nWarnung: Matrix_LA::Rechts_unten_einpassen");
    printf("\nEinpassen der Matrix nicht mglich, da Dimensionen zu gross. Abbruch.");
    exit(1);
    }

int i,j,l=0,r=0;

for(i=Zeilenzahl-z; i<Zeilenzahl; i++){
    for(j=Spaltenzahl-s; j<Spaltenzahl; j++){
        M[i][j]=N.M[l][r];
        r++;
        }
    r=0;
    l++;
    }
}

void Matrix_LA::Rechts_unten_auspassen(Matrix_LA& N, int z, int s){

if(z>Zeilenzahl || s>Spaltenzahl){
    printf("\nWarnung: Matrix_LA::Rechts_unten_auspassen");
    printf("\nEinpassen der Matrix nicht mglich, da Dimensionen zu gro. Abbruch.");
    exit(1);
    }

int i,j,l=0,r=0;

for(i=Zeilenzahl-z; i<Zeilenzahl; i++){
    for(j=Spaltenzahl-s; j<Spaltenzahl; j++){
        N.M[l][r]=M[i][j];
        r++;
        }
    r=0;
    l++;
    }
}

void Matrix_LA::Links_oben_einpassen(Matrix_LA& N){//Passe die Matrix N - soweit als mglich - links oben in die Scope-Matrix ein

int z=haupt::min(Zeilenzahl,N.Zeilenzahl);
int s=haupt::min(Spaltenzahl,N.Spaltenzahl);

int i,j;

for(i=0; i<z; i++){
    for(j=0; j<s; j++){
        M[i][j]=N.M[i][j];
        }
    }
}


void Matrix_LA::Rechts_einpassen(Matrix_LA& A){//Bilde neue Matrix durch Nebeneinanderstellen der Scope-Matrix und der Matrix A

if(Zeilenzahl!=A.Zeilenzahl){
    printf("\nNebeneinanderschreiben der Matrizen nicht mglich, da unterschiedliche Zeilenzahl. Abbruch.");
    exit(1);
    }

int i,j;

for(i=0; i<Zeilenzahl; i++){
    for(j=Spaltenzahl-A.Spaltenzahl; j<Spaltenzahl; j++){
        M[i][j]=A.M[i][j-Spaltenzahl+A.Spaltenzahl];
        }
    }
}

void Matrix_LA::SecWriteSpalte(std::vector<Share>& V, std::vector<Share>& B, int V_d, int B_d){//V in die in B kodierte Spalte schreiben

if(V_d!=Zeilenzahl){
    printf("\nWarnung: Matrix_LA::SecWriteSpalte");
    printf("\nSpalte kann nicht geschrieben werden, da Dimensionen von Vektor und Matrix verschieden. Abbruch.");
    exit(1);
    }

if(B_d!=Spaltenzahl){
    printf("\nWarnung: Matrix_LA::SecWriteSpalte");
    printf("\nSpalte kann nicht geschrieben werden, da Dimensionen von Bit-Vektor und Matrix verschieden. Abbruch.");
    exit(1);
    }

for(int i=0; i<Zeilenzahl; i++){
    for(int j=0; j<Spaltenzahl; j++){
        M[i][j]=B[j]*V[i]+(-B[j]+1)*M[i][j];
        }
    }

}

void Matrix_LA::SecWriteZeile(std::vector<Share>& V, std::vector<Share>& B, int V_d, int B_d){//V in die in B kodierte Zeile schreiben


if(V_d!=Spaltenzahl){
    printf("\nWarnung: Matrix_LA::SecWriteZeile");
    printf("\nZeile kann nicht geschrieben werden, da Dimensionen von Vektor und Matrix verschieden. Abbruch.");
    exit(1);
    }

if(B_d!=Zeilenzahl){
    printf("\nWarnung: Matrix_LA::SecWriteZeile");
    printf("\nZeile kann nicht geschrieben werden, da Dimensionen von Bit-Vektor und Matrix verschieden. Abbruch.");
    exit(1);
    }

for(int i=0; i<Zeilenzahl; i++){
    for(int j=0; j<Spaltenzahl; j++){
        M[i][j]=B[i]*V[j]+(-B[i]+1)*M[i][j];
        }
    }
}

Matrix_LA Matrix_LA::MatrixMultMatrix(const Matrix_LA& A){

int i,j,s=A.Spaltenzahl;

if(Spaltenzahl!=A.Zeilenzahl){
    printf("\nWarnung: Matrix_LA::MatrixMultMatrix");
    printf("\nMatrizenmultiplikation nicht mglich, da Matrizen nicht kompatibel. Abbruch");
    exit(1);
    }


Matrix_LA Ergebnis(this->Zeilenzahl, s, this->Spielerzahl, this->Schwellenwert, this->q, this->q_1, this->k, this->N);//Matrix, in die das Ergebnis geschrieben wird
Vektor_LA Hilfsvektor1(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);//Hilfsvektoren, in denen die Zeilen7Spalten der Scope-/Matrix2 gespeichert werden
Vektor_LA Hilfsvektor2(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);

for(i=0; i<Zeilenzahl; i++){
    Hilfsvektor1.getZeilenvektor(*this,i);
    for(j=0; j<s; j++){
        Hilfsvektor2.getSpaltenvektor(A,j);
        Hilfsvektor2.SkalarMult(Hilfsvektor1,Ergebnis.M[i][j]);//Ohne TruncPr, da hufig nicht-skalierte Matrizen verwendet werden (z.B. bei Matrix_zusammen)
        }
    }

return Ergebnis;
}

Matrix_LA Matrix_LA::MatrixMultMatrix_ab_Zeile_l(Matrix_LA& A, int l){//Die ersten l Zeilen des Ergebnisses sind mit den Ersten l Zeilen von A identisch. Voraussetzung: Scope-Matrix quadratisch

int i,j,s=A.Spaltenzahl;

Matrix_LA Ergebnis(Zeilenzahl, s, Spielerzahl, Schwellenwert, q, q_1, k, f);//Matrix, in die das Ergebnis geschrieben wird
Vektor_LA Hilfsvektor1(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);//Hilfsvektoren, in denen die Zeilen/Spalten der Scope-/Matrix2 gespeichert werden
Vektor_LA Hilfsvektor2(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);

for(i=0; i<l; i++){//Die ersten l Zeilen sind mit der Matrix A identisch
    for(j=0; j<s; j++){
        Ergebnis.M[i][j]=A.M[i][j];
        }
}

for(i=l; i<Zeilenzahl; i++){
    Hilfsvektor1.getZeilenvektor(*this,i);
    for(j=0; j<s; j++){
        Hilfsvektor2.getSpaltenvektor(A,j);
        Hilfsvektor2.SkalarMult(Hilfsvektor1,Ergebnis.M[i][j]);
        }
    }

return Ergebnis;
}

void Matrix_LA::Vorwaertselimination_SMPC_mit_Pivot(std::vector<Share>& b_Strich){//Kann nicht Vektor b bergeben, da dazu Vektor_LA.h in Matrix_LA.h eingebunden werden msste; berschneidungsgefahr!
//Erzeugung der LR-Zerlegung; L wird links unten und R rechts oben gespeichert
int l,i,j;

if(Zeilenzahl!=Spaltenzahl){
    printf("\nDie Matrix ist nicht quadratisch. Die Vorwaertselimination kann nicht durchgefhrt werden. Abbruch");
    exit(1);
    }

Vektor_LA Pivot(Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
Matrix_LA Nullmatrix(Spaltenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
Matrix_LA P_Strich(Spaltenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);//P_Strich wird in jeder Runde neu verwendet; P wird fr die Initialisierung von P' bentigt
Vektor_LA b(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
b.Vector_LA=b_Strich;

Share Null(Spielerzahl,Schwellenwert,q);
mpz_t null;
mpz_init_set_ui(null,0);
Null.Erzeugung(null,Schwellenwert,q);

Vektor_LA Hilfe(Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);//wird bentigt fr Rckgabe bei Matrix-Vektor-Multiplikation
Share Hilfsshare(Spielerzahl,Schwellenwert,q);

for(l=0; l<Zeilenzahl-1; l++){
    if(l>0){//in der ersten Iteration ist P'=0 nach Konstruktion
    P_Strich=Nullmatrix;}

    for(i=0; i<Zeilenzahl; i++){
        Pivot.Vector_LA.at(i)=M[i][l];
        }

    P_Strich.Erstelle_Permutationsmatrix(Pivot.Vector_LA,l);

    *this=P_Strich.MatrixMultMatrix_ab_Zeile_l(*this,l);

    b.MatrixMult_ab_Zeile(P_Strich,Hilfe,Zeilenzahl,Zeilenzahl,l);
    b=Hilfe;

    for(i=l+1; i<Zeilenzahl; i++){
        M[i][l].FPDiv(M[l][l],Schwellenwert,q,q_1,k);
        for(j=l+1; j<Spaltenzahl; j++){
            Hilfsshare=M[i][l];
            Hilfsshare*=M[l][j];
            Hilfsshare.TruncPr(k,f,Schwellenwert,q,q_1);
            M[i][j]=M[i][j]-Hilfsshare;
            }
        }
    }
b_Strich=b.Vector_LA;

mpz_clear(null);
}

void Matrix_LA::Vorwaertselimination_SMPC_mit_Pivot(Matrix_LA& P){//Kann nicht Vektor b bergeben, da dazu Vektor_LA.h in Matrix_LA.h eingebunden werden msste; berschneidungsgefahr!
//Erzeugung der LR-Zerlegung; L wird links unten und R rechts oben gespeichert
int l,i,j;

if(Zeilenzahl!=Spaltenzahl){
    printf("\nDie Matrix ist nicht quadratisch. Die Vorwaertselimination kann nicht durchgefhrt werden. Abbruch");
    exit(1);
    }

Vektor_LA Pivot(Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
Matrix_LA Nullmatrix(Spaltenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
Matrix_LA P_Strich(Spaltenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);//P_Strich wird in jeder Runde neu verwendet; P wird fr die Initialisierung von P' bentigt

Share Null(Spielerzahl,Schwellenwert,q);
mpz_t null;
mpz_init_set_ui(null,0);
Null.Erzeugung(null,Schwellenwert,q);

Share Hilfsshare(Spielerzahl,Schwellenwert,q);
Share Hilfe(Spielerzahl,Schwellenwert,q);
for(l=0; l<Zeilenzahl-1; l++){
    if(l>0){//in der ersten Iteration ist P'=0 nach Konstruktion
    P_Strich=Nullmatrix;}

    for(i=0; i<Zeilenzahl; i++){
        Pivot.Vector_LA.at(i)=M[i][l];
        }
    P_Strich.Erstelle_Permutationsmatrix(Pivot.Vector_LA,l);
    if(l==0){
        P=P_Strich;
        }
    else{
        P=P_Strich.MatrixMultMatrix_ab_Zeile_l(P,l);
        }

    *this=P_Strich.MatrixMultMatrix_ab_Zeile_l(*this,l);
    for(i=l+1; i<Zeilenzahl; i++){
        if(i==l+1){
            Hilfe=M[l][l];
            Hilfe.Inverses_ohne_Trunc(q_1,k);
            }
        M[i][l]*=Hilfe;
        M[i][l].TruncPr(3*k-f,2*k-f,Schwellenwert,q,q_1);
        for(j=l+1; j<Spaltenzahl; j++){
            Hilfsshare=M[i][l];
            Hilfsshare*=M[l][j];
            Hilfsshare.TruncPr(k,f,Schwellenwert,q,q_1);
            M[i][j]=M[i][j]-Hilfsshare;
            }
        }
    }
mpz_clear(null);
}

void Matrix_LA::Kopiere_links(Matrix_LA& L){//Kopiere Matrix L von links in den Scope (so weit wie mglich)

if(L.Zeilenzahl!=Zeilenzahl){
    printf("\nWarnung: Matrix_LA::Kopiere_links");
    printf("\nZeilenzahl von L stimmt nicht mit der des Scopes berein. Matrix kann nicht kopiert werden. Abbruch.");
    exit(1);
    }

int i,j;

int r=haupt::min(Spaltenzahl,L.Spaltenzahl);

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<r; j++){
        M[i][j]=L.M[i][j];
        }
    }
}

void Matrix_LA::Kopiere_rechts(Matrix_LA& L){//Kopiere Matrix L von links in den Scope (so weit wie mglich)

if(L.Zeilenzahl!=Zeilenzahl){
    printf("\nWarnung: Matrix_LA::Kopiere_rechts");
    printf("\nZeilenzahl von L stimmt nicht mit der des Scopes berein. Matrix kann nicht kopiert werden. Abbruch.");
    printf("\nZeilenzahl=%d, L.Zeilenzahl=%d",Zeilenzahl,L.Zeilenzahl);
    exit(1);
    }

int i,j;

int r=haupt::min(Spaltenzahl,L.Spaltenzahl);

for(i=0; i<Zeilenzahl; i++){
    for(j=0; j<r; j++){
        M[i][j]=L.M[i][L.Spaltenzahl-r+j];
        }
    }
}

void Matrix_LA::Oben_einpassen(Matrix_LA& A){//Kopiere den oberen Teil der Matrix A in den Scope

if(Spaltenzahl!=A.Spaltenzahl){
    printf("\nEinpassen der Matrix A von oben nicht mglich, da Spaltenzahl verschieden. Abbruch");
    exit(1);
    }

int i,j;
int l=haupt::min(Zeilenzahl,A.Zeilenzahl);

for(i=0; i<l; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j]=A.M[i][j];
        }
    }
}

void Matrix_LA::Unten_einpassen(Matrix_LA& A){//Kopiere den unteren Teil der Matrix A in den Scope

if(Spaltenzahl!=A.Spaltenzahl){
    printf("\nFehler: Matrix_LA::Unten_einpassen");
    printf("\nEinpassen der Matrix A von oben nicht mglich, da Spaltenzahl verschieden. Abbruch");
    printf("\nthis->Spaltenahl=%d, A.Spaltenzahl=%d",Spaltenzahl,A.Spaltenzahl);
    exit(1);
    }

int i,j;
int l=haupt::min(Zeilenzahl,A.Zeilenzahl);

for(i=0; i<l; i++){
    for(j=0; j<Spaltenzahl; j++){
        M[i][j]=A.M[A.Spaltenzahl-l+i][j];
        }
    }
}

void Matrix_LA::Matrix_Zusammenschieben(std::vector<Share>& Aktive_Menge_q, Matrix_LA& P){

Matrix_LA Hilfe(Zeilenzahl, Spaltenzahl, Spielerzahl, Schwellenwert, q, q_1, k, f);
Share B(Spielerzahl,Schwellenwert,q);
std::vector<Share> Spalte(Zeilenzahl,B);
Vektor_LA Index(Zeilenzahl, Spielerzahl,Schwellenwert,q,q_1,k,f);
Matrix_LA Null(Zeilenzahl,Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
Null=0;
P=0;

if(P.Zeilenzahl!=P.Spaltenzahl){
    printf("\nP nicht quadratisch. Aufstellen einer Permutationsmatrix in P nicht mglich. Abbruch");
    exit(1);
    }

if(P.Spaltenzahl!=Zeilenzahl){
    printf("\nDimensionen von P stimmen nicht mit denen der Scope-Matrix ueberein. Abbruch");
    exit(1);
    }

int i,j;
mpz_t eins;
mpz_init_set_ui(eins,1);

Index.Vector_LA.at(0).Erzeugung(eins,Schwellenwert,q);
Vektor_LA Aktive_Menge(Aktive_Menge_q.size(),Spielerzahl,Schwellenwert,q,q_1,k,f);
Aktive_Menge.Vector_LA=Aktive_Menge_q;

for(j=0; j<Zeilenzahl; j++){
    Aktive_Menge[j].EQZ(Schwellenwert,q,q_1,k);
    Aktive_Menge[j].Mul_Konst_int(-1,q);
    Aktive_Menge[j].Add_Konst_int(1,q);//Indikatorbit umdrehen

    for(i=0; i<Zeilenzahl; i++){
        Spalte.at(i)=Aktive_Menge[j]*Index.Vector_LA.at(i);
        }

    P.Ersetze_Spalte(Spalte, j, P.Zeilenzahl);

    Index.erhoehe_Index_konditional(Aktive_Menge[j],q);
    }

    Hilfe=P.MatrixMultMatrix(*this);

*this=Hilfe;
mpz_clear(eins);
Spalte.clear();
}

void Matrix_LA::Matrix_spaltenweise_zusammenschieben_binaer(std::vector<Share>& Aktive_Menge){

Matrix_LA A_t(Spaltenzahl,Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
A_t=Transponiere();
Matrix_LA P(Spaltenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
A_t.Matrix_Zusammenschieben_binaer(Aktive_Menge,P);
*this=A_t.Transponiere();
}

void Matrix_LA::Matrix_spaltenweise_zusammenschieben_binaer(std::vector<Share>& Aktive_Menge, Matrix_LA& P){

if(P.Zeilenzahl!=P.Spaltenzahl){
    printf("\nFehler: Matrix_LA::Matrix_Zusammenschieben_binaer");
    printf("\nP nicht quadratisch. Aufstellen einer Permutationsmatrix in P nicht mglich. Abbruch");
    exit(1);
    }

if(P.Zeilenzahl!=Spaltenzahl){
    printf("\nFehler: Matrix_LA::Matrix_Zusammenschieben_binaer");
    printf("\nDimensionen von P stimmen nicht mit denen der Scope-Matrix ueberein. Abbruch");
    printf("\nP.Zeilenzahl=%d, Spaltenzahl=%d",P.Zeilenzahl,Spaltenzahl);
    exit(1);
    }

Matrix_LA A_t(Spaltenzahl,Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
A_t=Transponiere();
A_t.Matrix_Zusammenschieben_binaer(Aktive_Menge,P);
*this=A_t.Transponiere();
}

void Matrix_LA::Matrix_Zusammenschieben_binaer(std::vector<Share>& Aktive_Menge_q, Matrix_LA& P){//Der Vektor Aktive_Menge_q wird als binr vorausgesetzt mit 1 an den Stellen der Nullzeilen der Scope-Matrix

Matrix_LA Hilfe(Zeilenzahl, Spaltenzahl, Spielerzahl, Schwellenwert, q, q_1, k, f);
Share B(Spielerzahl,Schwellenwert,q);
std::vector<Share> Spalte(Zeilenzahl,B);
Vektor_LA Index(Zeilenzahl, Spielerzahl,Schwellenwert,q,q_1,k,f);

P=0;

if(P.Zeilenzahl!=P.Spaltenzahl){
    printf("\nFehler: Matrix_LA::Matrix_Zusammenschieben_binaer");
    printf("\nP nicht quadratisch. Aufstellen einer Permutationsmatrix in P nicht mglich. Abbruch");
    exit(1);
    }

if(P.Spaltenzahl!=Zeilenzahl){
    printf("\nFehler: Matrix_LA::Matrix_Zusammenschieben_binaer");
    printf("\nDimensionen von P stimmen nicht mit denen der Scope-Matrix ueberein. Abbruch");
    printf("\nP.Spaltenzahl=%d, Zeilenzahl=%d",P.Spaltenzahl,Zeilenzahl);
    exit(1);
    }

int i,j;
mpz_t eins;
mpz_init_set_ui(eins,1);

Index.Vector_LA.at(0).Erzeugung(eins,Schwellenwert,q);
Vektor_LA Aktive_Menge(Aktive_Menge_q.size(),Spielerzahl,Schwellenwert,q,q_1,k,f);
Aktive_Menge.Vector_LA=Aktive_Menge_q;

for(j=0; j<Zeilenzahl; j++){
    for(i=0; i<Zeilenzahl; i++){
        Spalte[i]=Aktive_Menge[j]*Index.Vector_LA.at(i);
        }

    P.Ersetze_Spalte(Spalte, j, P.Zeilenzahl);

    Index.erhoehe_Index_konditional(Aktive_Menge[j],q);
    }
    Hilfe=P.MatrixMultMatrix(*this);

*this=Hilfe;
mpz_clear(eins);
Spalte.clear();
}

void Matrix_LA::Matrix_Zusammenschieben_LT(std::vector<Share>& Aktive_Menge_q){//wie Matrix_Zusammenschieben, nur da nicht auf Gleichheit mit Null, sondern auf Quadrat(Betrag)<2^-11 geprft wird

Matrix_LA A_Strich(Zeilenzahl, Spaltenzahl, Spielerzahl, Schwellenwert, q, q_1, k, f);
Matrix_LA Hilfe(Zeilenzahl, Spaltenzahl, Spielerzahl, Schwellenwert, q, q_1, k, f);
Matrix_LA A(Zeilenzahl, Zeilenzahl, Spielerzahl, Schwellenwert, q, q_1, k, f);//Permutationsmatrix
Vektor_LA Index(Zeilenzahl, Spielerzahl,Schwellenwert,q,q_1,k,f);
Matrix_LA Null(Zeilenzahl,Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
Null=0;


int i,j;
mpz_t eins;
mpz_init_set_ui(eins,1);
mpz_t Grenze;
mpz_init_set_ui(Grenze,1);
mpz_mul_2exp(Grenze,Grenze,f-11);

Index.Vector_LA.at(0).Erzeugung(eins,Schwellenwert,q);
Vektor_LA Aktive_Menge(Aktive_Menge_q.size(),Spielerzahl,Schwellenwert,q,q_1,k,f);
Aktive_Menge.Vector_LA=Aktive_Menge_q;

for(j=0; j<Zeilenzahl; j++){
    Aktive_Menge[j].Quadrat();
    Aktive_Menge[j].LT(Grenze,Schwellenwert,q,q_1,k);//berprfe, ob Wert^2 < Grenze
    Aktive_Menge[j].Mul_Konst_int(-1,q);
    Aktive_Menge[j].Add_Konst_int(1,q);//Indikatorbit umdrehen

    A=Null;

    for(i=0; i<Zeilenzahl; i++){
        A.M[i][j]=Aktive_Menge[j]*Index.Vector_LA.at(i);
        }

    Index.erhoehe_Index_konditional(Aktive_Menge[j],q);

    Hilfe=0;
    Hilfe=A.MatrixMultMatrix(*this);
    A_Strich+=Hilfe;
}

*this=A_Strich;
mpz_clear(eins);
mpz_clear(Grenze);
}

void Matrix_LA::Ersetze_Spalte(std::vector<Share>& Spalte, int Spaltenindex, int l){

if(l!=Zeilenzahl){
    printf("\nWarnung: Matrix_LA::Ersetze_Spalte");
    printf("\nErsetzen der Spalte nicht mglich. Dimensionen verschieden. Abbruch");
    exit(1);
    }

if(Spaltenindex>=Spaltenzahl){
    printf("\nWarnung: Matrix_LA::Ersetze_Spalte");
    printf("\nErsetzen der Spalte nicht mglich. Index zu hoch. Abbruch");
    exit(1);
    }

int i;

for(i=0; i<Zeilenzahl; i++){
    M[i][Spaltenindex]=Spalte[i];
    }
}

void Matrix_LA::Ersetze_Zeile(std::vector<Share>& Zeile, int Zeilenindex, int l){

if(l!=Zeilenzahl){
    printf("\nWarnung: Matrix_LA::Ersetze_Zeile");
    printf("\nErsetzen der Spalte nicht mglich. Dimensionen verschieden. Abbruch");
    exit(1);
    }

if(Zeilenindex>=Spaltenzahl){
    printf("\nWarnung: Matrix_LA::Ersetze_Zeile");
    printf("\nErsetzen der Spalte nicht mglich. Index zu hoch. Abbruch");
    exit(1);
    }

int i;

for(i=0; i<Spaltenzahl; i++){
    M[Zeilenindex][i]=Zeile[i];
    }
}

void Matrix_LA::Einheitsmatrix(mpz_t& Q){

int i;
Share Eins(Spielerzahl,Schwellenwert,Q);
mpz_t eins;
mpz_init_set_ui(eins,1);
Eins.Erzeugung(eins,Schwellenwert,Q);

if(Zeilenzahl!=Spaltenzahl){
    printf("\nZuweisung der Einheitsmatrix nicht mglich, da nicht quadratisch. Abbruch.");
    exit(1);
    }

for(i=0; i<Zeilenzahl; i++){
    M[i][i]=Eins;
    }

mpz_clear(eins);
}

void Matrix_LA::Einheitsmatrix_Fixpunkt(mpz_t& Q){

int i;
Share Eins(Spielerzahl,Schwellenwert,Q);
mpz_t eins;
mpz_init_set_ui(eins,1);
mpz_mul_2exp(eins,eins,f);
Eins.Erzeugung(eins,Schwellenwert,Q);

if(Zeilenzahl!=Spaltenzahl){
    printf("\nZuweisung der Einheitsmatrix nicht mglich, da nicht quadratisch. Abbruch.");
    exit(1);
    }

for(i=0; i<Zeilenzahl; i++){
    M[i][i]=Eins;
    }

mpz_clear(eins);
}

void Matrix_LA::Erstelle_Permutationsmatrix(std::vector<Share>& Spalte, int l){//Erstelle Permutationsmatrix zum Vertauschen der k-ten und der Zeile mit dem hchsten Betrag mit grerem Index

int i,z=0;
Vektor_LA Index(Zeilenzahl-l,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA Pivot(Zeilenzahl-l,Spielerzahl,Schwellenwert,q,q_1,k,f);

mpz_t eins;
mpz_init_set_ui(eins,1);
Share Eins(Spielerzahl,Schwellenwert,q);
Eins.Erzeugung(eins,Schwellenwert,q);

if(Zeilenzahl!=Spaltenzahl){
    printf("\nErstellen der Permutationsmatrix nicht mglich, da Matrix nicht quadratisch. Abbruch");
    exit(1);
    }

for(i=l; i<Zeilenzahl; i++){
    Pivot.Vector_LA.at(z)=Spalte.at(i);
    z++;
    }

Pivot.getPivot(Index);

for(i=0; i<l; i++){
    M[i][i]=Eins;
    }

z=0;

for(i=l; i<Zeilenzahl; i++){
    M[i][l]=Index.Vector_LA.at(z);
    M[l][i]=Index.Vector_LA.at(z);
    if(i!=l){
        M[i][i]=Eins-M[l][i];}
    z++;
    }
mpz_clear(eins);
}

void Matrix_LA::Matrix_aus_Vektor(std::vector<Share>& v_Strich, std::vector<Share>& w_Strich, int Laenge_v, int Laenge_w){//Berechne v'w'^t

Vektor_LA v(Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
v.Vector_LA=v_Strich;
Vektor_LA w(Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
w.Vector_LA=w_Strich;

int i,j;

for(i=0; i<Laenge_v; i++){
    for(j=0; j<Laenge_w; j++){
        M[i][j]=v[i]*w[j];
        M[i][j].TruncPr(k,f,Schwellenwert,q,q_1);
        }
    }
}

void Matrix_LA::Householder_Matrix_neu(std::vector<Share>& v, int d){

if(d!=Zeilenzahl || d!=Spaltenzahl){
    printf("\nDimensionen stimmen nicht berein. Erstellen der Householder-Matrix nicht mglich. Abbruch");
    exit(1);
    }

Einheitsmatrix_Fixpunkt(q);

Vektor_LA vv(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
vv.Vector_LA=v;
Share beta(Spielerzahl,Schwellenwert,q);
vv.house_neu();

Matrix_LA V(d,d,Spielerzahl,Schwellenwert,q,q_1,k,f);
V.Matrix_aus_Vektor(vv.Vector_LA,vv.Vector_LA,vv.d,vv.d);

Share Produkt(Spielerzahl,Schwellenwert,q);

Vektor_LA v2=vv;
Share Wert(Spielerzahl,Schwellenwert,q);
vv.SkalarMult(v2,Wert);
Wert.TruncPr(k,f,Schwellenwert,q,q_1);

Wert.Inverses_ohne_Trunc(q_1,k);

int i,j;
for(i=0; i<d; i++){
    for(j=0; j<d; j++){
        V.M[i][j]*=Wert;
        V.M[i][j]*=-2;
        V.M[i][j].TruncPr(3*k-f,2*k-f,Schwellenwert,q,q_1);
        }
    }

*this+=V;
v=vv.Vector_LA;
}

void Matrix_LA::Multipliziere_Householder_Prae(std::vector<Share>& v_Strich, Share& beta, int d){

Vektor_LA v(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
v.Vector_LA=v_Strich;
Vektor_LA v2=v;

Vektor_LA w(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);

v2.SkalarMult(v,beta);
beta.TruncPr(k,f,Schwellenwert,q,q_1);
beta.Inverses(q_1,k);
beta*=-2;

Matrix_LA A(Spaltenzahl,Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
A=Transponiere();

v.MatrixMult(A,w);

w*=beta;
w.TruncPr(2*k,2*f);
Matrix_LA B(Zeilenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
B.Matrix_aus_Vektor(v.Vector_LA,w.Vector_LA,v.d,w.d);
*this+=B;
}

void Matrix_LA::Multipliziere_Householder_Prae_beta_bekannt(std::vector<Share>& v_Strich, Share& beta, int d){

Vektor_LA v(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
v.Vector_LA=v_Strich;
Vektor_LA w(d,Spielerzahl,Schwellenwert,q,q_1,k,f);

Matrix_LA A=*this;
A.Transponiere_quadratisch();

v.MatrixMult(A,w);
w.TruncPr();
w*=beta;
w.TruncPr();
A.Matrix_aus_Vektor(v.Vector_LA,w.Vector_LA,v.d,w.d);
*this+=A;
}

void Matrix_LA::Multipliziere_Householder_Post_beta_bekannt(std::vector<Share>& v_Strich, Share& beta, int d){

Vektor_LA v(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
v.Vector_LA=v_Strich;

Vektor_LA w(d,Spielerzahl,Schwellenwert,q,q_1,k,f);

Matrix_LA A=*this;

v.MatrixMult(A,w);
w.TruncPr();
w*=beta;
w.TruncPr();
A.Matrix_aus_Vektor(w.Vector_LA,v.Vector_LA,w.d,v.d);
*this+=A;
}

void Matrix_LA::QR_Householder(Matrix_LA& Q){

if(Spaltenzahl>Zeilenzahl){
    printf("\nMehr Spalten als Zeilen im Scope. QR-Zerlegung nicht mglich. Abbruch");
    exit(1);
    }

if(Q.Zeilenzahl!=Zeilenzahl || Q.Spaltenzahl!=Spaltenzahl){
    printf("\nFalsche Dimensionen von Q. QR-Zerlegung nicht mglich. Abbruch");
    exit(1);
    }

int i,j;

Vektor_LA beta(Spaltenzahl-1,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA V(Spaltenzahl-1,Spielerzahl,Schwellenwert,q,q_1,k,f);

for(i=0; i<Spaltenzahl-1; i++){
    Vektor_LA v(Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    v.getSpaltenvektor_ab_Zeile(*this,i,i);
    v.house_neu();//Spare die Division in house() ein

    Matrix_LA N(Zeilenzahl-i,Spaltenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Rechts_unten_auspassen(N,Zeilenzahl-i,Spaltenzahl-i);
    N.Multipliziere_Householder_Prae(v.Vector_LA,beta[i],Zeilenzahl-i);
    Rechts_unten_einpassen(N,Zeilenzahl-i,Spaltenzahl-i);

    if(i<Zeilenzahl-1){
        V[i]=v[0];
        for(j=i+1; j<Zeilenzahl; j++){
            M[j][i]=v[j-i];
            }
        }
    }

Share Eins(Spielerzahl,Schwellenwert,q);
Eins.Erzeugung_double(1.0,Schwellenwert,q);
Q.Einheitsmatrix_Fixpunkt(q);

for(i=Spaltenzahl-2; i>=0; i--){
    Vektor_LA v(Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    v.getSpaltenvektor_ab_Zeile(*this,i,i);
    v[0]=V[i];
    Matrix_LA N(Zeilenzahl-i,Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Q.Rechts_unten_auspassen(N,Zeilenzahl-i,Zeilenzahl-i);
    N.Multipliziere_Householder_Prae_beta_bekannt(v.Vector_LA,beta[i],Zeilenzahl-i);
    Q.Rechts_unten_einpassen(N,Zeilenzahl-i,Zeilenzahl-i);
    }
}

void Matrix_LA::QR_Householder_neu_Rechtecksmatrix(std::vector<Share>& b, std::vector<Share>& beta){//Algorithmus nach Golub/Van Loan S.212

int i,j,n=Spielerzahl,t=Schwellenwert;

if(Spaltenzahl==1){
    Vektor_LA v(Zeilenzahl,n,t,q,q_1,k,f);
    v.getSpaltenvektor(*this,0);

    Vektor_LA w=v;
    v.house_neu();

    beta.at(0)=v[0];
    Share vtx(Spielerzahl,Schwellenwert,q);
    Share vtv(Spielerzahl,Schwellenwert,q);
    v.SkalarMult(w,vtx);
    vtx.TruncPr(k,f,t,q,q_1);

    v.SkalarMult(v,vtv);
    vtv.TruncPr(k,f,t,q,q_1);

    vtv.Inverses(q_1,k);
    vtv*=-2;

    b.at(0)=vtv;
    vtx*=vtv;
    vtx*=v[0];
    vtx.TruncPr(2*k,2*f,t,q,q_1);

    M[0][0]+=vtx;

    for(i=1; i<Zeilenzahl; i++){
        M[i][0]=v[i];
        }
    }

else{
    for(j=0; j<Spaltenzahl; j++){
        Vektor_LA v(Zeilenzahl-j,n,t,q,q_1,k,f);
        v.getSpaltenvektor_ab_Zeile(*this,j,j);
        v.house_neu();
        beta.at(j)=v[0];

        Matrix_LA N(Zeilenzahl-j,Spaltenzahl-j,n,t,q,q_1,k,f);
        Rechts_unten_auspassen(N,Zeilenzahl-j,Spaltenzahl-j);
        N.Multipliziere_Householder_Prae(v.Vector_LA,b.at(j),Zeilenzahl-j);
        Rechts_unten_einpassen(N,Zeilenzahl-j,Spaltenzahl-j);

        if(j<Zeilenzahl-1){
            for(i=j+1; i<Zeilenzahl; i++){
                M[i][j]=v[i-j];
                }
            }
        }
    }
}

void Matrix_LA::QR_Householder_neu(std::vector<Share>& b, std::vector<Share>& beta){//Algorithmus nach Golub/Van Loan S.212

int i,j,n=Spielerzahl,t=Schwellenwert;

for(j=0; j<Spaltenzahl-1; j++){
    Vektor_LA v(Zeilenzahl-j,n,t,q,q_1,k,f);
    v.getSpaltenvektor_ab_Zeile(*this,j,j);
    v.house_neu();
    beta.at(j)=v[0];

    Matrix_LA N(Zeilenzahl-j,Spaltenzahl-j,n,t,q,q_1,k,f);
    Rechts_unten_auspassen(N,Zeilenzahl-j,Spaltenzahl-j);
    N.Multipliziere_Householder_Prae(v.Vector_LA,b.at(j),Zeilenzahl-j);
    Rechts_unten_einpassen(N,Zeilenzahl-j,Spaltenzahl-j);

    if(j<Zeilenzahl-1){
        for(i=j+1; i<Zeilenzahl; i++){
            M[i][j]=v[i-j];
            }
        }
    }

}

void Matrix_LA::Berechne_Q_neu(std::vector<Share>& beta, std::vector<Share>& b){//in b sind die oberen Werte der HH-Vektoren gespeichert
//Argumente i.V. zu QR_Householder_neu vertauscht
Matrix_LA Q(Spaltenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
Q.Einheitsmatrix_Fixpunkt(q);// Q=E_n

int i,n=Spielerzahl,t=Schwellenwert;

Share Eins(n,t,q);
Eins.Erzeugung_double(1.0,t,q);

for(i=Zeilenzahl-2; i>=0; i--){
    Vektor_LA v(Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    v.getSpaltenvektor_ab_Zeile(*this,i,i);
    v[0]=b.at(i);
    Matrix_LA N(Zeilenzahl-i,Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Q.Rechts_unten_auspassen(N,Zeilenzahl-i,Zeilenzahl-i);
    N.Multipliziere_Householder_Prae_beta_bekannt(v.Vector_LA,beta.at(i),Zeilenzahl-i);
    Q.Rechts_unten_einpassen(N,Zeilenzahl-i,Zeilenzahl-i);
    }
*this=Q;
}

void Matrix_LA::Berechne_Q_neu(std::vector<Share>& beta, std::vector<Share>& b, int Anzahl_HH_Vektoren){//in b sind die oberen Werte der HH-Vektoren gespeichert; fr den Fall, da weniger als dim Q-1 Vektoren vorgesehen sind
//Argumente i.V. zu QR_Householder_neu vertauscht
Matrix_LA Q(Spaltenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
Q.Einheitsmatrix_Fixpunkt(q);// Q=E_n

int i,n=Spielerzahl,t=Schwellenwert;

Share Eins(n,t,q);
Eins.Erzeugung_double(1.0,t,q);

for(i=Anzahl_HH_Vektoren-1; i>=0; i--){
    Vektor_LA v(Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    v.getSpaltenvektor_ab_Zeile(*this,i,i);
    v[0]=b.at(i);
    Matrix_LA N(Zeilenzahl-i,Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Q.Rechts_unten_auspassen(N,Zeilenzahl-i,Zeilenzahl-i);

    N.Multipliziere_Householder_Prae_beta_bekannt(v.Vector_LA,beta.at(i),Zeilenzahl-i);
    Q.Rechts_unten_einpassen(N,Zeilenzahl-i,Zeilenzahl-i);
    }
*this=Q;
}

void Matrix_LA::Berechne_Q_konditional(std::vector<Share>& beta, std::vector<Share>& Index){//Wie Berechne_Q; allerdings werden nur die HH-Vektoren bercksichtigt, die in 'Index' vermerkt sind

Matrix_LA Q(Spaltenzahl,Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
Share Hilfe(Spielerzahl,Schwellenwert,q);
Q.Einheitsmatrix_Fixpunkt(q);// Q=E_n

int i,n=Spielerzahl,t=Schwellenwert;

Share Eins(n,t,q);
Eins.Erzeugung_double(1.0,t,q);

for(i=Zeilenzahl-2; i>=0; i--){
    Vektor_LA v(Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    v.getSpaltenvektor_ab_Zeile(*this,i,i);
    v[0]=Eins;
    Matrix_LA N(Zeilenzahl-i,Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Q.Rechts_unten_auspassen(N,Zeilenzahl-i,Zeilenzahl-i);
    Matrix_LA MM=N;
    N.Multipliziere_Householder_Prae_beta_bekannt(v.Vector_LA,beta.at(i),Zeilenzahl-i);
    Hilfe=Index.at(i);
    Hilfe*=-1;
    Hilfe+=1;
    N=N*Index.at(i)+MM*Hilfe;
    Q.Rechts_unten_einpassen(N,Zeilenzahl-i,Zeilenzahl-i);
    }
*this=Q;
}

void Matrix_LA::Cholesky(){//Berechne Cholesky-Zerlegung von s.p.d. Matrix G
//Die untere Dreiecksmatrix wird berschrieben; die obere wird nach der Berechnung aus der unteren kopiert
if(Zeilenzahl!=Spaltenzahl){
    printf("\nCholesky-Zerlegung nicht mglich, da Matrix nicht quadratisch. Abbruch.");
    exit(1);
    }

int i,j,l;

Share Inverses(Spielerzahl, Schwellenwert, q);
Share Hilfe(Spielerzahl, Schwellenwert, q);

for(l=0; l<Zeilenzahl; l++){
    M[l][l].Wurzel(k,q,q_1);
    Inverses=M[l][l];
    Inverses.Inverses(q_1,k);
    for(i=l+1; i<Zeilenzahl; i++){
        M[i][l]=M[i][l]*Inverses;
        M[i][l].TruncPr(k,f,Schwellenwert,q,q_1);
        }
    for(j=l+1; j<Zeilenzahl; j++){
        for(i=j; i<Zeilenzahl; i++){
            Hilfe=M[j][l]*M[i][l];
            Hilfe.TruncPr(k,f,Schwellenwert,q,q_1);
            M[i][j]=M[i][j]-Hilfe;
            }
        }
    }

for(l=0; l<=Zeilenzahl-2; l++){
    for(i=Spaltenzahl-1; i>=l+1; i--){
        M[l][i]=M[i][l];
        }
    }
}

void Matrix_LA::CG(std::vector<Share>& b_Strich, std::vector<Share>& x_Strich, int D){

Vektor_LA b(D,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA x(D,Spielerzahl,Schwellenwert,q,q_1,k,f);
b.Vector_LA=b_Strich;
x.Vector_LA=x_Strich;

if(x.d!=Zeilenzahl){
    printf("\nWarnung: Matrix_LA::CG");
    printf("\nAnwendung des CG-Algorithmus nicht mglich, da Dimension der Matrix von x und b verschieden. Abbruch");
    exit(1);
    }

x=0;//Startwert ist Null; spart eine Matrix-Vektor-Multiplikation
Vektor_LA d=b;
Vektor_LA g=b;
g*=-1;
Share alpha(Spielerzahl,Schwellenwert,q);
Share beta(Spielerzahl,Schwellenwert,q);
Share z(Spielerzahl,Schwellenwert,q);
Share z2(Spielerzahl,Schwellenwert,q);
Vektor_LA Ad(x.d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Share n(Spielerzahl, Schwellenwert, q);
Vektor_LA alpha_d(d.d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA alpha_Ad(d.d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA beta_d(d.d,Spielerzahl,Schwellenwert,q,q_1,k,f);

int i;

g.SkalarMult(g,z);
z.TruncPr(k,f,Schwellenwert,q,q_1);

for(i=0; i<x.d; i++){
    d.MatrixMult(*this,Ad);
    Ad.TruncPr(k,f);
    d.SkalarMult(Ad,n);
    n.TruncPr(k,f,Schwellenwert,q,q_1);
    alpha=z;
    alpha.FPDiv(n,Schwellenwert,q,q_1,k);
    alpha_d=d*alpha;
    alpha_d.TruncPr(k,f);
    x+=alpha_d;
    alpha_Ad=Ad*alpha;
    alpha_Ad.TruncPr(k,f);
    g+=alpha_Ad;
    g.SkalarMult(g,z2);
    z2.TruncPr(k,f,Schwellenwert,q,q_1);
    beta=z2;
    beta.FPDiv(z,Schwellenwert,q,q_1,k);
    z=z2;
    beta_d=d*beta;
    beta_d.TruncPr(k,f);
    d=beta_d-g;
    }

x_Strich=x.Vector_LA;
}

void Matrix_LA::Loesche_links_unten(){

if(Spaltenzahl!=Zeilenzahl){
    printf("\nLschen unterhalb der Diagonale nicht mglich, da Matrix nicht quadratisch. Abbruch.");
    exit(1);
    }
int i,j;

for(j=0; j<Spaltenzahl; j++){
    for(i=j+1;i<Zeilenzahl;i++){
        M[i][j]=0;
        }
    }
}

void Matrix_LA::Loese_GLS_LR(std::vector<Share>& b_Strich, int d){
if(d!=Spaltenzahl){
    printf("\nLoesen des GLS nicht moeglich, da Vektor nicht zu Matrix passt. Abbruch.");
    exit(1);
    }
Vektor_LA b(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
b.Vector_LA=b_Strich;

Matrix_LA P(Zeilenzahl,Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,N);
Vorwaertselimination_SMPC_mit_Pivot(P);
Vektor_LA HV(b.d,Spielerzahl,Schwellenwert,q,q_1,k,f);
b.MatrixMult(P,HV);
b=HV;

b.VWSubs(*this);

Vektor_LA x(b.d,Spielerzahl,Schwellenwert,q,q_1,k,f);

x.RWSubs(*this,b);
b_Strich=x.Vector_LA;
}

void Matrix_LA::Loese_GLS_QR(std::vector<Share>&b_Strich, int d){
    if(d!=Spaltenzahl){
    printf("\nLoesen des GLS nicht moeglich, da Vektor nicht zu Matrix passt. Abbruch.");
    exit(1);
    }
Vektor_LA b(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
b.Vector_LA=b_Strich;
Vektor_LA beta(d-1,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA B(d-1,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA x(d,Spielerzahl,Schwellenwert,q,q_1,k,f);

QR_Householder_neu(B.Vector_LA,beta.Vector_LA);

int i;

for(i=0; i<d-1; i++){
    Vektor_LA v(Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    v.getSpaltenvektor_ab_Zeile(*this,i,i);
    v[0]=beta[i];
    Vektor_LA bb(Zeilenzahl-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    bb.Kopiere_unten(b);
    bb.Mul_house_Pre_beta_bekannt(v,B[i]);
    b.Kopiere_unten(bb);
    }
x.RWSubs(*this,b);
b_Strich=x.Vector_LA;
}

void Matrix_LA::Loese_GLS_Cholesky(std::vector<Share>& b_Strich, int d){

if(Spaltenzahl!=Zeilenzahl){
    printf("\nLoesen des Gleichungssystems nicht moeglich, da Matrix nicht quadratisch. Abbruch.");
    exit(1);
    }

if(d!=Spaltenzahl){
    printf("\nLoesen des GLS nicht moeglich, da Vektor nicht zu Matrix passt. Abbruch.");
    exit(1);
    }

Vektor_LA b(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
b.Vector_LA=b_Strich;

Vektor_LA Hilfsvektor(b.d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Cholesky();
//Ausgabe_skaliert("C");
b.VWSubs_mit_Div(*this);
Hilfsvektor.RWSubs(*this,b);

b_Strich=Hilfsvektor.Vector_LA;
}

void Matrix_LA::Eins_Norm(Share& Norm){

int i,j;

Matrix_LA Kopie=*this;
Norm=0;
Vektor_LA V(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);

for(j=0; j<Spaltenzahl; j++){
    for(i=0; i<Zeilenzahl; i++){
        Kopie.M[i][j].Betrag(q_1,k);
        V[j]+=Kopie.M[i][j];
        }
    }

Vektor_LA Index(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
V.Max_Neu(Index,k);
Norm=Index[V];
}

void Matrix_LA::Normiere_Spaltenweise(){//Jede Spalte der Matrix auf den Bereich [-1,1] normieren

int i,j;
Vektor_LA H(Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,N);
Vektor_LA I(Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,N);
Share m(Spielerzahl,Schwellenwert,q);

for(j=0; j<Spaltenzahl; j++){//Fr jede Spalte
    H.getSpaltenvektor(*this,j);
    H.Max_Betrag(I);
    m=H[I];
    m.Inverses_ohne_Trunc(q_1,k);//Invertiere das maximale Element
    for(i=0; i<Zeilenzahl; i++){
        M[i][j]*=m;
        M[i][j].TruncPr(3*k-f,2*k-f,Schwellenwert,q,q_1);
        }
    }
}
