/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include <Matrix_mpf_t.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <sstream>
#include <ctime>
#include <Vektor_mpf_t.h>
#include <head.h>

Matrix_mpf_t::Matrix_mpf_t(int m, int n)
{
Hoehe=m;
Breite=n;

int i,j;
M=new mpf_t *[m];
for(i=0; i<m; i++){
    M[i]=new mpf_t[n];
    for(j=0; j<n; j++){
        mpf_init_set_si(M[i][j],0);
        }
    }
}

Matrix_mpf_t::Matrix_mpf_t(int m, int n, double** MM)
{
Hoehe=m;
Breite=n;

int i,j;
M=new mpf_t *[m];
for(i=0; i<m; i++){
    M[i]=new mpf_t[n];
    for(j=0; j<n; j++){
        mpf_init_set_d(M[i][j],MM[i][j]);
        }
    }
}

Matrix_mpf_t::~Matrix_mpf_t()
{

int i,j;

for(i=0; i<Hoehe; i++){
    for(j=0; j<Breite; j++){
        mpf_clear(M[i][j]);
        }
    delete []M[i];
    M[i]=NULL;
    }

delete[]M;
M=NULL;
}

Matrix_mpf_t::Matrix_mpf_t(const Matrix_mpf_t& other):Hoehe(other.Hoehe),Breite(other.Breite){
int i,j;

M=new mpf_t * [Hoehe];
if(&other!=this){
    for(i=0; i<Hoehe; i++){
        M[i]=new mpf_t[Breite];
        for(j=0; j<Breite; j++){
            mpf_init_set(M[i][j],other.M[i][j]);
            }
        }
    }
}

Matrix_mpf_t& Matrix_mpf_t::operator=(const Matrix_mpf_t& rhs)
{
    if (this == &rhs) return *this; // handle self assignment

    int i,j;

    for(i=0; i<Hoehe; i++){
        for(j=0; j<Breite; j++){
            mpf_set(M[i][j],rhs.M[i][j]);
            }
        }


    return *this;
}

Matrix_mpf_t& Matrix_mpf_t::operator=(int n){

int i,j;

for(i=0; i<Hoehe; i++){
    for(j=0; j<Breite; j++){
        mpf_set_si(M[i][j],n);
        }
    }

return *this;
}

Matrix_mpf_t& Matrix_mpf_t::operator-=(const Matrix_mpf_t& A){

int i,j;

for(i=0; i<Hoehe; i++){
    for(j=0; j<Breite; j++){
        mpf_sub(M[i][j],M[i][j],A.M[i][j]);
        }
    }

return *this;
}

Matrix_mpf_t& Matrix_mpf_t::operator+=(const Matrix_mpf_t& A){

int i,j;

for(i=0; i<Hoehe; i++){
    for(j=0; j<Breite; j++){
        mpf_add(M[i][j],M[i][j],A.M[i][j]);
        }
    }

return *this;
}

Matrix_mpf_t& Matrix_mpf_t::operator*=(double d){

int i,j;

mpf_t F;
mpf_init_set_d(F,d);

for(i=0; i<Hoehe; i++){
    for(j=0; j<Breite; j++){
        mpf_mul(M[i][j],M[i][j],F);
        }
    }

mpf_clear(F);
return *this;
}

Matrix_mpf_t Matrix_mpf_t::operator*(const Matrix_mpf_t& B){

if(Breite!=B.Hoehe){
    printf("\nWarnung: Matrix_mpf_t::operator*=");
    printf("\nMatrixmultiplikation nicht mglich, da Matrizen inkompatibel. Abbruch");
    exit(1);
    }

int i,j,k;

Matrix_mpf_t Rueckgabe(Hoehe,B.Breite);
mpf_t H;
mpf_init_set_d(H,0.0);

for(i=0; i<Hoehe; i++){
    for(j=0; j<B.Breite; j++){
        mpf_set_d(H,0.0);
        for(k=0; k<Breite; k++){
            mpf_mul(H,M[i][k],B.M[k][j]);
            mpf_add(Rueckgabe.M[i][j],Rueckgabe.M[i][j],H);
            }
        }
    }

mpf_clear(H);
return Rueckgabe;
}

void Matrix_mpf_t::Kopiere_rechts(Matrix_mpf_t& R){

if(Hoehe!=R.Hoehe){
    printf("\nFehler: Matrix_mpf_t::Kopiere_rechts");
    printf("\nMatrix kann nicht kopiert werden, da Hhe verschieden. Abbruch");
    printf("\nHoehe=%d, R.Hoehe=%d",Hoehe,R.Hoehe);
    exit(1);
    }

int r=haupt::min(Breite,R.Breite);

for(int i=0; i<Hoehe; i++){
    for(int j=0; j<r; j++){
        mpf_set(M[i][j],R.M[i][R.Breite-r+j]);
        }
    }
}

void Matrix_mpf_t::Kopiere_links(Matrix_mpf_t& L){

if(Hoehe!=L.Hoehe){
    printf("\nFehler: Matrix_mpf_t::Kopiere_links");
    printf("\nMatrix kann nicht kopiert werden, da Hoehe verschieden. Abbruch");
    printf("\nHoehe=%d, L.Hoehe=%d",Hoehe,L.Hoehe);
    exit(1);
    }

int r=haupt::min(Breite,L.Breite);

for(int i=0; i<Hoehe; i++){
    for(int j=0; j<r; j++){
        mpf_set(M[i][j],L.M[i][j]);
        }
    }
}

void Matrix_mpf_t::Einheitsmatrix(){

if(Hoehe!=Breite){
    printf("\nWarnung: Matrix_mpf_t::Einheitsmatrix");
    printf("\nEinheitsmatrix kann nicht erstellt werden, da Matrix nicht quadratisch. Abbruch");
    exit(1);
    }

int i;

*this=0;

for(i=0; i<Hoehe; i++){
    mpf_set_d(M[i][i],1.0);
    }
}

void Matrix_mpf_t::Zufallsmatrix(){

int i,j;

mpz_t seed_mpz;
mpz_init(seed_mpz);
time_t seed;
double seed_d;

seed=clock();
seed_d=(double) seed;
mpz_set_d(seed_mpz,seed_d);

gmp_randstate_t rState;
gmp_randinit_mt(rState);
gmp_randseed(rState,seed_mpz); //Erzeugen einer Zufallszahl*/

for(i=0; i<Hoehe; i++){
    for(j=0; j<Breite; j++){
        mpf_urandomb(M[i][j],rState,20);
        mpf_mul_2exp(M[i][j],M[i][j],11); //Erzeuge Zufallszahl in -1024..1023
        mpf_sub_ui(M[i][j],M[i][j],1024);
        }
    }

gmp_randclear(rState);
mpz_clear(seed_mpz);
}

void Matrix_mpf_t::SpalteEinfuegen(int Spalte, mpf_t* v, int Dimension){

if(Dimension!=Hoehe){
    printf("\nWarnung: Matrix_mpf_t::SpalteEinfuegen");
    printf("\nVektor kann nicht eingefuegt werden, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

int i;

for(i=0; i<Hoehe; i++){
    mpf_set(M[i][Spalte],v[i]);
    }
}

void Matrix_mpf_t::Norm_Frobenius(mpf_t& Norm){

int i,j;

mpf_t Hilfe;
mpf_init(Hilfe);
mpf_set_si(Norm,0);

for(i=0; i<Hoehe; i++){
    for(j=0; j<Breite; j++){
        mpf_set(Hilfe,M[i][j]);
        mpf_mul(Hilfe,Hilfe,Hilfe);
        mpf_add(Norm,Norm,Hilfe);
        }
    }

mpf_sqrt(Norm,Norm);

mpf_clear(Hilfe);
}

double Matrix_mpf_t::Kondition(){

double Kondition_Wert;
mpf_t Norm_Frobenius_Wert;
mpf_init(Norm_Frobenius_Wert);

Norm_Frobenius(Norm_Frobenius_Wert);
Kondition_Wert=mpf_get_d(Norm_Frobenius_Wert);

Matrix_mpf_t IM=*this;
IM.InverseMatrix();
IM.Norm_Frobenius(Norm_Frobenius_Wert);
Kondition_Wert*=mpf_get_d(Norm_Frobenius_Wert);

mpf_clear(Norm_Frobenius_Wert);
return Kondition_Wert;
}

void Matrix_mpf_t::VWElimination(mpf_t * b){

if(Hoehe!=Breite){
    printf("\nWarnung: Matrix_mpft::VWElimination");
    printf("\nVorwaertselimination nicht moeglich, da Matrix nicht quadratisch. Abbruch.");
    exit(1);
    }

int i,j,k;

mpf_t h;
mpf_init(h);
for(k=0; k<Hoehe-1; k++){
    ///Pivotisierung
    Matrix_mpf_t HM(Hoehe-k,Hoehe);
    Vektor_mpf_t HV(Hoehe-k);
    for(i=k; i<Hoehe; i++){
        for(j=0; j<Breite; j++){
            mpf_set(HM.M[i-k][j],M[i][j]);
            }
        mpf_set(HV.V[i-k],b[i]);
        }

    HM.Pivot(HV.V,k);
    for(i=k; i<Hoehe; i++){
        for(j=0; j<Breite; j++){
            mpf_set(M[i][j],HM.M[i-k][j]);
            }
        mpf_set(b[i],HV.V[i-k]);
        }

    ///Pivotisierung Ende
    for(i=k+1; i<Hoehe; i++){
        mpf_div(M[i][k],M[i][k],M[k][k]);
        for(j=k+1; j<Hoehe; j++){
            mpf_mul(h,M[i][k],M[k][j]);
            mpf_sub(M[i][j],M[i][j],h);
            }
        }
    }

mpf_clear(h);
}

void Matrix_mpf_t::Pivot(mpf_t * b, int k){

int j;

Vektor_mpf_t v(Hoehe);
for(j=0; j<Hoehe; j++){
    mpf_set(v.V[j],M[j][k]);
    }

mpf_t Tausch;
mpf_init(Tausch);
int Index=v.Max();

for(j=0; j<Breite; j++){
    mpf_set(Tausch,M[0][j]);
    mpf_set(M[0][j],M[Index][j]);
    mpf_set(M[Index][j],Tausch);
    }


mpf_set(Tausch,b[0]);
mpf_set(b[0],b[Index]);
mpf_set(b[Index],Tausch);
mpf_clear(Tausch);
}

void Matrix_mpf_t::GaussAlgorithmus(mpf_t * b){

if(Breite!=Hoehe){
    printf("\nWarnung: Matrix_mpf_t::GaussAlgortihmus");
    printf("\nGaussAlgortihmus nicht moeglich, da Matrix nicht quadratisch. Abbruch");
    exit(1);
    }

VWElimination(b);

Vektor_mpf_t B(Hoehe);
for(int i=0; i<Hoehe; i++){
    mpf_set(B.V[i],b[i]);
    }

B.VWSubs(*this);
B.RWSubs(*this);

for(int i=0; i<Hoehe; i++){
    mpf_set(b[i],B.V[i]);
    }
}

void Matrix_mpf_t::InverseMatrix(){

if(Hoehe!=Breite){
    printf("\nWarnung: Matrix_mpf_t::InverseMatrix");
    printf("\nMatrix nicht quadratisch. Berechnung der Inversen nicht mglich");
    }

Matrix_mpf_t E(Hoehe,Breite);
E.Einheitsmatrix();

int i;

Vektor_mpf_t HV(Hoehe);
Matrix_mpf_t HM=*this;
for(i=0; i<Hoehe; i++){
    *this=HM;
    HV.getSpaltenvektor(i,E);
    GaussAlgorithmus(HV.V);
    E.SpalteEinfuegen(i,HV.V,HV.Dimension);
    }

*this=E;
}

void Matrix_mpf_t::Ausgabe(std::string Name){

int i, j;

char * r;
r=new char[Name.size()+1];
strcpy(r,Name.c_str());

printf("\n\n");

for(i=0; i<Hoehe; i++){
    for(j=0; j<Breite; j++){
        gmp_printf(" %s[%d][%d]=%.4Ff",r,i,j,M[i][j]);
        }
    printf("\n");
    }
delete []r;
r=NULL;
}

void Matrix_mpf_t::Ausgabe_Datei(std::string Name, std::string DateiName){

int i, j;

FILE* datei1;

char *r;
char *s;
r=new char[Name.size()+1];
strcpy(r,Name.c_str());
s=new char[DateiName.size()+1];
strcpy(s,DateiName.c_str());

datei1=fopen(s,"a");

fprintf(datei1,"\n\n");

for(i=0; i<Hoehe; i++){
    for(j=0; j<Breite; j++){
        gmp_fprintf(datei1," %s[%d][%d]=%.4Ff",r,i,j,M[i][j]);
        }
    fprintf(datei1,"\n");
    }

fclose(datei1);
delete []r;
delete []s;
r=NULL;
s=NULL;
}

Matrix_mpf_t Matrix_mpf_t::Matrix_aus_Vektor(mpf_t * HH, mpf_t * HG){


Matrix_mpf_t B(Hoehe,Hoehe);

for(int i=0; i<Hoehe; i++){
    for(int j=0; j<Breite; j++){
        mpf_mul(B.M[i][j],HH[i],HG[j]);
        }
    }

return B;
}

Matrix_mpf_t Matrix_mpf_t::Transponiere(){

Matrix_mpf_t A(Breite,Hoehe);

int i,j;

for(i=0; i<Breite; i++){
    for(j=0; j<Hoehe; j++){
        mpf_set(A.M[i][j],M[j][i]);
        }
    }

return A;
}

void Matrix_mpf_t::Loesche_links_unten(){

for(int i=1; i<Hoehe; i++){
    for(int j=0; j<i; j++){
        mpf_set_d(M[i][j],0.0);
        }
    }
}

void Matrix_mpf_t::Rechts_unten_auspassen(Matrix_mpf_t& N, int z, int s){

if(z>Hoehe || s>Breite){
    printf("\nWarnung: Matrix_mpf_t::Rechts_unten_auspassen");
    printf("\nEinpassen der Matrix nicht mglich, da Dimensionen zu gro. Abbruch.");
    exit(1);
    }

int i,j,l=0,r=0;

for(i=Hoehe-z; i<Hoehe; i++){
    for(j=Breite-s; j<Breite; j++){
        mpf_set(N.M[l][r],M[i][j]);
        r++;
        }
    r=0;
    l++;
    }
}

void Matrix_mpf_t::Rechts_unten_einpassen(Matrix_mpf_t& N, int z, int s){

if(z>Hoehe || s>Breite){
    printf("\nWarnung: Matrix_mpf_t::Rechts_unten_einpassen");
    printf("\nEinpassen der Matrix nicht mglich, da Dimensionen zu gross. Abbruch.");
    exit(1);
    }

int i,j,l=0,r=0;

for(i=Hoehe-z; i<Hoehe; i++){
    for(j=Breite-s; j<Breite; j++){
        mpf_set(M[i][j],N.M[l][r]);
        r++;
        }
    r=0;
    l++;
    }
}

void Matrix_mpf_t::Multipliziere_Householder_Prae(mpf_t * v_Strich, int d){

Vektor_mpf_t v(d);
for(int i=0; i<d; i++){
    mpf_set(v.V[i],v_Strich[i]);
    }

Vektor_mpf_t v2=v;

mpf_t beta;
mpf_init(beta);

Vektor_mpf_t w(d);

v2.SkalarProdukt(v,beta);
mpf_ui_div(beta,2,beta);
mpf_neg(beta,beta);//beta<- -2/v^tv

Matrix_mpf_t A(Breite,Hoehe);
A=Transponiere();

v.MatrixMultVektor(A,w);
w*=beta;

Matrix_mpf_t B(v.Dimension,v.Dimension);
B=Matrix_aus_Vektor(v.V,w.V);
*this+=B;

mpf_clear(beta);
}

void Matrix_mpf_t::QR_Householder(Matrix_mpf_t& Q){

int i,j;

if(Hoehe!=Breite){
    printf("\nMehr Spalten als Zeilen im Scope. QR-Zerlegung nicht mglich. Abbruch");
    exit(1);
    }

if(Q.Hoehe!=Hoehe || Q.Breite!=Breite){
    printf("\nFalsche Dimensionen von Q. QR-Zerlegung nicht mglich. Abbruch");
    exit(1);
    }

for(i=0; i<Breite-1; i++){
    Vektor_mpf_t v(Hoehe-i);
    v.getSpaltenvektor_ab_Zeile(i,*this,i);
    v.Householder_Vektor();

    Matrix_mpf_t N(Hoehe-i,Hoehe-i);
    Rechts_unten_auspassen(N,Hoehe-i,Breite-i);
    N.Multipliziere_Householder_Prae(v.V,Hoehe-i);
    Rechts_unten_einpassen(N,Hoehe-i,Breite-i);

    if(i<Hoehe-1){
        for(j=i+1; j<Hoehe; j++){
            mpf_set(M[j][i],v.V[j-i]);
            }
        }
    }


Q.Einheitsmatrix();

for(i=Breite-2; i>=0; i--){
    Vektor_mpf_t v(Hoehe-i);
    v.getSpaltenvektor_ab_Zeile(i,*this,i);
    mpf_set_d(v.V[0],1.0);
    Matrix_mpf_t N(Hoehe-i,Hoehe-i);
    Q.Rechts_unten_auspassen(N,Hoehe-i,Hoehe-i);
    N.Multipliziere_Householder_Prae(v.V,Hoehe-i);
    Q.Rechts_unten_einpassen(N,Hoehe-i,Hoehe-i);
    }
}
