/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include "SVM.h"
#include <Vektor_LA.h>
#include <Matrix_mpf_t.h>
#include <cstdio>
#include <head.h>

SVM::SVM(int m, int n, int Anzahl_Spieler, int PolyGrad, mpz_t& Modulus, mpz_t& kleiner_Modulus, int kk, int ff)
:Daten(Matrix_LA(m,n,Anzahl_Spieler,PolyGrad,Modulus,kleiner_Modulus,kk,ff)),
Hoehe_Daten(m),
Breite_Daten(n),
Spielerzahl(Anzahl_Spieler),
Schwellenwert(PolyGrad),
k(kk),
N(ff),
K(Matrix_LA(m,m,Anzahl_Spieler,PolyGrad,Modulus,kleiner_Modulus,k,N)),//Kernel-Matrix
G(Matrix_LA(m,m,Anzahl_Spieler,PolyGrad,Modulus,kleiner_Modulus,k,N)),//Optimierungsmatrix
Y(Vektor_LA(m,Anzahl_Spieler,PolyGrad,Modulus,kleiner_Modulus,kk,ff)),//Klassenindikatoren;
Alpha(Vektor_LA(m,Anzahl_Spieler,PolyGrad,Modulus,kleiner_Modulus,k,N)),//Lsungsvektor fr die Lagrange-Multiplikatoren
SV(Vektor_LA(m,Anzahl_Spieler,PolyGrad,Modulus,kleiner_Modulus,k,N)),//Support-Vektoren; binr kodiert
b(Share(Anzahl_Spieler,PolyGrad,Modulus)),//Konstanter Term der ges. Hyperflche
Lambda(Share(Anzahl_Spieler,PolyGrad,Modulus)),//Toleranzparameter fr SM-SVM
Koeff(Share(Anzahl_Spieler,PolyGrad,Modulus)),
r(Share(Anzahl_Spieler,PolyGrad,Modulus))
{
mpz_init_set(q,Modulus);
mpz_init_set(q_1,kleiner_Modulus);
}

SVM::~SVM()
{
mpz_clear(q);
mpz_clear(q_1);
}

SVM::SVM(const SVM& other)
:Daten(Matrix_LA(other.Hoehe_Daten,other.Breite_Daten,other.Spielerzahl,other.Schwellenwert,other.q,other.q_1,other.k,other.N)),
K(Matrix_LA(other.Hoehe_Daten,other.Hoehe_Daten,other.Spielerzahl,other.Schwellenwert,other.q,other.q_1,other.k,other.N)),
G(Matrix_LA(other.Hoehe_Daten,other.Hoehe_Daten,other.Spielerzahl,other.Schwellenwert,other.q,other.q_1,other.k,other.N)),
Hoehe_Daten(other.Hoehe_Daten),
Breite_Daten(other.Breite_Daten),
Spielerzahl(other.Spielerzahl),
Schwellenwert(other.Schwellenwert),
k(other.k),
N(other.N)
{
    Daten=other.Daten;
    K=other.K;
    G=other.G;
    Y=other.Y;
    b=other.b;
    Lambda=other.Lambda;
    Alpha=other.Alpha;
    SV=other.SV;
    b=other.b;
    Koeff=other.Koeff;
    r=other.Koeff;
    Spielerzahl=other.Spielerzahl;
    Schwellenwert=other.Schwellenwert;
    mpz_init_set(q,other.q);
    mpz_init_set(q_1,other.q_1);
}

SVM& SVM::operator=(const SVM& rhs)
{
    if (this == &rhs) return *this; // handle self assignment
    //assignment operator
    Daten=rhs.Daten;
    K=rhs.K;
    G=rhs.G;
    Y=rhs.Y;
    Alpha=rhs.Alpha;
    SV=rhs.SV;
    b=rhs.b;
    Koeff=rhs.Koeff;
    r=rhs.Koeff;
    Spielerzahl=rhs.Spielerzahl;
    Schwellenwert=rhs.Schwellenwert;
    k=rhs.k;
    N=rhs.N;
    mpz_set(q,rhs.q);
    mpz_set(q_1,rhs.q_1);
    return *this;
}

SVM& SVM::operator=(const Matrix_mpf_t& rhs)
{
    if(Hoehe_Daten!=rhs.Hoehe){
        printf("\nFehler: SVM::operator=(const Matrix_mpf_t& rhs)");
        printf("\nZeilenzahl der Matrizen verschieden. Zuweisung nicht moeglich. Abbruch.");
        printf("\nZeilenzahl=%d, Hoehe=%d",Hoehe_Daten,rhs.Hoehe);
        exit(1);
        }
    if(Breite_Daten!=rhs.Breite){
        printf("\nFehler: SVM::operator=(const Matrix_mpf_t& rhs)");
        printf("\nSpaltenzahl der Matrizen verschieden. Zuweisung nicht moeglich. Abbruch.");
        printf("\nSpalenzahl=%d, Breite=%d",Breite_Daten,rhs.Breite);
        exit(1);
        }

    int i,j;

    double d;

    for(i=0; i<Hoehe_Daten; i++){
        for(j=0; j<Breite_Daten; j++){
            d=mpf_get_d(rhs.M[i][j]);
            Daten.M[i][j].Erzeugung_double(d,Schwellenwert,q);
            }
        }

    return *this;
}

void SVM::Linearer_Kernel(){//Kernel-Matrix fr den linearen Kernel berechnen

int i,j;

Vektor_LA HV1(Breite_Daten,Spielerzahl,Schwellenwert,q,q_1,k,N);
Vektor_LA HV2(Breite_Daten,Spielerzahl,Schwellenwert,q,q_1,k,N);

for(i=0; i<Hoehe_Daten; i++){
    for(j=0; j<Hoehe_Daten; j++){
        HV1.getZeilenvektor(Daten,i);//In der ersten Spalte der Matrix ist die Klassifizierung gespeichert
        HV2.getZeilenvektor(Daten,j);
        HV1.SkalarMult(HV2,K.M[i][j]);
        K.M[i][j].TruncPr(k,N,Schwellenwert,q,q_1);
        }
    }
}

void SVM::Quadratischer_Kernel(Share& gamma, Share& Konst){//Kernelfunktion gegeben durch (Koeff*K_ij+Konst)^2

Koeff=gamma;
r=Konst;

int i,j;

Linearer_Kernel();

for(i=0; i<Hoehe_Daten; i++){
    for(j=0; j<Hoehe_Daten; j++){
        K.M[i][j]*=Koeff;
        K.M[i][j].TruncPr(k,N,Schwellenwert,q,q_1);
        K.M[i][j]+=Konst;
        K.M[i][j].Quadrat();
        K.M[i][j].TruncPr(k,N,Schwellenwert,q,q_1);
        }
    }
}

void SVM::Quadratischer_Kernel(double Koeff, double Konst){//Kernelfunktion gegeben durch (Koeff*K_ij+Konst)^2

Share A(Spielerzahl,Schwellenwert,q);
Share B(Spielerzahl,Schwellenwert,q);

A.Erzeugung_double(Koeff,Schwellenwert,q);
B.Erzeugung_double(Konst,Schwellenwert,q);

Quadratischer_Kernel(A,B);
}

void SVM::Berechne_G(){

int i,j;

for(i=0; i<K.Zeilenzahl; i++){
    for(j=0; j<K.Spaltenzahl; j++){
        G.M[i][j]=K.M[i][j]*Y[i]*Y[j];
        }
    }
}

void SVM::Entscheidungsregel_linear(Share& Ergebnis, Vektor_LA& x){

int i;

Vektor_LA Kx(Alpha.d,Spielerzahl,Schwellenwert,q,q_1,k,N);
Vektor_LA HV(Breite_Daten,Spielerzahl,Schwellenwert,q,q_1,k,N);
for(i=0; i<Kx.d; i++){
    HV.getZeilenvektor(Daten,i);
    HV.SkalarMult(x,Kx[i]);
    Kx[i].TruncPr(k,N,Schwellenwert,q,q_1);
    }

Ergebnis=0;

for(i=0; i<Alpha.d; i++){
    Ergebnis+=(Alpha[i]*Y[i]*Kx[i]);
    }

Ergebnis.TruncPr(k,N,Schwellenwert,q,q_1);
//Ergebnis.Rekonstrukt_Ausgabe("E_i");
Ergebnis+=b;

//Ergebnis.Rekonstrukt_Ausgabe("E_i+b");

Ergebnis.Signum(Schwellenwert,q,q_1,k);
}

void SVM::Entscheidungsregel_quadratisch(Share& Ergebnis, Vektor_LA& x){

int i;

Vektor_LA Kx(Alpha.d,Spielerzahl,Schwellenwert,q,q_1,k,N);
Vektor_LA HV(Breite_Daten,Spielerzahl,Schwellenwert,q,q_1,k,N);

for(i=0; i<Kx.d; i++){
    HV.getZeilenvektor(Daten,i);
    HV.SkalarMult(x,Kx[i]);
    Kx[i].TruncPr(k,N,Schwellenwert,q,q_1);
    Kx[i]*=Koeff;
    Kx[i].TruncPr(k,N,Schwellenwert,q,q_1);
    Kx[i]+=r;
    Kx[i].Quadrat();
    Kx[i].TruncPr(k,N,Schwellenwert,q,q_1);
    }

Ergebnis=0;

for(i=0; i<Alpha.d; i++){
    Ergebnis+=(Alpha[i]*Y[i]*Kx[i]);
    }

Ergebnis.TruncPr(k,N,Schwellenwert,q,q_1);
//Ergebnis.Rekonstrukt_Ausgabe("E_i");
Ergebnis+=b;

//Ergebnis.Rekonstrukt_Ausgabe("E_i+b");

Ergebnis.Signum(Schwellenwert,q,q_1,k);
}

void SVM::Berechne_SV(){

int i;

mpz_t epsilon;
mpz_init_set_ui(epsilon,1);
mpz_mul_2exp(epsilon,epsilon,10);

for(i=0; i<SV.d; i++){
    SV[i]=Alpha[i];
    SV[i].Quadrat();
    SV[i].TruncPr(k,N,Schwellenwert,q,q_1);
    SV[i].GT(epsilon,Schwellenwert,q,q_1,k);//Komponente ist SV, wenn Quadrat grer als 2^{-f+10}
    }

mpz_clear(epsilon);
}

void SVM::Berechne_b(){

if(Hoehe_Daten!=Alpha.d){
    printf("\nFehler: SVM::Berechne_b");
    printf("\nDimensionen von Y und Alpha verschieden. Berechnung von b nicht mglich. Abbruch");
    printf("\nY.d=%d, Alpha.d=%d",Hoehe_Daten,Alpha.d);
    }

int i,j;

//Berechne_SV();

Share Hilfe(Spielerzahl,Schwellenwert,q);
Share Zaehler(Spielerzahl,Schwellenwert,q);
Vektor_LA Nenner(Alpha.d,Spielerzahl,Schwellenwert,q,q_1,k,N);
Share Nenner_Wert(Spielerzahl,Schwellenwert,q);

Vektor_LA Kopie_Alpha=Alpha;

for(i=0; i<Alpha.d; i++){
    Alpha[i]*=SV[i];//Fehler in Alpha korrigieren
    Nenner[i]=Lambda-Alpha[i];
    Nenner[i]*=Alpha[i];//es werden nur alphas betrachtet, die wirklich 0 sind
    Nenner[i].TruncPr(k,N,Schwellenwert,q,q_1);
    Nenner_Wert+=Nenner[i];
    }

for(i=0; i<Alpha.d; i++){
    Hilfe=0;
    for(j=0; j<Hoehe_Daten; j++){
        Hilfe+=(Alpha[j]*Y[j]*K.M[j][i]);
        }
    Hilfe.TruncPr(k,N,Schwellenwert,q,q_1);
    Hilfe=Y[i]*zwei_hoch_f-Hilfe;
    Zaehler+=(Nenner[i]*Hilfe);
    }
Zaehler.TruncPr(k,N,Schwellenwert,q,q_1);
Zaehler.FPDiv(Nenner_Wert,Schwellenwert,q,q_1,k);
//Zaehler*=-1;

b=Zaehler;
}

void SVM::Entscheidungsregel_Test(std::string Name){

Vektor_LA Test(Hoehe_Daten,Spielerzahl,Schwellenwert,q,q_1,k,N);
Vektor_LA HV(Breite_Daten,Spielerzahl,Schwellenwert,q,q_1,k,N);
int i;

mpz_t epsilon;
mpz_init_set_ui(epsilon,1);
mpz_mul_2exp(epsilon,epsilon,f-50);//Toleranzschwelle

Vektor_LA Alpha_test=Alpha;
for(i=0; i<Alpha.d; i++){
    Alpha_test[i].Quadrat();
    Alpha_test[i].TruncPr(k,N,Schwellenwert,q,q_1);
    Alpha_test[i].LT(epsilon,Schwellenwert,q,q_1,k);
    Alpha[i]*=(-Alpha_test[i]+1);
    }//Alpha, die fast null sind eliminieren

Berechne_b();

//b.Rekonstrukt_Ausgabe("b");

for(i=0; i<Test.d; i++){
    HV.getZeilenvektor(Daten,i);
    Entscheidungsregel_linear(Test[i],HV);
    Test[i].EQ(Y[i],Schwellenwert,q,q_1,k);
    }

Test.Ausgabe("T");

FILE * datei1;
char * FileName;
FileName=new char[Name.size()+1];
strcpy(FileName,Name.c_str());

datei1=fopen(FileName,"a");

if(!datei1){
    printf("\nFehler: SVM::Entscheindungsregel");
    printf("\nDatei %s konnte nicht geoeffnet werden. Abbruch",Name.c_str());
    exit(1);
    }
Test.Ausgabe_Datei("T",datei1);

fclose(datei1);

mpz_clear(epsilon);
delete FileName;
}

void SVM::SVM_Test_linear(Matrix_LA& Daten_Test, Vektor_LA& Y_Test,std::string Name){

if(Daten_Test.Zeilenzahl!=Y_Test.d){
    printf("\nFehler: SVM::SVM_Test_linear");
    printf("\nDaten und Y haben unterschiedliche Groee. Abbruch");
    printf("\nDaten_Test.Zeilenzahl=%d, Y_Tes.d=%d",Daten_Test.Zeilenzahl,Y_Test.d);
    }

Vektor_LA Test(Daten_Test.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,N);
Vektor_LA HV(Daten_Test.Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,N);
Share H(Spielerzahl,Schwellenwert,q);
int i;

for(i=0; i<Test.d; i++){
    HV.getZeilenvektor(Daten_Test,i);
    Entscheidungsregel_linear(Test[i],HV);
    Test[i].EQ(Y_Test[i],Schwellenwert,q,q_1,k);
    H+=Test[i];
    }

char * FileName;
FileName=new char[Name.size()+1];
strcpy(FileName,Name.c_str());

Test.Ausgabe("T");
FILE * datei;
datei=fopen(FileName,"a");
Test.Ausgabe_Datei("T",datei);

mpz_t ratio;
mpz_init(ratio);
mpf_t (ratio_f);
mpf_init(ratio_f);

H.Rekonstrukt_GRR_mit_VZ(ratio,q,Schwellenwert);
printf("\n H rekonstruiert");
mpf_set_z(ratio_f,ratio);
mpf_div_ui(ratio_f,ratio_f,Y_Test.d);
mpf_mul_ui(ratio_f,ratio_f,100);

gmp_printf("\n\nPunkte korrekt klassifiziert: %Ff %%",ratio_f);
gmp_fprintf(datei,"\n\nPunkte korrekt klassifiziert: %Ff %%",ratio_f);
fclose(datei);

mpz_clear(ratio);
mpf_clear(ratio_f);
delete [] FileName;
}

void SVM::SVM_Test_quadratisch(Matrix_LA& Daten_Test, Vektor_LA& Y_Test,std::string Name){
printf("\nQuadratischer Test begonnen");
if(Daten_Test.Zeilenzahl!=Y_Test.d){
    printf("\nFehler: SVM::SVM_Test_linear");
    printf("\nDaten und Y haben unterschiedliche Groee. Abbruch");
    printf("\nDaten_Test.Zeilenzahl=%d, Y_Tes.d=%d",Daten_Test.Zeilenzahl,Y_Test.d);
    }

Vektor_LA Test(Daten_Test.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,N);
Vektor_LA HV(Daten_Test.Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,N);
Share H(Spielerzahl,Schwellenwert,q);
int i;

for(i=0; i<Test.d; i++){
    HV.getZeilenvektor(Daten_Test,i);
    Entscheidungsregel_quadratisch(Test[i],HV);
    Test[i].EQ(Y_Test[i],Schwellenwert,q,q_1,k);
    H+=Test[i];
    }

char * FileName;
FileName=new char[Name.size()+1];
strcpy(FileName,Name.c_str());

Test.Ausgabe("T");
FILE * datei;
datei=fopen(FileName,"a");
Test.Ausgabe_Datei("T",datei);

mpz_t ratio;
mpz_init(ratio);
mpf_t ratio_f;
mpf_init(ratio_f);
printf("\nratios initialisiert");

H.Rekonstrukt_GRR_mit_VZ(ratio,q,Schwellenwert);
printf("\nH rekonstruiert");
mpf_set_z(ratio_f,ratio);
mpf_div_ui(ratio_f,ratio_f,Y_Test.d);
mpf_mul_ui(ratio_f,ratio_f,100);

gmp_printf("\n\nPunkte korrekt klassifiziert: %Ff %%",ratio_f);
gmp_fprintf(datei,"\n\nPunkte korrekt klassifiziert: %Ff %%",ratio_f);
fclose(datei);

mpz_clear(ratio);
mpf_clear(ratio_f);
delete [] FileName;
}

void SVM::Save_SVM(std::string Dateiname){

FILE * datei;
char * r;
int k;

std::string Daten_SVM=Dateiname;
Daten_SVM.append("_Daten.txt");
r= new char [Daten_SVM.size()+1];
strcpy(r,Daten_SVM.c_str());
datei=fopen(r,"a");
k=Daten.Ausgabe_skaliert_Datei("D", datei);
fclose(datei);
delete r;

std::string K_SVM=Dateiname;
K_SVM.append("_K.txt");
r= new char [K_SVM.size()+1];
strcpy(r,K_SVM.c_str());
datei=fopen(r,"a");
k=K.Ausgabe_skaliert_Datei("K", datei);
fclose(datei);
delete r;

std::string G_SVM=Dateiname;
G_SVM.append("_G.txt");
r= new char [G_SVM.size()+1];
strcpy(r,G_SVM.c_str());
datei=fopen(r,"a");
k=G.Ausgabe_skaliert_Datei("G", datei);
fclose(datei);
delete r;

std::string alpha_SVM=Dateiname;
alpha_SVM.append("_alpha.txt");
r= new char [alpha_SVM.size()+1];
strcpy(r,alpha_SVM.c_str());
datei=fopen(r,"a");
k=Alpha.Ausgabe_skaliert_Datei("alpha", datei);
fclose(datei);
delete r;

std::string Y_SVM=Dateiname;
Y_SVM.append("_Y.txt");
r= new char [Y_SVM.size()+1];
strcpy(r,Y_SVM.c_str());
datei=fopen(r,"a");
k=Y.Ausgabe_Datei("Y", datei);
fclose(datei);
delete r;

std::string b_SVM=Dateiname;
b_SVM.append("_b.txt");
r= new char [b_SVM.size()+1];
strcpy(r,b_SVM.c_str());
datei=fopen(r,"a");
k=b.Ausgabe_skaliert_Datei("b", datei);
fclose(datei);
delete r;

std::string lambda_SVM=Dateiname;
lambda_SVM.append("_lambda.txt");
r= new char [lambda_SVM.size()+1];
strcpy(r,lambda_SVM.c_str());
datei=fopen(r,"a");
k=Lambda.Ausgabe_skaliert_Datei("lambda", datei);
fclose(datei);
delete r;

std::string SV_SVM=Dateiname;
SV_SVM.append("_SV.txt");
r= new char [SV_SVM.size()+1];
strcpy(r,SV_SVM.c_str());
datei=fopen(r,"a");
k=SV.Ausgabe_Datei("SV", datei);
fclose(datei);
delete r;
}
