/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include<cstdio>
#include<cstdlib>
#include<ctime>
#include<cmath>
#include<iostream>
#include<Share.h>
#include<Matrix.h>
#include <Vektor.h>
#include <head.h> //enthlt gmp.h und Share_F2M.h
#include <GF256.h>
#include <Share_F2M.h>


void Share::Erzeugung(mpz_t& s, int t, mpz_t& q){
    //Erzeugung eines Sharings von s unter n Spielern mit Hilfe eines Polynoms vom Grad t<n mittels der dividierten Differenzen
    register int j,k;

    Vektor f_Strich(t+1);//Hilfsvektor, in dem an Stelle 0 der Wert s und an den Stellen 1-t zufllige Werte gespeichert werden
    f_Strich.Zufall(Anzahl_Spieler,t,q);

    Vektor g(t+1); //Hilfsvektor fr die Berechnung der dividierten Differenzen

    for (j=1; j<=t; j++){ //Schreibe die ersten t zuflligen Funktionswerte in den Share-Vektor f
        mpz_set(Vector[j],f_Strich.Vector[j]);
        }
    mpz_set(Vector[0],s);

    mpz_set(g.Vector[0],s);
    for(j=1; j<=t; j++){
        mpz_set(g.Vector[j],f_Strich.Vector[j]);
        for(k=j-1; k>=0; k--){
            mpz_sub(g.Vector[k], g.Vector[k+1], g.Vector[k]);	/* g[k] = g[k+1] - g[k] */
            mpz_mod(g.Vector[k], g.Vector[k], q);	        /* MODULO-Reduzierung */
            }
        }
        for(j=t+1; j<=Anzahl_Spieler; j++){
            for(k=0; k<t; k++){
                mpz_add(g.Vector[k+1], g.Vector[k+1], g.Vector[k]);	/* g[k+1] = g[k+1] + g[k] */
				mpz_mod(g.Vector[k+1], g.Vector[k+1], q);		/* MODULO-Reduzierung */
            }
            mpz_set(Vector[j],g.Vector[t]);
        }
}

void Share::Erzeugung_double(double d, int t, mpz_t& q){
    //Erzeugung eines Sharings von s unter n Spielern mit Hilfe eines Polynoms vom Grad t<n mittels der dividierten Differenzen
    mpz_t s;
    mpf_t ff;

    mpf_init_set_d(ff,d);
    mpf_mul_2exp(ff,ff,f);
    mpz_init(s);

    mpz_set_f(s,ff);
    register int j,k;
    Vektor f_Strich(t+1);//Hilfsvektor, in dem an Stelle 0 der Wert s und an den Stellen 1-t zufllige Werte gespeichert werden
    f_Strich.Zufall(Anzahl_Spieler,t,q);

    Vektor g(t+1); //Hilfsvektor fr die Berechnung der dividierten Differenzen

    for (j=1; j<=t; j++){ //Schreibe die ersten t zuflligen Funktionswerte in den Share-Vektor f
        mpz_set(Vector[j],f_Strich.Vector[j]);
        }
    mpz_set(Vector[0],s);

    mpz_set(g.Vector[0],s);
    for(j=1; j<=t; j++){
        mpz_set(g.Vector[j],f_Strich.Vector[j]);
        for(k=j-1; k>=0; k--){
            mpz_sub(g.Vector[k], g.Vector[k+1], g.Vector[k]);	/* g[k] = g[k+1] - g[k] */
            mpz_mod(g.Vector[k], g.Vector[k], q);	        /* MODULO-Reduzierung */
            }
        }
        for(j=t+1; j<=Anzahl_Spieler; j++){
            for(k=0; k<t; k++){
                mpz_add(g.Vector[k+1], g.Vector[k+1], g.Vector[k]);	/* g[k+1] = g[k+1] + g[k] */
				mpz_mod(g.Vector[k+1], g.Vector[k+1], q);		/* MODULO-Reduzierung */
            }
            mpz_set(Vector[j],g.Vector[t]);
        }
        mpz_clear(s);
        mpf_clear(ff);
}

void Share::Rekonstrukt(mpz_t& s, int t, mpz_t& q) //Abgeleitet von Phase 2 von Mul_2_Lory.c
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    int i,k;

    Vektor X(t+1);
    Vektor Hilfe(t+1);

    mpz_t R;
    mpz_init(R);

    for(i=1; i<=t+1; i++){
        mpz_set(Hilfe.Vector[i],Vector[i]);
        for(k=i-1; k>=1; k--){
            mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
	         }
        mpz_set(X.Vector[i], Hilfe.Vector[1]);				// X[i] = g[1]
        }

    for(k=t; k>=1; k--){
        mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
        }

    haupt::Waehle_Modulus(q,R);
    mpz_mod(s, X.Vector[1], q);
    k=mpz_cmp(s,R);
    if(k>0){
        mpz_sub(s,s,q);}

mpz_clear(R);
}

void Share::Rekonstrukt_GRR(mpz_t& s, mpz_t& q, int t){

int i;

mpz_set_ui(s,0);

if(t==Schwelle){
    for (i=1; i<=t+1; i++){
        mpz_addmul(s, Vector[i], lambda_Rekonstrukt[i]);
        }
    }

if (t==2*Schwelle){
    for (i=1; i<=t+1; i++){
        mpz_addmul(s, Vector[i], lambda[i]);
        }
    }

mpz_mod(s,s,q);
}

void Share::Rekonstrukt_GRR_mit_VZ(mpz_t& s, mpz_t& q, int t){

int i,k;

mpz_t R;
mpz_init(R);

mpz_set_ui(s,0);

if(t==Schwelle){
    for (i=1; i<=t+1; i++){
        mpz_addmul(s, Vector[i], lambda_Rekonstrukt[i]);
        }
    }

if (t==2*Schwelle){
    for (i=1; i<=t+1; i++){
        mpz_addmul(s, Vector[i], lambda[i]);
        }
    }

mpz_mod(s,s,q);
haupt::Waehle_Modulus(q,R);
k=mpz_cmp(s,R);
if(k>0){
    mpz_sub(s,s,q);
    }

mpz_clear(R);
}

void Share::Rekonstrukt_Ausgabe(std::string ShareName) //Abgeleitet von Phase 2 von Mul_2_Lory.c
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    register int i,k;

    mpz_t Ausgabe_z;
    mpz_init(Ausgabe_z);

    Vektor X(Anzahl_Spieler);
    Vektor Hilfe(Anzahl_Spieler);

    mpf_t Ausgabe;
    mpf_init(Ausgabe);

    mpz_t rem;
    mpz_init(rem);

    char * Name;
    Name=new char[ShareName.size()+1];
    strcpy(Name,ShareName.c_str());

   for(i=1; i<=Schwelle+1; i++){
        mpz_set(Hilfe.Vector[i],Vector[i]);
        for(k=i-1; k>=1; k--){
            mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
            }
        mpz_set(X.Vector[i], Hilfe.Vector[1]);				// X[i] = g[1]
        }
        for(k=Schwelle; k>=1; k--){
            mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
            }

        mpz_mod(Ausgabe_z, X.Vector[1], Modulus);

        k=mpz_cdiv_q_ui(rem,Modulus,2); //Zentriere Ausgabe um 0
        k=mpz_cmp(Ausgabe_z,rem);
        if(k>0){
            mpz_sub(Ausgabe_z,Ausgabe_z,Modulus);}

        mpf_set_z(Ausgabe,Ausgabe_z);
        mpf_div_2exp(Ausgabe,Ausgabe,f);
        gmp_printf("\n%s=%.10Ff",Name,Ausgabe);

delete [] Name;
Name=NULL;
mpz_clear(rem);
mpf_clear(Ausgabe);
mpz_clear(Ausgabe_z);
}

int Share::Ausgabe_skaliert_Datei(std::string ShareName, FILE * datei) //Abgeleitet von Phase 2 von Mul_2_Lory.c
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    register int i,k;

    mpz_t Ausgabe_z;
    mpz_init(Ausgabe_z);

    Vektor X(Anzahl_Spieler);
    Vektor Hilfe(Anzahl_Spieler);

    mpf_t Ausgabe;
    mpf_init(Ausgabe);

    mpz_t rem;
    mpz_init(rem);

    char * r;
    r= new char [ShareName.size()+1];
    strcpy(r,ShareName.c_str());

    if(datei==NULL){
    printf("\nFehler beim ffnen der Datei %s",ShareName.c_str());//Fehler beim ffnen
    return 1;
    }

   for(i=1; i<=Schwelle+1; i++){
        mpz_set(Hilfe.Vector[i],Vector[i]);
        for(k=i-1; k>=1; k--){
            mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
            }
        mpz_set(X.Vector[i], Hilfe.Vector[1]);				// X[i] = g[1]
        }
        for(k=Schwelle; k>=1; k--){
            mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
            }

        mpz_mod(Ausgabe_z, X.Vector[1], Modulus);

        k=mpz_cdiv_q_ui(rem,Modulus,2); //Zentriere Ausgabe um 0
        k=mpz_cmp(Ausgabe_z,rem);
        if(k>0){
            mpz_sub(Ausgabe_z,Ausgabe_z,Modulus);}

        mpf_set_z(Ausgabe,Ausgabe_z);
        mpf_div_2exp(Ausgabe,Ausgabe,f);
        gmp_fprintf(datei,"\n%s=%.10Ff",ShareName.c_str(),Ausgabe);

delete [] r;
mpz_clear(rem);
mpf_clear(Ausgabe);
mpz_clear(Ausgabe_z);
return 0;
}

void Share::Rekonstrukt_Ausgabe(std::string ShareName, int n) //Wie Rekonstrukt_Ausgabe, es kann aber ein Index mitgegeben werden. Hilfreich fr Ausgaben von Vektorkomponenten innerhalb von Schleifen
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    register int i,k;

    mpz_t Ausgabe_z;
    mpz_init(Ausgabe_z);

    Vektor X(Anzahl_Spieler);
    Vektor Hilfe(Anzahl_Spieler);

    mpf_t Ausgabe;
    mpf_init(Ausgabe);

    mpz_t rem;
    mpz_init(rem);

    char * Name;
    Name=new char[ShareName.size()+1];
    strcpy(Name,ShareName.c_str());

   for(i=1; i<=Schwelle+1; i++){
        mpz_set(Hilfe.Vector[i],Vector[i]);
        for(k=i-1; k>=1; k--){
            mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
            }
        mpz_set(X.Vector[i], Hilfe.Vector[1]);				// X[i] = g[1]
        }
        for(k=Schwelle; k>=1; k--){
            mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
            }

        mpz_mod(Ausgabe_z, X.Vector[1], Modulus);

        k=mpz_cdiv_q_ui(rem,Modulus,2); //Zentriere Ausgabe um 0
        k=mpz_cmp(Ausgabe_z,rem);
        if(k>0){
            mpz_sub(Ausgabe_z,Ausgabe_z,Modulus);}

        mpf_set_z(Ausgabe,Ausgabe_z);
        mpf_div_2exp(Ausgabe,Ausgabe,f);
        gmp_printf("\n%s(%d)=%.Ff",Name,n,Ausgabe);

delete [] Name;
Name=NULL;
mpz_clear(rem);
mpf_clear(Ausgabe);
mpz_clear(Ausgabe_z);
}

void Share::Rekonstrukt_Ausgabe_unskaliert(std::string ShareName) //Abgeleitet von Phase 2 von Mul_2_Lory.c
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    register int i,k;

    mpz_t Ausgabe_z;
    mpz_init(Ausgabe_z);

    Vektor X(Anzahl_Spieler);
    Vektor Hilfe(Anzahl_Spieler);

    mpz_t rem;
    mpz_init(rem);

    char * Name;
    Name=new char[ShareName.size()+1];
    strcpy(Name,ShareName.c_str());

   for(i=1; i<=Schwelle+1; i++){
        mpz_set(Hilfe.Vector[i],Vector[i]);
        for(k=i-1; k>=1; k--){
            mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
	         }
        mpz_set(X.Vector[i], Hilfe.Vector[1]);				// X[i] = g[1]
        }
        for(k=Schwelle; k>=1; k--){
            mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
        }

        mpz_mod(Ausgabe_z, X.Vector[1], Modulus);

        k=mpz_cdiv_q_ui(rem,Modulus,2); //Zentriere Ausgabe um 0
        k=mpz_cmp(Ausgabe_z,rem);
        if(k>0){
            mpz_sub(Ausgabe_z,Ausgabe_z,Modulus);}

        gmp_printf("\n%s=%.Zd",Name,Ausgabe_z);

delete [] Name;
Name=NULL;
mpz_clear(rem);
mpz_clear(Ausgabe_z);
}

void Share::Rekonstrukt_Ausgabe_unskaliert(std::string ShareName, mpz_t& p) //Abgeleitet von Phase 2 von Mul_2_Lory.c
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    register int i,k;

    mpz_t Ausgabe_z;
    mpz_init(Ausgabe_z);

    Vektor X(Anzahl_Spieler);
    Vektor Hilfe(Anzahl_Spieler);

    mpz_t rem;
    mpz_init(rem);

    char * Name;
    Name=new char[ShareName.size()+1];
    strcpy(Name,ShareName.c_str());

   for(i=1; i<=Schwelle+1; i++){
        mpz_set(Hilfe.Vector[i],Vector[i]);
        for(k=i-1; k>=1; k--){
            mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
	         }
        mpz_set(X.Vector[i], Hilfe.Vector[1]);				// X[i] = g[1]
        }
        for(k=Schwelle; k>=1; k--){
            mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
        }

        mpz_mod(Ausgabe_z, X.Vector[1], p);

        k=mpz_cdiv_q_ui(rem,p,2); //Zentriere Ausgabe um 0
        k=mpz_cmp(Ausgabe_z,rem);
        if(k>0){
            mpz_sub(Ausgabe_z,Ausgabe_z,p);
            }

        gmp_printf("\n%s=%.Zd",Name,Ausgabe_z);

delete [] Name;
Name=NULL;
mpz_clear(rem);
mpz_clear(Ausgabe_z);
}

void Share::Rekonstrukt_Ausgabe_unskaliert(std::string ShareName, int n) //Abgeleitet von Phase 2 von Mul_2_Lory.c
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    register int i,k;

    mpz_t Ausgabe_z;
    mpz_init(Ausgabe_z);

    Vektor X(Anzahl_Spieler);
    Vektor Hilfe(Anzahl_Spieler);

    mpz_t rem;
    mpz_init(rem);

    char * Name;
    Name=new char[ShareName.size()+1];
    strcpy(Name,ShareName.c_str());

   for(i=1; i<=Schwelle+1; i++){
        mpz_set(Hilfe.Vector[i],Vector[i]);
        for(k=i-1; k>=1; k--){
            mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
	         }
        mpz_set(X.Vector[i], Hilfe.Vector[1]);				// X[i] = g[1]
        }
        for(k=Schwelle; k>=1; k--){
            mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
        }

        mpz_mod(Ausgabe_z, X.Vector[1], Modulus);

        k=mpz_cdiv_q_ui(rem,Modulus,2); //Zentriere Ausgabe um 0
        k=mpz_cmp(Ausgabe_z,rem);
        if(k>0){
            mpz_sub(Ausgabe_z,Ausgabe_z,Modulus);}

        gmp_printf("\n%s(%d)=%.Zd",Name,n,Ausgabe_z);

delete [] Name;
Name=NULL;
mpz_clear(rem);
mpz_clear(Ausgabe_z);
}

void Share::Rekonstrukt_Ausgabe_skaliert_hexadezimal(std::string ShareName) //Abgeleitet von Phase 2 von Mul_2_Lory.c
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    register int i,k;

    mpz_t Ausgabe_z;
    mpz_init(Ausgabe_z);

    Vektor X(Anzahl_Spieler);
    Vektor Hilfe(Anzahl_Spieler);

    mpf_t Ausgabe;
    mpf_init(Ausgabe);

    mpz_t rem;
    mpz_init(rem);

    char * Name;
    Name=new char[ShareName.size()+1];
    strcpy(Name,ShareName.c_str());

   for(i=1; i<=Schwelle+1; i++){
        mpz_set(Hilfe.Vector[i],Vector[i]);
        for(k=i-1; k>=1; k--){
            mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
            }
        mpz_set(X.Vector[i], Hilfe.Vector[1]);				// X[i] = g[1]
        }
        for(k=Schwelle; k>=1; k--){
            mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
            }

        mpz_mod(Ausgabe_z, X.Vector[1], Modulus);

        k=mpz_cdiv_q_ui(rem,Modulus,2); //Zentriere Ausgabe um 0
        k=mpz_cmp(Ausgabe_z,rem);
        if(k>0){
            mpz_sub(Ausgabe_z,Ausgabe_z,Modulus);}

        mpf_set_z(Ausgabe,Ausgabe_z);
        mpf_div_2exp(Ausgabe,Ausgabe,f);
        gmp_printf("\n%s=%.Fa",Name,Ausgabe);

delete [] Name;
Name=NULL;
mpz_clear(rem);
mpf_clear(Ausgabe);
mpz_clear(Ausgabe_z);
}

void Share::Rekonstrukt_Spieler_beliebig(mpz_t& s, int * Spielerliste, int t, mpz_t& q){//Die an der Rekonstruktion teilnehmenden Spieler sind in *Spielerliste mit 1 kodiert; die anderen mit 0

register int i,k,j=1,l;

int Spielerliste_Liste[t+2], Differenzenmatrix[Anzahl_Spieler+1][Anzahl_Spieler+1]; //Auf den Positionen 1,...,t+1 werden die an der Rekonstruktion beteiligten  Spieler eingetragen
mpz_t invers;
mpz_init(invers);

    for(i=0; i<=Anzahl_Spieler; i++){
        Spielerliste_Liste[i]=0;
        if(Spielerliste[i]==1){
            Spielerliste_Liste[j]=i;
            j++;
        }
    }

    for(i=Anzahl_Spieler; i>=0; i--){ //Initialisierung der Differenzenmatrix
        for(j=0; j<=Anzahl_Spieler; j++){
        Differenzenmatrix[i][j]=i-j;}
    }

    Vektor X(Anzahl_Spieler);
    Vektor Hilfe(Anzahl_Spieler);

    mpz_t rem;
    mpz_init(rem);

     for(i=1; i<=t+1; i++){
        j=Spielerliste_Liste[i];
        mpz_set(Hilfe.Vector[i],Vector[j]); //Schiebe Sttzstellen und die zugehrigen Inversen zusammen
        for(k=i-1; k>=1; k--){
                mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
                l=Spielerliste_Liste[k];
                mpz_set_si(invers,Differenzenmatrix[j][l]);
                mpz_invert(invers,invers,q);
                mpz_mul(Hilfe.Vector[k], Hilfe.Vector[k], invers);//Division
                }
        mpz_set(X.Vector[i], Hilfe.Vector[1]);
        }			// X[i] = g[1]


    for(k=t; k>=1; k--){
        mpz_mul_si(X.Vector[k+1],X.Vector[k+1],Differenzenmatrix[Spielerliste_Liste[k]][0]);
        mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
        }

    mpz_mod(s, X.Vector[1], q);//Modulo-Reduzierung
    k=mpz_cdiv_q_ui(rem,q,2); //Zentriere Ausgabe um 0
    k=mpz_cmp(s,rem);
    if(k>0){
        mpz_sub(s,s,q);}

mpz_clear(rem);
mpz_clear(invers);
}

void Share::Rekonstrukt_Spieler_beliebig_Stelle_beliebig(mpz_t& s, int * Spielerliste, int t, int r, mpz_t& q){//Die an der Rekonstruktion teilnehmenden Spieler sind in *Spielerliste mit 1 kodiert; die anderen mit 0
//r: Stelle an der rekonstruiert werden soll
register int i,k,j=0,l;

int Spielerliste_Liste[t+2], Differenzenmatrix[Anzahl_Spieler+1][Anzahl_Spieler+1]; //Auf den Positionen 1,...,t+1 werden die an der Rekonstruktion beteiligten  Spieler eingetragen
mpz_t invers;
mpz_init(invers);

    for(i=0; i<=Anzahl_Spieler; i++){
        Spielerliste_Liste[i]=0;
        if(Spielerliste[i]==1){
            Spielerliste_Liste[j]=i;
            j++;
        }
    }


    for(i=Anzahl_Spieler; i>=0; i--){ //Initialisierung der Differenzenmatrix
        for(j=0; j<=Anzahl_Spieler; j++){
        Differenzenmatrix[i][j]=i-j;
        }
    }


    Vektor X(Anzahl_Spieler);
    Vektor Hilfe(Anzahl_Spieler);

    mpz_t rem;
    mpz_init(rem);

     for(i=0; i<=t; i++){
        j=Spielerliste_Liste[i];
        mpz_set(Hilfe.Vector[i],Vector[j]); //Schiebe Sttzstellen und die zugehrigen Inversen zusammen
        for(k=i-1; k>=0; k--){
                mpz_sub(Hilfe.Vector[k], Hilfe.Vector[k+1], Hilfe.Vector[k]);	// g[k] = g[k+1] - g[k]
                l=Spielerliste_Liste[k];
                mpz_set_si(invers,Differenzenmatrix[j][l]);
                mpz_invert(invers,invers,q);
                mpz_mul(Hilfe.Vector[k], Hilfe.Vector[k], invers);//Division
                }
        mpz_set(X.Vector[i], Hilfe.Vector[0]);
        }			// X[i] = g[1]


    for(k=t-1; k>=0; k--){
        mpz_mul_si(X.Vector[k+1],X.Vector[k+1],Differenzenmatrix[Spielerliste_Liste[k]][r]);
        mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		// X[k] = X[k] - X[k+1]
        }

    mpz_mod(s, X.Vector[0], q);//Modulo-Reduzierung
    k=mpz_cdiv_q_ui(rem,q,2); //Zentriere Ausgabe um 0
    k=mpz_cmp(s,rem);
    if(k>0){
        mpz_sub(s,s,q);}

mpz_clear(rem);
mpz_clear(invers);
}

void Share::Zufallszahl(int t, mpz_t& q){

mpz_t s;
mpz_init(s);

mpz_urandomm(s,rState1,q);

Erzeugung(s,t,q);

mpz_clear(s);
}

void Share::PRandInt(int t, int Laenge, mpz_t& q){

mpz_t l;

mpz_init(l);


Share R(Anzahl_Spieler, t, q);

mpz_urandomb(l,rState1,Laenge);//Msste bei echt verteilter Berechnung aufgespalten werden
R.Erzeugung(l,t,q);

mpz_clear(l);
}

void Share::Mul(mpz_t *b, int t, mpz_t& q) //Multiplikationsprotokoll von Peter Lory
    {
        /* Hilfs- und Zhlvariablen */
	register int i, j, k;
	mpz_t rem;

	/* Dynamische Speichererzeugung und Initialisierung der Vektoren
	 *  lambda, f und der Matrix h
	 *
	 *  lambda:	Zwischenspeicher fr die
	 *			Lagrange Interpolationskoeffizienten
	 *  f:		Zwischenspeicher fr die Funktionswerte der
	 *			Verteilungspolynome hi(x) (aller Spieler)
	 *  h:		Zwischenspeicher fr die verteilten Shares der einzelnen
	 *			Spieler
	 *  g:
	 *  H:
	*/

	Vektor g(2*t+1);							/* g */
	Matrix F(2*t+1, Anzahl_Spieler);					/* f */
	/* initMatrix(&h, 2*t+1, n); */				/* h */
	mpz_init(rem);
    Vektor X(Anzahl_Spieler);


	/* Initialisierung des Zwischenspeichers fr die Funktionswerte des
	 * Polynoms hi(x) mit Grad t
	 * hi(x) dient der Verteilung des Share ai*bi von Teilnehmer Pi
	 */

	/* Generierung von zufllig gewhlten Funktionswerten in Zq
	 * fr das Polynom hi(x) fr alle Spieler
	 */
	RandNumbers(F.M, t, q, Vector, b);

		   /*  Berechnung der Shares von ai*bi aller Teilnehmer
	    *
	    *  i: ist die Laufvariable fr die Spieler i
	    *  k: ist die Laufvariable fr die Funktionswerte des
	    *			Polynoms hi(x) von Spieler i
	    *  j: ist die Laufvariable fr die Spieler j
	    *  l: ist die Laufvariable fr die Berechnung des Shares hi(l)
	    *			fr Spieler l
	    */
	   for(i=1; i<=Anzahl_Spieler; i++){
	      mpz_set(g.Vector[0], F.M[i][0]);				/* g[0] = f[0] */
	      for(j=1; j<=t; j++){
		     mpz_set(g.Vector[j], F.M[i][j]);			/* g[j] = f[j] */
			 for(k=j-1; k>=0; k--){
		   	    mpz_sub(g.Vector[k], g.Vector[k+1], g.Vector[k]);	/* g[k] = g[k+1] - g[k] */
		   	    mpz_mod(g.Vector[k], g.Vector[k], q);	        /* MODULO-Reduzierung */
                }
            }

	      for(j=t+1; j<=Anzahl_Spieler; j++){
	         for(k=0; k<t; k++){
			    mpz_add(g.Vector[k+1], g.Vector[k+1], g.Vector[k]);	/* g[k+1] = g[k+1] + g[k] */
				mpz_mod(g.Vector[k+1], g.Vector[k+1], q);		/* MODULO-Reduzierung */
                }
	         mpz_set(F.M[i][j], g.Vector[t]);		/* f[j] = g[t] */
            }
	   }

		   /* Berechnung der resultierende Shares der Teilnehmer durch
	    * Aufsummierung der erhaltenen Teilshares, multipliziert mit den
	    * Lagrange Interpolationskoeffizienten
	    */

	   for(j=1; j<=Anzahl_Spieler; j++){
	      for(i=1; i<=2*t+1; i++){
	         mpz_set(g.Vector[i], F.M[i][j]);			/* g[i] = f[i][j] */
	         for(k=i-1; k>=1; k--){
			    mpz_sub(g.Vector[k], g.Vector[k+1], g.Vector[k]);	/* g[k] = g[k+1] - g[k] */
                }
            mpz_set(X.Vector[i], g.Vector[1]);				/* X[i] = g[1] */
            }

            for(k=2*t; k>=1; k--){
                mpz_sub(X.Vector[k], X.Vector[k], X.Vector[k+1]);		/* X[k] = X[k] - X[k+1] */
                }
            mpz_mod(Vector[j], X.Vector[1], q); //Ergebnis der Mutliplikation wird in den Share-Vektor des Scopes geschrieben

            k=mpz_cdiv_q_ui(rem,q,2); //Zentriere Ausgabe um 0
            k=mpz_cmp(Vector[j],rem);
            if(k>0){
                mpz_sub(Vector[j],Vector[j],q);
                }
	   }

	mpz_clear(rem);
    }

void Share::MUL_Lory3(mpz_t *b, int t, mpz_t& q){
	/* Hilfs- und Zhlvariablen */
	register int i, j, k;
	mpz_t temp;
	mpz_t rem;

	/* Dynamische Speichererzeugung und Initialisierung der Vektoren
	 *  lambda, f und der Matrix h
	 *
	 *  lambda:	Zwischenspeicher fr die
	 *			Lagrange Interpolationskoeffizienten
	 *  f:		Zwischenspeicher fr die Funktionswerte der
	 *			Verteilungspolynome hi(x) (aller Spieler)
	 *  h:		Zwischenspeicher fr die verteilten Shares der einzelnen
	 *			Spieler
	 *  g:
	 *  H:
	*/
	Vektor g(t);
	Matrix F(Anzahl_Spieler,Anzahl_Spieler);

	mpz_init(temp);								/* temp */
	mpz_init(rem);

    //Lagrangekoeffizienten sind in head.h vorausberechnet
	/* Initialisierung des Zwischenspeichers fr die Funktionswerte des
	 * Polynoms hi(x) mit Grad t
	 * hi(x) dient der Verteilung des Share ai*bi von Teilnehmer Pi
	 */

	/* Generierung von zufllig gewhlten Funktionswerten in Zq
	 * fr das Polynom hi(x) fr alle Spieler
	 */
	RandNumbers(F.M, t, q, Vector, b);

	   /*  Berechnung der Shares von ai*bi aller Teilnehmer
	    *
	    *  i: ist die Laufvariable fr die Spieler i
	    *  k: ist die Laufvariable fr die Funktionswerte des
	    *			Polynoms hi(x) von Spieler i
	    *  j: ist die Laufvariable fr die Spieler j
	    *  l: ist die Laufvariable fr die Berechnung des Shares hi(l)
	    *			fr Spieler l
	    */
    for(i=1; i<=Anzahl_Spieler; i++){
        mpz_set(g.Vector[0], F.M[i][0]);				/* g[0] = f[0] */
        for(j=1; j<=t; j++){
            mpz_set(g.Vector[j], F.M[i][j]);			/* g[j] = f[j] */
            for(k=j-1; k>=0; k--){
                mpz_sub(g.Vector[k], g.Vector[k+1], g.Vector[k]);	/* g[k] = g[k+1] - g[k] */
		   	    mpz_mod(g.Vector[k], g.Vector[k], q);	        /* MODULO-Reduzierung */
                }
            }
            for(j=t+1; j<=Anzahl_Spieler; j++){
                for(k=0; k<t; k++){
                    mpz_add(g.Vector[k+1], g.Vector[k+1], g.Vector[k]);	/* g[k+1] = g[k+1] + g[k] */
                    mpz_mod(g.Vector[k+1], g.Vector[k+1], q);		/* MODULO-Reduzierung */
                    }
                mpz_set(F.M[i][j], g.Vector[t]);		/* f[j] = g[t] */
            }
        }

	   /* Berechnung der resultierende Shares der Teilnehmer durch
	    * Aufsummierung der erhaltenen Teilshares, multipliziert mit den
	    * Lagrange Interpolationskoeffizienten
	    */
    for (j=1; j<=Anzahl_Spieler; j++){
        mpz_set_ui(temp,0);
        for (i=1; i<=2*t+1; i++){
	         /* mpz_addmul(temp, h[i][j], lambda[i]); */
	         /* void mpz_addmul (mpz t rop, mpz t op1, mpz t op2) */
	         /*	Set rop to rop + op1  op2. */
            mpz_addmul(temp, F.M[i][j], lambda[i]);
            }
        mpz_set(Vector[j], temp);					/* H[j] = temp; */
        k=mpz_cdiv_q_ui(rem,q,2); //Zentriere Ausgabe um 0
        k=mpz_cmp(Vector[j],rem);
            if(k>0){
                mpz_sub(Vector[j],Vector[j],q);}
	   }

	mpz_clear(rem);
	mpz_clear(temp);
}

void Share::MulPub(Share& b, mpz_t& Ausgabe){

int i;

Share Null(Anzahl_Spieler, 2*Schwelle, Modulus);
Share Hilfe(Anzahl_Spieler,Schwelle,Modulus);

mpz_t null;
mpz_init_set_ui(null,0);


Null.Erzeugung(null,2*Schwelle,Modulus);
for(i=1; i<=Anzahl_Spieler; i++){
    mpz_mul(Hilfe.Vector[i],Vector[i],b.Vector[i]);
    mpz_add(Hilfe.Vector[i],Hilfe.Vector[i],Null.Vector[i]);
    mpz_mod(Hilfe.Vector[i],Hilfe.Vector[i],Modulus);
    }

Hilfe.Rekonstrukt_GRR(Ausgabe, Modulus, 2*Schwelle);

mpz_clear(null);
}

void Share::Invertiere(){

mpz_t Ausgabe;
mpz_init(Ausgabe);

Share Hilfe(Anzahl_Spieler,Schwelle,Modulus);
Hilfe.Zufallszahl(Schwelle,Modulus);
Share Kopie=Hilfe;
Hilfe.MulPub(*this,Ausgabe);
mpz_invert(Ausgabe,Ausgabe,Modulus);
Kopie.Mul_Konst(Ausgabe,Modulus);
*this=Kopie;

mpz_clear(Ausgabe);
}

void Share::Div_Konst(mpz_t& d, int t, mpz_t& q, int k, mpz_t& q_1){

mpz_t invers_t;
mpf_t invers;
mpf_t zwei_hoch_f_f;

mpf_init(zwei_hoch_f_f);
mpf_set_z(zwei_hoch_f_f,zwei_hoch_f);
mpf_init(invers);
mpz_init(invers_t);

mpf_set_z(invers,d);
mpf_ui_div(invers,1,invers);

mpf_mul(invers, invers, zwei_hoch_f_f); //Berechne das Inverse von d und multipliziere es mit der Anzahl der Nachkommastellen

mpz_set_f(invers_t,invers); //caste invers nach mpz_t
Mul_Konst(invers_t,q); //Vector*invers_t

TruncPr(k,f,t,q,q_1); //Trunc

mpf_clear(zwei_hoch_f_f);
mpf_clear(invers);
mpz_clear(invers_t);
}

void Share::Div_Konst_int(int d, mpz_t& q, mpz_t& q_1, int k){

mpz_t div;
mpz_init_set_si(div,d);

Div_Konst(div, Schwelle, q, k, q_1);

mpz_clear(div);
}

void Share::RandNumbers(mpz_t **randNumbers, int t, mpz_t& q, mpz_t *a, mpz_t *b){
	/* Hilfs- und Zhlvariablen */
	int i, j;

    /* Variablen zur Erstellung von Zufallszahlen */
	mpz_t rand_Num;
	mpz_init(rand_Num);

	/* Generierung von Zufallszahlen fr alle Spieler */
	for(i=0; i<=Anzahl_Spieler; i++){
	   /* Generierung von Zufallszahlen von Spieler i */
	   for(j=0; j<=t ; j++){
	      if(j==0){
	         /* randNumbers[i][0] = fa(i)*fb(i) MOD q setzen */
	         mpz_mul(randNumbers[i][j], a[i], b[i]);
	         mpz_mod(randNumbers[i][j], randNumbers[i][j], q);
            }
	      /* randNumbers[i][j] mit Zufallszahlen besetzen */
	      else{
	         mpz_urandomm(rand_Num, rState1, q);
	         mpz_set(randNumbers[i][j], rand_Num);
	      }
	   }
	}
	/* Speicherfreigabe fr nicht mehr verwendete Variablen */
	mpz_clear(rand_Num);
}

void Share::RandBit(int t, mpz_t& q){
    //Erzeugung eines verteilten Zufallsbits nach Damgard/Fitzi/Kiltz/Nielsen/Toft
    mpz_t invers_zwei;
    mpz_t s;

    mpz_init_set_ui(invers_zwei,2);
    mpz_init(s);

    int Bedingung, i;

    Share a(Anzahl_Spieler,t,q);
    Vektor Hilfsvektor(Anzahl_Spieler); //Speichern von an fr sptere Verwendung

    do{
        a.Zufallszahl(t,q);

        for(i=0; i<=Anzahl_Spieler; i++){
            mpz_set(Hilfsvektor.Vector[i],a.Vector[i]);
            }

        a.MulPub(a,s);//s=a*a

        mpz_mod(s,s,q); //a%p =?
        Bedingung=mpz_cmp_ui(s,0);
        }
    while(Bedingung==0);

    haupt::Diskrete_Wurzel(&s,s,q);

    for(i=0; i<=Anzahl_Spieler; i++){//Wiederherstellung des alten a
        mpz_set(a.Vector[i],Hilfsvektor.Vector[i]);
        }

    mpz_invert(s,s,q);
    a.Mul_Konst(s,q);

    mpz_invert(invers_zwei,invers_zwei,q);

    a.Add_Konst_int(1,q);
    a.Mul_Konst(invers_zwei,q);
    for (i=0; i<=Anzahl_Spieler; i++){
        mpz_set(Vector[i], a.Vector[i]);
        }

mpz_clear(s);
mpz_clear(invers_zwei);
}

void Share::Umwandlung(int t, mpz_t& q, mpz_t& q_1){

    register int i,j;

    mpz_t p;
    mpz_t s;

    mpz_init(p);
    mpz_init(s);

    haupt::reset_RISS();

    Share RISS_q_1(Anzahl_Spieler,t,q_1);
    for(i=1; i<=Anzahl_Spieler; i++){ //RISS to Shamir q_1 nach Secure SCM-D.9.2; berechne fr jeden Spieler
        mpz_set_ui(s,0);
        for(j=1; j<= n_ueber_t; j++){
            mpz_set_ui(p,0);
            if(Teilmengen[j][i]==0){
                mpz_mul(p,RISS_Koeffizienten_q1[i][j],RISS[j]);
                }
        mpz_add(s,s,p);
        }
    mpz_mod(RISS_q_1.Vector[i],s,q_1);
    }

    Share RISS_q(Anzahl_Spieler,t,q); //RISS to Shamir q

    for(i=1; i<=Anzahl_Spieler; i++){
        mpz_set_ui(s,0);
        for(j=1; j<=n_ueber_t; j++){
            mpz_set_ui(p,0);
            if(Teilmengen[j][i]==0){
                mpz_mul(p,RISS_Koeffizienten_q[i][j],RISS[j]);
                }
        mpz_add(s,s,p);
        }
    mpz_mod(RISS_q.Vector[i],s,q);
    }

    Plus_Vektor(RISS_q_1.Vector,Anzahl_Spieler,q_1);
    Rekonstrukt_GRR(s,q_1,t);

    RISS_q.Mul_Konst_int(-1,q);
    RISS_q.Add_Konst(s,q);

    *this=RISS_q;

    mpz_clear(s);
    mpz_clear(p);
}

void Share::Mul_GRR(mpz_t * b, int t, mpz_t& q){//vgl. Wenzl

int i,j,l;

mpz_t temp;
mpz_init(temp);

Matrix koeff(Anzahl_Spieler,Anzahl_Spieler);
Matrix h(Anzahl_Spieler,Anzahl_Spieler);
RandNumbers(koeff.M, t, q, Vector, b);

for(i=1; i<=2*t+1; i++){
	      /* koeff[0]	(a_0)	berechnen!!
	       * erst mal x-beliebige Zahl aus Zq
	       * => Zufallszahl; 1 fr 1. Faktor (Alpha), 1 fr 2. Faktor (Beta)
	       * a_0 = share fr alpha * share fr beta
	       * bis jetzt koeff[i][0] zufllig gewhlt,
	       * spter Produkt aus Shares fr Alpha und Beta
	       */
	      /* Berechnung des Shares fr Spieler j */
	      for(j=1; j<=Anzahl_Spieler; j++){
	         /* Berechnung des Shares hi(j) fr Spieler j,
	          * wobei a[i]*b[i] freier Koeffizient von hi(x) ist */
	         mpz_set(temp, koeff.M[i][t]);//Berechnung der h_ij nach dem Hornerschema
	         for (l=t-1; l>=1; l--){
			    mpz_mul_ui(temp, temp, j);
				mpz_mod(temp, temp, q);				/* MODULO-Reduzierung */
				mpz_add(temp, temp, koeff.M[i][l]);
				mpz_mod(temp, temp, q);				/* MODULO-Reduzierung */
	         }
	         mpz_mul_ui(temp, temp, j);
	         mpz_mod(temp, temp, q);			/* MODULO-Reduzierung */
	         mpz_add(temp, temp, koeff.M[i][0]);
	         mpz_mod(h.M[i][j], temp, q);			/* MODULO-Reduzierung */
	      }
	   }

for (j=1; j<=Anzahl_Spieler; j++){
        mpz_set_ui(temp,0);
        for (i=1; i<=2*t+1; i++){
	         mpz_addmul(temp, h.M[i][j], lambda[i]);
            }
        mpz_mod(Vector[j], temp, q);	        /* MODULO-Reduzierung */
	   }

mpz_clear(temp);
}

void Share::TruncPr(int k, int m, int t, mpz_t& q, mpz_t& q_1){//Protokoll TruncPr aus "Secure Computation with Fixed-Point Numbers"
int i;

mpz_t zwei_hoch;
mpz_t c;

mpz_init(c);
mpz_init_set_ui(zwei_hoch,1);


Share b(Anzahl_Spieler,t,q);
mpz_mul_2exp(zwei_hoch,zwei_hoch,k-1);
for(i=1; i<=Anzahl_Spieler; i++){
    mpz_set(b.Vector[i],Vector[i]);
    }
b.Add_Konst(zwei_hoch,q);

mpz_set_ui(zwei_hoch,1);
Share r(Anzahl_Spieler,t,q_1);
Share r_Strich(Anzahl_Spieler,t,q);

for(i=0;i<m;i++){//Verwendung von q_1 spart ca. 6s\equiv 1/3 der Rechenzeit
    r.RandBit(t,q_1);
    r.Mul_Konst(zwei_hoch,q_1);
    r_Strich.Plus_Vektor(r.Vector,Anzahl_Spieler,q_1);
    mpz_mul_ui(zwei_hoch,zwei_hoch,2);
}

r_Strich.Umwandlung(t,q,q_1);

Share r_2Strich(Anzahl_Spieler,t,q);
r_2Strich.PRandInt(t,k+kappa-m,q);


Share R(Anzahl_Spieler,t,q);

r_2Strich.Mul_Konst(zwei_hoch,q);

R.Plus_Vektor(r_2Strich.Vector,Anzahl_Spieler,q);
R.Plus_Vektor(r_Strich.Vector,Anzahl_Spieler,q);
R.Plus_Vektor(b.Vector,Anzahl_Spieler,q);
R.Rekonstrukt_GRR_mit_VZ(c,q,t);

mpz_mod(c,c,zwei_hoch);

Share a_Strich(Anzahl_Spieler,t,q);
a_Strich.minus(r_Strich.Vector,Anzahl_Spieler,q);
a_Strich.Add_Konst(c,q);

minus(a_Strich.Vector,Anzahl_Spieler,q);

mpz_invert(zwei_hoch,zwei_hoch,q);
Mul_Konst(zwei_hoch,q);

mpz_clear(c);
mpz_clear(zwei_hoch);
}

void Share::Trunc_F28(int t, mpz_t& q, mpz_t& q_1, int k, int m){//Trunc Function; bentigt fr LTZ

mpz_t zwei_hoch;

mpz_init_set_ui(zwei_hoch,2);

mpz_pow_ui(zwei_hoch,zwei_hoch,m);

Share A=*this;//Merke das Ausgangssharing; es wird durch die Funktion Mod_2m berschrieben
Mod2m_F28(k,m,q,q_1,t);

mpz_invert(zwei_hoch,zwei_hoch,q);
A.minus(Vector,Anzahl_Spieler,q);
A.Mul_Konst(zwei_hoch,q);

*this=A;

mpz_clear(zwei_hoch);
}

void Share::Mod2m_F28(int k, int m, mpz_t& q, mpz_t& q_1, int t){

mpz_t zwei;
mpz_t zwei_hoch;
mpz_t c;

mpz_init(c);
mpz_init_set_si(zwei,2);
mpz_init(zwei_hoch);

std::vector<Share> r_Liste;
haupt::PRandBitM_SMPC(k,m,t,q,q_1,&r_Liste,Anzahl_Spieler);

Plus_Vektor(r_Liste.at(m).Vector,Anzahl_Spieler,q);//a+r' (r' an Stelle m in r_Liste gespeichert; vgl. head.h)
mpz_pow_ui(zwei_hoch,zwei,m);

r_Liste.at(m+1).Mul_Konst(zwei_hoch,q);
Plus_Vektor(r_Liste.at(m+1).Vector,Anzahl_Spieler,q);//+2^m*r'' (r'' ist an Stelle m+1 von r_Liste)

mpz_pow_ui(zwei_hoch,zwei,k-1);
Add_Konst(zwei_hoch,q);//+2^{k-1}

Rekonstrukt_GRR(c,q,t);
mpz_pow_ui(zwei_hoch,zwei,m);
mpz_mod(c,c,zwei_hoch);//c<-c mod 2^m

BitLT_F28(t,m,c,r_Liste,q_1);//BitLT wird in F_2^8 durchgefhrt

Mul_Konst(zwei_hoch,q);
Add_Konst(c,q);
minus(r_Liste.at(m).Vector,Anzahl_Spieler,q);

r_Liste.clear();
mpz_clear(zwei_hoch);
mpz_clear(zwei);
mpz_clear(c);
}

void Share::Mod2(int k, mpz_t& q, mpz_t& q_1, int t){

std::vector<Share> r_Liste;

haupt::PRandBitM_SMPC(k,1,t,q,q_1,&r_Liste, Anzahl_Spieler);

Plus_Vektor(r_Liste.at(1).Vector,Anzahl_Spieler,q);

mpz_t zwei_hoch;
mpz_init_set_ui(zwei_hoch,1);
mpz_mul_2exp(zwei_hoch,zwei_hoch,k-1);
Add_Konst(zwei_hoch,q);

r_Liste.at(2).Mul_Konst_int(2,q);
Plus_Vektor(r_Liste.at(2).Vector,Anzahl_Spieler,q);

mpz_t c;
mpz_init(c);
int c_0=0;

Rekonstrukt_GRR(c,q,t);
c_0=mpz_get_si(c);
c_0=c_0%2;

*this=r_Liste.at(1);
Add_Konst_int(c_0,q);
r_Liste.at(1).Mul_Konst_int(-2*c_0,q);
Plus_Vektor(r_Liste.at(1).Vector,Anzahl_Spieler,q);

r_Liste.clear();
mpz_clear(c);
mpz_clear(zwei_hoch);
}

void Share::BitLT_F28(int t, int k, mpz_t& Zahl_1, std::vector<Share>& Bits, mpz_t& q_1){//m=Lnge von Bits; Vergleich des Sharings im Scope mit der bitweise zerlegten Zahl Bits
//k hier entspricht vorherigem m bzw. f
register int i;

Share_F2M T(Anzahl_Spieler, Schwelle);
Share RISS_q(Anzahl_Spieler,Schwelle,q_1);
std::vector<Share_F2M> B(k,T);
mpz_t z;
mpz_init(z);

for(i=0; i<k; i++){
Bits.at(i).Umwandlung_Bit_Zq_F2M(B.at(i),&RISS_q,RISS_Koeffizienten_q1,Schwelle,q_1,z);
B.at(i).Add_Konst_int(1);//1-B(i)=1+B(i), da Arithmetik in F2^8
}

CarryOutL_cin_1offen_F2M(Zahl_1, B,t,k,1,q_1);//Ergebnis erscheint im Scope-Vector
B.at(0).Add_Konst_int(1);//B(0)=1-CarryOutL (-=+, da F2M)
F2MtoZq(Modulus,q_1,B.at(0),t);

mpz_clear(z);
B.clear();
}

void Share::CarryOutL_1offen(mpz_t& Argument_1, std::vector<Share>& Argument_2, int t, mpz_t& q, int k){//Berechne das oberste Bit der Summe Argument_1+Argument_2 (das "CarryOut-Bit")
//k=Anzahl der verteilten Bits vgl. SecureSCM-D.9.2; Ergebnis wird in Scope-Vector geschrieben
register int i=0;

Share A(Anzahl_Spieler, t,q);
std::vector<Share> d_1(k,A);
std::vector<Share> d_2(k,A);
std::vector<Share> Hilfe(k,A);


int Bits_1[k+1];

for(i=0; i<=k; i++){//Initialisiere Bit-Vektor von Argument_1
Bits_1[i]=0;
}

haupt::Bitzerlegung(Argument_1, k, Bits_1);

for(i=0; i<k; i++){//Berechne d_1 und d_2 (jeweils k-Tupel von Sharings) evtl. at(i) durch [] ersetzen

Hilfe[i]=Argument_2[i];//Speichere Argument 2 in einer Hilfsvariablen
d_1.at(i).Plus_Vektor(Argument_2.at(i).Vector,Anzahl_Spieler,q);
d_1.at(i).Add_Konst_int(Bits_1[i],q);
Hilfe.at(i).Mul_Konst_int(Bits_1[i],q);


d_2[i]=Hilfe[i];
Hilfe.at(i).Mul_Konst_int(-2,q);
d_1.at(i).Plus_Vektor(Hilfe.at(i).Vector,Anzahl_Spieler,q);
}

CarryOutAux(d_1,d_2,k,t,q);

*this=d_2[0];
Hilfe.clear();
d_2.clear();
d_1.clear();
}

void Share::CarryOutL_cin_1offen(mpz_t& Argument_1, std::vector<Share>& Argument_2, int t, mpz_t& q, int k, int carry_in){//Berechne das oberste Bit der Summe Argument_1+Argument_2 (das "CarryOut-Bit")
//k=Anzahl der verteilten Bits vgl. SecureSCM-D.9.2; Ergebnis wird in Scope-Vector geschrieben
register int i=0;

Share A(Anzahl_Spieler,t,q);
std::vector<Share> d_1=Argument_2;
std::vector<Share> d_2(k,A);
std::vector<Share> Hilfe(k,A);

int Bits_1[k+1];

for(i=0; i<=k; i++){//Initialisiere Bit-Vektor von Argument_1
Bits_1[i]=0;
}

haupt::Bitzerlegung(Argument_1, k, Bits_1);

for(i=0; i<k; i++){//Berechne d_1 und d_2 (jeweils k-Tupel von Sharings) evtl. at(i) durch [] ersetzen
Hilfe.at(i)=Argument_2.at(i);//Speichere Argument 2 in einer Hilfsvariablen
d_1.at(i).Add_Konst_int(Bits_1[i],q);
Hilfe.at(i).Mul_Konst_int(Bits_1[i],q);

d_2[i]=Hilfe[i];
Hilfe.at(i).Mul_Konst_int(-2,q);
d_1.at(i).Plus_Vektor(Hilfe.at(i).Vector,Anzahl_Spieler,q);
}

Share Hilfsvektor(Anzahl_Spieler,t,q);

Hilfsvektor=d_1[0];
Hilfsvektor.Mul_Konst_int(carry_in, q);
d_2.at(0).Plus_Vektor(Hilfsvektor.Vector,Anzahl_Spieler,q);

CarryOutAux(d_1,d_2,k,t,q);

*this=d_2[0];
Hilfe.clear();
d_2.clear();
d_1.clear();
}

void Share::CarryOutL_cin_1offen_F2M(mpz_t& Argument_1, std::vector<Share_F2M>& Argument_2, int t, int k, int carry_in, mpz_t& q_1){//Berechne das oberste Bit der Summe Argument_1+Argument_2 (das "CarryOut-Bit")
//k=Anzahl der verteilten Bits vgl. SecureSCM-D.9.2; Ergebnis wird in Scope-Vector geschrieben
register int i=0;

Share_F2M A(Anzahl_Spieler,t);
std::vector<Share_F2M> d_1(k,A);
std::vector<Share_F2M> d_2(k,A);
std::vector<Share_F2M> Hilfe(k,A);

int Bits_1[k+1];

for(i=0; i<=k; i++){//Initialisiere Bit-Vektor von Argument_1
Bits_1[i]=0;
}

haupt::Bitzerlegung(Argument_1, k, Bits_1);

for(i=0; i<k; i++){//Berechne d_1 und d_2 (jeweils k-Tupel von Sharings) evtl. at(i) durch [] ersetzen
    d_1.at(i)=Argument_2.at(i);
    d_1.at(i).Add_Konst_int(Bits_1[i]);//d_1(i)=Argument2(i)XOR Bits_1(i)
    d_2.at(i)=Argument_2.at(i);
    d_2.at(i).Mul_Konst_int(Bits_1[i]);
    }

Share_F2M Hilfsvektor(Anzahl_Spieler,t);
Hilfsvektor=d_1.at(0);//g_1<-g_1+carry_in*p_1
Hilfsvektor.Mul_Konst_int(carry_in);
d_2.at(0).Add(Hilfsvektor);

CarryOutAux_F2M(d_1,d_2,k,t);
Argument_2.at(0)=d_2.at(0);

Hilfe.clear();
d_2.clear();
d_1.clear();
}

void Share::CarryOutAux(std::vector<Share>& d_1, std::vector<Share>& d_2, int k, int t, mpz_t& q){//Hilfsprogramm; vgl. SecureSCM-D.9.2; Rckgabe ist d_1.at(1) und d_2.at(1)
register int i,j,l;//durch bergabe per Pointer kann Rckgabe vermieden werden
int r;

r=k%2;
if(r==0)//k<-int(k/2)
        l=k/2;
    else
        l=(k-1)/2;


std::vector<Share> u_1;
std::vector<Share> u_2;

u_1.reserve(k);
u_2.reserve(k);

//Initialisiere u_1 und u_2 an Stelle 0; diese werden fr die Flle k gerade UND ungerade bentigt; die Initialisierung an den Stellen 2,...,l nur fr k gerade
Share A(Anzahl_Spieler,t,q);
u_1.push_back(A);
u_2.push_back(A);

if (k>1){
    if(r==0){//d.h. k gerade; Ist k%2=0, so knnen jeweils 2 d's zu einem u zusammengefasst werden
        for(i=1; i<l; i++){//Initialisiere u_1 und u_2 an den Stellen 2 bis l
            u_1.push_back(A);
            u_2.push_back(A);
            }

        for(i=0; i<l; i++){
            j=2*(i)+1;//Schleife gegenber Vorgabe um 1 nach vorne verschoben; deswegen Indizes um 1 erhht
            k=2*(i);//k kann wiederverwendet werden

            u_1[i]=d_1[j];//(u_1,u_2)=(d_1^{2i}*d_1^{2i-1},d_2^{2i-1}+d_1^{2i-1}d_2^{2i})
            u_1.at(i).MUL_Lory3(d_1.at(k).Vector,t,q);

            u_2[i]=d_2[k];
            u_2.at(i).MUL_Lory3(d_1.at(j).Vector,t,q);
            u_2.at(i).Plus_Vektor(d_2.at(j).Vector,Anzahl_Spieler,q);
            }
        }

        else{
            u_1[0]=d_1[1];//(u_1,u_2)=(d_1^{2i}*d_1^{2i-1},d_2^{2i-1}+d_1^{2i-1}d_2^{2i})
            u_1.at(0).MUL_Lory3(d_1.at(0).Vector,t,q); //Berechne u^1=(u_(1,1),u_(2,1))

            u_2[0]=d_2[0];
            u_2.at(0).MUL_Lory3(d_1.at(1).Vector,t,q);
            u_2.at(0).Plus_Vektor(d_2.at(1).Vector,Anzahl_Spieler,q);

            l=k-1;


            for(i=2; i<k; i++){
                u_1.push_back(d_1.at(i));
                u_2.push_back(d_2.at(i));
            }
        }

    d_1=u_1;
    d_2=u_2;

    CarryOutAux(d_1,d_2,l,t,q);
    }

u_2.clear();
u_1.clear();
}

void Share::CarryOutAux_F2M(std::vector<Share_F2M>& d_1, std::vector<Share_F2M>& d_2, int k, int t){//Hilfsprogramm; vgl. SecureSCM-D.9.2; Rckgabe ist d_1.at(1) und d_2.at(1)

register int i,j,l;//durch bergabe per Pointer kann Rckgabe vermieden werden
int r=k%2;

if(r==0)//k<-int(k/2)
    l=k/2;
else
    l=(k-1)/2;


std::vector<Share_F2M> u_1;
std::vector<Share_F2M> u_2;

u_1.reserve(k);
u_2.reserve(k);

//Initialisiere u_1 und u_2 an Stelle 0; diese werden fr die Flle k gerade UND ungerade bentigt; die Initialisierung an den Stellen 2,...,l nur fr k gerade
Share_F2M A(Anzahl_Spieler,t);
u_1.push_back(A);
u_2.push_back(A);

if (k>1){
    if(r==0){//d.h. k gerade; Ist k%2=0, so knnen jeweils 2 d's zu einem u zusammengefasst werden
        for(i=1; i<l; i++){//Initialisiere u_1 und u_2 an den Stellen 2 bis l
            u_1.push_back(A);
            u_2.push_back(A);
            }

        for(i=0; i<l; i++){
            j=2*(i)+1;//Schleife gegenber Vorgabe um 1 nach vorne verschoben; deswegen Indizes um 1 erhht
            k=2*(i);//k kann wiederverwendet werden

            u_1.at(i)=d_1.at(j); //(u_1,u_2)=(d_1^{2i}*d_1^{2i-1},d_2^{2i-1}+d_1^{2i-1}d_2^{2i})
            u_1.at(i)*=d_1.at(k);//u_1(i)=d_1(2*i+1)*d_1(2*i)

            u_2.at(i)=d_2.at(k);
            u_2.at(i)*=d_1.at(j);
            u_2.at(i).Add(d_2.at(j));//d_2(i)=d_2(2*i+1)+d_1(2*i+1)*d_2(2*i)
            }
        }

        else{
            u_1.at(0)=d_1.at(1); //(u_1,u_2)=(d_1^{2i}*d_1^{2i-1},d_2^{2i-1}+d_1^{2i-1}d_2^{2i})
            u_1.at(0)*=d_1.at(0); //Berechne u^1=(u_(1,1),u_(2,1))

            u_2.at(0)=d_2.at(0);
            u_2.at(0)*=d_1.at(1);
            u_2.at(0).Add(d_2.at(1));

            l=k-1;

            for(i=2; i<k; i++){
                u_1.push_back(d_1.at(i));
                u_2.push_back(d_2.at(i));
                }
            }

        d_1=u_1;
        d_2=u_2;

        CarryOutAux_F2M(d_1,d_2,l,t);
        }

u_2.clear();
u_1.clear();
}

void Share::LTZ_F28(int t, mpz_t& q, mpz_t& q_1, int k){
Trunc_F28(t,q,q_1,k,k-1);
Mul_Konst_int(-1,q);
}

void Share::LT(Share& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k){
minus(Argument_1.Vector,Anzahl_Spieler,q);
LTZ_F28(t,q,q_1,k);
}

void Share::LQ(Share& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k){//<=?

GT(Argument_1,t,q,q_1,k);
Mul_Konst_int(-1,q);
Add_Konst_int(1,q);
}

void Share::LT(int c, int t, mpz_t& q, mpz_t& q_1, int k){

*this-=c;
LTZ_F28(t,q,q_1,k);
}

void Share::LT(mpz_t& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k){
*this-=Argument_1;
LTZ_F28(t,q,q_1,k);
}

void Share::LT_Konst_Betrag(mpz_t& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k){

Quadrat();
*this-=Argument_1;
LTZ_F28(t,q,q_1,k);
}

void Share::GTZ(int t,mpz_t& q, mpz_t& q_1, int k){
*this*=-1;
LTZ_F28(t,q,q_1,k);
}

void Share::GQ(int t, mpz_t& q, mpz_t& q_1,int k){//>=0?
LTZ_F28(t,q,q_1,k);
Mul_Konst_int(-1,q);
Add_Konst_int(1,q);
}

void Share::GT(Share& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k){
*this-=Argument_1;
GTZ(t,q,q_1,k);
}

void Share::GT(mpz_t& Argument_1, int t, mpz_t& q, mpz_t& q_1, int k){
*this-=Argument_1;
GTZ(t,q,q_1,k);
}

void Share::EQZ(int t, mpz_t& q, mpz_t& q_1, int k){//berprfe Gleichheit mit 0 des Sharings im Scope

Vektor_LA BitsV(k,Anzahl_Spieler,Schwelle,q,q_1,k,f);

std::vector<Share> Bits;

mpz_t zwei_hoch;
mpz_t c;

mpz_init_set_ui(zwei_hoch,1);
mpz_init(c);
register int i;

haupt::PRandBitM_SMPC(k,k,t,q,q_1,&Bits,Anzahl_Spieler);

Plus_Vektor(Bits.at(k).Vector,Anzahl_Spieler,q);//a+r' (r' an Stelle m in r_Liste gespeichert; vgl. head.h)
mpz_mul_2exp(zwei_hoch,zwei_hoch,k);

Bits.at(k+1).Mul_Konst(zwei_hoch,q);
Plus_Vektor(Bits.at(k+1).Vector,Anzahl_Spieler,q);
Rekonstrukt_GRR(c,q,t);

mpz_mod(c,c,zwei_hoch);

int Bits_1[k+1];
for(i=0; i<=k; i++){//Initialisiere Bit-Vektor von Argument_1
Bits_1[i]=0;
}
haupt::Bitzerlegung(c,k,Bits_1);

mpz_t s;
mpz_init(s);

Share Hilfsshare(Anzahl_Spieler,t,q_1);
Share Hilfsshare2(Anzahl_Spieler,t,q_1);
for(i=0;i<k;i++){
    Hilfsshare=Bits[i];
    Bits.at(i).Add_Konst_int(Bits_1[i],q_1);
    Hilfsshare.Mul_Konst_int(-2,q_1);
    Hilfsshare.Mul_Konst_int(Bits_1[i],q_1);
    Bits.at(i).Plus_Vektor(Hilfsshare.Vector,Anzahl_Spieler,q_1);
    BitsV[i]=Bits.at(i);
    Hilfsshare=BitsV[i];

    BitsV[i].Umwandlung(t,q,q_1);
    Hilfsshare2=BitsV[i];
    }

BitsV.KOR(*this,k,q);

Mul_Konst_int(-1,q);
Add_Konst_int(1,q);

mpz_clear(s);
mpz_clear(c);
mpz_clear(zwei_hoch);
Bits.clear();
}

int Share::EQZPub(){

Share r(Anzahl_Spieler, Schwelle, Modulus);
r.Zufallszahl(Schwelle, Modulus);

mpz_t c;
mpz_init(c);

MulPub(r,c);

int s;
if(mpz_cmp_ui(c,0)==0){
    s=1;
    }
else{
    s=0;
    }

mpz_clear(c);
return s;
}

void Share::EQ(Share& Argument, int t, mpz_t& q, mpz_t& q_1, int k){//Vergleich mit Argument
*this-=Argument;
EQZ(t,q,q_1,k);
}

void Share::EQ_mpz(mpz_t& Argument, int t, mpz_t&q, mpz_t& q_1, int k){
    *this-=Argument;
    EQZ(t,q,q_1,k);
    }

void Share::EQ_int(int Wert, int t, mpz_t& q, mpz_t& q_1, int k){//Vergleich mit "Wert"
mpz_t Subtrahend;
mpz_init_set_si(Subtrahend,Wert);

*this-=Subtrahend;
EQZ(t,q,q_1,k);

mpz_clear(Subtrahend);
}

void Share::GTZ_approximativ(int t, mpz_t& q, mpz_t& q_1, int k){//berprfe, ob Wert grer als 2^{-f+8} ist

mpz_t epsilon;
mpz_init_set_ui(epsilon,1);
mpz_mul_2exp(epsilon,epsilon,10);
*this-=epsilon;
GTZ(t,q,q_1,k);
mpz_clear(epsilon);
}

void Share::LTZ_approximativ(int t, mpz_t& q, mpz_t& q_1, int k){//berprfe, ob Wert kleiner als 2^{-f+8} ist

mpz_t epsilon;
mpz_init_set_ui(epsilon,1);
mpz_mul_2exp(epsilon,epsilon,15);
*this+=epsilon;
LTZ_F28(t,q,q_1,k);
mpz_clear(epsilon);
}

void Share::OR(Share& Argument){

*this+=Argument-*this*Argument;
}

void Share::AND(Share& Argument){

*this*=Argument;
}

void Share::Signum(int i, mpz_t& q, mpz_t& q_1, int k){

LTZ_F28(Schwelle,q,q_1,k);
Mul_Konst_int(-2,q);
Add_Konst_int(1,q);//*this=1-2*LTZ(*this)
}

void Share::min(const Share& a, const Share& b, mpz_t& q, mpz_t& q_1, int t, int k, Share& Hilfe){
    *this=a;
    *this-=b;
    LTZ_F28(t,q,q_1,k);
    Share A=*this;
    Hilfe=*this;
    Hilfe*=-1;
    Hilfe+=1;
    *this=*this*a+Hilfe*b;
    Hilfe=A;
    }

void Share::KOrCL(std::vector<Share> * Bits, int k, int t, mpz_t& q, mpz_t& q_1){//OR von Bits; Bits sind Sharings in Z/q_1Z

mpz_t zwei_hoch;
mpz_t c;

mpz_init_set_ui(zwei_hoch,1);
mpz_init(c);

int i,m;

double Log=(double) k;
Log=log(Log)/(log((double)2));
m=ceil(Log);
int Bits_1[m];

std::vector<Share> r;
haupt::PRandBitM_SMPC(k,m,t,q,q_1,&r,Anzahl_Spieler);

mpz_mul_2exp(zwei_hoch,zwei_hoch,m);
r.at(m+1).Mul_Konst(zwei_hoch,q);

Share Hilfe(Anzahl_Spieler,t,q);

for(i=0; i<k; i++){
Hilfe.Plus_Vektor(Bits->at(i).Vector,Anzahl_Spieler,q_1);
}

Hilfe.Umwandlung(t,q,q_1);
Hilfe+=r[m];
Hilfe.Plus_Vektor(r.at(m+1).Vector,Anzahl_Spieler,q);

Hilfe.Rekonstrukt_GRR(c,q,t);

for(i=0; i<m; i++){//Initialisiere Bits_1
    Bits_1[i]=0;
    }

haupt::Bitzerlegung(c,m,Bits_1);

for(i=0;i<m;i++){//r XOR Bits_1
    Hilfe=r[i];
    r[i].Add_Konst_int(Bits_1[i],q_1);
    Hilfe.Mul_Konst_int(-2,q_1);
    Hilfe.Mul_Konst_int(Bits_1[i],q_1);
    r[i].Plus_Vektor(Hilfe.Vector,Anzahl_Spieler,q_1);
    }

KOrCS(r,m,t,q_1);

*this=r[0];
Umwandlung(t,q,q_1);

mpz_clear(c);
mpz_clear(zwei_hoch);
r.clear();
}

void Share::KOrCS(std::vector<Share>& Bits, int k, int t, mpz_t& q){
register int i;
int l=k%2,j,m,r;

if(l==0)
    r=k/2;
else
    r=(k-1)/2;

std::vector<Share> d;

d.reserve(r);

Share A(Anzahl_Spieler,t,q);
d.push_back(A);
if(k>1){
if(l==0){//k gerade
    for(i=1; i<r; i++){
        d.push_back(A);
        }

    for(i=0; i<r; i++){
        j=2*i;
        m=2*i+1;
        d[i]=Bits[j];
        Bits[j].Plus_Vektor(Bits[m].Vector,Anzahl_Spieler,q);
        d.at(i).MUL_Lory3(Bits[m].Vector,t,q);
        d.at(i).Mul_Konst_int(-1,q);
        d.at(i).Plus_Vektor(Bits[j].Vector,Anzahl_Spieler,q);
        }
    k=r;
    }
else{
    d[0]=Bits[0];
    Bits[0].Plus_Vektor(Bits[1].Vector,Anzahl_Spieler,q);
    d.at(0).MUL_Lory3(Bits[1].Vector,t,q);
    d.at(0).Mul_Konst_int(-1,q);
    d.at(0).Plus_Vektor(Bits[0].Vector,Anzahl_Spieler,q);

    for(i=2; i<k; i++){
    d.push_back(Bits[i]);
    }

    k=k-1;
    }

Bits=d;
KOrCS(Bits,k,t,q);}
d.clear();
}

void Share::FPDiv(const Share& Nenner, int t, mpz_t& q, mpz_t& q_1, int k){

mpz_t alpha;

mpz_init_set_ui(alpha,1);

int i;

mpz_mul(alpha,zwei_hoch_f,zwei_hoch_f);

Share a=*this; //Zhler speichern
Share x=Nenner;

*this=Nenner;
AppRcr(k,t,q,q_1);

x.MUL_Lory3(Vector,t,q);
x.Mul_Konst_int(-1,q);
x.Add_Konst(alpha,q);//x=alpha-b*w

MUL_Lory3(a.Vector,t,q);
TruncPr(2*k,f,t,q,q_1);

for(i=1; i<Theta_Division; i++){//Theta_Division in head.h definiert
    a=x;
    a.Add_Konst(alpha,q);//a=x+alpha
    MUL_Lory3(a.Vector,t,q);//y<-y*(alpha+x)
    x.MUL_Lory3(x.Vector,t,q);//x<-x*x
    TruncPr(2*k,2*f,t,q,q_1);
    x.TruncPr(2*k,2*f,t,q,q_1);
    }

x.Add_Konst(alpha,q);
MUL_Lory3(x.Vector,t,q);//y<-y*(alpha+x)
TruncPr(2*k,2*f,t,q,q_1);

mpz_clear(alpha);
}

void Share::AppRcr(int k, int t, mpz_t& q, mpz_t& q_1){

Share v(Anzahl_Spieler,t,q);

Norm(v,k,t,q,q_1);//Ausgabe fr c wird in den Scope geschrieben, v in v gespeichert
Mul_Konst_int(-2,q);
Add_Konst(alpha_division,q);//d=alpha-2*c
MUL_Lory3(v.Vector,t,q);
TruncPr(2*k,2*(k-f),t,q,q_1);
}

void Share::Norm(Share& v, int k, int t, mpz_t& q, mpz_t& q_1){
int i;

mpz_t zwei_hoch;
mpz_init_set_ui(zwei_hoch,1);

Share x=*this;

x.LTZ_F28(t,q,q_1,k);

x.Mul_Konst_int(-2,q);
x.Add_Konst_int(1,q); //s=1-2*LTZ(b,k)
Share s=x;
x.MUL_Lory3(Vector,t,q);

Share_F2M A(Anzahl_Spieler,t);
std::vector<Share_F2M> Bits(k,A);

*this=x;//Schreibe x in den Scope-Vektor fr Bit-Dec
BitDec(&Bits,k,k,t,q,q_1);//BitDec verndert den Scope-Vector

A.PreOr_Bits_nicht_cR(Bits,k);

Share B(Anzahl_Spieler,t,q);
std::vector<Share> y(k,B);
std::vector<Share> z(k,B);


for(i=0; i<k; i++){
    y[i].F2MtoZq(q,q_1,Bits[i],t);
    }

for(i=0; i<k-1; i++){
    z[i]=y[i];
    z[i].minus(y[i+1].Vector,Anzahl_Spieler,q);
    }

z[k-1]=y[k-1];

for(i=k-1; i>=0; i--){
z[i].Mul_Konst(zwei_hoch,q);
mpz_mul_ui(zwei_hoch,zwei_hoch,2);
B.Plus_Vektor(z[i].Vector,Anzahl_Spieler,q);
}

x.MUL_Lory3(B.Vector,t,q);//x steht im Scope
s.MUL_Lory3(B.Vector,t,q);

v=s;
*this=x;

mpz_clear(zwei_hoch);
y.clear();
z.clear();
Bits.clear();
}

void Share::Norm_Find_First(Share& v, int k, int t, mpz_t& q, mpz_t& q_1){
int i;
mpz_t zwei_hoch;
mpz_init_set_ui(zwei_hoch,1);
Share x=*this;

x.LTZ_F28(t,q,q_1,k);

x.Mul_Konst_int(-2,q);
x.Add_Konst_int(1,q); //s=1-2*LTZ(b,k)
Share s=x;
x.MUL_Lory3(Vector,t,q);

Share_F2M A(Anzahl_Spieler,t);
std::vector<Share_F2M> Bits(k,A);
std::vector<Share_F2M> Index(k,A);

*this=x;//Schreibe x in den Scope-Vektor fr Bit-Dec
BitDec(&Bits,k,k,t,q,q_1);//BitDec verndert den Scope-Vector

A.FindFirst(Bits,Index,k);

Share B(Anzahl_Spieler,t,q);
std::vector<Share> y(k,B);

for(i=0; i<k; i++){
    y.at(i).F2MtoZq(q,q_1,Index.at(i),t);
    }

for(i=k-1; i>=0; i--){
y.at(i).Mul_Konst(zwei_hoch,q);
mpz_mul_ui(zwei_hoch,zwei_hoch,2);
B.Plus_Vektor(y.at(i).Vector,Anzahl_Spieler,q);
}

x.MUL_Lory3(B.Vector,t,q);//x steht im Scope
s.MUL_Lory3(B.Vector,t,q);

v=s;
*this=x;

mpz_clear(zwei_hoch);
y.clear();
Index.clear();
Bits.clear();
}

void Share::Norm_mit_Wurzel(Share& v, Share& SQ, Share& Wert, int k, int t, mpz_t& q, mpz_t& q_1){//SQ=Wurzel(2^m) oder (wenn m ungerade Wurzel 2^{m-1}), Wert=2^m

int i;
mpz_t zwei_hoch;

int Zahl=2;

mpz_init_set_ui(zwei_hoch,1);

Share x=*this;

/** Wir nehmen an, da der Radikand immer grer Null ist*/

Share_F2M A(Anzahl_Spieler,t);
std::vector<Share_F2M> Bits(k,A);

*this=x;//Schreibe x in den Scope-Vektor fr Bit-Dec
BitDec(&Bits,k,k,t,q,q_1);//BitDec verndert den Scope-Vector
A.PreOr_Bits_nicht_cR(Bits,k);

Share B(Anzahl_Spieler,t,q);
std::vector<Share> y(k,B);
std::vector<Share> z(k,B);
std::vector<Share> W(k,B);
std::vector<Share> W_2(k,B);

for(i=0; i<k; i++){
    y.at(i).F2MtoZq(q,q_1,Bits.at(i),t);
}

for(i=0; i<k-1; i++){
z.at(i)=y.at(i);
z.at(i).minus(y.at(i+1).Vector,Anzahl_Spieler,q);
W.at(i)=z.at(i);//wird fr Wurzelziehung von 2^m bentigt
W_2.at(i)=z.at(i);//wird fr Wurzelziehung von 2^m bentigt
}
z.at(k-1)=y.at(k-1);

Wert=z.at(0);
Share Hilfe(Anzahl_Spieler,Schwelle,Modulus);

for(i=1; i<k; i++){
    Hilfe=z.at(i);
    Hilfe*=Zahl;
    Wert+=Hilfe;
    Zahl+=1;
    }

W.at(k-1)=z.at(k-1);//wird fr Wurzelziehung von 2^m bentigt
W_2.at(k-1)=z.at(k-1);//wird fr Wurzelziehung von 2^m bentigt

int h;//Bestimme Wurzel(2^m) (bzw. Wurzel(2^{m-1}) durch Bitverschiebungen

Vektor_LA H(k,Anzahl_Spieler,t,q,q_1,k,f);

for(i=1; i<k; i=i+2){
h=(i+1)/2;
W.at(h)=W.at(i);
W.at(i)=0;
}

for(i=2; i<k; i=i+2){
h=(i)/2;
W_2.at(h)=W_2.at(i);
W_2.at(i)=0;
}

for(i=0; i<k/2; i++){
W.at(i)+=W_2.at(i);
}

for(i=k/2; i<k; i++){
    W.at(i)=0;
    }

mpz_set_ui(zwei_hoch,1);
for(i=0; i<k/2; i++){
    W.at(i)*=zwei_hoch;
    mpz_mul_ui(zwei_hoch,zwei_hoch,2);
    SQ+=W.at(i);
    }

B=0;
mpz_set_ui(zwei_hoch,1);
for(i=k-1; i>=0; i--){
z.at(i).Mul_Konst(zwei_hoch,q);
mpz_mul_ui(zwei_hoch,zwei_hoch,2);
B.Plus_Vektor(z.at(i).Vector,Anzahl_Spieler,q);
}

x.MUL_Lory3(B.Vector,t,q);//x steht im Scope

v=B;
*this=x;

y.clear();
W.clear();
W_2.clear();
z.clear();
Bits.clear();
mpz_clear(zwei_hoch);
}

void Share::BitDec(std::vector<Share_F2M> * Bits, int k, int m, int t, mpz_t& q, mpz_t& q_1){
int i,nu_int;

mpz_t zwei_hoch;
mpz_t c;

mpz_init_set_ui(zwei_hoch,1);
mpz_init(c);

Share A(Anzahl_Spieler,t,q);
Share_F2M B(Anzahl_Spieler,t);

std::vector<Share> r(m,A);
std::vector<Share_F2M> r_F2M(m,B);

Share R(Anzahl_Spieler,t,q);
Share r_Strich(Anzahl_Spieler,t,q);

for(i=0; i<m; i++){//Erzeuge zufllige Bits gleichzeitig in Z/qZ und F_2^8 und berechne sofort r'=\sum_{i=0}^{m-1}2^i*r(i)
    r[i].PRandBitD(r_F2M[i], t, q, q_1);
    r[i].Mul_Konst(zwei_hoch,q);
    r_Strich.Plus_Vektor(r[i].Vector,Anzahl_Spieler,q);

    mpz_mul_ui(zwei_hoch,zwei_hoch,2);//r'=\sum_{i=0}^{m-1}2^i*r.at(i)
    }

Share r_2Strich(Anzahl_Spieler,t,q);
r_2Strich.PRandInt(t,kappa+k-m,q);
r_2Strich.Mul_Konst(zwei_hoch,q);

R=r_2Strich;
R.Plus_Vektor(r_Strich.Vector,Anzahl_Spieler,q);

double nu_d=(double)n_ueber_t;
nu_d=log(nu_d)/log((double)2);
nu_d=ceil(nu_d);
nu_int=(int)nu_d;

mpz_set_ui(zwei_hoch,2);
mpz_mul_2exp(zwei_hoch,zwei_hoch,k-1);//2^{k}
Add_Konst(zwei_hoch,q);
mpz_mul_2exp(zwei_hoch,zwei_hoch,kappa+nu_int);
Add_Konst(zwei_hoch,q);
minus(R.Vector,Anzahl_Spieler,q);

Rekonstrukt_GRR(c,q,t);
int c_Bits[m]; //Zerlege c in Bits
for(i=0; i<m;i++){
    c_Bits[i]=0;}
haupt::Bitzerlegung(c,m,c_Bits);

Share_F2M cin(Anzahl_Spieler,t);
cin.BitAdd_F2M_offen(&r_F2M,c_Bits,m);

*Bits=r_F2M;

r.clear();
r_F2M.clear();
mpz_clear(zwei_hoch);
mpz_clear(c);
}

void Share::Umwandlung_Bit_Zq_F2M(Share_F2M& r_F2M, Share * RISS_q, mpz_t **RKq, int t, mpz_t& q, mpz_t& z){//Erzeugen von r_0,...,r_{m-1} Zufallsbits aus F_2^8, r'=\sum_{i=0}^{m-1}2^ir_i und r''<2^{k+kappa-m} (in dieser Reihenfolge)

    int Faktor_int;
    register int i,j;

    mpz_t Faktor;
    mpz_t s_mpz;
    mpz_t p_mpz;

    mpz_init(Faktor);
    mpz_init(s_mpz);
    mpz_init(p_mpz);

    GF256 s=0;
    GF256 p;
    GF256 Faktor_256;

    haupt::reset_RISS();

    Share_F2M RISS_F2M(Anzahl_Spieler,t);

    for(i=1; i<=Anzahl_Spieler; i++){ //RISS to Shamir F2M nach Secure SCM-D.9.2; berechne fr jeden Spieler?
        s=0;
        for(j=1; j<=n_ueber_t; j++){
            p=0;
            if(Teilmengen[j][i]==0){
                mpz_mod_ui(Faktor,RISS[j],2);
                Faktor_int=mpz_get_ui(Faktor);
                Faktor_256=Faktor_int;
                p=RISS_Koeffizienten_F2M[i][j]*Faktor_256;//p=RISS[j]*p (als Z-Modul)
                }
        s=s+p;
        }
    RISS_F2M.Vector[i]=s;
    }

    for(i=1; i<=Anzahl_Spieler; i++){
        mpz_set_ui(s_mpz,0);
        for(j=1; j<=n_ueber_t; j++){
            mpz_set_ui(p_mpz,0);
            if(Teilmengen[j][i]==0){
                mpz_mul(p_mpz,RKq[i][j],RISS[j]);
                }
        mpz_add(s_mpz,s_mpz,p_mpz);
        }
    mpz_mod(RISS_q->Vector[i],s_mpz,q);
    }


    Share Hilfe=*this;//*this ist umzuwandelndes Bit
    Hilfe.Plus_Vektor(RISS_q->Vector,Anzahl_Spieler,q);
    Hilfe.Rekonstrukt_GRR(s_mpz,q,t);
    mpz_set(z,s_mpz);//wird bentigt in PRandBitD

    mpz_mod_ui(s_mpz,s_mpz,2);
    Faktor_int=mpz_get_ui(s_mpz);

    RISS_F2M.Add_Konst_int(Faktor_int);//addiere letztes Bit; direkte Summe in F2M entspricht Summe
    r_F2M=RISS_F2M;

    mpz_clear(s_mpz);
    mpz_clear(p_mpz);
    mpz_clear(Faktor);
}

void Share::PRandBitD(Share_F2M& F2M_Bit, int t, mpz_t& q, mpz_t& q_1){//Erzeugung eines Bits sowohl in Z/qZ als auch in F_2^{8}; ersteres wird in den Scope-Vektor geschrieben

mpz_t s;
mpz_init(s);

RandBit(t,q_1);//Erzeugung eines zuflligen Bits ber Z/q_1Z; wird in den Scope-Vektor geschrieben;

Share RISS_q(Anzahl_Spieler,t,q);

Umwandlung_Bit_Zq_F2M(F2M_Bit,&RISS_q,RISS_Koeffizienten_q1,t,q_1,s);//Bit wird in Sharing ber F_{2^8} umgewandelt
*this=RISS_q;
Umwandlung(t,q,q_1);

Mul_Konst_int(-1,q);
Add_Konst(s,q);

mpz_clear(s);
}

void Share::F2MtoZq(mpz_t& q, mpz_t& q_1, Share_F2M& Bit, int t){

Share_F2M Bit_F2M(Anzahl_Spieler,t);
GF256 s;
int s_int;

mpz_t c;
mpz_init(c);

PRandBitD(Bit_F2M,t,q,q_1);
Bit.Add(Bit_F2M);

Bit.Rekonstrukt_GRR(s,t);
s_int=(int) s.value;

Share Hilfe=*this;

Add_Konst_int(s_int,q);
Hilfe.Mul_Konst_int(-2*s_int,q);
Plus_Vektor(Hilfe.Vector,Anzahl_Spieler,q);

mpz_clear(c);
}

Share Share::operator+(const Share& A){

Share Ergebnis=*this;
Ergebnis.Plus_Vektor(A.Vector, Anzahl_Spieler, Modulus);
return Ergebnis;
}

Share Share::operator+(int c){
    Share S=*this;
    S.Add_Konst_int(c,Modulus);
    return S;
    }

Share Share::operator+(mpz_t& c){
    Share S=*this;
    S.Add_Konst(c,Modulus);
    return S;
    }

Share& Share::operator+=(const Share& A){

this->Plus_Vektor(A.Vector, Anzahl_Spieler, Modulus);
return *this;
}

Share& Share::operator+=(int c){

Add_Konst_int(c,Modulus);
return *this;
}

Share& Share::operator+=(mpz_t& c){

Add_Konst(c,Modulus);
return *this;
}

Share Share::operator-(const Share& A){

Share Ergebnis=*this;
Ergebnis.minus(A.Vector, Anzahl_Spieler, Modulus);
return Ergebnis;
}

Share Share::operator-(){

Share Hilfe=*this;
Hilfe*=-1;
return Hilfe;
}

Share Share::operator-(int c){

Share Ergebnis=*this;
Ergebnis.Add_Konst_int(-c,Modulus);
return Ergebnis;
}

Share Share::operator-(mpz_t& c){
    Share S=*this;
    S.Minus_Konst(c,Anzahl_Spieler,Modulus);
    return S;
    }

Share& Share::operator-=(const Share& A){

this->minus(A.Vector, Anzahl_Spieler, Modulus);
return *this;
}

Share& Share::operator-=(mpz_t& Subtrahend){

for(int i=1; i<=Anzahl_Spieler; i++){
mpz_sub(Vector[i],Vector[i],Subtrahend);
}

return *this;
}

Share& Share::operator-=(int c){

mpz_t Subtrahend;
mpz_init_set_si(Subtrahend,c);

for(int i=1; i<=Anzahl_Spieler; i++){
mpz_sub(Vector[i],Vector[i],Subtrahend);
}

mpz_clear(Subtrahend);
return *this;
}

Share Share::operator*(const Share& A){

Share Ergebnis=*this;
Ergebnis.MUL_Lory3(A.Vector,Schwelle,Modulus);
return Ergebnis;
}

Share Share::operator*(mpz_t& a){

Share Ergebnis=*this;
Ergebnis.Mul_Konst(a,Modulus);
return Ergebnis;
}

Share Share::operator*(int n){

mpz_t c;
mpz_init_set_ui(c,n);
Share Ergebnis=*this;

Ergebnis.Mul_Konst(c,Modulus);

mpz_clear(c);
return Ergebnis;
}

Share& Share::operator*=(const Share& A){

this->MUL_Lory3(A.Vector,Schwelle,Modulus);
return *this;
}

Share& Share::operator*=(mpz_t& c){

this->Mul_Konst(c,Modulus);

return *this;
}

Share& Share::operator*=(int c){

mpz_t Konstante;
mpz_init_set_si(Konstante,c);

this->Mul_Konst(Konstante,Modulus);

mpz_clear(Konstante);
return *this;
}

Share& Share::operator=(int n){//Setze ein Sharing konstant auf n

int i;

for(i=1; i<=Anzahl_Spieler; i++){
    mpz_set_ui(this->Vector[i],n);
    }
return *this;
}

Share& Share::operator=(mpz_t n){//Setze ein Sharing konstant auf n

int i;

for(i=1; i<=Anzahl_Spieler; i++){
    mpz_set(this->Vector[i],n);
    }
return *this;
}

void Share::Quadrat(){
Share Hilfe=*this;

*this*=Hilfe;
}

void Share::Wurzel(int k, mpz_t& q, mpz_t& q_1){//Berechnung der Wurzel nach Goldschmidt

int i;
/** Berechne ursprngliche Appr. y von 1/sqrt(x),  d.d. 1/2<x*y^2<3/2 : 2* (-0.2928932190)*x+1.273986062 fr x\in(1,2] */

Share v(Anzahl_Spieler, Schwelle, Modulus);
Share SQ(Anzahl_Spieler, Schwelle, Modulus);//liefert Wurzel von 2^m zurck (bzw. von 2^{m-1})
Share r(Anzahl_Spieler, Schwelle, Modulus);

Share gh=*this;
Share Radikand=*this;
Share Wert(Anzahl_Spieler,Schwelle,q);

Share g=*this;//Kopie erstellen
Norm_mit_Wurzel(v,SQ,Wert,k,Schwelle,q,q_1);


Mul_Konst(C_1_wurzel,q);
Add_Konst(C_0_wurzel,q);

Wert.Mod2(k/2,q,q_1,Schwelle);
Share Eins_minus_Wert=Wert;
Eins_minus_Wert*=-1;
Eins_minus_Wert+=1;
MUL_Lory3(v.Vector,Schwelle,q);

TruncPr(3*k,3*k-2*f,Schwelle,q,q_1);

MUL_Lory3(SQ.Vector,Schwelle,q);
Div_Konst(Wurzel_zwei_hoch_f,Schwelle,q,k,q_1);

int t=Schwelle;
Wert*=*this;
*this=Eins_minus_Wert**this*zwei_hoch_f+Wert*Wurzel_2_z;//Multiplikation mit 2^f im 1.Summanden, da der zweite Summand auf jeden Fall mit 2^f skaliert ist
TruncPr(k,f,t,q,q_1);

g.MUL_Lory3(this->Vector,Schwelle,q);//g=b_0*y_0
g.TruncPr(k,f,Schwelle,q,q_1);

Div_Konst_int(2,q,q_1,k);

gh=*this;
gh.MUL_Lory3(g.Vector,Schwelle,q);//gh_0=g_0*h_0
gh.TruncPr(k,f,Schwelle,q,q_1);

/**Beginne die Iteration; vgl. Markstein */

for(i=0; i<Theta_wurzel-2; i++){
    gh*=-1;
    gh+=halb;
    r=gh;
    r.Add_Konst(eins_Wurzel,q);
    g.MUL_Lory3(r.Vector,Schwelle,q);//g_{i+1}=g_i+g_i*r_i

    MUL_Lory3(r.Vector,Schwelle,q);//h_{i+1}=h_i+h_i*r_i
    g.TruncPr(k,f,Schwelle,q,q_1);
    TruncPr(k,f,Schwelle,q,q_1);
    gh=*this*g;
    gh.TruncPr(k,f,Schwelle,q,q_1);
    }

///Letzte Iteration fr h

gh*=-1;
gh+=halb;
r=gh;
r.Add_Konst(eins_Wurzel,q);

MUL_Lory3(r.Vector,Schwelle,q);//h_{i+1}=h_i+h_i*r_i
TruncPr(k,f,Schwelle,q,q_1);
Mul_Konst_int(2,q);

SQ=*this;
SQ.Quadrat();
SQ*=Radikand;
SQ*=-1;
SQ+=Drei;
SQ*=*this;
SQ.Div_Konst_int(2,q,q_1,k);
SQ*=Radikand;
SQ.TruncPr(4*k,4*f,Schwelle,q,q_1);

*this=SQ;
}

void Share::DivNR(Share& Nenner, mpz_t& q_1, int k){

Share v(Anzahl_Spieler, Schwelle, Modulus);

Nenner.Norm(v, k, Schwelle, Modulus, q_1);
Nenner.RecItNR(k,q_1);
v.MUL_Lory3(Nenner.Vector, Schwelle, Modulus);
MUL_Lory3(v.Vector, Schwelle, Modulus);
TruncPr(3*k-f,2*k-f,Schwelle,Modulus,q_1);
}

void Share::RecItNR(int k, mpz_t& q_1){

int i,theta_int;

mpz_t beta;

mpz_init_set_ui(beta,2);

mpz_mul_2exp(beta,beta,2*k);
double theta_d=log((double)k+1)/log(3.5);
theta_d=ceil(theta_d);
theta_int=(int) theta_d;

Share Hilfe=*this;
Mul_Konst_int(-2,Modulus);
Add_Konst(alpha_division,Modulus);


Share Merke_x(Anzahl_Spieler,Schwelle,Modulus);
for(i=0; i<=theta_int; i++){
    Merke_x=*this;
    Merke_x.MUL_Lory3(Hilfe.Vector,Schwelle,Modulus);
    Merke_x.Mul_Konst_int(-1,Modulus);
    Merke_x.Add_Konst(beta,Modulus);
    MUL_Lory3(Merke_x.Vector,Schwelle,Modulus);
    TruncPr(3*k, 2*k,Schwelle,Modulus,q_1);
    }

mpz_clear(beta);
}

void Share::Inverses(mpz_t& q_1, int k){

Share v(Anzahl_Spieler,Schwelle,Modulus);

Norm(v,k,Schwelle,Modulus,q_1);
RecItNR(k,q_1);
MUL_Lory3(v.Vector,Schwelle,Modulus);
TruncPr(2*k-f,2*k-2*f,Schwelle,Modulus,q_1);
}

void Share::Inverses_ohne_Trunc(mpz_t& q_1, int k){

Share v(Anzahl_Spieler,Schwelle,Modulus);

Norm(v,k,Schwelle,Modulus,q_1);
RecItNR(k,q_1);
MUL_Lory3(v.Vector,Schwelle,Modulus);
}

void Share::Betrag(mpz_t& q_1, int k){

Share Hilfe=*this;

Hilfe.LTZ_F28(Schwelle,Modulus,q_1,k);
Hilfe.Mul_Konst_int(-2,Modulus);
Hilfe.Add_Konst_int(1,Modulus);

MUL_Lory3(Hilfe.Vector,Schwelle,Modulus);
}
