/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include<cstdlib>
#include<vector>
#include<ctime>
#include<string>
#include<unistd.h>

#include "GF256.h"
#include "Share_F2M.h"
#include "Share.h"
#include "gmp.h"

extern GF256 * lambda_256;
extern GF256 * lambda_256_Rekonstrukt;

Share_F2M::Share_F2M(int n, int t):Anzahl_Spieler(n),Schwellenwert(t)
{

    Vector=new GF256[n+1];
    for(int i=0; i<=n; i++){
    Vector[i]=0;
    }
}

Share_F2M::~Share_F2M()
{
    delete [] Vector;
}

Share_F2M::Share_F2M(const Share_F2M& other):Anzahl_Spieler(other.Anzahl_Spieler),Schwellenwert(other.Schwellenwert)
{
    Vector= new GF256[Anzahl_Spieler+1];
    assert(Anzahl_Spieler!=0);
    for(int i=0; i<=Anzahl_Spieler;i++){
    Vector[i]=other.Vector[i];
    }
}

Share_F2M& Share_F2M::operator=(const Share_F2M& rhs)
{
    if (this == &rhs) return *this; // handle self assignment

    for(int i=0; i<=Anzahl_Spieler; i++){
    this->Vector[i]=rhs.Vector[i];
    }
    return *this;
}

Share_F2M& Share_F2M::operator*=(int Faktor){

int i;

for(i=1; i<=Anzahl_Spieler; i++){
    Vector[i]*=Faktor;
    }

return *this;
}

Share_F2M& Share_F2M::operator+=(int Faktor){

int i;

for(i=1; i<=Anzahl_Spieler; i++){
    Vector[i]+=Faktor;
    }

return *this;
}

Share_F2M Share_F2M::operator*(const Share_F2M& Faktor){

Share_F2M Hilfe=*this;

Hilfe.Mul_GRR(Faktor,Schwellenwert);

return Hilfe;
}

Share_F2M& Share_F2M::operator*=(const Share_F2M& Faktor){

Mul_GRR(Faktor,Schwellenwert);

return *this;
}


void Share_F2M::Erzeugung(GF256 Wert){

register int j,k, hilfe_int, hilfe_k;
GF256 J,K,Hilfe;

Share_F2M f_Strich(Anzahl_Spieler, Schwellenwert);//Hilfsvektor, in dem an Stelle 0 der Wert s und an den Stellen 1-t zufllige Werte gespeichert werden
f_Strich.Zufall();
Share_F2M g(Anzahl_Spieler,Schwellenwert); //Hilfsvektor fr die Berechnung der dividierten Differenzen

    for (j=1; j<=Schwellenwert; j++){ //Schreibe die ersten t zuflligen Funktionswerte in den Share-Vektor f
    Vector[j]=f_Strich.Vector[j];
    }
    Vector[0]=Wert;

    g.Vector[0]=Wert;
    for(j=1; j<=Schwellenwert; j++){
        J=j;
        g.Vector[j]=f_Strich.Vector[j];
        for(k=j-1; k>=0; k--){
            K=k;
            g.Vector[k]=(g.Vector[k+1]-g.Vector[k])/(J-K);/* g[k] = g[k+1] - g[k] */
            }
        }

    for(j=Schwellenwert+1; j<=Anzahl_Spieler; j++){
        J=j;
        hilfe_int=j-Schwellenwert;
        for(k=0; k<Schwellenwert; k++){
            hilfe_k=hilfe_int+k;
            Hilfe=hilfe_k;
            g.Vector[k+1]=g.Vector[k+1]+g.Vector[k]*(J-Hilfe);/* g[k+1] = g[k+1] + g[k] */
        }
        Vector[j]=g.Vector[Schwellenwert];
    }

}

void Share_F2M::Erzeugung_GRR(GF256 Wert){

Share_F2M f_Strich(Anzahl_Spieler, Schwellenwert);
f_Strich.Zufall(); //Finde zuflllige Koeffizienten

int i,j;
GF256 I, Share_Wert, Hilfe;

Vector[0]=Wert;
I=1;
for(i=1; i<=Anzahl_Spieler; i++){
I=i;
Hilfe=I;
Share_Wert=0;
    for(j=1; j<=Schwellenwert; j++){
        Share_Wert=I*f_Strich.Vector[j];
        Hilfe=Hilfe*I;
        }
Vector[i]=Share_Wert+Wert;
}
}

void Share_F2M::Zufall(){

int i; //Klasse Erzeugung getestet in eigenem Programm am 12.1.2011
int zufall;
srand(clock() + getpid());

for(i=1; i<=Schwellenwert; i++)
    {
    zufall = rand() % 255;
    Vector[i]=zufall;
    }
}

void Share_F2M::Rekonstrukt(GF256& s, int t) //Abgeleitet von Phase 2 von Mul_2_Lory.c
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    int i,k;
    GF256 I,K,Hilfe_256;

    Share_F2M X(Anzahl_Spieler,t);
    Share_F2M Hilfe(Anzahl_Spieler,t);

    for(i=1; i<=t+1; i++){
        Hilfe.Vector[i]=Vector[i];
        I=i;
        for(k=i-1; k>=1; k--){
            K=k;
            Hilfe_256=I-K;
            Hilfe.Vector[k]=(Hilfe.Vector[k+1]-Hilfe.Vector[k])/Hilfe_256;
	         }
        X.Vector[i]=Hilfe.Vector[1];
        }

    for(k=t; k>=1; k--){
        K=k;
        X.Vector[k]=X.Vector[k]-X.Vector[k+1]*K;
        }

    s=X.Vector[1];
}

void Share_F2M::Rekonstrukt_Ausgabe(std::string N, int t) //Abgeleitet von Phase 2 von Mul_2_Lory.c
{//Anteile der ersten t+1 Spieler werden fr Rekonstruktion herangezogen
    int i,k;
    GF256 I,K,Hilfe_256;

    char * Name;
    Name=new char[N.size()+1];
    strcpy(Name,N.c_str());

    Share_F2M X(Anzahl_Spieler,t);
    Share_F2M Hilfe(Anzahl_Spieler,t);

    for(i=1; i<=t+1; i++){
        Hilfe.Vector[i]=Vector[i];
        I=i;
        for(k=i-1; k>=1; k--){
            K=k;
            Hilfe_256=I-K;
            Hilfe.Vector[k]=(Hilfe.Vector[k+1]-Hilfe.Vector[k])/Hilfe_256;
            }
        X.Vector[i]=Hilfe.Vector[1];
        }

    for(k=t; k>=1; k--){
        K=k;
        X.Vector[k]=X.Vector[k]-X.Vector[k+1]*K;
        }

    I=X.Vector[1];
    printf("\n%s=%d",Name,I.value);
}

void Share_F2M::Rekonstrukt_GRR(GF256& Wert, int t){

int i;

Wert=0;
for(i=1; i<=t+1; i++){
Wert=Wert+Vector[i]*lambda_256_Rekonstrukt[i];
}
}

void Share_F2M::Rekonstrukt_Spieler_beliebig(GF256 * s, int * Spielerliste, int t){//Die an der Rekonstruktion teilnehmenden Spieler sind in *Spielerliste mit 1 kodiert; die anderen mit 0

register int i,k,j=1,l;

int Spielerliste_Liste[t+1]; //Auf den Positionen 1,...,t+1 werden die an der Rekonstruktion beteiligten  Spieler eingetragen
GF256 inverses,I,J, Differenzenmatrix[Anzahl_Spieler+1][Anzahl_Spieler+1];

    for(i=0; i<=Anzahl_Spieler; i++){
        Spielerliste_Liste[i]=0;
        if(Spielerliste[i]==1){
            Spielerliste_Liste[j]=i;
            j++;
            }
        }

    for(i=Anzahl_Spieler; i>=0; i--){ //Initialisierung der Differenzenmatrix
        I=i;
        for(j=0; j<=Anzahl_Spieler; j++){
            J=j;
            Differenzenmatrix[i][j]=I-J;
            }
        }

    Share_F2M X(Anzahl_Spieler,Schwellenwert);
    Share_F2M Hilfe(Anzahl_Spieler,Schwellenwert);

     for(i=1; i<=t+1; i++){
        j=Spielerliste_Liste[i];
        Hilfe.Vector[i]=Vector[j]; //Schiebe Sttzstellen und die zugehrigen Inversen zusammen
        for(k=i-1; k>=1; k--){
                Hilfe.Vector[k]= Hilfe.Vector[k+1]- Hilfe.Vector[k];	// g[k] = g[k+1] - g[k]
                l=Spielerliste_Liste[k];
                inverses=Differenzenmatrix[j][l];
                inverses=inverses.inverse();
                Hilfe.Vector[k]= Hilfe.Vector[k]* inverses;//Division
                }
        X.Vector[i]= Hilfe.Vector[1];
        }			// X[i] = g[1]


    for(k=t; k>=1; k--){
        X.Vector[k+1]=X.Vector[k+1]*Differenzenmatrix[Spielerliste_Liste[k]][0];
        X.Vector[k]= X.Vector[k]- X.Vector[k+1];		// X[k] = X[k] - X[k+1]
        }

    *s=X.Vector[1];
}

void Share_F2M::Add(Share_F2M Summand_2){
int i;

for(i=1; i<=Anzahl_Spieler; i++){
Vector[i]=Vector[i]+Summand_2.Vector[i];
}
}

void Share_F2M::Add_Konst_int(int Konstante){

GF256 c;
c=Konstante;

for(int i=1; i<=Anzahl_Spieler; i++){
Vector[i]=Vector[i]+c;
}

}

void Share_F2M::Mul(const Share_F2M& Faktor_2){

register int i, j, k, hilfe_int, hilfe_k;

	GF256 I,J,K,Hilfe;


	Share_F2M g(2*Schwellenwert+1,0);							/* g */

	Share_F2M A(2*Schwellenwert+1,Schwellenwert);
	std::vector<Share_F2M> F(Anzahl_Spieler+1,A);					/* f */

    Share_F2M X(Anzahl_Spieler,0);

	RandNumbers(&F, Schwellenwert, *this, Faktor_2);

	   for(i=1; i<=Anzahl_Spieler; i++){
          g.Vector[0]=F.at(i).Vector[0];
	      for(j=1; j<=Schwellenwert; j++){
            J=j;
		     g.Vector[j]=F.at(i).Vector[j];
			 for(k=j-1; k>=0; k--){
			     K=k;
		   	    g.Vector[k]=(g.Vector[k+1]-g.Vector[k])/(J-K);
                }
            }
	      for(j=Schwellenwert+1; j<=Anzahl_Spieler; j++){
	          J=j;
	          hilfe_int=j-Schwellenwert;
	         for(k=0; k<Schwellenwert; k++){
                K=k;
                hilfe_k=hilfe_int+k;
                Hilfe=hilfe_k;//Hilfe = 2.Endunkt der Sttzstellen; 1. ist J
                g.Vector[k+1]=g.Vector[k+1]+g.Vector[k]*(J-Hilfe);
	         }
	         F.at(i).Vector[j]=g.Vector[Schwellenwert];
	      }
	   }

	   for(j=1; j<=Anzahl_Spieler; j++){
	      for(i=1; i<=2*Schwellenwert+1; i++){
	          I=i;
	          g.Vector[i]=F.at(i).Vector[j];
	         for(k=i-1; k>=1; k--){
	             K=k;
	             g.Vector[k]=(g.Vector[k+1]-g.Vector[k])/(I-K);
	         }
	         X.Vector[i]=g.Vector[1];
	      }
	      for(k=2*Schwellenwert; k>=1; k--){
	          K=k;
	          X.Vector[k]=X.Vector[k]-X.Vector[k+1]*(K);
	      }
	   Vector[j]=X.Vector[1];
	   }
F.clear();
}

void Share_F2M::Mul_Konst(GF256 Konstante){
int i;

for(i=0; i<=Anzahl_Spieler; i++){
Vector[i]=Vector[i] * Konstante;
}
}

void Share_F2M::Mul_Konst_int(int c){
GF256 Konstante=c;
Mul_Konst(Konstante);
}

void Share_F2M::RandNumbers(std::vector<Share_F2M> * randNumbers, int t, const Share_F2M& a, const Share_F2M& b){
	/* Hilfs- und Zhlvariablen */
	int i, j,zufall;
    /* Variablen zur Erstellung von Zufallszahlen */

	/* Generierung von Zufallszahlen in fr alle Spieler */
	for(i=0; i<2*t+2; i++){
	   /* Generierung von Zufallszahlen von Spieler i */
	   for(j=0; j<=t ; j++){
	      if(j==0){
	         randNumbers->at(i).Vector[j]=a.Vector[i] * b.Vector[i];
            }
	      else{
            zufall = rand() % 256;
            randNumbers->at(i).Vector[j]=zufall;
	      }
	   }
	}
}

void Share_F2M::PreOr_Bits_nicht_cR(std::vector<Share_F2M>& Faktoren, int Anzahl_Faktoren){//vgl. PreOrC in Catrina/de Hoogh, PreMulC hier nicht implementiert da zwar konstante Rundenzahl, aber mehr Mutliplikationen
//fr Bit-Input und F_2^8 vereinfacht
int i;
GF256 s;

Share_F2M Hilfe=Faktoren.at(Anzahl_Faktoren-1);

for(i=0; i<Anzahl_Faktoren; i++){
Faktoren.at(i).Add_Konst_int(1);
}

for(i=Anzahl_Faktoren-2; i>=0; i--){//Faktoren(i)=\prod_{j=i}^{Anzahl_Faktoren}Faktoren(j)
Faktoren.at(i).Mul_GRR(Faktoren.at(i+1),Schwellenwert);
}
Faktoren.at(Anzahl_Faktoren-1)=Hilfe;

for(i=Anzahl_Faktoren-2; i>=0; i--){
Faktoren.at(i).Add_Konst_int(1);
}
}

void Share_F2M::FindFirst(std::vector<Share_F2M>& VektorF2M, std::vector<Share_F2M>& Index, int l){

Share_F2M C(Anzahl_Spieler,Schwellenwert);
Share_F2M S(Anzahl_Spieler,Schwellenwert);
Share_F2M nll(Anzahl_Spieler,Schwellenwert);

int i;

for(i=0; i<l; i++){
    Index.at(i)=nll;
    Index.at(i)=VektorF2M.at(i);
    Index.at(i).Mul_GRR(S,Schwellenwert);//I(i)=V(i)*S
    C=VektorF2M.at(i);
    C*=-1;
    C+=1;//C=1-V[i]
    S=S*C;
    }
}

void Share_F2M::BitAdd_F2M(std::vector<Share_F2M>& a, std::vector<Share_F2M>& b, int k){//Bitweise Addition von a und b; Ergebnis wird in a geschrieben
int i;

Share_F2M A(Anzahl_Spieler,Schwellenwert);
std::vector<Share_F2M> p(k,A);
std::vector<Share_F2M> g(k,A);

p=a;
g=b;

g.at(0).Mul_GRR(a.at(0),Schwellenwert);
p.at(0).Add(b.at(0));
a.at(0)=p.at(0);
Share_F2M cin=g.at(0);


for(i=1; i<k; i++){//vgl. SecureSCM-D.9.2 S. 45 - angepasst
p.at(i).Add(b.at(i));//p=a\oplus b
g.at(i).Mul_GRR(a.at(i),Schwellenwert);//g=a\cdot b
cin.Mul_GRR(p.at(i),Schwellenwert);
cin.Add(g.at(i));
p.at(i).Add(cin);
a.at(i)=p.at(i);
}
a.push_back(cin);//letztes Bit anhngen

g.clear();
p.clear();
}

void Share_F2M::BitAdd_F2M_offen(std::vector<Share_F2M> *a, int * r, int k){
int i;

GF256 c;

Share_F2M A(Anzahl_Spieler,Schwellenwert);
std::vector<Share_F2M> p;
std::vector<Share_F2M> g;

p=*a;
g=*a;

*this=g.at(0);//*this=c_0

a->at(0).Add_Konst_int(r[0]);//LSB der Summe

/*Berechne erstes echtes Carry-Bit*/
Mul_Konst_int(r[0]);

Share_F2M Hilfe(Anzahl_Spieler,Schwellenwert);

for(i=1; i<k; i++){//vgl. SecureSCM-D.9.2 S. 45 - angepasst
    p.at(i).Add_Konst_int(r[i]);//p=a\oplus b

    Hilfe=p.at(i);//Merke p fr sptere Berechnung von cin(i+1)

    p.at(i).Add(*this);//im Scope ist cin gespeichert; das Summenbit erhlt man durch Addition der zewi neuen Bit
    a->at(i)=p.at(i);

    g.at(i).Mul_Konst_int(r[i]);//g=a\cdot b


    Mul_GRR(Hilfe,Schwellenwert);//cin_{i+1}=g(i)+cin_{i}*p(i+1)  <- hinterer Summand
    Add(g.at(i)); //cin_{i+1}=g(i)+cin_{i}*p(i+1)  <- vorderer Summand
    }

a->push_back(*this);//letztes Bit anhngen
g.clear();
p.clear();
}

void Share_F2M::Mul_GRR(const Share_F2M& b, int t){
int i,j,l;
GF256 temp=0,J;

Share_F2M A(Anzahl_Spieler,Schwellenwert);
std::vector<Share_F2M> koeff(Anzahl_Spieler+1,A);
std::vector<Share_F2M> h(Anzahl_Spieler+1,A);


RandNumbers(&koeff, t, *this, b);

for(i=0; i<=Anzahl_Spieler; i++){
	      for(j=1; j<=Anzahl_Spieler; j++){
	          J=j;
	         temp=koeff[i].Vector[t];
	         for (l=t-1; l>=1; l--){
                temp=temp*J;
			    temp=temp+koeff[i].Vector[l];
                }
	         temp=temp*J;
	         temp=temp+koeff[i].Vector[0];
	         h[i].Vector[j]=temp;
	      }
	   }

    for (j=1; j<=Anzahl_Spieler; j++){
        temp=0;
        for (i=1; i<=2*t+1; i++){
            temp=temp+(h[i].Vector[j]*lambda_256[i]);
            }
        Vector[j]=temp;
	   }

h.clear();
koeff.clear();
}

