/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include <Simplex_Startwert_Catrina.h>
#include <Matrix_LA.h>
#include <head.h>
#include <cstdlib>
#include <cstdio>

Simplex_Startwert_Catrina::Simplex_Startwert_Catrina(int m, int n, int Anzahl_Ungleichheitsbedingungen, int Anzahl_Spieler, int Schwelle, mpz_t& qmod, mpz_t& q_1mod, int kl, int ffl)://m: Anzahl NBen, n Anzahl Variablen
Matrix_LA(m+1, n+1, Anzahl_Spieler, Schwelle, qmod, q_1mod, kl, ffl),//Erstellen des Simplex-Tableaus
Tableau_Zeilen(m+1),
Tableau_Spalten(n+1),
Spielerzahl(Anzahl_Spieler),
Schwellenwert(Schwelle),
k(kl),
ff(ffl)
{
if(Anzahl_Ungleichheitsbedingungen>m){
    printf("\nSimplex-Tableau kann nicht erstellt werden, mehr Ungleichungen als Zeilen. Abbruch");
    exit(1);
    }
mpz_init_set(q,qmod);
mpz_init_set(q_1,q_1mod);
}

Simplex_Startwert_Catrina::~Simplex_Startwert_Catrina()
{
mpz_clear(q_1);
mpz_clear(q);
}

Simplex_Startwert_Catrina::Simplex_Startwert_Catrina(const Simplex_Startwert_Catrina& other):
Matrix_LA(other.Tableau_Zeilen, other.Tableau_Spalten, other.Spielerzahl, other.Schwellenwert, other.q, other.q_1, other.k, other.ff),
Tableau_Zeilen(other.Tableau_Zeilen),
Tableau_Spalten(other.Tableau_Spalten),
Spielerzahl(other.Spielerzahl),
Schwellenwert(other.Schwellenwert),
k(other.k),
ff(other.ff)
{
    int i,j;

    mpz_init_set(q,other.q);
    mpz_init_set(q_1,other.q_1);

    for(i=0; i<Zeilenzahl; i++){
        for(j=0; j<Spaltenzahl; j++){
            M[i][j]=other.M[i][j];
            M[i][j].Laenge=other.M[i][j].Laenge;
            M[i][j].Anzahl_Spieler=other.M[i][j].Anzahl_Spieler;
            M[i][j].Schwelle=other.M[i][j].Schwelle;
            mpz_init_set(M[i][j].Modulus,q);
            }
        }
}

Simplex_Startwert_Catrina& Simplex_Startwert_Catrina::operator=(const Simplex_Startwert_Catrina& rhs)
{
    if (this == &rhs) return *this; // handle self assignment

    if((Tableau_Zeilen!=rhs.Tableau_Zeilen) || (Tableau_Spalten!=rhs.Tableau_Spalten)){
        printf("\nZuweisung nicht mglich. Dimensionen verschieden.");
        exit(1);
        }

    M=rhs.M;

    return *this;
}

void Simplex_Startwert_Catrina::Tableau_Erzeugen(Matrix_LA& A_alt, Vektor_LA& b, Vektor_LA& b_Signum, Vektor_LA&c, Vektor_LA& x0, int Anzahl_Gleichheitsbedingungen, int Anzahl_Ungleichheitsbedingungen){

if(b.d!=A_alt.Zeilenzahl){
    printf("A und b inkonsistent. Abbruch");
    exit(1);
    }

if(A_alt.Zeilenzahl!=Anzahl_Gleichheitsbedingungen+Anzahl_Ungleichheitsbedingungen || Tableau_Zeilen-1!=Anzahl_Gleichheitsbedingungen+Anzahl_Ungleichheitsbedingungen){
    printf("\nAnzahl Bedingungen nicht mit Tableau kompatibel. Abbruch");
    exit(1);
    }

int i,j;

Share Eins(Spielerzahl,Schwellenwert,q);
Eins.Erzeugung_double(1.0,Schwellenwert,q);


for(i=0; i<Anzahl_Gleichheitsbedingungen+Anzahl_Ungleichheitsbedingungen; i++){
    for(j=0; j<A_alt.Spaltenzahl; j++){
        M[i][j]=A_alt.M[i][j]*b_Signum[i];
        }
    }

for(i=0; i<b.d; i++){
    M[i][Tableau_Spalten-1]=b[i]*b_Signum[i];
    }

for(i=0; i<Spaltenzahl-1; i++){
    M[Tableau_Zeilen-1][i]=c[i];
    }

Vektor_LA X(Tableau_Spalten-1,Spielerzahl,Schwellenwert,q,q_1,k,N);
x0.Oben_einpassen(b);
X.Oben_einpassen(x0);
X.SkalarMult(c,M[Tableau_Zeilen-1][Tableau_Spalten-1]);
M[Tableau_Zeilen-1][Tableau_Spalten-1].TruncPr(k,N,Schwellenwert,q,q_1);
}

int Simplex_Startwert_Catrina::Simplex_Algorithmus(Matrix_LA& A_alt, Vektor_LA&b, Vektor_LA& c, Vektor_LA& x0, Vektor_LA& X, int Anzahl_Gleichheitsbedingungen, int Anzahl_Ungleichheitsbedingungen, Vektor_LA& B, Vektor_LA& N){//Simplex-Algorithmus mit Startwert x0


int i,Anzahl_Iterationen=0;

if(X.d!=Spaltenzahl-1){
    printf("\n\nSimplex_Startwert_Catrina::Simplex_Algorithmus");
    printf("\nAchtung: Der Loesungsvektor X hat die falsche Groesse!. Abbruch");
    exit(1);
    }

Vektor_LA b_kleiner_Null=b;
Vektor_LA b_Signum(b.d,Spielerzahl,Schwellenwert,q,q_1,k,ff);

for(i=0; i<b_kleiner_Null.d; i++){
    b_kleiner_Null[i].LTZ_F28(Schwellenwert,q,q_1,k);
    b_Signum[i]=-b_kleiner_Null[i]*2+1;
    }

Tableau_Erzeugen(A_alt, b, b_Signum, c, x0, Anzahl_Gleichheitsbedingungen, Anzahl_Ungleichheitsbedingungen);

mpz_t eins;
mpz_init_set_ui(eins,1);

Vektor_LA V(Tableau_Spalten-1,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA V_ganz(Tableau_Spalten,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA C(Tableau_Zeilen-1,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA C_ganz(Tableau_Zeilen,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA W(Tableau_Zeilen-1,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA W_ganz(Tableau_Zeilen,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA R(Tableau_Spalten-1,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA R_ganz(Tableau_Spalten,Spielerzahl,Schwellenwert,q,q_1,k,f);
Share p(Spielerzahl, Schwellenwert, q);

mpz_t Zahl;
mpz_init(Zahl);

for(i=0; i<B.d; i++){
    mpz_set_ui(Zahl,A_alt.Spaltenzahl+i);
    B[i].Erzeugung(Zahl,Schwellenwert,q);
    }
for(i=0; i<N.d; i++){
    mpz_set_ui(Zahl,i);
    N[i].Erzeugung(Zahl,Schwellenwert,q);
    }


int s;

Matrix_LA A(Tableau_Zeilen-1,Tableau_Spalten-1,Spielerzahl,Schwellenwert,q,q_1,k,ff);
A.Links_oben_einpassen(*this);
TransformCost(A,b_kleiner_Null);
Vektor_LA HH(Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);


for(i=0; i<1000; i++){
    Anzahl_Iterationen++;
    s=GetPivColumn(V);

    V_ganz.Oben_einpassen(V);

    if(s==0){//Optimale Lsung
        break;
        }

    C_ganz.Spalte_auslesen(*this,V_ganz,Tableau_Zeilen);
    C.Oben_einpassen(C_ganz);

    s=GetPivRow(C,W);

    W_ganz.Oben_einpassen(W);

    if(s==1){
        Ausgabe_skaliert("T");
        N.Ausgabe_skaliert("N");
        printf("\nDas Problem ist unbeschraenkt");
        break;
        }
    R_ganz.SecReadRow(*this,W_ganz,Tableau_Spalten);
    R.Oben_einpassen(R_ganz);

    p=R[V];

    UpdTab(C_ganz, R_ganz, V_ganz, W_ganz, p);
    UpdVar(B,N,V,W);
    }

if(i==999){
    printf("\n Eine zulssige Lsung konnte in 1000 Iterationen nicht gefunden werden. Abbruch.");
    exit(1);
    }

GetSolution(B,X);

return Anzahl_Iterationen;
}

void Simplex_Startwert_Catrina::UpdCost(Vektor_LA& B, Matrix_LA& A, Vektor_LA b, int d){

Vektor_LA X(A.Spaltenzahl+A.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Vektor_LA x(A.Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Vektor_LA r_VGL_NULL(b.d,Spielerzahl,Schwellenwert,q,q_1,k,ff);

mpz_t z;
mpz_init_set_ui(z,1);
mpz_mul_2exp(z,z,8);

GetSolution(B,X);
x.Oben_einpassen(X);

Vektor_LA X_VGL_NULL=x;
x.MatrixMult(A,r_VGL_NULL);
r_VGL_NULL.TruncPr();
r_VGL_NULL=b-r_VGL_NULL; //r=b-Ax

Vektor_LA Hilfe_x(X_VGL_NULL.d,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Hilfe_x.getZeilenvektor_bis_Spalte(*this,Tableau_Zeilen-1,Hilfe_x.d);
Share H(Spielerzahl,Schwellenwert,q);

for(int i=0; i<x.d; i++){
    X_VGL_NULL[i].LTZ_F28(Schwellenwert,q,q_1,k);
    }

Vektor_LA Hilfe_r(r_VGL_NULL.d,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Vektor_LA rr(Hilfe_r.d,Spielerzahl,Schwellenwert,q,q_1,k,ff);

for(int i=0; i<r_VGL_NULL.d; i++){
    rr[i]=r_VGL_NULL[i];
    r_VGL_NULL[i].Signum(Schwellenwert,q,q_1,k);
    rr[i].Quadrat();
    rr[i].TruncPr(k,ff,Schwellenwert,q,q_1);
    rr[i].LT(z,Schwellenwert,q,q_1,k);//Prfe, ob rr[i]=0
    }

for(int i=0; i<x.d; i++){
    M[Tableau_Zeilen-1][i]=-X_VGL_NULL[i]*zwei_hoch_f;
    }

for(int i=x.d; i<Tableau_Spalten-1; i++){
    M[Tableau_Zeilen-1][i]=r_VGL_NULL[i-x.d]*zwei_hoch_f;
    M[Tableau_Zeilen-1][i]*=(-rr[i-x.d]+1);
}

mpz_clear(z);
}

void Simplex_Startwert_Catrina::TransformCost(Vektor_LA& W, Matrix_LA& A, Vektor_LA& b_Signum){

int i;

Matrix_LA P(A.Spaltenzahl,A.Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Matrix_LA P2(A.Spaltenzahl,A.Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Matrix_LA P3(A.Zeilenzahl,A.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Matrix_LA N(A.Zeilenzahl,A.Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Matrix_LA A_B(A.Zeilenzahl,A.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Matrix_LA A_N(A.Zeilenzahl,A.Spaltenzahl-A.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Matrix_LA Kopie_A=A;

Vektor_LA Z(A.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Vektor_LA Hilfe_Z=Z;
Vektor_LA c(Spaltenzahl-1,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Vektor_LA c_N(A.Spaltenzahl-A.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Vektor_LA c_B(A.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);

mpz_t eins;
mpz_init_set_ui(eins,1);

Vektor_LA V(W.d,Spielerzahl,Schwellenwert,q,q_1,k,ff);
for(i=0; i<V.d; i++){
    V[i].Erzeugung(eins,Schwellenwert,q);
    }
V=V-W;//Nicht-Basisvariablen

A.Multipliziere_Spaltenweise(W.Vector_LA);
N=Kopie_A-A;
A.Matrix_spaltenweise_zusammenschieben_binaer(W.Vector_LA,P);
A_B.Kopiere_links(A);

N.Multipliziere_Spaltenweise(V.Vector_LA);
N.Matrix_spaltenweise_zusammenschieben_binaer(V.Vector_LA,P2);
A_N.Kopiere_links(N);

A_B.Vorwaertselimination_SMPC_mit_Pivot(P3);
for(i=0; i<A_N.Spaltenzahl; i++){
    Z.getSpaltenvektor(A_N,i);
    Z.VWSubs(A_B);
    Hilfe_Z.RWSubs(A_B,Z);
    A_N.Ersetze_Spalte(Hilfe_Z.Vector_LA,i,Hilfe_Z.d);
    }//Berechne A_B^{-1}*A_N
c.getZeilenvektor_bis_Spalte(*this,Tableau_Zeilen-1,Tableau_Spalten-1);//Kostenfunktion extrahieren
Vektor_LA c2=c;//Wird fr c_B bentigt
c.Multipliziere_Zeilenweise(V);//=c_N

c2=c2-c;//c_B=c-c_N

c.MatrixMult(P2,c);
c_N.Oben_einpassen(c);

c2.MatrixMult(P,c2);
c_B.Oben_einpassen(c2);
c_B.Multipliziere_Zeilenweise(b_Signum);//VZ von b, das in A_B auftaucht, bercksichtigen

Matrix_LA A_N_T(A_N.Spaltenzahl,A_N.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Vektor_LA y(A_N_T.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
A_N_T=A_N.Transponiere();

c_B.MatrixMult(A_N_T,y);
y.TruncPr();
c_N=c_N-y;
c2=0;
c2.Oben_einpassen(c_N);
P=P2;
P.Transponiere_quadratisch();
c2.MatrixMult(P,c);

for(i=0; i<Tableau_Spalten-1; i++){
    M[Tableau_Zeilen-1][i]=c[i];
    }

mpz_clear(eins);
}

void Simplex_Startwert_Catrina::TransformCost(Matrix_LA& A, Vektor_LA& b_kleiner_Null){

int i;

Vektor_LA c_N(A.Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Vektor_LA c_B(A.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);

for(i=0; i<c_B.d; i++){
    c_B[i]=b_kleiner_Null[i];//Wenn b<0, d.h. 0 in Komponente i nicht zulssig, minimiere das Residuum; der Faktor -1 ist bereits in A_N_T enthalten
    }


Matrix_LA A_N_T(Spaltenzahl-1,Zeilenzahl-1,Spielerzahl,Schwellenwert,q,q_1,k,ff);
Vektor_LA y(A_N_T.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,ff);
A_N_T=A.Transponiere();

c_B.MatrixMult(A_N_T,y);
y*=-1;
c_N=y;

for(i=0; i<Tableau_Spalten-1; i++){
    M[Tableau_Zeilen-1][i]=c_N[i];
    }
}

int Simplex_Startwert_Catrina::GetPivColumn(Vektor_LA& V){//gib s (Anzeige, ob Pivotwahl mglich) zurck

mpz_t z;
mpz_init_set_ui(z,1);
mpz_mul_2exp(z,z,8);

Vektor_LA D(Tableau_Spalten-1,Spielerzahl,Schwellenwert,q,q_1,k,f);
Share Summe(Spielerzahl,Schwellenwert,q);
for(int i=0; i<Tableau_Spalten-1; i++){
    D[i]=M[Tableau_Zeilen-1][i];
    D[i]+=z;
    D[i].LTZ_F28(Schwellenwert,q,q_1,k);//LTZ approximativ bestimmen
    Summe+=D[i];
    }

int s=1-Summe.EQZPub();

V=D;
if(s==0){
    mpz_clear(z);
    return s;
    }
else{
    V.SelectFirst();

    mpz_clear(z);
    return s;
    }
}

int Simplex_Startwert_Catrina::GetPivRow(Vektor_LA& C, Vektor_LA& W){//W=Rckgabe, B ist in der Tabelle im Scope enthalten

int i,s=0;

mpz_t unendlich;
mpz_init_set_ui(unendlich,1);
mpz_mul_2exp(unendlich,unendlich,k-1);
mpz_sub_ui(unendlich,unendlich,1);

mpz_t z;
mpz_init_set_ui(z,1);
mpz_mul_2exp(z,z,8);

Share T(Spielerzahl,Schwellenwert,q);
Vektor_LA D(Tableau_Zeilen-1,Spielerzahl,Schwellenwert,q,q_1,k,N);
Vektor_LA Werte_D=D;

Vektor_LA B(Tableau_Zeilen-1,Spielerzahl,Schwellenwert,q,q_1,k,N);
B.getSpaltenvektor_bis_Zeile(*this,Tableau_Spalten-1,Tableau_Zeilen-1);//letzte Spalte (entspricht b) des Tableaus auslesen
Vektor_LA Werte_B=B;

for(i=0; i<Tableau_Zeilen-1; i++){
    D[i]=C[i];
    Werte_D[i]=D[i];
    D[i]-=z;
    D[i].GTZ(Schwellenwert,q,q_1,k);
    }

for(i=0; i<D.d; i++){
    Werte_D[i]=D[i]*Werte_D[i]+(-D[i]+1)*zwei_hoch_f;//Wenn Nenner nicht grer 0, ersetze Nenner durch FP(1)(=2^f)
    Werte_B[i]=D[i]*Werte_B[i]+(-D[i]+1)*unendlich;//Wenn Nenner nicht ger 0, ersete Zhler durch unendlich
    T+=D[i];
    }

s=T.EQZPub();

if(s==1){
    W=D;
    return s;
    }

Werte_B.Min_Bruch(Werte_D,W);

mpz_clear(unendlich);
mpz_clear(z);
return s;
}

void Simplex_Startwert_Catrina::UpdTab(Vektor_LA& C, Vektor_LA& R, Vektor_LA& V, Vektor_LA& W, Share& p){

Share y=p;
y.Inverses_ohne_Trunc(q_1,k);
Share Hilfe=p+zwei_hoch_f;
R.SecWrite(V,Hilfe);
Hilfe=p-zwei_hoch_f;
C.SecWrite(W,Hilfe);

int i,j;
for(i=0; i<C.d; i++){
    C[i]*=y;
    }

Simplex_Startwert_Catrina Tableau_Kopie=*this;
for(i=0; i<C.d; i++){
    for(j=0; j<R.d; j++){
        M[i][j]=C[i]*R[j];
        M[i][j].TruncPr(3*k,2*k,Schwellenwert,q,q_1);
        M[i][j]=Tableau_Kopie.M[i][j]-M[i][j];
        }
    }
}

void Simplex_Startwert_Catrina::UpdVar(Vektor_LA& B, Vektor_LA& N, Vektor_LA& V, Vektor_LA& W){

Share s(Spielerzahl,Schwellenwert,q);
s=B[W];

Share u(Spielerzahl,Schwellenwert,q);
u=N[V];

B.SecWrite(W,u);
N.SecWrite(V,s);
}

void Simplex_Startwert_Catrina::GetSolution(Vektor_LA& B, Vektor_LA& X){

int i;

X=0;
Vektor_LA V(Zeilenzahl+Spaltenzahl-2,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA x=V;
for(i=0; i<Zeilenzahl-1; i++){
    V=0;
    V.BitMask(B[i]);
    x.SecWrite(V,M[i][Tableau_Spalten-1]);
    }

X.Oben_einpassen(x);
}

