/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include<cstdio>
#include<cstdlib>
#include<vector>
#include <head.h>
#include "gmp.h"
#include "Matrix.h"
#include "Vektor.h"

Vektor::Vektor(int n):Laenge(n){ /* Funktion zur Speichererzeugung und Initialisierung eines Vektors */

Vector = new mpz_t[n+1];

for (int i=0; i<=n; i++){
    mpz_init_set_ui(Vector[i],0);
    }
}

Vektor::Vektor(){Laenge=0;}

Vektor::~Vektor(){

for(int i=0; i<Laenge+1; i++){
    mpz_clear(Vector[i]);
    }

delete []Vector;
Vector=NULL;
}

Vektor::Vektor(const Vektor& Ursprungsvektor):Laenge(Ursprungsvektor.Laenge)
{

Vector=new mpz_t[Laenge+1];
assert(Laenge!=0); //berprfe, ob Lnge korrekt erzeugt

for(int i=0; i<=Laenge; i++){
    mpz_init_set(Vector[i],Ursprungsvektor.Vector[i]);
    }
}

Vektor& Vektor::operator=(const Vektor& Ursprungsvektor){

if(&Ursprungsvektor!=this){//Vermeide Selbstzuweisung
    for(int i=0; i<=Laenge; i++){
        mpz_set(Vector[i],Ursprungsvektor.Vector[i]);
        }
    }

return *this;
}

void Vektor::print(char * name, int n){

int x;

printf("\n\n%s = \n", name);

for (x=0; x<=n; x++){
        gmp_printf("(%d,%Zd)\n",x,Vector[x]);
        }
}

void Vektor::Plus_Vektor(mpz_t * Summand2, int n, mpz_t& q){

for(int i=1; i<=n; i++){
    mpz_add(Vector[i],Vector[i],Summand2[i]);
    mpz_mod(Vector[i],Vector[i],q);
    }
}

void Vektor::minus(mpz_t * Summand2, int n, mpz_t& q){

for(int i=1; i<=n; i++){
    mpz_sub(Vector[i],Vector[i],Summand2[i]);
    mpz_mod(Vector[i],Vector[i],q);
    }
}

void Vektor::Mul_Konst(mpz_t& Konstante, int n, mpz_t& q){

for(int i=1; i<=n; i++){
    mpz_mul(Vector[i],Vector[i],Konstante);
    mpz_mod(Vector[i],Vector[i],q);
    }
}

void Vektor::Mul_Konst_int(int c, int n, mpz_t& q){

mpz_t Konstante;
mpz_init_set_si(Konstante,c);
Mul_Konst(Konstante,n,q);
mpz_clear(Konstante);
}

void Vektor::Add_Konst(mpz_t& Konstante, int n, mpz_t& q){

for(int i=1; i<=n; i++){
    mpz_add(Vector[i],Vector[i],Konstante);
    mpz_mod(Vector[i],Vector[i],q);
    }
}

void Vektor::Minus_Konst(mpz_t& Konstante, int n, mpz_t& q){

for(int i=1; i<=n; i++){
    mpz_sub(Vector[i],Vector[i], Konstante);
    mpz_mod(Vector[i],Vector[i],q);
    }
}

void Vektor::Zufall(int n, int t, mpz_t& p){//Erzeugung eines Vektors mit zuflligen Eintrgen auf den Positionen 1,...,t (<n)
register int i;

if ((t+1)>n){
    printf("\nFehler: Vektor::Zufall");
    printf("\n\nAchtung: t+1>n. Abbruch");
    printf("\n t=%d, n=%d",t,n);
    exit(1);
    }

for (i=1; i<=t; i++)
    {
    mpz_urandomm(Vector[i],rState1,p);
    }
}
