/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include <Vektor_LA.h>
#include <Matrix.h>
#include <Share_F2M.h>
#include <head.h>

#include<cstdlib>
#include<cassert>
#include<cstring>
#include<ctime>

extern mpz_t * lambda;

Vektor_LA::Vektor_LA(int Laenge, int n, int t, mpz_t& grosser_Modulus, mpz_t& kleiner_Modulus, int Bitlaenge, int Nachkommastellen)
{
    d=Laenge;
    Spielerzahl=n;
    Schwellenwert=t;
    mpz_init_set(q,grosser_Modulus);
    mpz_init_set(q_1,kleiner_Modulus);
    k=Bitlaenge;
    N=Nachkommastellen;

    Share A(n,t,q);
    Vector_LA.assign(Laenge,A);
}

Vektor_LA::~Vektor_LA()
{
    Vector_LA.clear();

    mpz_clear(q);
    mpz_clear(q_1);
}

Vektor_LA::Vektor_LA(const Vektor_LA& other):d(other.d), Spielerzahl(other.Spielerzahl), Schwellenwert(other.Schwellenwert), k(other.k), N(other.N)
{
    Vector_LA=other.Vector_LA;

    mpz_init_set(q,other.q);
    mpz_init_set(q_1,other.q_1);
}

Vektor_LA& Vektor_LA::operator=(const Vektor_LA& rhs)
{
    if (this == &rhs) return *this; // handle self assignment

    if(d!=rhs.d){
        printf("\nDimensionen verschieden. Zuweisung nicht moeglich.");
        exit(1);
        }

    Vector_LA=rhs.Vector_LA;

    return *this;
}

void Vektor_LA::operator=(int n)
{
    int i,j;

    for(i=0; i<d; i++){
        for(j=0; j<=Spielerzahl; j++){
            mpz_set_ui(this->Vector_LA.at(i).Vector[j],n);
            }
        }
}

Vektor_LA& Vektor_LA::operator=(const Vektor_mpf_t& Init){

if(d!=Init.Dimension){
    printf("\nWarnung: Vektor_LA::operator=(const Vektor_mpf_t& Init)");
    printf("\nZuweisung des Vektors Init nicht moeglich, da Dimension verschieden. Abbruch.");
    printf("\nd=%d, Init.d=%d",d,Init.Dimension);
    exit(1);
    }

int i;

double DD;
for(i=0; i<d; i++){
    DD=mpf_get_d(Init.V[i]);
    Vector_LA.at(i).Erzeugung_double(DD,Schwellenwert,q);
    }

return *this;
}

void Vektor_LA::Unskaliert_einlesen(const Vektor_mpf_t& rhs){

if(rhs.Dimension!=d){
    printf("\nWarnung: Vektor_LA::Unskaliert_einlesen");
    printf("\nZuweisung nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

int i,l;
mpz_t g;
mpz_init(g);

for(i=0; i<d; i++){
    l=mpf_get_si(rhs.V[i]);
    mpz_set_si(g,l);
    Vector_LA.at(i).Erzeugung(g,Schwellenwert,q);
    }

mpz_clear(g);
}

Vektor_LA Vektor_LA::operator+(const Vektor_LA& VLA){

assert(d==VLA.d);//berprfe, ob Dimension gleich

int i;

Vektor_LA Ergebnis=*this;

for(i=0; i<d; i++){
Ergebnis.Vector_LA.at(i).Plus_Vektor(VLA.Vector_LA.at(i).Vector,Spielerzahl,q);
}
return Ergebnis;
}

void Vektor_LA::operator+=(const Vektor_LA& VLA){

assert(d==VLA.d);//berprfe, ob Dimension gleich

int i;

for(i=0; i<d; i++){
Vector_LA.at(i).Plus_Vektor(VLA.Vector_LA.at(i).Vector,Spielerzahl,q);
}
}

Vektor_LA Vektor_LA::operator-(const Vektor_LA& VLA){

assert(d==VLA.d);//berprfe, ob Dimension gleich

int i;

Vektor_LA Ergebnis=*this;

for(i=0; i<d; i++){
Ergebnis.Vector_LA.at(i).minus(VLA.Vector_LA.at(i).Vector,Spielerzahl,q);
}

return Ergebnis;
}

void Vektor_LA::operator-=(const Vektor_LA& VLA){

assert(d==VLA.d);//berprfe, ob Dimension gleich

int i;

for(i=0; i<d; i++){
this->Vector_LA.at(i).minus(VLA.Vector_LA.at(i).Vector,Spielerzahl,q);
}
}

Vektor_LA Vektor_LA::operator*(mpz_t& Skalar){

int i;

Vektor_LA Ergebnis=*this;

for(i=0; i<d; i++){
Ergebnis.Vector_LA.at(i).Mul_Konst(Skalar,q);
}

return Ergebnis;
}

Vektor_LA Vektor_LA::operator*(int c){

mpz_t Konstante;
mpz_init_set_si(Konstante,c);

Vektor_LA Ergebnis=*this;

for(int i=0; i<d; i++){
Ergebnis.Vector_LA.at(i).Mul_Konst(Konstante,q);
}

mpz_clear(Konstante);
return Ergebnis;
}

Vektor_LA& Vektor_LA::operator*=(mpz_t& Skalar){

int i;

for(i=0; i<d; i++){
Vector_LA.at(i).Mul_Konst(Skalar,q);
}

return *this;
}

Vektor_LA& Vektor_LA::operator*=(int c){

mpz_t Konstante;
mpz_init_set_si(Konstante,c);

*this*=Konstante;

mpz_clear(Konstante);
return *this;
}

Vektor_LA Vektor_LA::operator*(Share& Faktor){//Multiplikation mit verteiltem Faktor

int i;

Vektor_LA Ergebnis=*this;

for(i=0; i<d; i++){
Ergebnis.Vector_LA.at(i).MUL_Lory3(Faktor.Vector,Schwellenwert,q);
}

return Ergebnis;
}

Vektor_LA& Vektor_LA::operator*=(Share& Faktor){//Multiplikation mit verteiltem Faktor; Ergebnis wird in den Scope geschrieben

int i;

for(i=0; i<d; i++){
this->Vector_LA.at(i).MUL_Lory3(Faktor.Vector,Schwellenwert,q);
}

return *this;
}

Share& Vektor_LA::operator[](int i){

return this->Vector_LA.at(i);
}

Share Vektor_LA::operator[](const Vektor_LA& Index){//Zugriff auf das in Index binr gespeicherte Element

int i;

Share Wert(Spielerzahl,Schwellenwert,q);

for(i=0; i<d; i++){
Wert+=Vector_LA.at(i)*Index.Vector_LA.at(i);
}

return Wert;
}

void Vektor_LA::Summe_Vektor(Share& Summe, mpz_t& Modulus){

int i,j;

Summe=0;
for(i=1; i<Spielerzahl; i++){
    for(j=0; j<d; j++){
        mpz_add(Summe.Vector[i],Summe.Vector[i],Vector_LA.at(j).Vector[i]);
        }
    mpz_mod(Summe.Vector[i],Summe.Vector[i],Modulus);
    }
}

void Vektor_LA::Multipliziere_Zeilenweise(Vektor_LA& Binaerer_Vektor){

if(d!=Binaerer_Vektor.d){
    printf("\n\nFehler: Vektor_LA::Multipliziere_Zeilenweise");
    printf("\nDie Dimension des Scope-Vektors ist %d, die Dimension des Binren Vektors ist %d. Zeilenweise Mutliplikation nicht moeglich. Abbruch.",d, Binaerer_Vektor.d);
    exit(1);
    }

int i;

for(i=0; i<d; i++){
    Vector_LA[i]*=Binaerer_Vektor.Vector_LA.at(i);
    }
}

void Vektor_LA::SkalarMult(const Vektor_LA& b, Share& Ergebnis){//Nach Catrina/de Hoogh: Improved Primitives; s. auch MUL_Lory3 in Share.cpp; Lagrange-Koeffizienten werden vorausberechnet in main

if(d!=b.d){
    printf("\nFehler: Vektor_LA::SkalarMult");
    printf("\nBerechnung des Skalarprodukts nicht moeglich, da Dimensionen verschieden. Abbruch");
    printf("\na.d=%d, b.d=%d",d,b.d);
    exit(1);
    }

int i,j,l;

mpz_t rand_Num,temp;

mpz_init(rand_Num);
mpz_init(temp);


Matrix D(Spielerzahl,Spielerzahl);
Vektor_LA Hilfsvektor(d,Spielerzahl,Schwellenwert,q,q_1,k,f);

for(i=1; i<=Spielerzahl; i++){//Fr alle Spieler
    for(j=0; j<d; j++){//Fr alle Komponenten
        mpz_mul(Hilfsvektor.Vector_LA.at(j).Vector[i],Vector_LA.at(j).Vector[i],b.Vector_LA.at(j).Vector[i]);
        mpz_add(D.M[i][0],D.M[i][0],Hilfsvektor.Vector_LA.at(j).Vector[i]);
        }
}

	/* Generierung von Zufallszahlen fr alle Spieler */
	for(i=1; i<=Spielerzahl; i++){//Erzeuge Zufallszahlen fr Sharing von d
	   /* Generierung von Zufallszahlen von Spieler i */
	   for(j=1; j<=Schwellenwert ; j++){
            mpz_urandomm(rand_Num, rState1, q);
            mpz_set(D.M[i][j], rand_Num);
            }
        }

Share g(Spielerzahl,Schwellenwert,q);

for(i=1; i<=2*Schwellenwert+1; i++){//Erzeuge Sharing von d
    mpz_set(g.Vector[0], D.M[i][0]);				/* g[0] = f[0] */
    for(j=1; j<=Schwellenwert; j++){
        mpz_set(g.Vector[j], D.M[i][j]);			/* g[j] = f[j] */
        for(l=j-1; l>=0; l--){
            mpz_sub(g.Vector[l], g.Vector[l+1], g.Vector[l]);	/* g[k] = g[k+1] - g[k] */
            mpz_mod(g.Vector[l], g.Vector[l], q);	        /* MODULO-Reduzierung */
            }
        }
    for(j=Schwellenwert+1; j<=Spielerzahl; j++){
        for(l=0; l<Schwellenwert; l++){
            mpz_add(g.Vector[l+1], g.Vector[l+1], g.Vector[l]);	/* g[k+1] = g[k+1] + g[k] */
            mpz_mod(g.Vector[l+1], g.Vector[l+1], q);		/* MODULO-Reduzierung */
            }
        mpz_set(D.M[i][j], g.Vector[Schwellenwert]);		/* f[j] = g[t] */
        }
    }

for (j=1; j<=Spielerzahl; j++){
    mpz_set_ui(temp,0);
    for (i=1; i<=2*Schwellenwert+1; i++){
        mpz_addmul(temp, D.M[i][j], lambda[i]);
        }
mpz_mod(Ergebnis.Vector[j], temp, q); 			/* MODULO-Reduzierung */
}

mpz_clear(temp);
mpz_clear(rand_Num);
}

void Vektor_LA::MatrixMult(const Matrix_LA& A, Vektor_LA& Ergebnis_Vektor){//m,n: Dimensionen: n=d!

if(A.Spaltenzahl!=d){
printf("\n\nFehler: Vektor_LA::MatrixMult");
printf("\nMultiplikation nicht moeglich, da Dimensionen verschieden. Abbruch");
printf("\nA.Spaltenzahl=%d, d=%d",A.Spaltenzahl,d);
exit(1);
}

if(A.Zeilenzahl!=Ergebnis_Vektor.d){
printf("\n\nFehler: Vektor_LA::MatrixMult");
printf("\nMultiplikation nicht moeglich, da Dimensionen verschieden. Abbruch");
printf("\nA.Zeilenzahl=%d, Ergebnis_Vektor.d=%d",A.Zeilenzahl,Ergebnis_Vektor.d);
exit(1);
}

int i;

Vektor_LA Hilfe=*this;//Hilfsvektor
Vektor_LA Arbeitsvektor(d,Spielerzahl,Schwellenwert,q,q_1,k,f);

for(i=0; i<A.Zeilenzahl; i++){
    Arbeitsvektor.getZeilenvektor(A,i);
    Arbeitsvektor.SkalarMult(Hilfe, Ergebnis_Vektor[i]);
    }
}

void Vektor_LA::MatrixMult_ab_Zeile(const Matrix_LA& A, Vektor_LA& Ergebnis_Vektor, int m, int n, int Zeilenzahl){//m,n: Dimensionen: n=d!
//Multipliziere erst ab Zeile Zeilenzahl
if(n!=d){
printf("\nMultiplikation nicht moeglich, da Dimensionen verschieden. Abbruch");
exit(1);
}

int i;

Vektor_LA Hilfe=*this;//Hilfsvektor
Vektor_LA Arbeitsvektor(d,Spielerzahl,Schwellenwert,q,q_1,k,f);

for(i=0; i<Zeilenzahl; i++){
    Ergebnis_Vektor[i]=Vector_LA.at(i);
    }

for(i=Zeilenzahl; i<m; i++){
    Arbeitsvektor.getZeilenvektor(A,i);
    Arbeitsvektor.SkalarMult(Hilfe,Ergebnis_Vektor[i]);
    }
}

void Vektor_LA::getZeilenvektor(const Matrix_LA& Ma, int Zeilenindex){
int j;

if(Ma.Spaltenzahl!=d){
    printf("\n\nVektor_LA::getZeilenvektor");
    printf("\nM.Spaltenzahl=%d, d=%d",Ma.Spaltenzahl,d);
    printf("\nFehler: Kopieren von Matrixzeile in Vektor nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

for(j=0; j<d; j++){
Vector_LA.at(j)=Ma.M[Zeilenindex][j];
}
}

void Vektor_LA::getZeilenvektor_ab_Spalte(const Matrix_LA& Ma, int Zeilenindex, int Spaltenindex){

int j;

if(Ma.Spaltenzahl-Spaltenindex!=d){
    printf("\nFehler: Vektor_LA::getZeilenvektor_ab_Spalte.");
    printf("\nKopieren von Matrixzeile in Vektor nicht moeglich, da Dimensionen verschieden. Abbruch");
    printf("\nSpaltenzahl=%d, Spaltenindex=%d, d=%d",Ma.Spaltenzahl,Spaltenindex,d);
    exit(1);
    }

for(j=0; j<d; j++){
Vector_LA.at(j)=Ma.M[Zeilenindex][j+Spaltenindex];
}
}

void Vektor_LA::getZeilenvektor_bis_Spalte(const Matrix_LA& Ma, int Zeilenindex, int Spaltenindex){

int j;

if(Spaltenindex!=d){
    printf("\nWarnung: Vektor_LA::getZeilenvektor_bis_Spalte");
    printf("\nFehler: Kopieren von Matrixzeile in Vektor nicht moeglich, da Dimensionen verschieden. Abbruch");
    printf("\nSpaltenindex=%d, d=%d",Spaltenindex,d);
    exit(1);
    }

for(j=0; j<d; j++){
Vector_LA.at(j)=Ma.M[Zeilenindex][j];
}
}

void Vektor_LA::getSpaltenvektor(const Matrix_LA& Ma, int Spaltenindex){

int i;

if(Ma.Zeilenzahl!=d){
    printf("\nWarnung: Vektor_LA::getSpaltenvektor");
    printf("\nFehler: Kopieren von Matrixspalte in Vektor nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

for(i=0; i<d; i++){
Vector_LA[i]=Ma.M[i][Spaltenindex];
}
}

void Vektor_LA::getSpaltenvektor_ab_Zeile(const Matrix_LA& Ma, int Spaltenindex, int Zeilenindex){
int i;

if(d!=Ma.Zeilenzahl-Zeilenindex){
    printf("\nWarnung: Vektor_LA::getSpaltenvektor_ab_Zeile");
    printf("\nFehler: Kopieren von Matrixteilspalte in Vektor nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

for(i=0; i<d; i++){
Vector_LA[i]=Ma.M[i+Zeilenindex][Spaltenindex];
}
}

void Vektor_LA::getSpaltenvektor_bis_Zeile(const Matrix_LA& Ma, int Spaltenindex, int Zeilenindex){
int i;

if(d!=Zeilenindex){
    printf("\nWarnung: Vektor_LA::getSpaltenvektor_bis_Zeile");
    printf("\nFehler: Kopieren von Matrixteilspalte in Vektor nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

for(i=0; i<Zeilenindex; i++){
Vector_LA[i]=Ma.M[i][Spaltenindex];
}
}

void Vektor_LA::getSpaltenvektor_ab_Zeile(Vektor_LA& b, int Zeile){

if(d!=b.d-Zeile){
    printf("\nWarnung: Vektor_LA::getSpaltenvektor_ab_Zeile");
    printf("\nFehler: Kopieren von Vektorteilspalte in Vektor nicht mglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

for(int i=0; i<d; i++){
    Vector_LA[i]=b[i+Zeile];
    }
}

void Vektor_LA::getSpaltenvektor_bis_Zeile(Vektor_LA& b, int Zeile){

if(d!=Zeile){
    printf("\nWarnung: Vektor_LA::getSpaltenvektor_bis_Zeile");
    printf("\nFehler: Kopieren von Vektorteilspalte in Vektor nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

for(int i=0; i<Zeile; i++){
    Vector_LA[i]=b[i];
    }
}

void Vektor_LA::Ausgabe(std::string Name){

int i;
mpz_t s;
mpz_init(s);

printf("\n");

char * c;
c= new char [Name.size()+1];
strcpy(c,Name.c_str());

for(i=0; i<d; i++){
Vector_LA.at(i).Rekonstrukt_GRR_mit_VZ(s,q,Schwellenwert);
gmp_printf("\n%s(%d)=%Zd",c,i,s);
}

delete []c;
c=NULL;
mpz_clear(s);
}

void Vektor_LA::Ausgabe(std::string Name, mpz_t& p){

int i;
mpz_t s;
mpz_init(s);

printf("\n");

char * c;
c= new char [Name.size()+1];
strcpy(c,Name.c_str());

for(i=0; i<d; i++){
Vector_LA.at(i).Rekonstrukt(s,Schwellenwert,p);
gmp_printf("\n%s(%d)=%Zd",c,i,s);
}

delete []c;
c=NULL;
mpz_clear(s);
}

void Vektor_LA::Ausgabe_skaliert(std::string Name){

int i,z;

mpz_t s;
mpf_t c;

mpz_init(s);
mpf_init(c);

char * r;
r= new char [Name.size()+1];
strcpy(r,Name.c_str());

printf("\n");

for(i=0; i<d; i++){
Vector_LA.at(i).Rekonstrukt_GRR_mit_VZ(s,q,Schwellenwert);
mpf_set_z(c,s);
mpf_div_2exp(c,c,(mp_bitcnt_t)f);
z=gmp_printf("\n%s(%d)=%.10Ff",r,i,c);
}

delete[]r;
r=NULL;
mpf_clear(c);
mpz_clear(s);
}

void Vektor_LA::Ausgabe_skaliert_Vektor_mpf_t(Vektor_mpf_t& V){

if(V.Dimension!=d){
    printf("\nWarnung: Vektor_LA::Ausgabe_skaliert_Vektor_mpf_t");
    printf("\n Ausgabe in Vektor_mpf_t nicht moeglich, da Dimensionen verschieden. Abbruch.");
    exit(1);
    }

int i;
mpz_t s;
mpf_t c;
mpz_init(s);
mpf_init(c);

for(i=0; i<d; i++){
Vector_LA.at(i).Rekonstrukt_GRR_mit_VZ(s,q,Schwellenwert);
mpf_set_z(c,s);
mpf_div_2exp(c,c,(mp_bitcnt_t)f);
mpf_set(V.V[i],c);
}

mpf_clear(c);
mpz_clear(s);
}

int Vektor_LA::Ausgabe_skaliert_Datei(std::string Name, FILE* datei1){

int i,z;
mpz_t s;
mpf_t c;
mpz_init(s);
mpf_init(c);

char * r;
r= new char [Name.size()+1];
strcpy(r,Name.c_str());

if(datei1==NULL){
    printf("\nFehler beim ffnen der Datei Ausgabe.txt");//Fehler beim ffnen
    return 1;
    }

for(i=0; i<d; i++){
Vector_LA.at(i).Rekonstrukt_GRR_mit_VZ(s,q,Schwellenwert);
mpf_set_z(c,s);
mpf_div_2exp(c,c,(mp_bitcnt_t)f);
z=gmp_fprintf(datei1,"\n%s(%d)=%.10Ff",r,i,c);
}

fprintf(datei1,"\n");

delete[]r;
r=NULL;
mpf_clear(c);
mpz_clear(s);
return 0;
}

int Vektor_LA::Ausgabe_Datei(std::string Name, FILE* datei1){

int i,z;
mpz_t s;
mpz_init(s);

char * r;
r= new char [Name.size()+1];
strcpy(r,Name.c_str());

if(datei1==NULL){
    printf("\nFehler: Vektor_LA::Ausgabe_Datei");
    printf("\nDatei %s ist nicht geoeffnet. Abbruch.",Name.c_str());//Fehler beim ffnen
    return 1;
    }

fprintf(datei1,"\n");

for(i=0; i<d; i++){
Vector_LA.at(i).Rekonstrukt_GRR_mit_VZ(s,q,Schwellenwert);
z=gmp_fprintf(datei1,"\n%s(%d)=%Zd",r,i,s);
}


delete[]r;
r=NULL;
mpz_clear(s);
return 0;
}

void Vektor_LA::Ausgabe_Shares(std::string Name){

char * r;
r= new char [Name.size()+1];
strcpy(r,Name.c_str());

int i;

for(i=0; i<d; i++){
    Vector_LA[i].print("r",Spielerzahl);
    }
delete []r;
r=NULL;
}

void Vektor_LA::TruncPr(){//Komponentenweises Abschneiden des Vektors mit Hilfe des TruncPr-Protokolls
int i;

for(i=0; i<d; i++){
Vector_LA.at(i).TruncPr(k,f,Schwellenwert,q,q_1);
}
}

void Vektor_LA::TruncPr(int l, int r){//Komponentenweises Abschneiden des Vektors mit Hilfe des TruncPr-Protokolls,; Lnge der Zahl: k Bits, davon f Nachkommastellen
int i;

for(i=0; i<d; i++){
Vector_LA.at(i).TruncPr(l,r,Schwellenwert,q,q_1);
}
}

void Vektor_LA::VWSubs(const Matrix_LA& L){

int l;

for(l=1; l<d; l++){
    Vektor_LA B(l+1,Spielerzahl,Schwellenwert,q,q_1,k,N);
    Vektor_LA v(l+1,Spielerzahl,Schwellenwert,q,q_1,k,N);
    Vektor_LA H(l,Spielerzahl,Schwellenwert,q,q_1,k,f);

    ///Spaltenvektor aus b extrahieren
    H.getSpaltenvektor_bis_Zeile(*this,l);
    B.Kopiere_unten(H);
    B*=-1;
    B[0]=Vector_LA[l];

    ///Teilzeile aus L extrahieren
    H.getZeilenvektor_bis_Spalte(L,l,l);
    v.Kopiere_unten(H);
    v[0].Erzeugung_double(1.0,Schwellenwert,q);

    ///Wert fr b[l+1] berechnen
    v.SkalarMult(B,Vector_LA[l]);
    Vector_LA[l].TruncPr(k,f,Schwellenwert,q,q_1);
    }
}

void Vektor_LA::VWSubs_mit_Div(const Matrix_LA& L){

int l;

Vector_LA.at(0).FPDiv(L.M[0][0],Schwellenwert,q,q_1,k);

for(l=1; l<d; l++){
    Vektor_LA B(l+1,Spielerzahl,Schwellenwert,q,q_1,k,N);
    Vektor_LA v(l+1,Spielerzahl,Schwellenwert,q,q_1,k,N);
    Vektor_LA H(l,Spielerzahl,Schwellenwert,q,q_1,k,f);

    ///Spaltenvektor aus b extrahieren
    H.getSpaltenvektor_bis_Zeile(*this,l);
    B.Kopiere_unten(H);
    B*=-1;
    B[0]=Vector_LA[l];

    ///Teilzeile aus L extrahieren
    H.getZeilenvektor_bis_Spalte(L,l,l);
    v.Kopiere_unten(H);
    v[0].Erzeugung_double(1.0,Schwellenwert,q);

    ///Wert fr b[l+1] berechnen
    v.SkalarMult(B,Vector_LA[l]);
    Vector_LA[l].TruncPr(k,f,Schwellenwert,q,q_1);

    ///Division durch Diagonalelement
    Vector_LA.at(l).FPDiv(L.M[l][l],Schwellenwert,q,q_1,k);
    }
}


void Vektor_LA::RWSubs(const Matrix_LA& A, Vektor_LA& b){

int l;

Share Hilfe(Spielerzahl, Schwellenwert, q);

Vector_LA.at(d-1)=b[d-1];
Vector_LA[d-1].FPDiv(A.M[d-1][d-1],Schwellenwert,q,q_1,k);

for(l=d-2; l>=0; l--){
    Vektor_LA v(d-l,Spielerzahl,Schwellenwert,q,q_1,k,N);
    v.getZeilenvektor_ab_Spalte(A,l,l);
    v*=-1;
    v[0].Erzeugung_double(1.0,Schwellenwert,q);
    Vektor_LA w(d-l,Spielerzahl,Schwellenwert,q,q_1,k,N);
    w.getSpaltenvektor_ab_Zeile(*this,l);
    w[0]=b[l];
    v.SkalarMult(w,Vector_LA[l]);
    Vector_LA[l].TruncPr(k,N,Schwellenwert,q,q_1);
    Vector_LA.at(l).FPDiv(A.M[l][l],Schwellenwert,q,q_1,k);
    }
}

void Vektor_LA::getPivot(Vektor_LA& Index){

Vektor_LA Kopie=*this;//Kopie erzeugen; da fr Pivotisierung das betragsmig grte Element bentigt wird, mssen die Elemente quadriert werden; der Scope soll aber nicht vernder werden

int i;

for(i=0; i<d; i++){
Kopie.Vector_LA.at(i).MUL_Lory3(Kopie.Vector_LA.at(i).Vector,Schwellenwert,q);//v(i)<-v(i)*v(i)
Kopie[i].TruncPr(k,f,Schwellenwert,q,q_1);//Abschneiden und Maximum mit halber Lnge der Werte ausrechnen ist billiger als auf Abschneiden verzichten und mit doppelter Lnge das Maximum zu bestimmen
}//vgl. Fixpunktarithmetik_Kosten_Neu.mw
Kopie.Max_Neu(Index,k);
}

void Vektor_LA::binaerer_Index(Share * Index, int Obergrenze){//Wandelt den in "Index" gespeicherten Index in einen binren Vektor um, der im Scope gespeichert wird
//"Index" mu berall mit 0 initialisiert sein
int i;

Share Hilfe(Spielerzahl, Schwellenwert, q);

for(i=0; i<Obergrenze; i++){
Hilfe=*Index;
Hilfe.Add_Konst_int(-i,q);
Hilfe.EQZ(Schwellenwert,q,q_1,k);
Vector_LA.at(i)=Hilfe;
}
}

void Vektor_LA::erhoehe_Index(){//Schiebt alle Elemente im Scope-Vektor um 1 nach oben; ist dieser ein binrer Indikator-Vektor wird so der Index um 1 erhht

int i;

Share Hilfe(Spielerzahl,Schwellenwert,q);
Share Hilfe_2(Spielerzahl,Schwellenwert,q);

for(i=0; i<d; i++){
Hilfe=Vector_LA.at(i);
Vector_LA.at(i)=Hilfe_2;
Hilfe_2=Hilfe;
}
}

void Vektor_LA::erhoehe_Index_konditional(Share& Bedingung, mpz_t& q){//Schiebt alle Elemente im Scope-Vektor um 1 nach oben; ist dieser ein binrer Indikator-Vektor wird so der Index um 1 erhht
//Dies geschieht nur, wenn das Bit Bedingung den Wert 1 besitzt
int i;
mpz_t eins;
mpz_init_set_ui(eins,1);

Share Eins(Spielerzahl,Schwellenwert,q);
Eins.Erzeugung(eins,Schwellenwert,q);
Eins.minus(Bedingung.Vector,Spielerzahl,q);
Share Hilfe(Spielerzahl,Schwellenwert,q);
Share Hilfe_2(Spielerzahl,Schwellenwert,q);

for(i=0; i<d; i++){
    Hilfe=Vector_LA.at(i);
    Vector_LA.at(i)=Hilfe_2;
    Vector_LA.at(i)=Bedingung*Vector_LA[i]+Eins*Hilfe;
    Hilfe_2=Hilfe;
    }

mpz_clear(eins);
}

void Vektor_LA::Max_Betrag(Vektor_LA& Index){//Index des betragsmig grten Werts identifizieren

Vektor_LA Kopie=*this;
for(int i=0; i<d; i++){
    Kopie[i].Quadrat();
    Kopie[i].TruncPr(k,f,Schwellenwert,q,q_1);
    }

Kopie.Max_Neu(Index,k);
}

void Vektor_LA::Max_Neu(Vektor_LA& Index, int K){//Bestimme das Maximum des Scope-Vektors und speichere den Index binr in "Index"; der Scope-Vektor wird verndert; K gibt die Skalierung der Werte an; mu nicht Standardskalierung sein! (vgl. getPivot)
//vgl. Share::CarryOutAux
int i,Merke_d,r;
r=d%2;
mpz_t c;
mpz_init(c);

Vektor_LA Merke_this=*this;

if(d==1){
    Index.Vector_LA.at(0).gemeinsame_1(q);
    }

if(d>1){
    int l;

    if(r==0){
        l=d/2;}
    else l=(d+1)/2;

    Share Hilfe(Spielerzahl,Schwellenwert,q);
    Share Summand_1(Spielerzahl,Schwellenwert,q);
    Vektor_LA B(l,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Vektor_LA A_Strich(l,Spielerzahl,Schwellenwert,q,q_1,k,f);

    if(r==0){
        for(i=0; i<l; i++){
            B[i]=Vector_LA[2*i];
            B[i].LT(Vector_LA[2*i+1],Schwellenwert,q,q_1,K);

            Hilfe=B[i];
            Hilfe*=-1;
            Hilfe+=1;//Hilfe=1-B[i]
            Hilfe.MUL_Lory3(Vector_LA[2*i].Vector,Schwellenwert,q);
            Summand_1=Vector_LA[2*i+1];
            Summand_1*=B[i];
            Summand_1.Plus_Vektor(Hilfe.Vector,Spielerzahl,q);
            A_Strich[i]=Summand_1;//A'(i)=B(i)*A(2*i)+(1-B(i))*A(2*i-1)
            }
        Vector_LA=A_Strich.Vector_LA;
        }
    else{
        B[0]=Vector_LA.at(0);
        for(i=1; i<l; i++){
            B[i]=Vector_LA[2*i-1];
            B[i].LT(Vector_LA.at(2*i),Schwellenwert,q,q_1,K);

            Hilfe=B[i];
            Hilfe*=-1;
            Hilfe+=1;//Hilfe=1-B[i]
            Hilfe.MUL_Lory3(Vector_LA.at(2*i-1).Vector,Schwellenwert,q);
            Summand_1=Vector_LA[2*i];
            Summand_1*=B[i];
            Summand_1.Plus_Vektor(Hilfe.Vector,Spielerzahl,q);
            A_Strich[i]=Summand_1;//A'(i)=B(i)*A(2*i-1)+(1-B(i))*A(2*i)
            }

        for(i=1; i<l; i++){//Hnge die hinteren d-2 Elemente an A_Strich an (nur fr die ersten beiden wurde das Maximum gebildet!)
            Vector_LA.at(i)=A_Strich[i];
            }
        }
    Merke_d=d;
    d=l;
    Max_Neu(Index,K);
    Vektor_LA Index_Strich(Index.d,Spielerzahl,Schwellenwert,q,q_1,k,f);
    d=Merke_d;

    if(r==0){
        for(i=0; i<l; i++){
            Index_Strich[2*i+1]=B[i];
            Index_Strich.Vector_LA.at(2*i+1).MUL_Lory3(Index.Vector_LA.at(i).Vector,Schwellenwert,q);//I'=B(i)*I(i)

            Index_Strich[2*i]=Index[i];
            Index_Strich.Vector_LA.at(2*i).minus(Index_Strich.Vector_LA.at(2*i+1).Vector,Spielerzahl,q);//I'(2*i)=I(i)-I'(2*i+1)=(1-B(i))*I'(i), d.h. das umgekehrte Bit von I'(i)
            }
            Index=Index_Strich;
        }

    else{
        for(i=1; i<l; i++){
            Index_Strich[2*i]=B[i];
            Index_Strich.Vector_LA.at(2*i).MUL_Lory3(Index.Vector_LA.at(i).Vector,Schwellenwert,q);//I'=B(i)*I(i)

            Index_Strich[2*i-1]=Index[i];
            Index_Strich.Vector_LA.at(2*i-1).minus(Index_Strich.Vector_LA.at(2*i).Vector,Spielerzahl,q);//I'(2*i)=I(i)-I'(2*i+1)=(1-B(i))*I'(i), d.h. das umgekehrte Bit von I'(i)
            }
        for(i=1; i<Index.d; i++){
            Index[i]=Index_Strich[i];
            }
        }
    }
Vector_LA=Merke_this.Vector_LA;

mpz_clear(c);
}

void Vektor_LA::Max_Bruch(Vektor_LA& B, Vektor_LA& Index){//Bestimme das Maximum einer Reihe von Brchen A_i/B_i (A im Scope). durch a/b<c/d <=> ad<cb
//vgl. Share::CarryOutAux; MinCons in Catrina/de Hoogh: Secure Multiparty Linear Programming Using Fixed Point Arithmetic
Vektor_LA Merke_this=*this;
Vektor_LA Merke_B=B;

int i,Merke_d,r;
r=d%2;
mpz_t c;
mpz_init(c);

if(d==1){
    Index.Vector_LA.at(0).gemeinsame_1(q);
    }

if(d>1){
    int l;

    if(r==0){
        l=d/2;}
    else l=(d-1);

    Share Hilfe(Spielerzahl,Schwellenwert,q);
    Share Hilfe2(Spielerzahl,Schwellenwert,q);
    Share Summand_1(Spielerzahl,Schwellenwert,q);
    Vektor_LA D(l,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Vektor_LA A_Strich(l,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Vektor_LA B_Strich(l,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Vektor_LA x(d,Spielerzahl,Schwellenwert,q,q_1,k,f);


    if(r==0){
        for(i=0; i<d; i++){//Um Brche zu vergleichen, vergleiche ad und bc (s.o.)
            if(i%2==0){
                x[i]=Vector_LA.at(i)*B[i+1];
                }
            else{
                x[i]=Vector_LA.at(i)*B[i-1];
                }
            }

        for(i=0; i<l; i++){
            D[i]=x[2*i];
            D[i].LT(x[2*i+1],Schwellenwert,q,q_1,2*k);

            Hilfe=D[i];
            Hilfe*=-1;
            Hilfe+=1;//Hilfe=1-D[i]

            Hilfe2=Hilfe;

            Hilfe.MUL_Lory3(Vector_LA.at(2*i).Vector,Schwellenwert,q);

            Summand_1=Vector_LA[2*i+1];
            Summand_1*=D[i];
            Summand_1.Plus_Vektor(Hilfe.Vector,Spielerzahl,q);
            A_Strich[i]=Summand_1;//A'(i)=D(i)*A(2*i)+(1-D(i))*A(2*i-1)

            Hilfe2*=B[2*i];
            Summand_1=B[2*i+1]*D[i]+Hilfe2;
            B_Strich[i]=Summand_1;//A'(i)=D(i)*A(2*i)+(1-D(i))*A(2*i-1)
            }
        Vector_LA=A_Strich.Vector_LA;
        B.Vector_LA=B_Strich.Vector_LA;
        }

    else{
        x[0]=Vector_LA.at(0)*B[1];
        x[1]=Vector_LA.at(1)*B[0];

        D[0]=x[0];
        D[0].LT(x[1],Schwellenwert,q,q_1,2*k);

        Hilfe=D[0];
        Hilfe.Mul_Konst_int(-1,q);
        Hilfe.Add_Konst_int(1,q);
        Hilfe2=Hilfe;
        Hilfe.MUL_Lory3(Vector_LA.at(0).Vector,Schwellenwert,q);
        Summand_1=Vector_LA[1];
        Summand_1*=D.Vector_LA.at(0);
        Summand_1.Plus_Vektor(Hilfe.Vector,Spielerzahl,q);

        A_Strich[0]=Summand_1;//A'(0)=D(0)*A(0)+(1-D(0))*A(1)
        Summand_1=B[1]*D[0]+Hilfe2*B[0];
        B_Strich[0]=Summand_1;//A'(0)=D(0)*A(0)+(1-D(0))*A(1)

        for(i=1; i<d-1; i++){//Hnge die hinteren d-2 Elemente an A_Strich an (nur fr die ersten beiden wurde das Maximum gebildet!)
            Vector_LA.at(i)=Vector_LA.at(i+1);
            B.Vector_LA.at(i)=B.Vector_LA.at(i+1);
            }
        Vector_LA.at(0)=A_Strich.Vector_LA.at(0);
        B[0]=B_Strich[0];
        }
    Merke_d=d;
    d=l;
    B.d=l;
    Max_Bruch(B, Index);
    Vektor_LA Index_Strich(Index.d,Spielerzahl,Schwellenwert,q,q_1,k,f);
    d=Merke_d;
    B.d=Merke_d;

    if(r==0){
        for(i=0; i<l; i++){
            Index_Strich[2*i+1]=D[i];
            Index_Strich[2*i+1].MUL_Lory3(Index[i].Vector,Schwellenwert,q);//I'=D(i)*I(i)

            Index_Strich[2*i]=Index[i];
            Index_Strich[2*i].minus(Index_Strich[2*i+1].Vector,Spielerzahl,q);//I'(2*i)=I(i)-I'(2*i+1)=(1-D(i))*I'(i), d.h. das umgekehrte Bit von I'(i)
            }
        Index=Index_Strich;
        }

    else{
        Index_Strich[1]=D[0];
        Index_Strich.Vector_LA.at(1).MUL_Lory3(Index.Vector_LA.at(0).Vector,Schwellenwert,q);//I'(0)=D(0)*I(0)

        Index_Strich[0]=Index[0];
        Index_Strich.Vector_LA.at(0).minus(Index_Strich.Vector_LA.at(1).Vector,Spielerzahl,q);//I'(0)=I(0)-I'(1)=(1-D(0))*I'(0), d.h. das umgekehrte Bit von I'(i)

        for(i=Index.d-2; i>0; i--){
            Index.Vector_LA.at(i+1)=Index.Vector_LA.at(i);//Schiebe Elemente des Indexvektors um 1 nach hinten um vorne Platz fr 1 neues zu machen
            }
        Index.Vector_LA.at(1)=Index_Strich.Vector_LA.at(1);//Fge die neugeschaffenen Elemente ein
        Index.Vector_LA.at(0)=Index_Strich.Vector_LA.at(0);
        }
    }
B.Vector_LA=Merke_B.Vector_LA;
B.d=Merke_B.d;
Vector_LA=Merke_this.Vector_LA;
d=Merke_this.d;
mpz_clear(c);
}

void Vektor_LA::Min_Bruch(Vektor_LA& C, Vektor_LA& W){

Vektor_LA Kopie=*this;
Kopie*=-1;
Kopie.Max_Bruch(C,W);
}

void Vektor_LA::Min(Vektor_LA& Index, int K){

Vektor_LA Hilfsvektor=*this;
Hilfsvektor*=(-1);

Hilfsvektor.Max_Neu(Index,K);
}

void Vektor_LA::Concat_Vektor(Vektor_LA * a, Vektor_LA *b){

int i,j=0;

for(i=0; i<a->d; i++){
Vector_LA.at(i)=a->Vector_LA.at(i);
}
for(i=a->d; i<d; i++){
Vector_LA.at(i)=b->Vector_LA.at(j);
j++;
}
}

void Vektor_LA::Gleich_Null(Share& Summe){//Berechne Gleichheit mit Null anhand der euklidischen Norm

SkalarMult(*this,Summe);
Summe.TruncPr(k,f,Schwellenwert,q,q_1);
Summe.EQZ(Schwellenwert,q,q_1,k);
}

void Vektor_LA::Gleich_Null_approximativ(Share& Bit){//Prfe ob, die Summe der Quadrate aller Komponenten nicht weiter als 2^-10 von Null entfernt ist

mpz_t Schwelle;
mpz_init_set_ui(Schwelle,1);
mpz_mul_2exp(Schwelle,Schwelle,10);//Schwelle=2^-11

SkalarMult(*this,Bit);
Bit.TruncPr(k,f,Schwellenwert,q,q_1);
Bit-=Schwelle;
Bit.LTZ_F28(Schwellenwert,q,q_1,k);

mpz_clear(Schwelle);
}

void Vektor_LA::Vektor_Zusammenschieben(std::vector<Share>& Aktive_Menge_q, Matrix_LA& P){

if(P.Zeilenzahl!=d || P.Spaltenzahl!=d){
    printf("\nKompaktifizieren des Vektors nicht mglich, da Dimensionen falsch. Abbruch");
    exit(1);
    }

P=0;

Vektor_LA A_Strich(d, Spielerzahl, Schwellenwert, q, q_1, k, f);
Vektor_LA Ergebnis(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Matrix_LA A(d, d, Spielerzahl, Schwellenwert, q, q_1, k, f);//Permutationsmatrix
Vektor_LA Index(d, Spielerzahl,Schwellenwert,q,q_1,k,f);
Matrix_LA Null(d,d,Spielerzahl,Schwellenwert,q,q_1,k,f);

Null=0;
Vektor_LA Kopie=*this;


int i;
mpz_t eins;
mpz_init_set_ui(eins,1);

Index.Vector_LA.at(0).Erzeugung(eins,Schwellenwert,q);
Vektor_LA Aktive_Menge(Aktive_Menge_q.size(),Spielerzahl,Schwellenwert,q,q_1,k,f);
Aktive_Menge.Vector_LA=Aktive_Menge_q;

for(int j=0; j<d; j++){
    Aktive_Menge[j].EQZ(Schwellenwert,q,q_1,k);
    Aktive_Menge[j].Mul_Konst_int(-1,q);
    Aktive_Menge[j].Add_Konst_int(1,q);//Indikatorbit umdrehen

    A=Null;

    for(i=0; i<d; i++){
        A.M[i][j]=Aktive_Menge[j]*Index.Vector_LA[i];
        }

    Index.erhoehe_Index_konditional(Aktive_Menge.Vector_LA[j],q);

    Kopie.MatrixMult(A,Ergebnis);
    A_Strich+=Ergebnis;
    Kopie=*this;
    P+=A;
}

*this=A_Strich;
mpz_clear(eins);
}

void Vektor_LA::Vektor_Zusammenschieben_binaer(std::vector<Share>& Aktive_Menge_q, Matrix_LA& P){//In aktive_Menge_q sind bereits die aktiven Zeilen binr kodiert

if(P.Zeilenzahl!=d || P.Spaltenzahl!=d){
    printf("\nKompaktifizieren des Vektors nicht mglich, da Dimensionen falsch. Abbruch");
    exit(1);
    }

P=0;

Vektor_LA Index(d, Spielerzahl,Schwellenwert,q,q_1,k,f);

int i;
mpz_t eins;
mpz_init_set_ui(eins,1);

Index.Vector_LA.at(0).Erzeugung(eins,Schwellenwert,q);
Vektor_LA Aktive_Menge(Aktive_Menge_q.size(),Spielerzahl,Schwellenwert,q,q_1,k,f);
Aktive_Menge.Vector_LA=Aktive_Menge_q;

for(int j=0; j<d; j++){
    for(i=0; i<d; i++){
        P.M[i][j]=Aktive_Menge[j]*Index.Vector_LA[i];
        }
    Index.erhoehe_Index_konditional(Aktive_Menge.Vector_LA[j],q);
    }

MatrixMult(P,*this);

mpz_clear(eins);
}

void Vektor_LA::OR(Vektor_LA& Argument_2){//OR von verteilten Vektoren

int i;

if(Argument_2.d!=d){
    printf("\nDimensionen verschieden. OR nicht mglich. Abbruch");
    exit(1);
    }

for(i=0; i<d; i++){
    Vector_LA[i]=Vector_LA[i]+Argument_2[i]-Vector_LA[i]*Argument_2[i];
    }
}

void Vektor_LA::AND(Vektor_LA& Argument_2){//OR von verteilten Vektoren

int i;

if(Argument_2.d!=d){
    printf("\n\nFehler: Vektor_LA::AND");
    printf("\nDimensionen verschieden. AND nicht mglich. Abbruch");
    exit(1);
    }

for(i=0; i<d; i++){
    Vector_LA[i]*=Argument_2[i];
    }
}

void Vektor_LA::KOR(Share& Ausgabe, int r, mpz_t& p){

int l=r%2,m,i;

if(l==0)
    m=r/2;
else
    m=(r-1)/2;

Share A(Spielerzahl,Schwellenwert,p);
std::vector<Share> V(m,A);

if(r==1){
    Ausgabe=Vector_LA[0];
    }
else{
    if(l==0){
        for(i=0; i<m; i++){
            A=Vector_LA[2*i];
            A.OR(Vector_LA[2*i+1]);
            V[i]=A;
            }
        }
    else{
        A=Vector_LA[0];
        A.OR(Vector_LA[1]);
        V[0]=A;

        for(i=1; i<m; i++){
            V[i]=Vector_LA[i+1];
            }
        }
    Vector_LA=V;
    int Merke_d=d;
    d=m;
    KOR(Ausgabe,m,p);
    d=Merke_d;
    }
}

void Vektor_LA::Norm(Share& N){

Vektor_LA Kopie=*this;
SkalarMult(Kopie, N);
N.TruncPr(k,f,Schwellenwert,q,q_1);
N.Wurzel(k,q,q_1);
}

void Vektor_LA::Abstand_zu_Ebene(Vektor_LA& Hesseform, Share& b, Share& Abstand){//Berechnet den Abstand eines Punktes von der in Heseform bergebenen Ebene; b steht auf der rechten Seite der HNF

if(d!=Hesseform.d){
    printf("\nWanung: Vektor_LA::Abstand_zu_Ebene");
    printf("\nDimension der Ebene ungleich der Anzahl der Komponenten des Punktes. Abbruch.");
    exit(1);
    }

int i;

Abstand=0;
Share Hilfe(Spielerzahl,Schwellenwert,q);
Share Hilfe_Nenner(Spielerzahl,Schwellenwert,q);
Share Summe_Nenner(Spielerzahl,Schwellenwert,q);
for(i=0; i<d; i++){
    Hilfe=Vector_LA.at(i)*Hesseform.Vector_LA.at(i);
    Hilfe.TruncPr(k,f,Schwellenwert,q,q_1);
    Abstand+=Hilfe;
    Hilfe_Nenner=Hesseform[i]*Hesseform[i];
    Hilfe_Nenner.TruncPr(k,f,Schwellenwert,q,q_1);
    Summe_Nenner+=Hilfe_Nenner;
    }

Abstand-=b;
Abstand.Betrag(q_1,k);
Summe_Nenner.Wurzel(k,q,q_1);
Abstand.FPDiv(Summe_Nenner,Schwellenwert,q,q_1,k);
}

void Vektor_LA::house_neu(){

int n=Spielerzahl, t=Schwellenwert;
Share mu(n,t,q);
Share VZ=Vector_LA.at(0);
Vektor_LA Kopie=*this;

Kopie.Norm(mu);
VZ.LTZ_F28(Schwellenwert,q,q_1,k);
VZ*=-2;
VZ+=1; //VZ=1-2*LTZ(VZ), d.h. VZ=VZ(x[0])

Vector_LA.at(0)+=VZ*mu;
}

void Vektor_LA::Mul_house_Pre_beta_bekannt(Vektor_LA& v, Share& beta){

Vektor_LA x=*this;
Share Ergebnis(Spielerzahl,Schwellenwert,q);

x.SkalarMult(v,Ergebnis);
v*=beta;
v*=Ergebnis;
v.TruncPr(3*k,3*f);
*this+=v;
}

void Vektor_LA::Qt_mult_b(Vektor_LA& beta, Matrix_LA& M, int d){

if(M.Zeilenzahl!=d){
    printf("\n Dimension von b ungleich Zeilenzahl der Matrix. Abbruch");
    exit(1);
    }

Share Eins(Spielerzahl,Schwellenwert,q);
Eins.Erzeugung_double(1.0,Schwellenwert,q);


for(int i=0; i<=d-2; i++){
    Vektor_LA v(d-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Vektor_LA b_Hilfe(d-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    v.getSpaltenvektor_ab_Zeile(M,i,i);
    v[0]=Eins;
    b_Hilfe.Kopiere_unten(*this);
    b_Hilfe.Mul_house_Pre_beta_bekannt(v,beta[i]);
    Kopiere_unten(b_Hilfe);
    }
}

void Vektor_LA::Qt_mult_b_neu(Vektor_LA& beta, Vektor_LA& beta_HH, Matrix_LA& M, int d){

if(M.Zeilenzahl!=d){
    printf("\n Dimension von b ungleich Zeilenzahl der Matrix. Abbruch");
    exit(1);
    }

Share Eins(Spielerzahl,Schwellenwert,q);
Eins.Erzeugung_double(1.0,Schwellenwert,q);


for(int i=0; i<=d-2; i++){
    Vektor_LA v(d-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    Vektor_LA b_Hilfe(d-i,Spielerzahl,Schwellenwert,q,q_1,k,f);
    v.getSpaltenvektor_ab_Zeile(M,i,i);
    v[0]=beta_HH[i];
    b_Hilfe.Kopiere_unten(*this);
    b_Hilfe.Mul_house_Pre_beta_bekannt(v,beta[i]);
    Kopiere_unten(b_Hilfe);
    }
}

void Vektor_LA::Kopiere_unten(Vektor_LA& V){//den Vektor V von unten in den Scope schreiben;

int r=haupt::min(d,V.d);

for(int i=0; i<r; i++){
    Vector_LA.at(d-i-1)=V[V.d-i-1];
    }
}


void Vektor_LA::Oben_einpassen(Vektor_LA& V){//Schreibt den Vektor V soweit wie mglich in den Vektor im Scope hinein

int s=haupt::min(d,V.d);
for(int i=0; i<s; i++){
    Vector_LA.at(i)=V[i];
    }
}

void Vektor_LA::Umdrehen(){//Reihenfolge umdrehen

Vektor_LA V=*this;

for(int i=0; i<d; i++){
    V[i]=Vector_LA[d-i-1];
    }

*this=V;
}

void Vektor_LA::Pre_Mul_C(){

int i,j;

Vektor_LA r(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA s(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA v(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA w(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA z(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
Vektor_LA p(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
mpz_t * u;
mpz_t * m;
mpz_t * pm;
u=new mpz_t[d];
m=new mpz_t[d];
pm=new mpz_t[d];

for(i=0; i<d; i++){
    mpz_init_set_ui(u[i],0);
    mpz_init_set_ui(m[i],0);
    r[i].Zufallszahl(Schwellenwert,q);
    s[i].Zufallszahl(Schwellenwert,q);
    s[i].MulPub(r[i],u[i]);
    }

for(i=0; i<d-1; i++){
    v[i]=r[i+1]*s[i];
    }

w[0]=r[0];

for(i=1; i<d; i++){
    mpz_invert(u[i-1],u[i-1],q);
    w[i]=v[i-1]*u[i-1];
    z[i-1]=s[i-1]*u[i-1];
    }
mpz_invert(u[d-1],u[d-1],q);
z[d-1]=s[d-1]*u[d-1];

for(i=0; i<d; i++){
    w[i].MulPub(Vector_LA.at(i),m[i]);
    }

mpz_init_set(pm[0],m[0]);
for(i=1; i<d; i++){
    mpz_init_set(pm[i],m[i]);
    mpz_mul(pm[i],pm[i],pm[i-1]);
    }

p[0]=Vector_LA.at(0);
for(j=1; j<d; j++){
    p[j]=z[j]*pm[j];
    }

*this=p;
for(i=0; i<d; i++){
    mpz_clear(u[i]);
    mpz_clear(m[i]);
    mpz_clear(pm[i]);
    }

delete[]u;
delete[]m;
delete[]pm;
u=NULL;
m=NULL;
pm=NULL;
}

void Vektor_LA::PreOrC(){

int i;
Share Hilfe=Vector_LA.at(0);
for(i=0; i<d; i++){
    Vector_LA.at(i)+=1;
    }
Pre_Mul_C();
Vector_LA.at(0)=Hilfe;
for(i=1; i<d; i++){
    Vector_LA.at(i).Mod2(k,q,q_1,Schwellenwert);
    Vector_LA.at(i)*=-1;
    Vector_LA.at(i)+=1;
    }
}

void Vektor_LA::Spalte_auslesen(Matrix_LA& A, Vektor_LA& I, int l){//Auslesen der im bergebenen Vektor I binr codierten Spalte bis zum Element l<d, l<A.Zeilenzahl

if(I.d!=A.Spaltenzahl){
    printf("\nWarnung: Vektor_LA::Spalte_auslesen");
    printf("\nAuslesen der Spalte nicht moelich, da I.d und Spaltenzahl verschieden. Abbruch");
    exit(1);
    }

if(l>d){
    printf("\nWarnung: Vektor_LA::Spalte_auslesen");
    printf("\nAuslesen der Spalte nicht moelich, da l>d. Abbruch");
    exit(1);
    }

if(l>A.Zeilenzahl){
    printf("\nWarnung: Vektor_LA::Spalte_auslesen");
    printf("\nAuslesen der Spalte nicht moeglich, da l>A.Zeilenzahl. Abbruch");
    exit(1);
    }

Vektor_LA Hilfsvektor(A.Spaltenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
for(int i=0; i<l; i++){
    Hilfsvektor.getZeilenvektor(A,i);
    Hilfsvektor.SkalarMult(I,Vector_LA.at(i));
    }
}

void Vektor_LA::SecWrite(Vektor_LA& V, Share& s){//vgl. Catrina/de Hoogh: Simplex

if(V.d!=d){
    printf("\n\nFehler: Vektor_LA::SecWrite");
    printf("\nDimensionen verschieden. Sicheres Schreiben nicht moeglich. Abbruch.");
    exit(1);
    }

for(int i=0; i<d; i++){
    Vector_LA.at(i)=Vector_LA.at(i)+V[i]*(s-Vector_LA.at(i));
    }
}

void Vektor_LA::SelectFirst(){//vgl. Catrina Simplex

PreOrC();

Vektor_LA V(d,Spielerzahl,Schwellenwert,q,q_1,k,f);
V[0]=Vector_LA.at(0);
for(int i=1; i<d; i++){
    V[i]=Vector_LA.at(i)-Vector_LA.at(i-1);
    }
*this=V;
}

void Vektor_LA::FindFirst(Vektor_LA& Index){//Finde erstes Nicht-Null-Element

if(d!=Index.d){
    printf("\nWarnung: Vektor_LA::FindFirst");
    printf("\nLaengen der Vektoen verschieden. FindFirst kann nicht angewandt werden. Abbruch");
    exit(1);
    }

Share Zaehler(Spielerzahl,Schwellenwert,q);
Zaehler.gemeinsame_1(q);

Share Hilfe(Spielerzahl,Schwellenwert,q);

for(int i=0; i<d; i++){//
    Index[i]=Vector_LA[i]*Zaehler;
    Hilfe=-Vector_LA[i]+1;
    Zaehler=Zaehler*Hilfe;
    }
}

void Vektor_LA::SecReadRow(Matrix_LA& A, Vektor_LA& I, int l){//Sicheres Lesen der Zeile im Scope bis Spalte l.

if(l>d){
    printf("\nAuslesen der Zeile nicht moeglich, da l>d. Abbruch");
    exit(1);
    }

if(l>A.Spaltenzahl){
    printf("\nAuslesen der Zeile nicht moeglich, da l>A.Zeilenzahl. Abbruch");
    exit(1);
    }

Vektor_LA Hilfsvektor(A.Zeilenzahl,Spielerzahl,Schwellenwert,q,q_1,k,f);
for(int i=0; i<l; i++){
    Hilfsvektor.getSpaltenvektor(A,i);
    Hilfsvektor.SkalarMult(I,Vector_LA.at(i));
    }
}

void Vektor_LA::BitMask(Share& x){//vgl. Secure SCM 3.2 1.1 Bit Mask, aber verbessert, da nur d Multiplikationen
//Koeffizienten evtl. im Voraus global berechnen
Vektor_LA X(d-1,Spielerzahl,Schwellenwert,q,q_1,k,f);
int i,j;

mpz_t * Koeff;
Koeff=new mpz_t[d];

for(i=0; i<d; i++){
    mpz_init_set_ui(Koeff[i],1);
    }

for(i=0; i<d; i++){//fr alle Polynome
    haupt::Koeffizienten_Bit_Mask(d,i,q,Koeff);
    Vector_LA.at(i)=Koeff[d-1];
    for(j=d-2; j>0; j--){//Auswertung jedes Polynoms
        if(j>i)
            Vector_LA.at(i)=Vector_LA.at(i)*(x-j)+Koeff[j];
        if(j<=i)
            Vector_LA.at(i)=Vector_LA.at(i)*(x-(j-1))+Koeff[j];
        }
    Vector_LA.at(i)=Vector_LA.at(i)*(x-i)+Koeff[0];
    }

delete[]Koeff;
Koeff=NULL;
}
