/*
DEUTSCH

Der in dieser Datei enthaltene Programmcode ist geistiges Eigentum des Autors (Manuel Liedel, Manuel.Liedel "at" gmail.com). Jeder Gebrauch - auer zur Verifikation experimenteller Ergebnisse der Dissertationsschrift "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - ist ohne explizite Genehmigung des Autors untersagt. Dies schliet insbesondere kommerzielle Nutzung, Weitergabe, Modifikation und Erweiterungen ein. Auch die Verwendung zu wissenschaftlichen Zwecken bedarf der Genehmigung des Autors.

ENGLISH

The source code contained in this file is the intellectual property of the author (Manuel Liedel, Manuel.Liedel "at" gmail.com). Any use - except to verify experimental results contained in the author's PhD-Dissertation "Sichere Mehrparteienberechnungen und datenschutzfreundliche Klassifikation auf Basis horizontal partitionierter Datenbanken" - is prohibited without the author's explicit consent. This includes - but is no limited to - commercial use, redistribution, modifications and extensions. Scientific use also requires the author's explicit consent.
*/

#include "Vektor_mpf_t.h"
#include <gmp.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <sstream>
#include <ctime>
#include <Matrix_mpf_t.h>

Vektor_mpf_t::Vektor_mpf_t(int d)
{
Dimension=d;
V=new mpf_t[d];

for(int i=0; i<d; i++){
    mpf_init_set_d(V[i],0.0);
    }
}

Vektor_mpf_t::Vektor_mpf_t(int d, double*w)
{
Dimension=d;
V=new mpf_t[d];

for(int i=0; i<d; i++){
    mpf_init_set_d(V[i],w[i]);
    }
}

Vektor_mpf_t::Vektor_mpf_t(int d, mpf_t*w)
{
Dimension=d;
V=new mpf_t[d];

for(int i=0; i<d; i++){
    mpf_init_set(V[i],w[i]);
    }
}

Vektor_mpf_t::~Vektor_mpf_t()
{
for(int i=0; i<Dimension; i++){
    mpf_clear(V[i]);
    }
delete []V;
V=NULL;
}

Vektor_mpf_t::Vektor_mpf_t(const Vektor_mpf_t& other)
{
Dimension=other.Dimension;
V=new mpf_t[Dimension];

for(int i=0; i<Dimension; i++){
    mpf_init_set(V[i],other.V[i]);
    }
}

Vektor_mpf_t& Vektor_mpf_t::operator=(const Vektor_mpf_t& rhs)
{
    if (this == &rhs) return *this; // handle self assignment

    if(rhs.Dimension!=Dimension){
    printf("\nWarnung: Vektor_mpf_t::operator=.");
    printf("\nZuweisung nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

    for(int i=0; i<Dimension; i++){
        mpf_set(V[i],rhs.V[i]);
        }

    return *this;
}

void Vektor_mpf_t::operator=(int n){

for(int i=0; i<Dimension; i++){
    mpf_set_si(V[i],n);
    }

}

void Vektor_mpf_t::getSpaltenvektor(int Spalte, Matrix_mpf_t& A){

if(Dimension!=A.Hoehe){
    printf("\nWarnung: Vektor_mpf_t::getSpaltenvektor");
    printf("\nVektor kann nicht ausgelesen werden, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

int i;

for(i=0; i<Dimension; i++){
    mpf_set(V[i],A.M[i][Spalte]);
    }
}

void Vektor_mpf_t::SkalarProdukt(Vektor_mpf_t& v2, mpf_t& Ergebnis){

if(v2.Dimension!=Dimension){
    printf("\nWarnung: Vektor_mpf_t::SkalarProdukt");
    printf("\nBerechnen des Skalarprodukts nicht mglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

mpf_t h;
mpf_init(h);

mpf_set_ui(Ergebnis,0);

for(int i=0; i<Dimension; i++){
    mpf_mul(h,V[i],v2.V[i]);
    mpf_add(Ergebnis,Ergebnis,h);
    }

mpf_clear(h);
}

void Vektor_mpf_t::getSpaltenvektor_ab_Zeile(int Spalte, Matrix_mpf_t& A, int Zeile){

if(Dimension!=(A.Hoehe-Zeile)){
    printf("\nWarnung: Vektor_mpf_t::getSpaltenvektor_ab_Zeile");
    printf("\nVektor kann nicht ausgelesen werden, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

int i;

for(i=Zeile; i<A.Hoehe; i++){
    mpf_set(V[i-Zeile],A.M[i][Spalte]);
    }
}

void Vektor_mpf_t::MatrixMultVektor(Matrix_mpf_t& M, Vektor_mpf_t& Ergebnis){

if(Dimension!=M.Breite){
    printf("\nWarnung: Vektor_mpf_t::MatrixMultVektor");
    printf("\nMatrix-Vektor-Multiplikation nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

int i,j;

mpf_t h;
mpf_init(h);

for(i=0; i<M.Hoehe; i++){
    mpf_set_d(Ergebnis.V[i],0.0);
    for(j=0; j<M.Breite; j++){
        mpf_mul(h,M.M[i][j],V[j]);
        mpf_add(Ergebnis.V[i],Ergebnis.V[i],h);
        }
    }

mpf_clear(h);
}

void Vektor_mpf_t::Zufallsvektor(){

int i;

mpz_t seed_mpz;
mpz_init(seed_mpz);
time_t seed;
double seed_d;

seed=clock();
seed_d=(double) seed;
seed_d=seed_d*seed_d;
mpz_set_d(seed_mpz,seed_d);

gmp_randstate_t rState;
gmp_randinit_mt(rState);
gmp_randseed(rState,seed_mpz); //Erzeugen einer Zufallszahl*/

for(i=0; i<Dimension; i++){
    mpf_urandomb(V[i],rState,20);
    mpf_mul_2exp(V[i],V[i],11); //Erzeuge Zufallszahl in -1024..1023
    mpf_sub_ui(V[i],V[i],1024);
    }

gmp_randclear(rState);
mpz_clear(seed_mpz);
}

double Vektor_mpf_t::Norm(){

double N=0.0;
mpf_t h;
mpf_init_set_d(h,0.0);
mpf_t g;
mpf_init_set_d(g,0.0);

for(int i=0; i<Dimension; i++){
    mpf_mul(g,V[i],V[i]);
    mpf_add(h,h,g);
    }

mpf_sqrt(h,h);
N=mpf_get_d(h);

mpf_clear(g);
mpf_clear(h);

return N;
}

Vektor_mpf_t Vektor_mpf_t::operator+(const Vektor_mpf_t& W){

if(W.Dimension!=Dimension){
    printf("\nWarnung: Vektor_mpf_t::operator+.");
    printf("\nAddition nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

Vektor_mpf_t Ergebnis(Dimension);

for(int i=0; i<Dimension; i++){
    mpf_add(Ergebnis.V[i],V[i],W.V[i]);
    }

return Ergebnis;
}

Vektor_mpf_t& Vektor_mpf_t::operator+=(const Vektor_mpf_t& W){

if(W.Dimension!=Dimension){
    printf("\nWarnung: Vektor_mpf_t::operator+=.");
    printf("\nAddition nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

for(int i=0; i<Dimension; i++){
    mpf_add(V[i],V[i],W.V[i]);
    }

return *this;
}

Vektor_mpf_t Vektor_mpf_t::operator-(const Vektor_mpf_t& W){

if(W.Dimension!=Dimension){
    printf("\nWarnung: Vektor_mpf_t::operator-.");
    printf("\nSubtraktion nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

Vektor_mpf_t Ergebnis(Dimension);

for(int i=0; i<Dimension; i++){
    mpf_sub(Ergebnis.V[i],V[i],W.V[i]);
    }

return Ergebnis;
}

Vektor_mpf_t& Vektor_mpf_t::operator-=(const Vektor_mpf_t& W){

if(W.Dimension!=Dimension){
    printf("\nWarnung: Vektor_mpf_t::operator-=.");
    printf("\nSubtraktion nicht moeglich, da Dimensionen verschieden. Abbruch");
    exit(1);
    }

for(int i=0; i<Dimension; i++){
    mpf_sub(V[i],V[i],W.V[i]);
    }

return *this;
}

Vektor_mpf_t& Vektor_mpf_t::operator*=(mpf_t& Faktor){

for(int i=0; i<Dimension; i++){
    mpf_mul(V[i],V[i],Faktor);
    }

return *this;
}

void Vektor_mpf_t::VWSubs(Matrix_mpf_t& L){

if(Dimension!=L.Hoehe){
    printf("\nWarnung: Vektor_mpf_t::VWSubs");
    printf("\nVWSubs nicht moeglich, da Dimensionen von Vektor und Matrix verschieden. Abbruch");
    exit(1);
    }

if(L.Breite!=L.Hoehe){
    printf("\nWarnung: Vektor_mpf_t::VWSubs");
    printf("\nVWSubs nicht moeglich, da Matrix nicht quadratisch. Abbruch");
    exit(1);
    }

int i,k;

mpf_t h;
mpf_init(h);

for(k=0; k<Dimension-1; k++){
    for(i=k+1; i<Dimension; i++){
        mpf_mul(h,L.M[i][k],V[k]);
        mpf_sub(V[i],V[i],h);
        }
    }

mpf_clear(h);
}

void Vektor_mpf_t::RWSubs(Matrix_mpf_t& L){

if(Dimension!=L.Hoehe){
    printf("\nWarnung: Vektor_mpf_t::RWSubs");
    printf("\nRWSubs nicht moeglich, da Dimensionen von Vektor und Matrix verschieden. Abbruch");
    exit(1);
    }

if(L.Breite!=L.Hoehe){
    printf("\nWarnung: Vektor_mpf_t::RWSubs");
    printf("\nRWSubs nicht moeglich, da Matrix nicht quadratisch. Abbruch");
    exit(1);
    }

int i,k;

mpf_t h;
mpf_init(h);

Vektor_mpf_t x(Dimension);

for(k=Dimension-1; k>=0; k--){
    mpf_set(x.V[k],V[k]);
    for(i=k+1; i<Dimension; i++){
        mpf_mul(h,x.V[i],L.M[k][i]);
        mpf_sub(V[k],V[k],h);
        }
    mpf_div(x.V[k],V[k],L.M[k][k]);
    }

*this=x;
mpf_clear(h);
}

void Vektor_mpf_t::Ausgabe(std::string Name){

int i,z;

char * r;
r= new char [Name.size()+1];
strcpy(r,Name.c_str());

printf("\n");

for(i=0; i<Dimension; i++){
z=gmp_printf("\n%s(%d)=%.10Ff",r,i,V[i]);
}

delete[]r;
r=NULL;
}

void Vektor_mpf_t::Ausgabe_wissenschaftlich(std::string Name){

int i,z;

char * r;
r= new char [Name.size()+1];
strcpy(r,Name.c_str());

printf("\n");

for(i=0; i<Dimension; i++){
z=gmp_printf("\n%s(%d)=%.10Fe",r,i,V[i]);
}

delete[]r;
r=NULL;
}

void Vektor_mpf_t::Ausgabe_Datei(std::string Name, FILE* datei){

int i,z;

char * r;
r= new char [Name.size()+1];
strcpy(r,Name.c_str());

fprintf(datei,"\n");

for(i=0; i<Dimension; i++){
z=gmp_fprintf(datei,"\n%s(%d)=%.10Ff",r,i,V[i]);
}

delete[]r;
r=NULL;
}

void Vektor_mpf_t::Ausgabe_wissenschaftlich_Datei(std::string Name, FILE* datei){

int i,z;

char * r;
r= new char [Name.size()+1];
strcpy(r,Name.c_str());

fprintf(datei,"\n");

for(i=0; i<Dimension; i++){
z=gmp_fprintf(datei,"\n%s(%d)=%.10Fe",r,i,V[i]);
}

delete[]r;
r=NULL;
}

int Vektor_mpf_t::Max(){

int i;

mpf_t max;
mpf_init(max);

mpf_t h;
mpf_init(h);

int Index=0;

mpf_mul(h,V[0],V[0]);
mpf_set(max,h);
for(i=1; i<Dimension; i++){
    mpf_mul(h,V[i],V[i]);
    if(mpf_cmp(h,max)>0){
        mpf_set(max,h);
        Index=i;
        }
    }

mpf_clear(h);
mpf_clear(max);

return Index;
}

void Vektor_mpf_t::Householder_Vektor(){

mpf_t beta;
mpf_t mu;
mpf_t VZ;

mpf_init_set(beta,V[0]);
mpf_init(mu);
mpf_init(VZ);

Vektor_mpf_t Kopie=*this;
double n=Kopie.Norm();
mpf_set_d(mu,n);

if(mpf_cmp_si(V[0],0)<0){
    mpf_set_d(VZ,-1.0);
    }
else{
    mpf_set_d(VZ,1.0);
    }

mpf_mul(mu,mu,VZ);//mu<-mu*VZ
mpf_add(beta,beta,mu);//beta<-beta+mu*VZ

for(int i=1; i<Dimension; i++){
    mpf_div(V[i],V[i],beta);
    }

mpf_set_d(V[0],1.0);


mpf_clear(VZ);
mpf_clear(mu);
mpf_clear(beta);
}
/*
Share mu(n,t,q);
    Share VZ=Vector_LA.at(0);
    Share beta=Vector_LA.at(0);
    Vektor_LA Kopie=*this;

    Kopie.Norm(mu);
    VZ.LTZ_F28(Schwellenwert,q,q_1,k);
    VZ*=-2;
    VZ+=1; //VZ=1-2*LTZ(VZ), d.h. VZ=VZ(x[0])

    beta+=VZ*mu;
    beta.Inverses_ohne_Trunc(q_1,k);

    for(int i=1; i<d; i++){
        Vector_LA[i]*=beta;
        Vector_LA[i].TruncPr(3*k-f,2*k-f,Schwellenwert,q,q_1);
        }

    Vector_LA[0].Erzeugung_double(1.0,Schwellenwert,q);*/
